/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <dbus/dbus-glib.h>

#include "glib-util.h"
#include "pkgmgr-if.h"
#include "package.h"
#include "pkgmgrd-dbus.h"
#include "dbus-marshal.h"
#include "utils.h"
#include "../daemon/debug.h"

typedef	gboolean (*CommandFunc) (int argc,char* argv[]);

typedef struct _Command {
	const gchar *name;
	CommandFunc func;
} Command;
/* Common options*/
gboolean verbose = FALSE;
static GOptionEntry common_opts[] = {
	{ "verbose", 'V', 0, G_OPTION_ARG_NONE, &verbose, "Print verbose output", NULL },
	{ NULL }
};

static char *
make_flag_string (const char *flag, ...)
{
  const char *flags[100];
  int n_flags = 0;

  va_list ap;
  va_start (ap, flag);
  while (flag && n_flags < 99)
    {
      gboolean val = va_arg (ap, int);
      if (val)
        flags[n_flags++] = flag;
      flag = va_arg (ap, const char *);
    }
  flags[n_flags++] = NULL;

  return g_strjoinv (", ", (gchar **)flags);
}
  

static GMainLoop *loop = NULL;

gboolean need_line_feed = FALSE;

static void
print_usage(const gchar *name);

static void
__print(const gchar *fmt, ...)
{
	if (need_line_feed) {
		printf("\n");
		need_line_feed = FALSE;
	}
	va_list args;
	va_start (args, fmt);
	vfprintf (stdout, fmt, args);
	va_end (args);
}

static void
__download_progress_handler(DBusGProxy *proxy,
			    const gchar* operation,
			    const gchar* packagename,
			    const gchar* version,
			gint already,
			gint total,
			gpointer data)
{
	static gboolean download_in_progress = FALSE;
	/* Simple percentage progress print */
	if (already == 0 && total == 0) {
		/* Nothing to download */
	} else if (already == total) {
		/* we are done downloading */
		if (!download_in_progress && need_line_feed) {
			g_print("\n");
		}
		g_print ("\rDownloading: 100 %%\n");
		download_in_progress = FALSE;
		need_line_feed = FALSE;
	} else if (already >= 0 && total > 0) {
		/* print progress */
		if (!download_in_progress && need_line_feed) {
			g_print("\n");
		}
		gint pct = (((gint64) already) * 100) / total;
		g_print("\rDownloading: %3d %%", pct);
		need_line_feed = TRUE;
		download_in_progress = TRUE;
	} else {
		/* what ?*/
		if (need_line_feed) {
			g_print("\n");
		}
		/* g_print("already: %d, total: %d\n", already, total); */
	}
}

static void
__operation_started_handler(DBusGProxy *proxy,
			const gchar* op,
			const gchar* pkg,
			const gchar* version,
			gpointer data)
{
    /* g_print("Operation started. Operation: %s, Package: %s, Version: %s\n", op, pkg, version); */
}

static void
__operation_progress_handler(DBusGProxy *proxy,
			const gchar* op,
			const gchar* pkg,
			const gchar* version,
			gint percentage,
			gpointer data)
{
	static gboolean operation_in_progress = FALSE;
	if (percentage >= 100) {
		/* we are done downloading */
		if (!operation_in_progress && need_line_feed) {
			g_print("\n");
		}
		g_print ("\r%s %s: 100 %%\n", op, pkg);
		operation_in_progress = FALSE;
		need_line_feed = FALSE;
	} else if (percentage >= 0) {
		/* print progress */
		if (!operation_in_progress && need_line_feed) {
			g_print("\n");
		}
		g_print("\r%s %s: %3d %%", op, pkg, percentage);
		need_line_feed = TRUE;
		operation_in_progress = TRUE;
	}
}

static void
__upgrades_available_handler(DBusGProxy *proxy,
			gpointer data)
{
}

static void
connect_signals(DBusGProxy *proxy)
{
	dbus_g_object_register_marshaller (__marshaller_VOID__STRING_STRING_STRING_INT_INT,
                                           G_TYPE_NONE,
                                           G_TYPE_STRING,
                                           G_TYPE_STRING,
                                           G_TYPE_STRING,
                                           G_TYPE_INT,
                                           G_TYPE_INT,
                                           G_TYPE_INVALID);
	dbus_g_object_register_marshaller (__marshaller_VOID__STRING_STRING_STRING_INT,
					G_TYPE_NONE,
					G_TYPE_STRING,
					G_TYPE_STRING,
					G_TYPE_STRING,
					G_TYPE_INT,
					G_TYPE_INVALID);
	dbus_g_object_register_marshaller (__marshaller_VOID__VOID,
					G_TYPE_NONE,
					G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy,
				"download_progress",
				G_TYPE_STRING,
				G_TYPE_STRING,
				G_TYPE_STRING,
				G_TYPE_INT,
				G_TYPE_INT,
				G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy,
				"operation_started",
				G_TYPE_STRING,
				G_TYPE_STRING,
				G_TYPE_STRING,
				G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy,
				"operation_progress",
				G_TYPE_STRING,
				G_TYPE_STRING,
				G_TYPE_STRING,
				G_TYPE_INT,
				G_TYPE_INVALID);
	dbus_g_proxy_add_signal(proxy,
				"package_list_updated",
				G_TYPE_INVALID);
	dbus_g_proxy_connect_signal (proxy, "download_progress",
				G_CALLBACK(__download_progress_handler),
				NULL, NULL);
	dbus_g_proxy_connect_signal (proxy, "operation_started",
				G_CALLBACK(__operation_started_handler),
				NULL, NULL);
	dbus_g_proxy_connect_signal (proxy, "operation_progress",
				G_CALLBACK(__operation_progress_handler),
				NULL, NULL);
	dbus_g_proxy_connect_signal (proxy, "package_list_updated",
				G_CALLBACK(__upgrades_available_handler),
				NULL, NULL);
}

static gboolean
list_available_packages(DBusGProxy *proxy, const gchar* category)
{
	GPtrArray *packages = NULL;
	GError *error = NULL;
	if(!com_nokia_package_manager_fetch_packages (proxy, category, &packages, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		print_packages_map(packages);
	}
	return TRUE;
}

static gboolean
list_installed_packages(DBusGProxy *proxy)
{
	GPtrArray *packages = NULL;
	GError *error = NULL;
	if(!com_nokia_package_manager_fetch_installed (proxy, &packages, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		print_packages_map(packages);
	}
	return TRUE;
}

typedef struct _ListAvailableUpdatesData {
	GPtrArray *packages;
} ListAvailableUpdatesData;

static void
list_available_updates_reply (DBusGProxy *proxy, GPtrArray *OUT_packages, GError *error, gpointer userdata)
{
	if (error) {
		print_error (error);
		g_error_free (error);
		exit (1);
	}
	else {
		print_packages_map (OUT_packages);
	}

/* 	g_free (OUT_packages); */
	if (loop)
		g_main_loop_quit (loop);
}

static gboolean
list_available_updates (DBusGProxy *proxy, const gchar* category)
{
/* 	ListAvailableUpdatesData *data = g_new0 (ListAvailableUpdatesData, 1); */
/* 	GPtrArray *packages = NULL; */
/* 	GError *error = NULL; */
	if (!com_nokia_package_manager_fetch_available_updates_async (proxy,
								      category,
								      &list_available_updates_reply,
								      NULL))
	{
		__print ("Failed to make async call!\n");
		exit (2);
		return TRUE;
	}
	return FALSE;
}

/*
static gboolean
list_catalogues(DBusGProxy *proxy)
{
	GPtrArray *catalogues = NULL;
	GError *error = NULL;
	if(!com_nokia_package_manager_fetch_catalogues (proxy, &catalogues, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		print_catalogues_map(catalogues);
	}
	return TRUE;
}
*/

typedef struct _InstallFileData {
	gchar *filename;
} InstallFileData;

static void
install_file_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	InstallFileData* data = (InstallFileData*)userdata;
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("%s installed successfully!\n", data->filename);
	}
	g_free (data);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
install_file(DBusGProxy *proxy, const gchar* filename)
{
	InstallFileData* data = g_new0(InstallFileData, 1);
	data->filename = g_strdup(filename);
	if (!com_nokia_package_manager_install_file_async (proxy,
									filename,
									&install_file_reply,
									data)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

typedef struct _InstallData {
	const gchar *pkgname;
	const gchar *version;
} InstallData;

static void
refresh_package_cache_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("Cache refreshed successfully!\n");
	}
	g_free (userdata);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
refresh_package_cache(DBusGProxy *proxy, const char *domain, const char *flags)
{
	if (!com_nokia_package_manager_refresh_package_cache_async (proxy,
                                                                    domain,
                                                                    flags,
                                                                    &refresh_package_cache_reply,
                                                                    NULL)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

static void
ignore_update_package_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	InstallData* data = (InstallData*)userdata;
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("Package '%s' added to ignore list!\n", data->pkgname);
	}
	g_free (data);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
ignore_update_package(DBusGProxy *proxy, const gchar* pkgname, const gchar *version)
{
	InstallData* data = g_new0(InstallData, 1);
	data->pkgname = pkgname;
	data->version = version;
	if (!com_nokia_package_manager_ignore_update_package_async (proxy,
						      pkgname,
						      version,
						      &ignore_update_package_reply,
						      data)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

static void
install_package_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	InstallData* data = (InstallData*)userdata;
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("Package '%s' installed successfully!\n", data->pkgname);
	}
	g_free (data);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
install_package(DBusGProxy *proxy, const gchar* pkgname, const gchar *version)
{
	InstallData* data = g_new0(InstallData, 1);
	data->pkgname = pkgname;
	data->version = version;
	if (!com_nokia_package_manager_install_async (proxy,
						      pkgname,
						      version,
						      &install_package_reply,
						      data)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

typedef struct _UpgradeData {
	const gchar *pkgname;
} UpgradeData;

static void
upgrade_package_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	UpgradeData* data = (UpgradeData*)userdata;
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("Package '%s' upgraded successfully!\n", data->pkgname);
	}
	g_free (data);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
upgrade_package(DBusGProxy *proxy, const gchar* pkgname, const gchar *version)
{
	UpgradeData* data = g_new0(UpgradeData, 1);
	data->pkgname = pkgname;
	if (!com_nokia_package_manager_upgrade_async (proxy,
						      pkgname,
						      version,
						      &upgrade_package_reply,
						      data)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

static void
upgrade_all_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
/* 	UpgradeData* data = (UpgradeData*)userdata; */
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("All packages upgraded successfully!\n");
	}
/* 	g_free (data); */
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
upgrade_all(DBusGProxy *proxy, gboolean allow_system_upgrade)
{
/* 	UpgradeData* data = g_new0(UpgradeData, 1); */
/* 	data->pkgname = pkgname; */
	if (!com_nokia_package_manager_upgrade_all_async (proxy,
							  allow_system_upgrade,
							  &upgrade_all_reply,
							  NULL)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

static void
download_system_upgrade_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	UpgradeData* data = (UpgradeData*)userdata;
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("Package '%s' downloaded successfully!\n", data->pkgname);
	}
	g_free (data);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
download_system_upgrade(DBusGProxy *proxy, const gchar* pkgname, const gchar *version)
{
	UpgradeData* data = g_new0(UpgradeData, 1);
	data->pkgname = pkgname;
	if (!com_nokia_package_manager_download_system_upgrade_async (proxy,
						      pkgname,
						      version,
						      &download_system_upgrade_reply,
						      data)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

static void
upgrade_system_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	UpgradeData* data = (UpgradeData*)userdata;
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("Package '%s' upgraded successfully!\n", data->pkgname);
	}
	g_free (data);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
upgrade_system(DBusGProxy *proxy, const gchar* pkgname, const gchar *version)
{
	UpgradeData* data = g_new0(UpgradeData, 1);
	data->pkgname = pkgname;
	if (!com_nokia_package_manager_upgrade_system_async (proxy,
						      pkgname,
						      version,
						      &upgrade_system_reply,
						      data)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

typedef struct _UninstallData {
	const gchar *pkgname;
} UninstallData;

static void
uninstall_package_reply(DBusGProxy *proxy, GError *error, gpointer userdata)
{
	UninstallData* data = (UninstallData*)userdata;
	if (error) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		__print("Package '%s' uninstalled successfully!\n", data->pkgname);
	}
	g_free (data);
	if (loop) {
		g_main_loop_quit (loop);
	}
}

static gboolean
uninstall_package(DBusGProxy *proxy, const gchar* pkgname)
{
	UninstallData* data = g_new0(UninstallData, 1);
	data->pkgname = pkgname;
	if (!com_nokia_package_manager_uninstall_async (proxy,
								pkgname,
								&uninstall_package_reply,
								data)) {
		__print("Failed to make async call!\n");
		exit(2);
		return TRUE;
	}
	return FALSE;
}

static gboolean
file_details (DBusGProxy *proxy, const gchar* filename)
{
	GError *error = NULL;
	GHashTable *map = g_hash_table_new(g_str_hash, g_str_equal);
	if (!com_nokia_package_manager_fetch_package_data_file (proxy, filename, &map, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		print_package_map (map);
	}
	return TRUE;
}

static gboolean
package_details (DBusGProxy *proxy, const gchar* name, const  gchar* version)
{
	GError *error = NULL;
	GHashTable *map = g_hash_table_new(g_str_hash, g_str_equal);
	if (!com_nokia_package_manager_fetch_package_data (proxy, name, version, &map, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		print_package_map (map);
	}
	return TRUE;
}

static gboolean
package_files (DBusGProxy *proxy, const gchar* name)
{
	GError *error = NULL;
	char **buffer;

	if (!com_nokia_package_manager_fetch_files_of_installed_package(proxy, name, &buffer, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {

	    int i;
	    for(i=0;buffer[i];i++)
	    {
		g_print("%s\n",buffer[i]);
		//g_free(buffer[i]);
	    }
	    g_strfreev(buffer);

//		print_package_map (map);
	}
	return TRUE;
}

static gboolean
package_dep_details (DBusGProxy *proxy, const gchar* name, const  gchar* version)
{
	GError *error = NULL;
	GPtrArray *packages = NULL;
	int i;

	if (!com_nokia_package_manager_fetch_package_dependency_data (proxy, name, version, &packages, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
		for (i = 0; i < packages->len; ++i)
			print_package_map ((GHashTable *) g_ptr_array_index (packages, i));
	}
	return TRUE;
}

/*
static gboolean
upgrade_info (DBusGProxy *proxy, const gchar* name, const  gchar* version)
{
	GError *error = NULL;
	GHashTable *data = g_hash_table_new(g_str_hash, g_str_equal);

	if (!com_nokia_package_manager_fetch_upgrade_info (proxy, name, version, &data, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
        print_data_map( data );
	}
	
	return TRUE;
}
*/

/*
static gboolean
add_catalogue (DBusGProxy *proxy,
	       const gchar* name,
	       const gchar* uri,
	       const gchar* dist,
	       const gchar* comps)
{
	GError *error = NULL;
	if (!com_nokia_package_manager_add_catalogue (proxy, name, uri, dist, comps, &error)) {
		print_error (error);
		g_error_free (error);
		exit (1);
	} else {
		// show nothing?
	}
	return TRUE;
}

static gboolean
remove_catalogue (DBusGProxy *proxy,
		  const gchar* name)
{
	GError *error = NULL;
	if (!com_nokia_package_manager_remove_catalogue (proxy, name, &error)) {
		print_error (error);
		g_error_free (error);
		exit (1);
	} else {
		// show nothing?
	}
	return TRUE;
}

static gboolean
enable_catalogue (DBusGProxy *proxy,
		  const gchar* name,
		  gboolean enable)
{
	GError *error = NULL;
	if (!com_nokia_package_manager_enable_catalogue (proxy, name, enable, &error)) {
		print_error (error);
		g_error_free (error);
		exit (1);
	} else {
		// show nothing?
	}
	return TRUE;
}*/

static DBusGProxy*
connect_dbus ()
{
    GError *error;
    DBusGProxy *proxy;
    DBusGConnection* conn;
    error = NULL;
    // connect to the bus
    conn = dbus_g_bus_get(DBUS_BUS_SYSTEM, &error);

    if (conn == NULL) {
	    __print ("Failed to open connection to bus: %s\n",
		    error->message);
	    g_error_free (error);
	    exit(3);
	    return NULL;
    }
    proxy = dbus_g_proxy_new_for_name (conn,
				    PKGMGR_DBUS_SERVICE,
				    PKGMGR_DBUS_PATH,
				    PKGMGR_DBUS_INTERFACE);

    dbus_g_proxy_set_default_timeout (proxy, 24*60*60*1000);
    connect_signals(proxy);
    return proxy;
}
static gboolean
cmd_list_available(int argc, char* argv[])
{
	static const gchar* category = NULL;

	static GOptionEntry entries[] = {
		{ "category", 'c', 0, G_OPTION_ARG_STRING, &category, "Category to list", "<category>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- list available packages");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!category) {
		category = "";
	}
	DBusGProxy *proxy = connect_dbus ();
	list_available_packages (proxy, category);
	return TRUE;
}
static gboolean
cmd_list_installed(int argc, char* argv[])
{
	static GOptionEntry entries[] = {
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- list installed packages");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}
	DBusGProxy *proxy = connect_dbus();
    return list_installed_packages(proxy);
}
static gboolean
cmd_list_updates(int argc, char* argv[])
{
	static const gchar* category = NULL;
	static gboolean refresh = FALSE;

	static GOptionEntry entries[] = {
		{ "category", 'c', 0, G_OPTION_ARG_STRING, &category, "Category to list", "<category>" },
		{ "refresh", 'r', 0, G_OPTION_ARG_NONE, &refresh, "Refresh updates from network", NULL },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- list available updates");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!category) {
		category = "";
	}
	DBusGProxy *proxy = connect_dbus ();
	return list_available_updates (proxy, category);
}

/*
static gboolean
cmd_list_catalogues(int argc, char* argv[])
{
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- list catalogues");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}
	DBusGProxy *proxy = connect_dbus();
	return list_catalogues(proxy);
}
*/

static gboolean
cmd_install_file(int argc, char* argv[])
{
	static const gchar* filename = NULL;
	static GOptionEntry entries[] = {
		{ "file", 'f', 0, G_OPTION_ARG_FILENAME, &filename, "Debian package to install", "<file>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;
	gchar *abs_filename = NULL;

	context = g_option_context_new ("- install standalone package");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!filename) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	DBusGProxy *proxy = connect_dbus ();
	abs_filename = get_absolute_filename(filename);
	gboolean res = install_file(proxy, abs_filename);
	g_free (abs_filename);
	return res;
}

static gboolean
cmd_ignore_update_package(int argc, char* argv[])
{
    static const gchar* pkgname = NULL;
	static const gchar* version = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to add into the ignore list", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &version, "Package version to add into the ignore list", "<version>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- ignore a package update");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	DBusGProxy *proxy = connect_dbus ();
	return ignore_update_package(proxy, pkgname, version);
}

static gboolean
cmd_refresh_cache(int argc, char* argv[])
{
    static const gchar* domain = NULL;
    static gboolean clear_ignored = FALSE;
    static gboolean automatic = FALSE;
	static GOptionEntry entries[] = {
		{ "clear-ignored", 'c', 0, G_OPTION_ARG_NONE, &clear_ignored, "Clear an ignored list before refreshing", NULL },
		{ "automatic", 'a', 0, G_OPTION_ARG_NONE, &automatic, "Do not try to bring up the network", NULL },
		{ "domain", 'd', 0, G_OPTION_ARG_STRING, &domain, "Refresh only the given domain", "<domain>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- refresh a cache");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	DBusGProxy *proxy = connect_dbus ();
        char *flags = make_flag_string ("clearignored", clear_ignored,
                                        "automatic", automatic,
                                        NULL);
        fprintf (stderr, "flags: %s\n", flags);
        gboolean res = refresh_package_cache(proxy, domain, flags);
        free (flags);
        return res;
}

static gboolean
cmd_install(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static const gchar* version = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to install", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &version, "Package version to install", "<version>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- install a package");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	DBusGProxy *proxy = connect_dbus ();
	return install_package(proxy, pkgname, version);
}
static gboolean
cmd_upgrade(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static const gchar* version = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to upgrade", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &version, "Package version to install", "<version>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- upgrade a package");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	DBusGProxy *proxy = connect_dbus ();
	return upgrade_package(proxy, pkgname, version);
}

static gboolean
cmd_upgrade_all(int argc, char* argv[])
{
    static gboolean allow_sys_upgrade = FALSE;
	static GOptionEntry entries[] = {
		{ "allow-sys-upgrade", 'a', 0, G_OPTION_ARG_NONE, &allow_sys_upgrade, "Allow system upgrade", NULL },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- upgrade all packages");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	DBusGProxy *proxy = connect_dbus ();
	return upgrade_all(proxy, allow_sys_upgrade);
}

static gboolean
cmd_download_system_upgrade(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static const gchar* version = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to download", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &version, "Package version to download", "<version>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- download a system package");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	DBusGProxy *proxy = connect_dbus ();
	return download_system_upgrade(proxy, pkgname, version);
}

static gboolean
cmd_upgrade_system(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static const gchar* version = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "System package to upgrade", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &version, "System package version to install", "<version>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- upgrade a system package");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	DBusGProxy *proxy = connect_dbus ();
	return upgrade_system(proxy, pkgname, version);
}

static gboolean
cmd_uninstall(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to remove", "<pkgname>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- remove a package");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_add_main_entries (context, common_opts, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	DBusGProxy *proxy = connect_dbus ();
	return uninstall_package(proxy, pkgname);
}
static gboolean
cmd_file_details(int argc, char* argv[])
{
	static const gchar* filename = NULL;
	static GOptionEntry entries[] = {
		{ "file", 'f', 0, G_OPTION_ARG_FILENAME, &filename, "Debian package to show details from", "<file>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;
	gchar *abs_filename = NULL;

	context = g_option_context_new ("- show package details");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!filename) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}

	// default to verbose
	verbose = TRUE;

	DBusGProxy *proxy = connect_dbus ();
	abs_filename = get_absolute_filename(filename);
	file_details(proxy, abs_filename);
	g_free (abs_filename);
	return TRUE;
}
static gboolean
cmd_package_details(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static const gchar* pkgversion = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to show details from", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &pkgversion, "Package version to show details from", "<pkgversion>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- show package details");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname || !pkgversion) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return package_details(proxy, pkgname, pkgversion);
}
static gboolean
cmd_list_package(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to show details from", "<pkgname>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- show package files");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return package_files(proxy, pkgname);
}

static gboolean
cmd_package_dep_details(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static const gchar* pkgversion = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package to show details from", "<pkgname>" },
		/* { "version", 'v', 0, G_OPTION_ARG_STRING, &pkgversion, "Package version to show details from", "<pkgversion>" }, */
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- show package details with dependencies");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname /* || !pkgversion */) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return package_dep_details(proxy, pkgname, pkgversion);
}

/*
static gboolean
cmd_upgrade_info(int argc, char* argv[])
{
	static const gchar* pkgname = NULL;
	static const gchar* pkgversion = NULL;
	static GOptionEntry entries[] = {
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Upgrade package to show info from", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &pkgversion, "Upgrade package version to show info from", "<pkgversion>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- show upgrade package info");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!pkgname || !pkgversion ) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return upgrade_info(proxy, pkgname, pkgversion);
}
*/

/*
static gboolean
cmd_add_catalogue(int argc, char *argv[])
{
	static const gchar *name = NULL;
	static const gchar *uri = NULL;
	static const gchar *dist = NULL;
	static const gchar *comps = NULL;
	static GOptionEntry entries[] = {
		{ "name", 'n', 0, G_OPTION_ARG_STRING, &name, "Catalogue name", "<name>" },
		{ "uri", 'u', 0, G_OPTION_ARG_STRING, &uri, "Catalogue URI", "<uri>" },
		{ "dist", 'd', 0, G_OPTION_ARG_STRING, &dist, "Catalogue distribution", "<dist>" },
		{ "comps", 'c', 0, G_OPTION_ARG_STRING, &comps, "Catalogue components", "<comps>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- add catalogue");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!name || !uri || !dist || !comps) {
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return add_catalogue (proxy, name, uri, dist, comps);
}

static gboolean
cmd_remove_catalogue(int argc, char *argv[])
{
	static const gchar *name = NULL;
	static GOptionEntry entries[] = {
		{ "name", 'n', 0, G_OPTION_ARG_STRING, &name, "Catalogue name", "<name>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- remove catalogue");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!name) {
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return remove_catalogue (proxy, name);
}

static gboolean
cmd_enable_catalogue(int argc, char *argv[])
{
	static const gchar *name = NULL;
	static GOptionEntry entries[] = {
		{ "name", 'n', 0, G_OPTION_ARG_STRING, &name, "Catalogue name", "<name>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- enable catalogue");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!name) {
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return enable_catalogue (proxy, name, TRUE);
}

static gboolean
cmd_disable_catalogue(int argc, char *argv[])
{
	static const gchar *name = NULL;
	static GOptionEntry entries[] = {
		{ "name", 'n', 0, G_OPTION_ARG_STRING, &name, "Catalogue name", "<name>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- disable catalogue");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!name) {
		__print ("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
	return enable_catalogue (proxy, name, FALSE);
}
*/

static void
print_field (GHashTable *op, const char *field)
{
	GValue *val = g_hash_table_lookup (op, field);
	if (val) {
		const gchar *str = g_value_get_string (val);
		g_print ("%s: %s\n", field, str);
	}
}

static void
print_field_int (GHashTable *op, const char *field)
{
	GValue *val = g_hash_table_lookup (op, field);
	if (val) {
		int i = g_value_get_int (val);
		g_print ("%s: %d\n", field, i);
	}
}

static void
print_field_int64 (GHashTable *op, const char *field)
{
	GValue *val = g_hash_table_lookup (op, field);
	if (val) {
		gint64 i = g_value_get_int64 (val);
		g_print ("%s: %Ld\n", field, i);
	}
}

static void
print_current_operation_data (GHashTable *op)
{
	GValue *val = g_hash_table_lookup (op, "Operation");
	if (!val)
		g_print ("No current operation\n");
	else {
		print_field (op, "Operation");
		print_field (op, "Name");
		print_field (op, "Version");
		print_field_int64 (op, "DownloadCurrentBytes");
		print_field_int64 (op, "DownloadTotalBytes");
		print_field_int (op, "OperationProgressPct");
	}
}

static gboolean
cmd_current_operation (int argc, char *argv[])
{
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();

	GError *error = NULL;
	GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);
	if (!com_nokia_package_manager_fetch_current_operation_data (proxy, &map, &error)) {
		print_error (error);
		g_error_free (error);
		exit (1);
	} else {
		print_current_operation_data (map);
	}
	return TRUE;
}

static gboolean
cmd_simulate_operation(int argc, char* argv[])
{
	static const gchar* operation = NULL;
	static const gchar* pkgname = NULL;
	static const gchar* pkgversion = NULL;
	static const gchar* flags = NULL;
	static GOptionEntry entries[] = {
	    { "operation", 'o', 0, G_OPTION_ARG_STRING, &operation, "Operation to simulate", "<operation>" },
		{ "package", 'p', 0, G_OPTION_ARG_STRING, &pkgname, "Package", "<pkgname>" },
		{ "version", 'v', 0, G_OPTION_ARG_STRING, &pkgversion, "Package version", "<pkgversion>" },
		{ "flags", 'f', 0, G_OPTION_ARG_STRING, &pkgversion, "", "Flags <flags>" },
		{ NULL }
	};
	GError *error = NULL;
	GOptionContext *context;

	context = g_option_context_new ("- simulates a specified function");
	g_option_context_add_main_entries (context, entries, "pkgmgr");
	g_option_context_set_help_enabled (context, TRUE);
	if (!g_option_context_parse (context, &argc, &argv, &error))
	{
		__print ("option parsing failed: %s\n", error->message);
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit (1);
	}

	if (!operation || !pkgname || !pkgversion) {
		__print("%s", g_option_context_get_help (context, TRUE, NULL));
		exit(1);
	}
	if ( !flags ) {
	    flags = g_strdup("");
	}
	// default to verbose
	verbose = TRUE;
	DBusGProxy *proxy = connect_dbus ();
    GHashTable *result = NULL;
    
	if (!com_nokia_package_manager_simulate_operation (proxy, operation, pkgname, pkgversion, flags, &result, &error)) {
		print_error(error);
		g_error_free (error);
		exit(1);
	} else {
        print_operation_map ( result );
	}
	
	return TRUE;
}

static Command commands[] =
{
	{"list-installed", &cmd_list_installed},
	{"list-updates", &cmd_list_updates},
	{"list-available", &cmd_list_available},
/*	{"list-catalogues", &cmd_list_catalogues},*/
	{"install-file", &cmd_install_file},
	{"file-details", &cmd_file_details},
	{"package-details", &cmd_package_details},
	{"package-dep-details", &cmd_package_dep_details},
//	{"upgrade-info", &cmd_upgrade_info},
	{"list-package",&cmd_list_package},
	{"ignore-update-package",&cmd_ignore_update_package},
	{"refresh-cache",&cmd_refresh_cache},
	{"install", &cmd_install},
	{"uninstall", cmd_uninstall},
	{"upgrade", &cmd_upgrade},
	{"upgrade-all", &cmd_upgrade_all},
/*	{"add-catalogue", &cmd_add_catalogue},
	{"remove-catalogue", &cmd_remove_catalogue},
	{"enable-catalogue", &cmd_enable_catalogue},
	{"disable-catalogue", &cmd_disable_catalogue},*/
	{"download-system-upgrade", &cmd_download_system_upgrade},
	{"upgrade-system", &cmd_upgrade_system},
	{"current-operation", &cmd_current_operation},
	{"simulate-operation", &cmd_simulate_operation},
	{NULL,}
};

static void
print_usage(const gchar *name)
{
	__print("Usage: %s <command> <options>\n"
		"\n"
		"Commands:\n"
		,name);
	Command *cmds  = commands;
	while(cmds->name) {
		__print("  %s\n", cmds->name);
		cmds++;
	}
	__print("\n");
}


int main(int argc, char* argv[])
{
	const gchar *name = argv[0];
	g_type_init();
	Command *cmds  = commands;
	CommandFunc func = NULL;
	gchar* cmd_str = argv[1];
	if (argc < 2
		|| strcmp (cmd_str, "--help") == 0
		|| strcmp (cmd_str, "-h") == 0) {
		print_usage(argv[0]);
		return 0;
	}
	while (cmds->name) {
/* 		__print("%s\n", cmds->name); */
		if (strcmp(cmd_str, cmds->name) == 0) {
			func = cmds->func;
			break;
		}
		cmds++;
	}
	argc--;
	argv++;
	if (func) {
		loop = g_main_loop_new (NULL, FALSE);
		if (!func(argc, argv)) {
			g_main_loop_run (loop);
		}
		g_main_loop_unref (loop);
		loop = NULL;
	} else {
		__print("Unknown command: '%s'\n\n", cmd_str);
		print_usage (name);
		exit(1);
	}
	return 0;
}

