/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#define _GNU_SOURCE

#include <stdlib.h>
#include <string.h>

#include "package.h"
#include "catalogue.h"
#include "utils.h"
#include "glib-util.h"

extern gboolean verbose;
extern gboolean need_line_feed;

static void
print_map_value_error (GValue *val, const gchar* key, GType wanted)
{
	g_print("Invalid value in map (%s): Should be %s but was %s\n"
		, key, g_type_name(wanted), G_VALUE_TYPE_NAME(val));
}

gchar*
get_absolute_filename(const gchar* filename)
{
	if (g_path_is_absolute(filename)) {
		return g_strdup(filename);
	} else {
		gchar *cur_dir = g_get_current_dir();
		gchar *abs_filename = g_build_filename(cur_dir, filename, NULL);
		g_free (cur_dir);
		return abs_filename;
	}
}
static gchar*
__get_map_string(GHashTable *map, const gchar* key)
{
	GValue* val = g_hash_table_lookup (map, key);
	if (!val) {
		return NULL;
	}
	if (!G_VALUE_HOLDS_STRING(val) ) {
		print_map_value_error (val, key, G_TYPE_STRING);
		exit(1);
	}
	return g_value_dup_string(val);
}
static gint64
__get_map_int64(GHashTable *map, const gchar* key)
{
	GValue* val = g_hash_table_lookup (map, key);
	if (!val) {
		return 0;
	}
	if (!G_VALUE_HOLDS_INT64(val) ) {
		print_map_value_error (val, key, G_TYPE_INT64);
		exit(1);
	}
	return g_value_get_int64(val);
}

/* static gint */
/* __get_map_int(GHashTable *map, const gchar* key) */
/* { */
/* 	GValue* val = g_hash_table_lookup (map, key); */
/* 	if (!val) { */
/* 		return 0; */
/* 	} */
/* 	if (!G_VALUE_HOLDS_INT(val) ) { */
/* 		print_map_value_error (val, key, G_TYPE_INT); */
/* 		exit(1); */
/* 	} */
/* 	return g_value_get_int(val); */
/* } */

static gboolean
__get_map_boolean(GHashTable *map, const gchar* key)
{
	GValue* val = g_hash_table_lookup (map, key);
	if (!val) {
		return FALSE;
	}
	if (!G_VALUE_HOLDS_BOOLEAN(val) ) {
		print_map_value_error (val, key, G_TYPE_BOOLEAN);
		exit(1);
	}
	return g_value_get_boolean(val);
}
static PkgType
__package_type_string_to_enum(const gchar* status)
{
	PkgType s = PKG_TYPE_UNKNOWN;
	if (g_strcmp0(status, "Installed") == 0) {
		s = PKG_TYPE_INSTALLED;
	} else if (g_strcmp0(status, "NotInstalled") == 0) {
		s = PKG_TYPE_NOT_INSTALLED;
	} else if (g_strcmp0(status, "Update") == 0) {
		s = PKG_TYPE_UPDATE;
	} else if (g_strcmp0(status, "OlderThanInstalled") == 0) {
		s = PKG_TYPE_OLDER_THAN_INSTALLED;
	}
	return s;
}
static Package*
__unpack_package_map(GHashTable *map)
{
	Package *pkg = package_new();
	pkg->name = __get_map_string(map, "Name");
/* 	pkg->is_broken = __get_boolean(a, 1); */
	pkg->version = __get_map_string(map, "Version");
	pkg->download_size = __get_map_int64(map, "Size");
	pkg->installed_size = __get_map_int64(map, "InstalledSize");
	pkg->installed_timestamp = __get_map_int64(map, "InstalledTimestamp");
	pkg->available_timestamp = __get_map_int64(map, "AvailableTimestamp");
	pkg->display_name = __get_map_string(map, "DisplayName");
	pkg->description = __get_map_string(map, "Description");
	pkg->category = __get_map_string(map, "Category");
	pkg->repository = __get_map_string(map, "Repository");
	pkg->icon_data = __get_map_string(map, "IconData");
	gchar *type = __get_map_string(map, "Type");
	pkg->type = __package_type_string_to_enum(type);
	g_free (type);
	pkg->flag_reboot = __get_map_boolean(map, "FlagReboot");
	pkg->flag_ignored = __get_map_boolean(map, "FlagIgnored");
	pkg->flag_system_update = __get_map_boolean(map, "FlagSystemUpdate");
	pkg->flag_local_file = __get_map_boolean(map, "FlagLocalFile");
	return pkg;
}
static gchar *__get_short_desc(gchar *desc)
{
	gchar *end;
	gint len = 80;
	if (!desc) {
		return NULL;
	}
	end = g_strstr_len(desc, len, "\n");
	if (end) {
		len = end-desc;
	}
	return g_strndup(desc, len);
}
static const gchar*
__boolean_str(gboolean b)
{
	return b ? "True" : "False";
}
static void
print_package(Package *pkg)
{
	if (verbose) {
		g_print("Package: %s (%s)\n", pkg->name, pkg->version);
		if( pkg->display_name && strnlen(pkg->display_name, 1) > 0 ) {
			g_print("DisplayName: %s\n", pkg->display_name);
		}
		g_print("Type: %s\n", package_type_to_string(pkg->type));
		g_print("Repository: %s\n", pkg->repository);
		g_print("Category: %s\n", pkg->category);
		g_print("Description: %s\n", pkg->description);
		g_print("Size: %"G_GINT64_FORMAT"\n", pkg->download_size);
		g_print("InstalledSize: %"G_GINT64_FORMAT"\n", pkg->installed_size);
		g_print("InstalledTimestamp: %ld\n", pkg->installed_timestamp);
		g_print("AvailableTimestamp: %ld\n", pkg->available_timestamp);
		if (pkg->flag_reboot) {
			g_print("FlagReboot: %s\n", __boolean_str(pkg->flag_reboot));
		}
		if (pkg->flag_ignored) {
			g_print("FlagIgnored: %s\n", __boolean_str(pkg->flag_ignored));
		}
		if (pkg->flag_system_update) {
			g_print("FlagSystemUpdate: %s\n", __boolean_str(pkg->flag_system_update));
		}
		if (pkg->flag_local_file) {
			g_print("FlagLocalFile: %s\n", __boolean_str(pkg->flag_local_file));
		}
/* 		g_print("IconData: %s\n", pkg->icon_data); */
		/* g_print("\n"); */
	} else {
		gchar *short_desc = __get_short_desc(pkg->description);
		if (short_desc) {
			g_print("%s (%s) - %s\n",pkg->name, pkg->version, short_desc);
			g_free (short_desc);
		} else {
			g_print("%s (%s)\n",pkg->name, pkg->version);
		}
	}
}

static const char *
_find_string (const char *strings[], const char *s)
{
	int i;
	for (i = 0; strings[i] != 0; ++i)
	{
		if (!strcmp (strings[i], s))
			return strings[i];
	}

	return 0;
}

static void
_print_strv(GValue *v)
{
	gchar **sv = (gchar **) g_value_get_boxed (v);

	if (!*sv)
		return;

	g_print("%s", *sv);

	for (++sv; *sv != NULL; ++sv) {
		g_print(" %s", *sv);
	}
}

static void
_print_additional_fields (GHashTable *map)
{
	static const char *skip_fields[] = {
		"Category",
		"Description",
		"DisplayName",
		"FlagReboot",
		"FlagIgnored",
		"FlagSystemUpdate",
		"FlagLocalFile",
		"IconData",
		"InstalledSize",
		"InstalledTimestamp",
		"AvailableTimestamp",
		"IsBroken",
		"Name",
		"Repository",
		"Type",
		"Version",
		"ErrorName",
		"ErrorMessage",
		"DependsSystemUpgradeVersion",
		"DependsSystemUpgradeName",
		0
	};

	static const char *int64_fields[] = {
	    "TotalInstalledSize",
	    "DownloadSize",
	    "Size",
	    0
	};

	static const char *strv_fields[] = {
	    "InstallDeps",
	    "UpgradeDeps",
	    "RemovePkgs",
	    0
	};

	GHashTableIter iter;
	gpointer key, value;
	const gchar *str;
	gint64 i;

	g_hash_table_iter_init (&iter, map);

	while (g_hash_table_iter_next (&iter, &key, &value))
	{
		if (!_find_string (skip_fields, (const char *) key)) {
		    if ( _find_string(int64_fields, (const char *) key) ) {
		        i = g_hash_table_get_int64_value(map, (const gchar *) key);
		        g_print("%s: %lld\n", (const gchar *) key, i);
		    } else if (_find_string(strv_fields, (const char *) key) ) {
			    g_print("%s: ", (const char *) key);
			    _print_strv((GValue *) g_hash_table_lookup (map, key));
			    g_print("\n");

		    } else {
    			str = g_hash_table_get_string_value (map,
    							     (const gchar *) key);

    			g_print ("%s: %s\n",
    				 (const gchar *) key,
    				 str);
            }
		}
	}
}

void
print_package_map(GHashTable *map)
{
	Package *pkg = __unpack_package_map (map);
	print_package (pkg);
	package_free (pkg);

	_print_additional_fields (map);
	g_print ("\n");

	/* if (verbose) { */
	/* } */
	/* else { */
	/* 	gchar *desc = g_hash_table_dup_string_value (map, */
	/* 						     "Description"); */

	/* 	gchar *short_desc = __get_short_desc (desc); */
	/* 	if (short_desc) { */
	/* 		g_print("%s (%s) - %s\n",pkg->name, pkg->version, short_desc); */
	/* 		g_free (short_desc); */
	/* 	} else { */
	/* 		g_print("%s (%s)\n", pkg->name, pkg->version); */
	/* 	} */
	/* 	g_free (desc); */
	/* } */
}

static void
print_strv_if_exists (GHashTable *map, const char *key)
{
  gpointer orig_key, value;
  if (g_hash_table_lookup_extended (map, key, &orig_key, &value))
    {
      g_print (" %s: ", key);
      _print_strv ((GValue *) value);
      g_print ("\n");
    }
}

void
print_operation_map (GHashTable *map)
{
    int download_size = __get_map_int64(map, "DownloadSize");
    int size = __get_map_int64(map, "Size");
    int install_size = __get_map_int64(map, "TotalInstalledSize");

    /* gchar *install_deps = __get_map_string(map, "InstallDeps"); */
    /* gchar *upgrade_deps = __get_map_string(map, "UpgradeDeps"); */
    /* gchar *remove_pkgs = __get_map_string(map, "RemovePkgs"); */
    gchar *error = __get_map_string(map, "Error");
    gchar *error_name = __get_map_string(map, "ErrorName");

    if ( error_name )
        g_print(" ErrorName: %s\n", error_name);
    if ( error )
        g_print(" Error: %s\n", error);

    print_strv_if_exists (map, "InstallUserDeps");
    print_strv_if_exists (map, "UpgradeUserDeps");
    print_strv_if_exists (map, "RemoveUserPkgs");
    print_strv_if_exists (map, "RevertUserDeps");
    print_strv_if_exists (map, "InstallDeps");
    print_strv_if_exists (map, "UpgradeDeps");
    print_strv_if_exists (map, "RemovePkgs");
    print_strv_if_exists (map, "RevertDeps");
    g_print(" DownloadSize: %d\n", download_size);
    g_print(" Size: %d\n", size);
    g_print(" TotalInstalledSize: %d\n", install_size);
    if (__get_map_boolean(map, "FlagTrusted"))
	g_print(" FlagTrusted: 1\n");
}

void
print_data_map(GHashTable *map)
{
	gboolean reboot = __get_map_boolean(map, "Reboot");

	int download_size = __get_map_int64(map, "DownloadSize");
	int size = __get_map_int64(map, "Size");
	int install_size = __get_map_int64(map, "TotalInstalledSize");

    gchar *install_deps = __get_map_string(map, "InstallDeps");
    gchar *upgrade_deps = __get_map_string(map, "UpgradeDeps");
    gchar *remove_pkgs = __get_map_string(map, "RemovePkgs");

    g_print(" Reboot: %s\n", reboot?"True":"False");
    g_print(" InstallDeps: %s\n", install_deps);
    g_print(" UpgradeDeps: %s\n", upgrade_deps);
    g_print(" RemovePkgs: %s\n", remove_pkgs);
    g_print(" DownloadSize: %d\n", download_size);
    g_print(" Size: %d\n", size);
    g_print(" TotalInstalledSize: %d\n", install_size);
}

void
print_packages_map(GPtrArray *packages)
{
	gint i=0;
	for(i=0; i < packages->len; i++) {
		GHashTable *map = packages->pdata[i];
		print_package_map (map);
	}
}

static Catalogue*
__unpack_catalogue_map(GHashTable *map)
{
	Catalogue *cat = catalogue_new();
	cat->name = __get_map_string(map, "Name");
/* 	cat->is_broken = __get_boolean(a, 1); */
	cat->uri = __get_map_string(map, "URI");
	cat->dist = __get_map_string(map, "Dist");
	cat->components = __get_map_string(map, "Components");
	cat->enabled = __get_map_boolean(map, "Enabled");
	cat->essential = __get_map_boolean(map, "Essential");
	return cat;
}
static void
print_catalogue(Catalogue *cat)
{
	if (verbose) {
		g_print("Name: %s\n", cat->name);
		g_print("URI: %s\n", cat->uri);
		g_print("Dist: %s\n", cat->dist);
		g_print("Components: %s\n", cat->components);
		g_print("Enabled: %s\n", cat->enabled ? "True" : "False");
		g_print("Essential: %s\n", cat->essential ? "True" : "False");
		g_print("\n");
	} else {
		g_print("%s (%s)\n",cat->name, cat->enabled ? "Enabled" : "Disabled");
	}
}
void
print_catalogue_map(GHashTable *map)
{
	Catalogue *cat = __unpack_catalogue_map (map);
	print_catalogue(cat);
	catalogue_free(cat);
}

void
print_catalogues_map(GPtrArray *catalogues)
{
	gint i=0;
	for(i=0; i < catalogues->len; i++) {
		GHashTable *map = catalogues->pdata[i];
		Catalogue *cat = __unpack_catalogue_map(map);
		print_catalogue(cat);
		catalogue_free(cat);
	}
}
void
print_error(GError *error)
{
	if (need_line_feed)
		g_print("\n");

	if (error->domain == DBUS_GERROR && error->code == DBUS_GERROR_REMOTE_EXCEPTION) {
		g_printerr ("Error: %s\n", /*dbus_g_error_get_name (error),*/ error->message);
	} else {
		g_printerr ("Error: %s\n", error->message);
	}
}
