/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "algorithms.h"
#include "apt-worker-cache.h"
#include "problemresolver.h"

#include <apt-pkg/configuration.h>
#include <apt-pkg/debversion.h>
#include <apt-pkg/sptr.h>
#include <apt-pkg/version.h>
#include <list>

using std::list;
typedef pkgCache::DepIterator DepIterator;
typedef pkgCache::PkgIterator PkgIterator;
typedef pkgCache::VerIterator VerIterator;
typedef pkgCache::Version Version;

void ProtectInstallDeps(const pkgCache::PkgIterator &Pkg,
			ProblemResolver &Fix)
{
   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
   MarkPolicyDepCache &DCache = awc->cache->dep_cache();
   pkgDepCache::StateCache &P = DCache[Pkg];
   VerIterator Cand = P.InstVerIter(DCache);
   DepIterator Dep = Cand.DependsList();
   const bool Debug = _config->FindB("Debug::Algorithms");

   for (; Dep.end() != true;)
   {
      // Grok or groups
      DepIterator Start = Dep;

      // bool Result = true;
      unsigned Ors = 0;
      for (bool LastOR = true; Dep.end() == false && LastOR == true; Dep++,Ors++)
      {
	 LastOR = (Dep->CompareOp & pkgCache::Dep::Or) == pkgCache::Dep::Or;

	 if (Start->Type != pkgCache::Dep::PreDepends
	      && Start->Type != pkgCache::Dep::Depends)
	    continue;

	 if ((DCache[Dep] & pkgDepCache::DepInstall) == pkgDepCache::DepInstall)
	 {
	    PkgIterator P = Dep.TargetPkg();
	    // VerIterator Cand = awc->cache->get_candidate_ver(P, Start);
	    // if (Cand.end())
	    //    continue;

	    if (!DCache[P].Install())
	       continue;

	    if (Debug)
	       std::clog << "ProtectDeps " << pkgCache::DepType(Start->Type)
			 << ' ' << P.Name() << std::endl;

	    Fix.Protect(P);
	 }
      }
   }   
}

void MarkDepsWithBetterCandidates(const pkgCache::PkgIterator &Pkg,
				  ProblemResolver &Fix)
{
   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
   MarkPolicyDepCache &DCache = awc->cache->dep_cache();
   pkgDepCache::StateCache &P = DCache[Pkg];
   VerIterator Cand = P.InstVerIter(DCache);
   DepIterator Dep = Cand.DependsList();
   const bool Debug = _config->FindB("Debug::Algorithms");

   for (; Dep.end() != true;)
   {
      // Grok or groups
      DepIterator Start = Dep;

      // bool Result = true;
      unsigned Ors = 0;
      for (bool LastOR = true; Dep.end() == false && LastOR == true; Dep++,Ors++)
      {
	 LastOR = (Dep->CompareOp & pkgCache::Dep::Or) == pkgCache::Dep::Or;
	 if (Start->Type != pkgCache::Dep::PreDepends
	     && Start->Type != pkgCache::Dep::Depends)
	    continue;

	 if ((DCache[Dep] & pkgDepCache::DepInstall) == pkgDepCache::DepInstall)
	 {
	    PkgIterator P = Dep.TargetPkg();
	    VerIterator Cur = P.CurrentVer();
	    if (Cur.end())
	       continue;

	    VerIterator Cand = awc->cache->get_candidate_ver(P, Start);
	    if (Cand.end())
	       continue;

	    domain_t CurDom = awc->cache->get_current_domain(P);
	    domain_t CandDom = awc->cache->find_best_domain(Cand);

	    if (domain_get_trust_level(CandDom) > domain_get_trust_level(CurDom))
	    {
	       if (Debug)
		  std::clog << "MarkDepsWithBetterCandidates "
			    << P.Name() << ' '
			    << Cur.VerStr() << " (from '" << domain_get_name(CurDom) << "')"
			    << " => "
			    << Cand.VerStr() << " (from '" << domain_get_name(CandDom) << "')"
			    << std::endl;

	       DCache.SetCandidateVersion(Cand);
	       DCache.MarkInstall1(P, true, 0, false);
	       Fix.Protect(P);
	       // DCache[Dep] &= ~(pkgDepCache::DepInstall | pkgDepCache::DepCVer);
	    }
	 }
      }
   }

}

bool ForceDeps(const pkgCache::PkgIterator &Pkg, ProblemResolver &Fix)
{
   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
   MarkPolicyDepCache &DCache = awc->cache->dep_cache();
   pkgCache *Cache = &DCache.GetCache();
   list<pkgCache::Version *> Deps;
   pkgDepCache::StateCache &P = DCache[Pkg];
   VerIterator Cand = P.InstVerIter(DCache);
   DepIterator Dep = Cand.DependsList();
   const bool ForceImportantDeps = false;
   bool AllOk = true;
   const bool Debug = _config->FindB("Debug::Algorithms");

   // Mark dependencies with a better candidate version for reinstall.
   // (This is now done in MarkDepsWithBetterCandidates)

   // for (; Dep.end() != true;)
   // {
   //    // Grok or groups
   //    DepIterator Start = Dep;
   //    // bool Result = true;
   //    unsigned Ors = 0;
   //    for (bool LastOR = true; Dep.end() == false && LastOR == true; Dep++,Ors++)
   //    {
   // 	 LastOR = (Dep->CompareOp & pkgCache::Dep::Or) == pkgCache::Dep::Or;

   // 	 if ((DCache[Dep] & pkgDepCache::DepInstall) == pkgDepCache::DepInstall)
   // 	 {
   // 	    PkgIterator P = Dep.TargetPkg();
   // 	    VerIterator Cand = awc->cache->get_candidate_ver(P, Start);
   // 	    VerIterator Cur = P.CurrentVer();
   // 	    if (!Cur.end() && strcmp(Cand.VerStr(), Cur.VerStr()))
   // 	       DCache[Dep] &= ~(pkgDepCache::DepInstall | pkgDepCache::DepCVer);
   // 	 }
   //    }
   // }

   // Collect candidate versions of broken dependencies.
   Dep = Cand.DependsList();

   for (; Dep.end() != true;)
   {
      // Grok or groups
      DepIterator Start = Dep;
      bool Result = true;
      unsigned Ors = 0;
      for (bool LastOR = true; Dep.end() == false && LastOR == true; Dep++,Ors++)
      {
	 LastOR = (Dep->CompareOp & pkgCache::Dep::Or) == pkgCache::Dep::Or;

	 if ((DCache[Dep] & pkgDepCache::DepInstall) == pkgDepCache::DepInstall)
	    Result = false;
      }
      
      // Dep is satisfied okay.
      if (Result == false)
	 continue;

      /* Check if this dep should be consider for install. If it is a user
         defined important dep and we are installed a new package then 
	 it will be installed. Otherwise we only check for important
         deps that have changed from the installed version
      */
      if (DCache.IsImportantDep(Start) == false)
	 continue;
      
      /* Check if any ImportantDep() (but not Critical) were added
       * since we installed the package.  Also check for deps that
       * were satisfied in the past: for instance, if a version
       * restriction in a Recommends was tightened, upgrading the
       * package should follow that Recommends rather than causing the
       * dependency to be removed. (bug #470115)
       */
      bool isNewImportantDep = false;
      bool isPreviouslySatisfiedImportantDep = false;
      if(!ForceImportantDeps && !Start.IsCritical())
      {
	 bool found=false;
	 VerIterator instVer = Pkg.CurrentVer();
	 if(!instVer.end())
	 {
	   for (DepIterator D = instVer.DependsList(); D.end() != true; D++)
	     {
	       //FIXME: deal better with or-groups(?)
	       DepIterator LocalStart = D;

	       if(DCache.IsImportantDep(D) && !D.IsCritical() &&
		  Start.TargetPkg() == D.TargetPkg())
		 {
		   if(!isPreviouslySatisfiedImportantDep)
		     {
		       DepIterator D2 = D;
		       while((D2->CompareOp & pkgCache::Dep::Or) != 0)
			 ++D2;

		       isPreviouslySatisfiedImportantDep =
			  ((DCache[D2] & pkgDepCache::DepGNow) != 0);
		     }

		   found=true;
		 }
	     }
	    // this is a new dep if it was not found to be already
	    // a important dep of the installed pacakge
	    isNewImportantDep = !found;
	 }
      }
      // if(isNewImportantDep)
      // 	 if(DebugAutoInstall == true)
      // 	    std::clog << OutputInDepth(Depth) << "new important dependency: "
      // 		      << Start.TargetPkg().Name() << std::endl;
      // if(isPreviouslySatisfiedImportantDep)
      // 	if(DebugAutoInstall == true)
      // 	  std::clog << OutputInDepth(Depth) << "previously satisfied important dependency on "
      // 		    << Start.TargetPkg().Name() << std::endl;

      // skip important deps if the package is already installed
      if (Pkg->CurrentVer != 0 && Start.IsCritical() == false 
	  && !isNewImportantDep && !isPreviouslySatisfiedImportantDep
	  && !ForceImportantDeps)
	 continue;
      
      /* If we are in an or group locate the first or that can 
         succeed. We have already cached this.. */
      for (; Ors > 1 && (DCache[Start] & pkgDepCache::DepCVer) != pkgDepCache::DepCVer; Ors--)
	 Start++;

      SPtrArray<Version *> List = Start.AllTargets();
      if (Start->Type != pkgCache::Dep::DpkgBreaks &&
	  (DCache[Start] & pkgDepCache::DepCVer) != pkgDepCache::DepCVer)
      {
	 // Find matching candidate version.
	 Version **Cur = List;
	 PkgIterator P = Start.TargetPkg();
	 PkgIterator InstPkg(*Cache,0);
	 bool Ok = false;
	 
	 // See if there are direct matches (at the start of the list)
	 for (; *Cur != 0 && (*Cur)->ParentPkg == P.Index(); Cur++)
	 {
	    PkgIterator Pkg(*Cache,Cache->PkgP + (*Cur)->ParentPkg);
	    if (DCache[Pkg].CandidateVer == *Cur)
		// && DCache[Pkg].Install())
	    {
	       Ok = true;
	       Fix.Protect(Pkg);

	       if (Debug)
	       {
	       	  VerIterator Ver(*Cache, *Cur);
	       	  std::clog << "ForceDeps: already ok: "
	       		    << Pkg.Name() << ' '
	       		    << Ver.VerStr() << std::endl;
	       }
	       break;
	    }

	    VerIterator Ver(*Cache, *Cur);
	    VerIterator Cand = awc->cache->get_candidate_ver(P, Start);
	    if (Cand.end ())
	    {
	       if (Debug)
	       {
		  std::clog << "ForceDeps: no candidate for "
			    << Pkg.Name() << std::endl;
	       }

	       continue;
	    }

	    Deps.push_back(Cand);
	    Ok = true;
	    break;
	 }
	 AllOk &= Ok;
      }
   }

   if (!AllOk)
      return false;

   // Mark specific dependency versions for install.

   for (list<Version *>::iterator i = Deps.begin(); i != Deps.end(); ++i)
   {
      VerIterator Ver(*Cache, *i);
      PkgIterator Pkg = Ver.ParentPkg();

      if (Debug)
      {
	 domain_t dom = awc->cache->find_best_domain(Ver);
	 std::clog << "ForceDeps "
		   << Pkg.Name() << '=' << Ver.VerStr()
		   << " (from '" << domain_get_name(dom) << "')"
		   << std::endl;
      }

      DCache.SetCandidateVersion(Ver);
      DCache.MarkInstall1(Pkg, false, 0, false);
      Fix.Protect(Pkg);
   }

   return !P.InstBroken() && !P.InstPolicyBroken();
}
