/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <sys/stat.h>
#include <sys/types.h>
#include <errno.h>
#include <dirent.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>

#include <apt-pkg/debversion.h>
#include <apt-pkg/policy.h>
#include <apt-pkg/pkgrecords.h>
#include <apt-pkg/error.h>
#include <apt-pkg/configuration.h>
#include <apt-pkg/init.h>
#include <apt-pkg/acquire-item.h>

#include "archivecleaner.h"
#include "debug.h"
#include "error.h"
#include "package.h"
#include "apt-worker-utils.h"
#include "apt-worker-cache.h"
#include "apt-worker.h"
#include "apt-worker-internal.h"
#include "failed-installs.h"

/** CACHE HANDLING

    This section contains some general purpose functions to maintain
    the cache of the package database.

    The package cache can represent both the 'current' situation
    (i.e., the union of the information from /var/lib/dpkg/status and
    the various Packages files downloaded from repositories) and a
    'desired' situation.

    A operation such as installing a package is performed by modifying
    the 'desired' situation in the cache and if that leads to a
    consistent configuration, the 'current' situation is brought in
    line with the 'desired' one by downloading the needed archives and
    running dpkg in an approriate way.

    We have our own idea of what should happen when a new package (or
    a new version of a package) is installed, for example, and the
    functions in this section implement this idea.  These principal
    functions are available:

    We remember which operation the cache currently represents.  That
    way, we can avoid recomputing it when the frontend requests the
    same operation multiple times in a row (which it likes to do).

    - cache_init

    This function creates or recreates the cache from
    /var/lib/dpkg/status and the various Packages file from the
    repositories.

    - ensure_cache

    This function tries to make sure that there is a valid
    PACKAGE_CACHE to work with.  It returns true when it succeeds and
    PACKAGE_CACHE is non-NULL then.  The idea is that if the cache
    couldn't be created in the past because of some transient error,
    it might be able to create it now.  Thus, every command handler
    that needs a cache should call ensure_cache.  When ensure_cache
    actually does some work, it will send STATUS messages if it was
    specified with its only parameter.

    - cache_reset ()

    This function resets the 'desired' state of the cache to be
    identical to the 'current' one.

    - mark_for_install ()

    This function modifies the 'desired' state of the cache to reflect
    the installation of the given package.  It will try to achieve a
    consistent 'desired' configuration by installing missing
    dependencies etc.  In general, it implements our own installation
    smartness.

    - mark_for_remove ()

    This function modifies the 'desired' state of the cache to reflect
    the removal of the given package.  As with mark_for_install,
    mark_for_removal implements our own removal smartness.
 */

AptWorkerCache *AptWorkerCache::current = 0;

bool AptWorkerCache::global_initialized = false;

GStaticRWLock cache_rw_lock_;

/* Setting this to true will not ignore package versions from a wrong
   domain.
 */
// bool flag_allow_wrong_domains = false;

/* We only report real progress information when reconstructing the
   cache and during downloads.  Only downloads can be cancelled.

   The following two classes allow us to hook into libapt-pkgs
   progress reporting mechanism.  Instances of UPDATE_PROGESS are used
   for cache related activities, and instances of DOWNLOAD_STATUS are
   used when 'acquiring' things.
*/

class UpdateProgress : public OpProgress
{
  bool with_status;

  virtual void
  Update ()
  {
//     if (with_status)
//       send_status (OP_GENERAL, (int)Percent, 100, 5);
  }

public:
  UpdateProgress (bool ws) : with_status (ws) { }
};

class myPolicy : public pkgPolicy {

protected:
  myCacheFile *my_cache_file;
  domain_t *pf_domain;

public:
  myPolicy (pkgCache *Owner, myCacheFile *cache_file)
    : pkgPolicy (Owner), my_cache_file (cache_file)
  {
    pf_domain = new int[Owner->Head().PackageFileCount];
  }
  ~myPolicy ()
  {
    delete [] pf_domain;
  }

  void InitDomains ();
  void SetDomain (const pkgCache::PkgFileIterator &pf, domain_t domain);
  domain_t GetDomain (const pkgCache::PkgFileIterator &pf);

  virtual pkgCache::VerIterator GetCandidateVer(pkgCache::PkgIterator Pkg);
  pkgCache::VerIterator GetCandidateVer(const pkgCache::PkgIterator &Pkg,
					const pkgCache::DepIterator &Dep);
};


void
myPolicy::InitDomains ()
{
  pkgSourceList List;
  if (!List.ReadMainList ())
    {
      _error->DumpErrors();
      return;
    }

  set_sources_for_get_domain (&List);

  for (pkgCache::PkgFileIterator I = Cache->FileBegin();
       I != Cache->FileEnd(); I++)
    {
      // Locate the associated index files so we can find its domain
      pkgIndexFile *Indx;
      if (List.FindIndex(I,Indx) || _system->FindIndex(I,Indx))
	{
	  pf_domain[I->ID] = get_domain (Indx);
	  DEBUG ("%s: %s", Indx->Describe(true).c_str(),
		  domain_get_name(pf_domain[I->ID]));
	}
      else
        pf_domain[I->ID] = find_domain_by_name ("");
    }
  
  set_sources_for_get_domain (NULL);
}

void
myPolicy::SetDomain (const pkgCache::PkgFileIterator &pf, domain_t d)
{
  pf_domain[pf->ID] = d;
}

domain_t
myPolicy::GetDomain (const pkgCache::PkgFileIterator &pf)
{
  return pf_domain[pf->ID];
}

pkgCache::VerIterator 
myPolicy::GetCandidateVer (pkgCache::PkgIterator Pkg)
{
  return GetCandidateVer (Pkg, pkgCache::DepIterator ());
}


	    // if (!DCache.VS().CheckDep(Start.TargetVer(), Start->CompareOp & 0x0F, Cand.VerStr()))
	    //    continue;

pkgCache::VerIterator 
myPolicy::GetCandidateVer (const pkgCache::PkgIterator &Pkg, const pkgCache::DepIterator &Dep)
{
  if (getenv ("AEGIS_FIXED_ORIGIN"))
    return pkgPolicy::GetCandidateVer (Pkg);

  // Look for a package pin and evaluate it.
  signed Max = GetPriority(Pkg);
  pkgCache::VerIterator Pref = GetMatch(Pkg);

  int max_level = 0;
  
  /* Falling through to the default version.. Setting Max to zero
     effectively excludes everything <= 0 which are the non-automatic
     priorities.. The status file is given a prio of 100 which will exclude
     not-automatic sources, except in a single shot not-installed mode.
     The second pseduo-status file is at prio 1000, above which will permit
     the user to force-downgrade things.
     
     The user pin is subject to the same priority rules as default 
     selections. Thus there are two ways to create a pin - a pin that
     tracks the default when the default is taken away, and a permanent
     pin that stays at that setting.
  */

  for (pkgCache::VerIterator Ver = Pkg.VersionList(); Ver.end() == false; Ver++)
    {
      if (!Dep.end())
	{
	  if (!debVS.CheckDep (Dep.TargetVer (), Dep->CompareOp & 0xF,
			       Ver.VerStr ()))
	    continue;
	}

// 	    DEBUG("pkg: %s-%s", Pkg.Name(), Ver.VerStr());
      for (pkgCache::VerFileIterator VF = Ver.FileList(); VF.end() == false;
	   VF++)
	{
	  /* If this is the status file, and the current version is not the
	     version in the status file (ie it is not installed, or somesuch)
	     then it is not a candidate for installation, ever. This weeds
	     out bogus entries that may be due to config-file states, or
	     other. */

	  if ((VF.File()->Flags & pkgCache::Flag::NotSource) == pkgCache::Flag::NotSource &&
	      Pkg.CurrentVer() != Ver)
	    continue;

	  /* Skip versions from the wrong domain, but only for
	     sources, and only for packages that are actually installed.
	   */

	  if (!Pkg.CurrentVer().end()
              && ((VF.File()->Flags & pkgCache::Flag::NotSource) != pkgCache::Flag::NotSource))
	    {
              domain_t inst = my_cache_file->extra_info[Pkg->ID].domain;
	      if (inst != DOMAIN_INVALID
                  && !domain_dominates_or_is_equal (pf_domain[VF.File()->ID], inst))
		{
		  log_stderr ("Ignoring version from wrong domain: %s %s",
			      Pkg.Name(), Ver.VerStr());
                  log_stderr ("  Installed from %s, offered from %s",
                              domain_get_name (my_cache_file->extra_info[Pkg->ID].domain),
                              domain_get_name (pf_domain[VF.File()->ID]));
		  continue;
		}
	    }
	  
          int level = domain_get_trust_level (pf_domain[VF.File()->ID]);
	  signed Prio = PFPriority[VF.File()->ID];
	  if (level > max_level || (level == max_level && Prio > Max))
	    {
	      Pref = Ver;
              max_level = level;
	      Max = Prio;
	    }	 
	}      
      
      if (Pkg.CurrentVer() == Ver && Max < 1000)
	{
	  /* Elevate our current selection (or the status file itself)
	     to the Pseudo-status priority. */
	  if (Pref.end() == true)
	    Pref = Ver;
	  Max = 1000;
	  
	  // Fast path optimize.
	  if (StatusOverride == false)
	    break;
	}       
    }
  return Pref;
}


bool
myCacheFile::Open (OpProgress &Progress, bool WithLock)
{
  if (BuildCaches(Progress,WithLock) == false)
    return false;
  
  // The policy engine
  myPolicy *pol = new myPolicy (Cache, this);
  Policy = pol;
  if (_error->PendingError() == true)
    return false;
  if (ReadPinFile(*Policy) == false)
    return false;
  
  pol->InitDomains ();
  load_extra_info ();

  // Create the dependency cache
  DCache = new MarkPolicyDepCache(Cache,Policy);
  if (_error->PendingError() == true)
    return false;
  
  DCache->Init(&Progress);
  Progress.Done();
  if (_error->PendingError() == true)
    return false;
  
  return true;
}

bool
myCacheFile::AddExtraIndex(string Extra, domain_t domain, OpProgress &Progress, bool WithLock)
{
  int old_package_count = (*Cache).Head().PackageCount;

  if (Map->PrivateSize() > (unsigned long)_config->FindI("APT::Cache-Extra-Limit", 1024*1024))
    {
      Close();
      if (!Open (Progress, WithLock))
        {
          _error->DumpErrors();
          return false;
        }
    }

  MergeExtraIndex (Extra, Progress, WithLock);

  delete DCache;
  delete Policy;

  DCache = NULL;
  Policy = NULL;

  adjust_extra_info (old_package_count);
  
  myPolicy *pol = new myPolicy(Cache, this);
  Policy = pol;
  if (_error->PendingError() == true)
    return false;
  
  if (ReadPinFile(*Policy) == false || ReadPinDir(*Policy) == false)
    return false;

  pol->InitDomains ();
  pol->SetDomain (GetExtraPackageFile (), domain);

  DCache = new MarkPolicyDepCache(Cache,Policy);
  if (_error->PendingError() == true)
    return false;
   
  DCache->Init(&Progress);
  Progress.Done();
  if (_error->PendingError() == true)
    return false;

  return true;
}

/* Save the 'extra_info' of the cache.  We first make a copy of the
   Auto flags in our own extra_info storage so that CACHE_RESET
   will reset the Auto flags to the state last saved with this
   function.
*/

void
myCacheFile::save_extra_info ()
{
  if (mkdir ("/var/lib/package-manager", 0777) < 0
      && errno != EEXIST)
    {
      log_stderr ("/var/lib/package-manager: %m");
      return;
    }

  failed_installs.write ();
  failed_upgrades.write ();

  int dirfd = open ("/var/lib/package-manager", O_RDONLY);
  fsync (dirfd);
  close (dirfd);
}

/* Load the 'extra_info'.  You need to call CACHE_RESET to
   transfer the auto flag into the actual cache.  */

void
myCacheFile::load_extra_info ()
{
  if (mkdir ("/var/lib/package-manager", 0777) < 0
      && errno != EEXIST)
    {
      log_stderr ("/var/lib/package-manager: %m");
      return;
    }

  pkgCache &cache = *Cache;

  int package_count = cache.Head().PackageCount;

  if (extra_info)
    free (extra_info);

  extra_info = static_cast<extra_info_struct *> (malloc (sizeof (extra_info_struct) * package_count));

  for (int i = 0; i < package_count; i++)
    {
      extra_info[i].domain = DOMAIN_INVALID;
    }

  pkg_ts_storage->readFromFile( INSTALLED_PACKAGES_FILE );

  failed_installs.read ();
  failed_upgrades.read ();

    // Read current domains from /var/lib/aegis/restok/restok.conf.
    // XXX - use a official API for this.

  string restok_conf = _config->FindFile ("Dir::Aegis::Restok::conf");

    FileFd restok_fd;
    if (restok_fd.Open (restok_conf, FileFd::ReadOnly))
      {
        pkgTagFile restok_file (&restok_fd);
        pkgTagSection section;
        while (restok_file.Step (section))
          {
            string package = section.FindS ("Package");
            string source = section.FindS ("Source");

            pkgCache::PkgIterator pkg = cache.FindPkg (package);
            if (pkg.end())
              continue;

            extra_info[pkg->ID].domain = find_domain_by_name (source.c_str());
            // DEBUG ("%s: %s", pkg.Name(), domain_get_name (domain));
          }
      }
}

void
myCacheFile::adjust_extra_info (int old_package_count)
{
  pkgCache &cache = *Cache;

  int package_count = cache.Head().PackageCount;

  DEBUG("going from %d to %d packages", old_package_count, package_count);

  extra_info = (extra_info_struct *)realloc (extra_info, sizeof(extra_info_struct)*package_count);

  for (int i = old_package_count; i < package_count; i++)
    {
      /* We assume that any completely new packages are not installed
         and we thus don't care about their domain.
      */
      extra_info[i].domain = DOMAIN_INVALID;
    }
}

AptWorkerCache::AptWorkerCache ()
  : init_cache_after_request (false), cache (0), action_group (0)
{
}

AptWorkerCache *
AptWorkerCache::GetCurrent (void)
{
  return current;
}

static int
trust_level_hook_trampoline (pkgIndexFile *index,
                             const pkgCache::VerIterator &ver)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  return awc->cache->trust_level_hook (index, ver);
}

/* Initialization of apt worker cache. It initializes the
 * APT subsystem, and then gets the current instance. */
void
AptWorkerCache::Initialize (void)
{
  if (!global_initialized)
    {
      if (pkgInitConfig (*_config) == false ||
          pkgInitSystem (*_config, _system) == false)
        {
          _error->DumpErrors ();
          return;
        }

      _config->Set ("DPkg::Options::", "--force-confold");
      _config->Set ("Dir::Log", "var/log");
      _config->Set ("Dir::Log::Terminal", "");
      _config->Set ("APT::Get::Purge", 1);

      _config->Set ("Dir::Bin::pm-query", "/usr/lib/package-manager/pm-query");
      _config->Set ("Dir::Cache::pkgcache-read", "pkgcache.bin.read");
      _config->Set ("Dir::State::status-read", "status.read");
      _config->Set ("Dir::Aegis", "var/lib/aegis");
      _config->Set ("Dir::Aegis::Restok", "restok");
      _config->Set ("Dir::Aegis::Restok::conf", "restok.conf");
      _config->Set ("Dir::Aegis::Restok::conf-read", "restok.conf.read");

      apt_set_index_trust_level_for_package_hook (trust_level_hook_trampoline);

      global_initialized = true;
    }

  current = new AptWorkerCache;
}

void
cache_r_lock ()
{
  g_static_rw_lock_reader_lock (&cache_rw_lock_);
}

void
cache_r_unlock ()
{
  g_static_rw_lock_reader_unlock (&cache_rw_lock_);
}

void
cache_rw_lock ()
{
  g_static_rw_lock_writer_lock (&cache_rw_lock_);
}

void
cache_rw_unlock ()
{
  g_static_rw_lock_writer_unlock (&cache_rw_lock_);
}

static void
clear_dpkg_updates ()
{
  string File = flNotFile(_config->Find("Dir::State::status")) + "updates/";
  DIR *DirP = opendir(File.c_str());
  if (DirP == 0)
    return;
   
  /* We ignore any files that are not all digits, this skips .,.. and 
     some tmp files dpkg will leave behind.. */

  for (struct dirent *Ent = readdir(DirP); Ent != 0; Ent = readdir(DirP))
    {
      bool ignore = false;

      for (unsigned int I = 0; Ent->d_name[I] != 0; I++)
	{
	  // Check if its not a digit..
	  if (isdigit(Ent->d_name[I]) == 0)
	    {
	      ignore = true;
	      break;
	    }
	}

      if (!ignore)
	{
	  g_message ("Running 'dpkg --configure dpkg' "
		     "to clean up the journal.");
	  system ("dpkg --configure dpkg");
	  break;
	}
    }

   closedir(DirP);
}

static bool
copy_file (const char *src, const char *dest, bool update = true)
{
  struct stat st0, st1;
  if (stat (src, &st0) == -1)
    return false;

  bool dest_exists = true;
  if (stat (dest, &st1) == -1)
    {
      if (errno != ENOENT)
	return false;
      dest_exists = false;
    }

  if (dest_exists && update && st0.st_mtime <= st1.st_mtime)
    return true;

  string cmd;
  strprintf (cmd, "cp %s %s", src, dest);

  return system (cmd.c_str ()) == 0;
}

void
cache_init (bool with_status, bool with_lock)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

  cache_rw_lock ();
  // sleep(3);

  /* Closes the cache, to prevent getting blocked by other locks in
   * dpkg structures. If we don't do it, changing the apt worker state
   * does not remove the dpkg state lock and then fails on trying to
   * run dpkg */
  /* @todo do we really keep doing this? */
  if (awc->cache)
    {
      cache_close ();
      cache_free ();
    }

  /* We need to dump the errors here since any pending errors will
     cause the following operations to fail.
  */
  _error->DumpErrors ();

  /* Clear out the dpkg journal before construction the cache.
   */
  clear_dpkg_updates ();

  UpdateProgress progress (with_status);
  awc->cache = new myCacheFile;

  DEBUG ("init.");
  if (!awc->cache->Open (progress, with_lock))
    {
      DEBUG ("failed.");
      _error->DumpErrors ();
      delete awc->cache;
      awc->cache = 0;
    }

  copy_file (_config->FindFile ("Dir::Cache::pkgcache").c_str (),
	     _config->FindFile ("Dir::Cache::pkgcache-read").c_str ());
  copy_file (_config->FindFile ("Dir::State::status").c_str (),
	     _config->FindFile ("Dir::State::status-read").c_str ());
  copy_file (_config->FindFile ("Dir::Aegis::Restok::conf").c_str (),
	     _config->FindFile ("Dir::Aegis::Restok::conf-read").c_str ());

  if (awc->cache)
    {
      /* We create an ActionGroup here that is active for the whole
	 lifetime of the cache.  This prevents libapt-pkg from
	 performing certain expensive operations that we don't need to
	 be done intermediately, such as garbage collection.
      */
      pkgDepCache &cache = *awc->cache;
      delete awc->action_group;
      awc->action_group = new pkgDepCache::ActionGroup (cache);
    }

  cache_reset ();

  // if (awc->cache)
  //   write_available_updates_file ();

  cache_rw_unlock ();
}

void
cache_close ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  DEBUG ("closing");
  delete awc->action_group;
  DEBUG ("action group gone");
  awc->action_group = 0;
  DEBUG ("action group null");
  awc->cache->Close ();
  DEBUG ("closed");
}

void
cache_free ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

  delete awc->cache;
  awc->cache = 0;
  DEBUG ("done");
}

pkgCache::PkgFileIterator
cache_add_extra_index (string extra, domain_t domain, bool with_status)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

  delete awc->action_group;

  UpdateProgress progress (with_status);
  if (!awc->cache->AddExtraIndex (extra, domain, progress))
    {
      delete awc->cache;
      awc->cache = NULL;
      return pkgCache::PkgFileIterator();
    }

  pkgDepCache &cache = *awc->cache;
  awc->action_group = new pkgDepCache::ActionGroup (cache);

  return awc->cache->GetExtraPackageFile ();
}

bool
ensure_cache (bool with_status)
{
  AptWorkerCache * awc = 0;

  awc = AptWorkerCache::GetCurrent ();
  if (awc->cache == NULL)
    cache_init (with_status);

  return awc->cache != NULL;
}

/* The operation represented by the cache.
 */
OperationData *current_operation = NULL;
OperationData *compound_operation = NULL;
static GStaticMutex current_operation_mutex = G_STATIC_MUTEX_INIT;

OperationData*
cache_get_current_operation()
{
	return current_operation;
}

void
cache_set_current_operation(OperationData *op)
{
	if (current_operation)
		operation_data_free(current_operation);

	current_operation = op;
}

OperationData*
cache_get_compound_operation()
{
	return compound_operation;
}

void cache_operation_lock ()
{
	g_static_mutex_lock (&current_operation_mutex);
}

void cache_operation_unlock ()
{
	g_static_mutex_unlock (&current_operation_mutex);
}

void
cache_compound_operation_reset (OperationData *op)
{
	cache_operation_lock ();
	operation_data_free (compound_operation);
	compound_operation = op;
	cache_operation_unlock ();
}

bool
cache_operation_start (const char *package, const char *version, const char *domain, const char *filename, OperationType op_type, GList *pending, GList *rejected)
{
    send_operation_progress_signal (-1);

	if (current_operation
		&& current_operation->type == op_type
		&& strcmp (current_operation->pkg, package) == 0)
		return true;

	if (current_operation) {
		cache_reset ();
	}

	OperationData *op = operation_data_new ();
	op->pkg = g_strdup (package);
	op->version = g_strdup(version);
	op->domain = g_strdup(domain);
	op->filename = g_strdup(filename);
	op->type = op_type;
	op->pending_pkgs = copy_package_version_list (pending);
	op->rejected_pkgs = copy_package_version_list (rejected);
	cache_operation_reset (op);

	return false;
}

void
cache_operation_set_package (const gchar *package)
{
	g_return_if_fail(current_operation);
	g_return_if_fail(package);

	cache_operation_lock ();
	if (current_operation->pkg)
		g_free (current_operation->pkg);
	current_operation->pkg = g_strdup (package);
	cache_operation_unlock ();
}

void
cache_operation_set_version (const gchar *version)
{
	g_return_if_fail(current_operation);
	g_return_if_fail(version);

	cache_operation_lock ();
	if (current_operation->version)
		g_free (current_operation->version);
	current_operation->version = g_strdup (version);
	cache_operation_unlock ();
}

void
cache_operation_set_domain (const gchar *domain)
{
	g_return_if_fail(current_operation);

	cache_operation_lock ();
	if (current_operation->domain)
		g_free (current_operation->domain);
	current_operation->domain = g_strdup (domain);
	cache_operation_unlock ();
}

void
cache_operation_add_user_dependency(const gchar *package, const gchar *version, gboolean update)
{
	g_return_if_fail(current_operation);
	PackageVersion *pv = package_version_new ();
	pv->pkg = g_strdup (package);
	pv->version = g_strdup (version);

	cache_operation_lock ();
	if (update) {
		current_operation->updated_user_deps = g_list_append(current_operation->updated_user_deps, pv);
	} else {
		current_operation->installed_user_deps = g_list_append(current_operation->installed_user_deps, pv);
	}
	cache_operation_unlock ();
}

void
cache_operation_add_remove_dependency(const gchar *package, const gchar *version)
{
	g_return_if_fail(current_operation);
	PackageVersion *pv = package_version_new ();
	pv->pkg = g_strdup (package);
	pv->version = g_strdup (version);

	cache_operation_lock ();
	current_operation->removed_user_deps = g_list_append(current_operation->removed_user_deps, pv);
	cache_operation_unlock ();
}

static void
clean_archives ()
{
  _config->Set ("Debug::ArchiveCleaner", true);

  myCacheFile &cache = *AptWorkerCache::GetCurrent ()->cache;
  ArchiveCleaner cleaner (*cache);
  cleaner.Go (_config->FindFile ("Dir::Cache::Archives"), cache);
  cleaner.Go (_config->FindFile ("Dir::Cache::Archives") + "partial", cache);
}

void
cache_operation_complete (GError *error, int pkgflags)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

    if (current_operation) {
        DEBUG("complete");
        cache_operation_lock ();
        OperationData *op = operation_data_copy (current_operation);
        cache_operation_unlock ();

        awc->cache->package_installed(error);

        apt_worker_reset_state ();
        awc->init_cache_after_request = false;

        if (error) {
            op->error = pkgmgr_error_to_dbus_error(PkgmgrdError(error->code));
        } else {
	    if (op->type != OPERATION_DOWNLOAD_SYSTEM_UPGRADE)
		clean_archives ();
	}
        op->pkgflags = pkgflags;
        DEBUG("%s", op->error ? op->error : "OK");
        send_operation_complete_signal(op);
        operation_data_free (op);

	// Reset operation as the last step to avoid hitting a race in
	// UI logic...
        cache_operation_reset (NULL);
    } else
        DEBUG("complete but no current op\n");
}

bool
cache_remove_failed_installation(OperationType op_type, const gchar* package, const gchar* version)
{
    bool ret = false;
    DEBUG("Pkg: %s (%s)", package, version);

    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    myCacheFile &cache = *(awc->cache);

    cache.failed_installs.read();
    cache.failed_upgrades.read();

    bool need_set_interesting = false;

    if ( op_type == OPERATION_INSTALL || op_type == OPERATION_INSTALL_FILE ) {
        if ( op_type == OPERATION_INSTALL ) {
            ret = cache.failed_installs.clear(package);
        } else {
            ret = cache.failed_installs.clear_by_filename(package);
        }
        if ( ret ) {
            cache.failed_installs.write();
            need_set_interesting = true;
        }
    } else if ( op_type == OPERATION_UPGRADE ) {
        ret = cache.failed_upgrades.clear(package);
        if ( ret ) {
            cache.failed_upgrades.write();
            need_set_interesting = true;
        }
    } else DEBUG("Unknown operation");

    if (need_set_interesting) {
        GList *packages = g_list_concat (cache.failed_installs.get_packages(),
                                         cache.failed_upgrades.get_packages());
        set_interesting_packages (packages);
        g_list_free (packages);
    }

//    DEBUG("ret: %s", ret?"true":"false");
    if ( ret ) {
        OperationData* od = operation_data_new ();
        if ( op_type == OPERATION_INSTALL_FILE ) {
            od->pkg = g_strdup(""); // FIXME: the package name is missing. Any need for it?
            od->filename = g_strdup(package);
        } else {
            od->pkg = g_strdup(package);
            od->filename = g_strdup("");
        }

        od->version = g_strdup(version);
        od->type = op_type;
        od->error = pkgmgr_error_to_dbus_error(PkgmgrdError( ERROR_OPERATION_CANCELED ));
        od->installed_user_deps = NULL;
        od->updated_user_deps = NULL;
        od->removed_user_deps = NULL;
        od->pending_pkgs = NULL;
        od->rejected_pkgs = NULL;
        od->pkgflags = od->pkgflags;
        od->update_failed_installs = od->update_failed_installs;
        od->cancel_flag = TRUE;

        send_operation_complete_signal( od );
        operation_data_free( od );
    }

    return ret;
}

/* Revert the cache to its initial state.  More concretely, all
   packages are marked as 'keep' and 'unrelated'.
*/
void
cache_reset_package (pkgCache::PkgIterator &pkg)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  cache.MarkKeep (pkg);
  cache.SetReInstall(pkg, false);

  awc->cache->extra_info[pkg->ID].related = false;
  awc->cache->extra_info[pkg->ID].soft = false;
}

void
cache_reset_packages ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  if (awc->cache == NULL)
    return;

  pkgDepCache &cache = *(awc->cache);

  int n_non_kept = 0;
  for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
    {
      if (!cache[pkg].Keep())
        {
          n_non_kept++;
          if (n_non_kept > 10)
            {
              cache.Init (NULL);
              int package_count = cache.Head().PackageCount;
              for (int i = 0; i < package_count; i++)
                {
                  awc->cache->extra_info[i].related = false;
                  awc->cache->extra_info[i].soft = false;
                }
              return;
            }
        }

      cache_reset_package (pkg);
    }
}

void
cache_operation_reset (OperationData *op)
{
  cache_operation_lock ();
  operation_data_free (current_operation);
  current_operation = op;
  cache_operation_unlock ();
}

void 
cache_operation_update_pkg(const char *package, const char *version)
{
    if (! current_operation ) {
        return;
    }
    g_free( current_operation->pkg );
    current_operation->pkg = g_strdup(package);
    g_free( current_operation->version );
    current_operation->version = g_strdup(version);
}

void
cache_reset ()
{
  cache_reset_packages ();
  cache_operation_reset (NULL);
}

/* Commands can request the package cache to be refreshed by calling
   NEED_CACHE_INIT before they return.  The cache will then be
   reconstructed after sending the response and before starting to
   handle the next command.  In this way, the cache reconstruction
   happens in the background.

   XXX - However, APTCMD_STATUS messages are still being sent when the
         cache is reconstructed in the background and the UI has some
         ugly logic to deal with that.
*/

void
need_cache_init ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  awc->init_cache_after_request = true;
}

time_t
myCacheFile::package_timestamp(const gchar* package, bool *ok)
{
    return pkg_ts_storage->getPackageTimestamp(package, ok);
}

const gchar*
myCacheFile::package_filename(const gchar* package, bool *ok)
{
    return pkg_ts_storage->getPackageFilename(package, ok);
}

//! \todo complete list
static bool
is_tracked_error_code (int code)
{
  return code == ERROR_OPERATION_FAILED
    || code == ERROR_UPGRADE_FAILED
    || code == ERROR_NOT_ENOUGH_SPACE
    || code == ERROR_LOW_BATTERY_LEVEL
    || code == ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE
    || code == ERROR_DOWNLOAD_FAILED
    || code == ERROR_REMOVED_BY_SYSTEM_UPGRADE;
}

bool
myCacheFile::track_package_error (OperationType type, const gchar *pkg, const gchar *version, const char *domain, const char *filename, GError *error)
{
  bool dirty = false;
  if ((type == OPERATION_INSTALL ||
       type == OPERATION_RESTORE_INSTALL ||
       type == OPERATION_RESTORE_UPGRADE ||
       type == OPERATION_UPGRADE ||
       type == OPERATION_UPGRADE_ALL ||
       type == OPERATION_SYSTEM_UPGRADE ||
       type == OPERATION_DOWNLOAD_SYSTEM_UPGRADE)
      && (!error || is_tracked_error_code (error->code)))
    {
      bool upgrade = type == OPERATION_UPGRADE
        || type == OPERATION_RESTORE_UPGRADE
	|| type == OPERATION_UPGRADE_ALL
	|| type == OPERATION_SYSTEM_UPGRADE
	|| type == OPERATION_DOWNLOAD_SYSTEM_UPGRADE;

      if (error)
	{
	  if (!upgrade || (upgrade && error->code == ERROR_UPGRADE_FAILED))
	    failed_installs.store (pkg, version, NULL, domain, error);
	  if (upgrade)
	    failed_upgrades.store (pkg, version, NULL, domain, error);
	}
      else
	{
	  failed_installs.clear(pkg);
	  const char *v = failed_upgrades.failed_version (pkg);
	  if (upgrade || (v && debVS.CmpVersion (version, v) >= 0))
	    failed_upgrades.clear (pkg);
	}
      dirty = true;
    }
  else if (type == OPERATION_INSTALL_FILE
	   && (error == NULL || is_tracked_error_code (error->code)))
    {
      if (error)
	failed_installs.store (pkg, version, filename, domain, error);
      else
	failed_installs.clear (pkg);
      dirty = true;
    }
  else if (type == OPERATION_UNINSTALL && !error)
    {
      
      if (failed_installs.clear (pkg))
	dirty = true;
      if (failed_upgrades.clear (pkg))
	dirty = true;
    }
  set_interesting_packages (failed_installs.get_packages());

  return dirty;
}

void
myCacheFile::track_error(GError *error)
{
    OperationData *op = current_operation;
    
    if ( op == NULL ) {
        DEBUG("op NULL");
        return;
    }

    if (error)
	DEBUG("error: %s", !error ? "" : pkgmgr_error_to_dbus_error((PkgmgrdError) error->code));

    if (op->update_failed_installs)
        // Store or clear install or upgrade failures for resolvable
        // errors.
	if (track_package_error (op->type, op->pkg, op->version, op->domain, op->filename, error))
	{
	    failed_installs.write();
	    failed_upgrades.write();
	}

    DEBUG("out");
}

bool
myCacheFile::previous_error_was (PkgmgrdError error)
{
    OperationData *op = current_operation;

    const char *error_code;
    const char *error_message;
    const char *filename;
    const char *domain;

    if (!failed_installs.get_data (op->pkg, op->version, error_code, error_message, filename, domain))
      if (!failed_upgrades.get_data (op->pkg, op->version, error_code, error_message, filename, domain))
        return false;

    return strcmp (pkgmgr_error_to_dbus_error (error), error_code) == 0;
}

void
myCacheFile::package_installed(GError *error)
{
    if (error)
	DEBUG("error: %s", error ? error->message : "");

    track_error(error);

    if (!error && current_operation
	&& current_operation->update_failed_installs
	&& current_operation->type == OPERATION_SYSTEM_UPGRADE)
    {
	DEBUG("track removed");
	bool dirty = false;
	for (GList *p = current_operation->removed_user_deps; p != NULL; p = g_list_next (p))
	{
	    PackageVersion *pv = static_cast<PackageVersion *>(p->data);
	    GError *e = 0;
	    g_set_error(&e, PKGMGRD_ERROR, ERROR_REMOVED_BY_SYSTEM_UPGRADE,
			"Package '%s' (%s) was removed by system upgrade",
			pv->pkg, pv->version);
	    dirty |= track_package_error(OPERATION_INSTALL,
					 pv->pkg, pv->version, 0, 0, e);
	    g_error_free(e);
	}

	if (dirty)
	{
	    failed_installs.write();
	    failed_upgrades.write();
	}
    }

    if ( current_operation == NULL || error || operation_failed(current_operation)) {
        return;
    }
    
    OperationType type = current_operation->type;
    if ( current_operation->revert_install ) {
        type = OPERATION_UNINSTALL;
    }

    bool need_save = false;
    for (GList *p = current_operation->removed_user_deps; p != NULL; p = g_list_next (p)) {
        PackageVersion *pv = static_cast<PackageVersion *>(p->data);
        if ( pkg_ts_storage->removePackage(pv->pkg) )
            need_save = true;
    }

    bool ok = false;
    if ( type == OPERATION_INSTALL ) { // Install
        ok = false;
        need_save = true;
        pkg_ts_storage->getPackageTimestamp(current_operation->pkg, current_operation->version, &ok); // TODO: Should it be without a version info?
        if ( !ok ) {
            pkg_ts_storage->addPackage(current_operation->pkg, current_operation->version);
        }

        GList* deps = current_operation->installed_user_deps;
        while ( deps ) {
            PackageVersion *pv = static_cast<PackageVersion *>(deps->data);
            pkg_ts_storage->addPackage(pv->pkg, pv->version);
            deps = g_list_next(deps);
        }
    } else if ( type == OPERATION_INSTALL_FILE || ( type == OPERATION_RESTORE_INSTALL && current_operation->filename && strcmp(current_operation->filename, "") != 0 ) ) { // Install or restore a file
        ok = false;
        need_save = true;
        pkg_ts_storage->getFileTimestamp(current_operation->filename, &ok);
        if ( ok ) { // Replace the timestamp anyway.
            pkg_ts_storage->removeFile(current_operation->filename);
        }

        pkg_ts_storage->addFile(current_operation->pkg, current_operation->version, current_operation->filename);
        GList* deps = current_operation->installed_user_deps;
        while ( deps ) {
            PackageVersion *pv = static_cast<PackageVersion *>(deps->data);
            pkg_ts_storage->addPackage(pv->pkg, pv->version);
            deps = g_list_next(deps);
        }
    } else if ( type == OPERATION_UNINSTALL ) { // Uninstall -> remove the package timestamp from the list.
        if ( pkg_ts_storage->removePackage(current_operation->pkg) || pkg_ts_storage->removeFile(current_operation->pkg) ) {
            need_save = true;
        }
    } else if ( type == OPERATION_RESTORE_INSTALL || type == OPERATION_RESTORE_UPGRADE || type == OPERATION_UPGRADE ) { // Restore/upgrade -> replace the timestamp.
        ok = false;
        need_save = true;
        time_t ts = pkg_ts_storage->getPackageTimestamp(current_operation->pkg, &ok);
        if ( ok ) {
            pkg_ts_storage->removePackage(current_operation->pkg); // Replace the package
            pkg_ts_storage->addPackage(current_operation->pkg, current_operation->version, ts); // Do not change the timestamp but just update the version
        } else {
            pkg_ts_storage->addPackage(current_operation->pkg, current_operation->version); // The timestamp could not be found -> Just add a new timestamp.
        }
    }
    
    if ( need_save ) {
        pkg_ts_storage->saveToFile( INSTALLED_PACKAGES_FILE );
    }
}

bool
myCacheFile::check_domain (const pkgCache::VerIterator &ver,
                           domain_t wanted,
                           GError **error)
{
  if (ver.end())
    return true;

  // We will always use the best domain, so check whether that is the
  // one that the user wants.
  //
  // XXX - allow using the wanted domain, even if it isn't the best.
  //       A VerIterator is not enough for this, however.  See
  //       NB#213278.
  
  domain_t best = find_best_domain (ver);

  if (wanted != DOMAIN_INVALID && best != wanted)
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_VERSION_NOT_FOUND,
                   "Package version not in wanted domain.");
      return false;
    }

  return true;
}

bool
myCacheFile::check_compatible_domains (const pkgCache::VerIterator &ver,
                                       GError **error)
{
  if (ver.end())
    return true;

  const pkgCache::PkgIterator &pkg = ver.ParentPkg();

  // If the package is not installed, or our argument is the installed
  // version, everything is good.
  //
  if (pkg.CurrentVer().end()
      || pkg.CurrentVer() == ver)
    return true;

  domain_t best = find_best_domain (ver);
  domain_t inst = extra_info[pkg->ID].domain;

  if (inst != DOMAIN_INVALID && !domain_dominates_or_is_equal (best, inst))
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_VERSION_NOT_FOUND,
                   "Package version in wrong domain: installed from '%s', not updating from '%s'.",
                   domain_get_name (inst), domain_get_name (best));
      return false;
    }

  return true;
}

bool
myCacheFile::find_version (const pkgCache::PkgIterator &pkg,
                           const char *version,
                           const char *domain,
                           const pkgCache::PkgFileIterator &pf,
                           pkgCache::VerIterator &ver,
                           GError **error)
{
  if (pkg.end())
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
		   "Package not found");
      return false;
    }

  myPolicy *policy = (myPolicy *)Policy;
  domain_t wanted = DOMAIN_INVALID;

  if (domain && *domain)
    {
      wanted = find_domain_by_name (domain);
      if (wanted == DOMAIN_INVALID)
        {
          g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_VERSION_NOT_FOUND,
                      "Domain '%s' not found", domain);
          return false;
        }
    }

  domain_t best = DOMAIN_INVALID;
  int best_level = -1;

  if (version && *version)
    {
      for (pkgCache::VerIterator cand = pkg.VersionList (); !cand.end (); cand++)
        {
          if (!strcmp (cand.VerStr (), version))
            {
              for (pkgCache::VerFileIterator VF = cand.FileList(); VF.end() == false; VF++)
                {
                  if (VF.File()->Flags & pkgCache::Flag::Ignore)
                    continue;

                  domain_t d = policy->GetDomain (VF.File());

                  if (wanted != DOMAIN_INVALID && d != wanted)
                    continue;

                  if (!pf.end() && VF.File() != pf)
                    continue;

                  int l = domain_get_trust_level (d);
                  if (l > best_level)
                    {
                      ver = cand;
                      best = d;
                      best_level = l;
                    }
                }
            }
        }
    }
  else
    {
      pkgDepCache &cache = (*this);
      ver = cache[pkg].CandidateVerIter (cache);
    }

  if (ver.end())
    {
      if (version)
        g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_VERSION_NOT_FOUND,
                    "Package '%s' version '%s' not found", pkg.Name (), version);
      else
        g_set_error(error, PKGMGRD_ERROR, ERROR_NO_CANDIDATE,
                    "No candidate for package '%s'", pkg.Name ());
      return false;
    }

  if (!check_domain (ver, wanted, error))
    {
      ver = pkgCache::VerIterator();
      return false;
    }
    
  return true;
}

bool
myCacheFile::find_version (const pkgCache::PkgIterator &pkg,
                           const char *version,
                           const char *domain,
                           pkgCache::VerIterator &ver,
                           GError **error)
{
  return find_version (pkg, version, domain, pkgCache::PkgFileIterator(), ver, error);
}

bool
myCacheFile::find_version (const char *package,
                           const char *version,
                           const char *domain,
                           pkgCache::PkgIterator &pkg,
                           pkgCache::VerIterator &ver,
                           GError **error)
{
  if (package == NULL)
    return false;

  pkg = Cache->FindPkg (package);
  if (pkg.end())
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
		   "Package '%s' not found", package);
      return false;
    }

  return find_version (pkg, version, domain, ver, error);
}

bool
myCacheFile::find_version (const char *package,
                           const char *version,
                           const char *domain,
                           const pkgCache::PkgFileIterator &pf,
                           pkgCache::VerIterator &ver,
                           GError **error)
{
  if (package == NULL)
    return false;

  pkgCache::PkgIterator pkg = Cache->FindPkg (package);
  if (pkg.end())
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
		   "Package '%s' not found", package);
      return false;
    }

  return find_version (pkg, version, domain, pf, ver, error);
}

bool
myCacheFile::find_version (const char *package,
                           const char *version,
                           const char *domain,
                           pkgCache::VerIterator &ver,
                           GError **error)
{
  return find_version (package, version, domain, pkgCache::PkgFileIterator(), ver, error);
}

bool
myCacheFile::find_candidate (const char *package,
                             const char *domain,
                             pkgCache::VerIterator &ver,
                             GError **error)
{
  return find_version (package, NULL, domain, ver, error);
}

bool
myCacheFile::find_candidate (const pkgCache::PkgIterator &pkg,
                             const char *domain,
                             pkgCache::VerIterator &ver,
                             GError **error)
{
  return find_version (pkg, NULL, domain, ver, error);
}

bool
myCacheFile::find_failed_install_version (const char *package,
                                          pkgCache::VerIterator &ver,
                                          GError **error)
{
  const char *version = failed_installs.failed_version (package);
  const char *domain = failed_installs.failed_domain (package);
  if (version == NULL)
    return false;

  return find_version (package, version, domain, ver, error);
}

bool
myCacheFile::find_failed_install_version (const pkgCache::PkgIterator &pkg,
                                          pkgCache::VerIterator &ver,
                                          GError **error)
{
  return find_failed_install_version (pkg.Name(), ver, error);
}

int
myCacheFile::trust_level_hook (pkgIndexFile *index,
                               const pkgCache::VerIterator &ver)
{
  if (Cache != ((pkgCache::VerIterator &)ver).Cache())
    abort ();

  pkgCache::PkgFileIterator pf = index->FindInCache(*Cache);
  domain_t d = ((myPolicy *)Policy)->GetDomain (pf);
  return domain_get_trust_level (d);
}

domain_t
myCacheFile::find_best_domain(const pkgCache::VerIterator &ver)
{
  int best_level = -1;
  int best_domain = DOMAIN_INVALID;
  int n_version_files = 0;
  for (pkgCache::VerFileIterator vf = ver.FileList(); vf.end() == false; vf++) 
    {
      n_version_files++;

      domain_t d;
      if ((vf.File()->Flags & pkgCache::Flag::NotSource)
          && ver == ver.ParentPkg().CurrentVer())
        {
          d = extra_info[ver.ParentPkg()->ID].domain;
          if (d == DOMAIN_INVALID)
            d = find_domain_by_name ("");
        }
      else
        d = ((myPolicy *)Policy)->GetDomain (vf.File());

      int l = domain_get_trust_level (d);
      if (l > best_level)
        {
          best_level = l;
          best_domain = d;
        }
    }
    
  if (n_version_files == 0)
    {
      DEBUG("Whoa, version without sources: %s %s", ver.ParentPkg().Name(), ver.VerStr());
      best_domain = find_domain_by_name ("");
      best_level = domain_get_trust_level (best_domain);
    }

  /* We expect apt to be configured so that each VerIterator has only
     one VerFileIterator, by setting APT::Use-Version-Hashes to false.
  */
  if (n_version_files > 1)
    DEBUG("Whoa, version with more than one source: %s %s", ver.ParentPkg().Name(), ver.VerStr());

  // DEBUG("best for %s %s: '%s' %d %d", ver.ParentPkg().Name(), ver.VerStr(),
  //          domain_get_name (best_domain), best_level, domain_is_certified (best_domain));
    
  return best_domain;
}

domain_t
myCacheFile::get_current_domain(const pkgCache::PkgIterator &pkg)
{
  if (pkg.end())
    return DOMAIN_INVALID;

  return extra_info[pkg->ID].domain;
}

const char*
myCacheFile::get_domain_name(const pkgCache::VerIterator &ver)
{
  return domain_get_name (find_best_domain(ver));
}

bool
myCacheFile::is_trusted (const pkgCache::VerIterator &ver)
{
  /* Find the first VerFile with the highest trust level and check the
     'certified' flag of its domain.

     The 'trust_level_hook' (above) makes sure that APT will choose
     the VerFile with the highest trust level as well, so this matches
     what will actually happen.

     XXX - This is a bit tricky. Instead, we should tell APT directly
           which VerFile to use, via a modified SetCandidate.
  */
  return domain_is_certified (find_best_domain(ver));
}

pkgCache::VerIterator myCacheFile::get_candidate_ver (const pkgCache::PkgIterator &pkg,
						      const pkgCache::DepIterator &dep)
{
  myPolicy &policy = static_cast<myPolicy &>(static_cast<pkgDepCache &> (*this).GetPolicy ());
  return policy.GetCandidateVer (pkg, dep);
}

string
os_meta_package_name ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  if (!awc->cache->os_meta_package_name.empty ())
    return awc->cache->os_meta_package_name;

  for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++)
    {
      // pkgDepCache::StateCache &s = cache[p];

      pkgCache::VerIterator cur = p.CurrentVer ();
      if (cur.end ())
	continue;

      if (!is_os_meta_package (cur))
	continue;

      awc->cache->os_meta_package_name = p.Name();
      break;
    }

  return awc->cache->os_meta_package_name;
}

pkgCache::VerIterator
cache_get_candidate_ver (const pkgCache::PkgIterator &pkg,
			 const pkgCache::DepIterator &dep)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  return awc->cache->get_candidate_ver (pkg, dep);
}

