/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#ifndef APT_WORKER_CACHE_H
#define APT_WORKER_CACHE_H

#include <apt-pkg/pkgcache.h>
#include <apt-pkg/cachefile.h>
#include <apt-pkg/debsystem.h>
#include <cstdlib>

#include "domain.h"
#include "error.h"
#include "apt-worker-operation.h"
#include "inst-pkg-ts-storage.h"
#include "failed-installs.h"
#include "markpolicydepcache.h"

extern OperationData *current_operation;

/* This struct describes some status flags for specific packages.
 * myCacheFile includes an array of these, with an entry per
 * package.
 */
struct extra_info_struct
{
  bool related : 1;
  bool soft : 1;
  domain_t domain;
};


class myCacheFile : public pkgCacheFile {

public:
  bool Open (OpProgress &Progress, bool WithLock = true);
  bool AddExtraIndex(string Extra, domain_t domain, OpProgress &Progress, bool WithLock = true);

  MarkPolicyDepCache &dep_cache () { return static_cast<MarkPolicyDepCache &>(*DCache); }

  void load_extra_info ();
  void adjust_extra_info (int old_package_count);
  void save_extra_info ();
  void package_installed(GError *error);
  time_t package_timestamp(const gchar* package, bool *ok);
  const gchar* package_filename(const gchar* package, bool *ok);
  bool track_package_error (OperationType type, const gchar *pkg, const gchar *version, const char *domain, const char *filename, GError *error);
  void track_error(GError *error);
  bool previous_error_was (PkgmgrdError error);

  bool check_domain (const pkgCache::VerIterator &ver,
                     domain_t wanted,
                     GError **error = NULL);

  /* Check whether VER is compatible with the installed version, from
     the point of view of the domain model.
  */
  bool check_compatible_domains (const pkgCache::VerIterator &ver,
                                 GError **error = NULL);

  /* We find the archives we should work on based on name, version and
     domain.

     When VERSION is NULL, the candidate version of the package is
     used.
     
     When DOMAIN is NULL, the 'best' domain is used that is compatible
     with the installed version of the package.

     XXX - explain algorithm, once that is clear.
  */

  bool find_version (const pkgCache::PkgIterator &pkg,
                     const char *version,
                     const char *domain,
                     const pkgCache::PkgFileIterator &pf,
                     pkgCache::VerIterator &ver_out,
                     GError **error = NULL);

  bool find_version (const pkgCache::PkgIterator &pkg,
                     const char *version,
                     const char *domain,
                     pkgCache::VerIterator &ver_out,
                     GError **error = NULL);

  bool find_version (const char *package,
                     const char *version,
                     const char *domain,
                     pkgCache::VerIterator &ver_out,
                     GError **error = NULL);

  bool find_version (const char *package,
                     const char *version,
                     const char *domain,
                     const pkgCache::PkgFileIterator &pf,
                     pkgCache::VerIterator &ver_out,
                     GError **error = NULL);

  bool find_version (const char *package,
                     const char *version,
                     pkgCache::VerIterator &ver_out,
                     GError **error = NULL);

  bool find_version (const char *package,
                     const char *version,
                     const char *domain,
                     pkgCache::PkgIterator &pkg_out,
                     pkgCache::VerIterator &ver_out,
                     GError **error = NULL);

  bool find_candidate (const char *package,
                       const char *domain,
                       pkgCache::VerIterator &ver_out,
                       GError **error = NULL);

  bool find_candidate (const pkgCache::PkgIterator &pkg,
                       const char *domain,
                       pkgCache::VerIterator &ver_out,
                       GError **error = NULL);

  bool find_failed_install_version (const pkgCache::PkgIterator &pkg,
                                    pkgCache::VerIterator &ver_out,
                                    GError **error = NULL);

  bool find_failed_install_version (const char *package,
                                    pkgCache::VerIterator &ver_out,
                                    GError **error = NULL);

  int trust_level_hook (pkgIndexFile *index,
                        const pkgCache::VerIterator &ver);

  domain_t find_best_domain(const pkgCache::VerIterator &ver);

  domain_t get_current_domain(const pkgCache::PkgIterator &pkg);

  bool is_trusted (const pkgCache::VerIterator &ver);
  
  const char* get_domain_name(const pkgCache::VerIterator &ver);

  pkgCache::VerIterator get_candidate_ver (const pkgCache::PkgIterator &pkg,
					   const pkgCache::DepIterator &dep);

  extra_info_struct *extra_info;

  myCacheFile ()
    : failed_upgrades (FAILED_UPGRADES_FILE)
  {
    extra_info = NULL;
    pkg_ts_storage = new PkgTsStorage();
  }

  ~myCacheFile ()
  {
    free (extra_info);
    delete pkg_ts_storage;
  }

public:
  FailedInstalls failed_installs;
  FailedInstalls failed_upgrades;
  string os_meta_package_name;

private:
    PkgTsStorage* pkg_ts_storage;
};

/* This class implements a global static cache managing for
 *  apt-worker. */
class AptWorkerCache
{
public:
  AptWorkerCache ();
  static void Initialize ();
  static AptWorkerCache * GetCurrent ();

  bool init_cache_after_request;
  myCacheFile *cache;
  pkgDepCache::ActionGroup *action_group;
  static AptWorkerCache *current;
  static bool global_initialized;
};

void
cache_r_lock ();

void
cache_r_unlock ();

void
cache_rw_lock ();

void
cache_rw_unlock ();

struct scoped_cache_rw_lock {
    scoped_cache_rw_lock() { cache_rw_lock(); }
    ~scoped_cache_rw_lock() { cache_rw_unlock(); }
};

void
cache_init (bool with_status, bool with_lock = true);

void
cache_close ();

void
cache_free ();

pkgCache::PkgFileIterator
cache_add_extra_index (string extra, domain_t domain, bool with_status);

bool
ensure_cache (bool with_status);

OperationData*
cache_get_current_operation();

OperationData*
cache_get_compound_operation();

void
cache_set_current_operation(OperationData *op);

void
cache_operation_lock ();

void
cache_operation_unlock ();

void
cache_operation_reset (OperationData *op);

void
cache_operation_update_pkg (const char *package, const char *version);

void
cache_compound_operation_reset (OperationData *op);

bool
cache_operation_start (const char *package, const char *version, const char *domain, const char *filename, OperationType op_type, GList *pending = NULL, GList *rejected = NULL);

void
cache_operation_set_package (const gchar *package);

void
cache_operation_set_version (const gchar *version);

void
cache_operation_set_domain (const gchar *version);

void
cache_operation_add_user_dependency(const gchar *package, const gchar *version, gboolean update);

void
cache_operation_add_remove_dependency(const gchar *package, const gchar *version);

void
cache_operation_complete (GError *error, int pkgflags = 0);

bool
cache_remove_failed_installation(OperationType op_type, const gchar* package, const gchar* version);

void
cache_reset_package (pkgCache::PkgIterator &pkg);

void
cache_reset_packages ();

void
cache_reset ();

void
need_cache_init ();

string
os_meta_package_name ();

pkgCache::VerIterator
cache_get_candidate_ver (const pkgCache::PkgIterator &pkg,
			 const pkgCache::DepIterator &dep);

#endif /* APT_WORKER_CACHE_H */
