/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#ifndef APT_WORKER_CMD_H // -*- Mode: C++; c-file-style: "linux" -*-
#define APT_WORKER_CMD_H

#include <glib.h>
#include <string.h>
#include "apt-worker.h"
#include "apt-worker-struct.h"
#include "package-list.h"
#include "package-map.h"

struct apt_worker_cmd
{
	typedef void (*callback_t) (apt_worker_cmd *cmd);

	apt_worker_cmd ()
		: apt_worker(0),
		  callback (0),
		  user_data (0),
		  error (0),
		  success(FALSE),
		  need_state_reset(FALSE)
		{ }

	virtual ~apt_worker_cmd () {};

	virtual gboolean execute () = 0;

	AptWorker *apt_worker;
	callback_t callback;
	gpointer user_data;
	GError *error;
	gboolean success;
	gboolean need_state_reset;
};

struct apt_worker_batch_cmd : apt_worker_cmd
{
	typedef void (*callback_t) (apt_worker_cmd *cmd);

	apt_worker_batch_cmd (GList *cmds)
		: cmds(cmds)
		{ }

	~apt_worker_batch_cmd ()
	{
		GList *list = cmds;
		while (list) {
			apt_worker_cmd *cmd = reinterpret_cast<apt_worker_cmd *>(list->data);
			delete cmd;
			list = g_list_next(list);
		}
		g_list_free (cmds);
	}

	virtual gboolean execute ()
	{
		GList *list = cmds;
		gboolean result = TRUE;
		success = TRUE;
		while (list) {
			if (need_state_reset) {
				apt_worker->reset_state();
			}
			apt_worker_cmd *cmd = reinterpret_cast<apt_worker_cmd *>(list->data);
			need_state_reset = cmd->need_state_reset;
			cmd->apt_worker = apt_worker;
			cmd->success = cmd->execute();
			if (!cmd->success) {
				success = FALSE;
				result = FALSE;
				break;
			}
			list = g_list_next(list);
		}
		return TRUE;
	}

	GList *cmds;
};

struct apt_worker_cmd_get_package_list : apt_worker_cmd
{
	apt_worker_cmd_get_package_list (const gchar *category,
					 PkgType type,
					 const gchar *pattern,
					 PkgSort sort)
		: category (g_strdup (category)),
		  type (type),
		  pattern (g_strdup (pattern)),
		  sort (sort),
		  packages (0)
	{
	}

	~apt_worker_cmd_get_package_list ()
	{
		g_free (category);
		g_free (pattern);
		package_list_free (packages);
	}

	virtual gboolean execute ()
	{
		apt_worker->cmd_get_package_list (category,
						  type,
						  pattern,
						  &packages,
						  sort,
						  &error);
		return error != NULL;
	}

	gchar *category;
	PkgType type;
	gchar *pattern;
	PkgSort sort;

	GList *packages;
};

struct apt_worker_cmd_get_file_details : apt_worker_cmd
{
	apt_worker_cmd_get_file_details (const char *filename)
		: filename (g_strdup (filename)),
		  retval (0)
	{
	}

	~apt_worker_cmd_get_file_details ()
	{
		g_free (filename);
		package_free (retval);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_get_file_details (filename,
							   &error);
		return retval != NULL;
	}

	gchar *filename;

	Package *retval;
};

struct apt_worker_cmd_get_file_details_map : apt_worker_cmd
{
	apt_worker_cmd_get_file_details_map (const char *filename)
		: filename (g_strdup (filename)),
		  retval (0)
	{
	}

	~apt_worker_cmd_get_file_details_map ()
	{
		g_free (filename);
		package_hash_table_free (retval);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_get_file_details_map (filename,
							       &error);
		return retval != NULL;
	}

	gchar *filename;

	GHashTable *retval;
};

struct apt_worker_cmd_get_package_details_map : apt_worker_cmd
{
	apt_worker_cmd_get_package_details_map (const char *package,
						const char *version)
		: package (g_strdup (package)),
		  version (g_strdup (version)),
		  installed (FALSE),
		  retval (0)
	{
	}

	apt_worker_cmd_get_package_details_map (const char *package,
						gboolean installed)
		: package (g_strdup (package)),
		  version (NULL),
		  installed (installed),
		  retval (0)
	{
	}

	~apt_worker_cmd_get_package_details_map ()
	{
		g_free (package);
		g_free (version);
		package_hash_table_free (retval);
	}

	virtual gboolean execute ()
	{
		if (version && strncmp(version,"", 1) == 0) {
			g_free (version);
			version = NULL;
		}
		retval = apt_worker->cmd_get_package_details_map (package,
								  version,
                                                                  NULL,
								  installed,
								  &error);
		return retval != NULL;
	}

	gchar *package;
	gchar *version;
	gboolean installed;

	GHashTable *retval;
};

struct apt_worker_cmd_get_package_dep_details : apt_worker_cmd
{
	apt_worker_cmd_get_package_dep_details (const char *package,
						const char *version)
		: package (g_strdup (package)),
		  version (g_strdup (version)),
		  retval (0)
	{
	}

	~apt_worker_cmd_get_package_dep_details ()
	{
		g_free (package);
		g_free (version);

		for (GList *i = g_list_first (retval); i != NULL; i = g_list_next (i))
			package_hash_table_free (static_cast<GHashTable *> (i->data));
		g_list_free( retval );
	}

	virtual gboolean execute ()
	{
		if (version && strncmp(version, "", 1) == 0) {
			g_free (version);
			version = NULL;
		}
		retval = apt_worker->cmd_get_package_dep_details (package,
								  version,
								  &error);
		return retval != NULL;
	}

	gchar *package;
	gchar *version;

	GList *retval;
};

struct apt_worker_cmd_get_restore_status : apt_worker_cmd
{
	apt_worker_cmd_get_restore_status ()
		: retval (0)
	{
	}

	~apt_worker_cmd_get_restore_status ()
	{
        g_free ( retval );
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_get_restore_status (&error);
		return retval != NULL;
	}

	gchar *retval;
};

struct apt_worker_cmd_get_restorable_packages : apt_worker_cmd
{
	apt_worker_cmd_get_restorable_packages ()
		: retval (0)
	{
	}

	~apt_worker_cmd_get_restorable_packages ()
	{
	    for (GList *i = g_list_first (retval); i != NULL; i = g_list_next (i))
            g_free ( i->data );
            g_list_free ( retval );
	}

	virtual gboolean execute ()
	{
		return apt_worker->cmd_get_restorable_packages (&retval, &error);
	}

	GList *retval;
};

struct apt_worker_cmd_restore_package : apt_worker_cmd
{
	apt_worker_cmd_restore_package (const char *package,
					const char *version)
		: package (g_strdup (package)),
		  version (g_strdup (version)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_restore_package ()
	{
		g_free (package);
		g_free (version);
	}


	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_restore_package (package,
							  version,
                                                          NULL,
							  &error);
		return retval;
	}

	gchar *package;
	gchar *version;

	gboolean retval;
};

struct apt_worker_cmd_ignore_restore_package : apt_worker_cmd
{
	apt_worker_cmd_ignore_restore_package (const char *package)
		: package (g_strdup (package))
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_ignore_restore_package ()
	{
        g_free (package);
	}

	virtual gboolean execute ()
	{
		apt_worker->cmd_ignore_restore_package (package, &error);
		return TRUE;
	}

    gchar *package;
};

struct apt_worker_cmd_get_current_operation_data : apt_worker_cmd
{
	apt_worker_cmd_get_current_operation_data ()
		: retval (0)
	{
	}

	~apt_worker_cmd_get_current_operation_data ()
	{
		package_hash_table_free (retval);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_get_current_operation_data (&error);

		return retval != NULL;
	}

	GHashTable *retval;
};

struct apt_worker_cmd_simulate_operation : apt_worker_cmd
{
	apt_worker_cmd_simulate_operation (const gchar *operation, const gchar *packageidentifier, const gchar *version, const gchar *flags)
        : operation (g_strdup(operation)),
          packageidentifier (g_strdup(packageidentifier)),
          version (g_strdup(version)),
          flags (g_strdup(flags)),
          retval (0)
	{
	}

	~apt_worker_cmd_simulate_operation ()
	{
	    g_free(operation);
	    g_free(packageidentifier);
	    g_free(version);
	    g_free(flags);
		package_hash_table_free (retval);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_simulate_operation (operation, packageidentifier, version, NULL, flags, &error);

		return retval != NULL;
	}

	gchar *operation;
	gchar *packageidentifier;
	gchar *version;
	gchar *flags;

    GHashTable *retval;
};


struct apt_worker_cmd_install_file : apt_worker_cmd
{
	apt_worker_cmd_install_file (const char *filename)
		: filename (g_strdup (filename)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_install_file ()
	{
		g_free (filename);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_install_file (filename,
						       &error);
		return retval;
	}

	gchar *filename;

	gboolean retval;
};

struct apt_worker_cmd_ignore_update_package : apt_worker_cmd
{
	apt_worker_cmd_ignore_update_package (const char *package,
					const char *version)
		: package (g_strdup (package)),
		  version (g_strdup (version))
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_ignore_update_package ()
	{
                g_free (package);
		g_free (version);
	}

	virtual gboolean execute ()
	{
		apt_worker->cmd_ignore_update_package (package, version,
							  &error);
		return TRUE;
	}

        gchar *package;
	gchar *version;
};

struct apt_worker_cmd_refresh_cache : apt_worker_cmd
{
	apt_worker_cmd_refresh_cache (const char *only_domain, const char *flags)
		: only_domain (g_strdup (only_domain)),
		  flags(g_strdup(flags)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_refresh_cache ()
	{
            g_free(only_domain);
            g_free(flags);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_refresh_cache (only_domain,
		                                        flags,
                                                        &error);
		return retval;
	}

        gchar* only_domain;
        gchar* flags;

	gboolean retval;
};

struct apt_worker_cmd_install_package : apt_worker_cmd
{
	apt_worker_cmd_install_package (const char *package,
					const char *version,
					const char *alt_download_root)
		: package (g_strdup (package)),
		  version (g_strdup (version)),
		  alt_download_root (g_strdup (alt_download_root)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_install_package ()
	{
		g_free (package);
		g_free (version);
		g_free (alt_download_root);
	}


	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_install_package (package,
							  version,
                                                          NULL,
							  alt_download_root,
							  &error);
		return retval;
	}

	gchar *package;
	gchar *version;
	gchar *alt_download_root;

	gboolean retval;
};

struct apt_worker_cmd_uninstall_package : apt_worker_cmd
{
	apt_worker_cmd_uninstall_package (const char *package)
		: package (g_strdup (package)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_uninstall_package ()
	{
		g_free (package);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_uninstall_package (package,
							    &error);
		return retval;
	}

	gchar *package;

	gboolean retval;
};

struct apt_worker_cmd_pkgmgr_fetch_files_of_installed_package : apt_worker_cmd
{
        apt_worker_cmd_pkgmgr_fetch_files_of_installed_package (const char *package)
                : package (g_strdup (package)),
                  retval (FALSE)
        {
                need_state_reset = FALSE;
        }

        ~apt_worker_cmd_pkgmgr_fetch_files_of_installed_package ()
        {
                g_free (package);
                g_list_free(files);
        }

        virtual gboolean execute ()
        {
                retval = apt_worker->cmd_pkgmgr_fetch_files_of_installed_package (package,&files,&error);
                return retval;
        }

        gchar *package;
	GList *files;
	gboolean retval;
};

struct apt_worker_cmd_upgrade_package : apt_worker_cmd
{
	apt_worker_cmd_upgrade_package (const char *package,
					const char *version,
					const char *alt_download_root)
		: package (g_strdup (package)),
		  version (g_strdup (version)),
		  alt_download_root (g_strdup (alt_download_root)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_upgrade_package ()
	{
		g_free (package);
		g_free (version);
		g_free (alt_download_root);
	}


	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_upgrade_package (package,
							  version,
                                                          NULL,
							  alt_download_root,
							  &error);
		return retval;
	}

	gchar *package;
	gchar *version;
	gchar *alt_download_root;

	gboolean retval;
};

struct apt_worker_cmd_download_system_upgrade : apt_worker_cmd
{
	apt_worker_cmd_download_system_upgrade (const char *package,
					const char *version,
					const char *alt_download_root)
		: package (g_strdup (package)),
		  version (g_strdup (version)),
		  alt_download_root (g_strdup (alt_download_root)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_download_system_upgrade ()
	{
		g_free (package);
		g_free (version);
		g_free (alt_download_root);
	}


	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_download_system_upgrade (package,
							  version,
                                                          NULL,
							  alt_download_root,
							  &error);
		return retval;
	}

	gchar *package;
	gchar *version;
	gchar *alt_download_root;

	gboolean retval;
};

struct apt_worker_cmd_upgrade_system : apt_worker_cmd
{
	apt_worker_cmd_upgrade_system (const char *package,
				       const char *version,
				       const char *alt_download_root)
		: package (g_strdup (package)),
		  version (g_strdup (version)),
		  alt_download_root (g_strdup (alt_download_root)),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_upgrade_system ()
	{
		g_free (package);
		g_free (version);
		g_free (alt_download_root);
	}


	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_upgrade_system (package,
							 version,
                                                         NULL,
							 alt_download_root,
							 &error);
		return retval;
	}

	gchar *package;
	gchar *version;
	gchar *alt_download_root;

	gboolean retval;
};

struct apt_worker_cmd_upgrade_all : apt_worker_cmd
{
	apt_worker_cmd_upgrade_all (gboolean allow_system_upgrade)
		: allow_system_upgrade (allow_system_upgrade),
		  retval (FALSE)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_upgrade_all ()
	{
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_upgrade_all (allow_system_upgrade,
						      &error);
		return retval;
	}

	gboolean allow_system_upgrade;

	gboolean retval;
};

struct apt_worker_cmd_check_updates : apt_worker_cmd
{
	apt_worker_cmd_check_updates (gboolean with_status)
		: with_status (with_status),
		  retval (rescode_failure)
	{
		need_state_reset = TRUE;
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_check_updates (with_status, NULL, false);
		return retval == rescode_success || retval == rescode_partial_success;
	}

	gboolean with_status;

	ResultCode retval;
};

struct apt_worker_cmd_get_catalogues : apt_worker_cmd
{
	apt_worker_cmd_get_catalogues ()
		: retval (FALSE),
		  catalogues (NULL)
	{
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_get_catalogues (&catalogues, &error);
		return retval;
	}

	gboolean retval;
	GList *catalogues;
};

struct apt_worker_cmd_add_catalogue : apt_worker_cmd
{
	apt_worker_cmd_add_catalogue (const gchar *name,
				      const gchar *uri,
				      const gchar *dist,
				      const gchar *comps)
		: retval (FALSE),
		  name (g_strdup (name)),
		  uri (g_strdup (uri)),
		  dist (g_strdup (dist)),
		  comps (g_strdup (comps))
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_add_catalogue ()
	{
		g_free (name);
		g_free (uri);
		g_free (dist);
		g_free (comps);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_add_catalogue (name,
							uri,
							dist,
							comps,
							&error);
		return retval;
	}

	gboolean retval;
	gchar *name;
	gchar *uri;
	gchar *dist;
	gchar *comps;
};

struct apt_worker_cmd_remove_catalogue : apt_worker_cmd
{
	apt_worker_cmd_remove_catalogue (const gchar *name)
		: retval (FALSE),
		  name (g_strdup (name))
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_remove_catalogue ()
	{
		g_free (name);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_remove_catalogue (name,
							   &error);
		return retval;
	}

	gboolean retval;
	gchar *name;
};

struct apt_worker_cmd_enable_catalogue : apt_worker_cmd
{
	apt_worker_cmd_enable_catalogue (const gchar *name, gboolean enable)
		: retval (FALSE),
		  name (g_strdup (name)),
		  enable (enable)
	{
		need_state_reset = TRUE;
	}

	~apt_worker_cmd_enable_catalogue ()
	{
		g_free (name);
	}

	virtual gboolean execute ()
	{
		retval = apt_worker->cmd_enable_catalogue (name,
							   enable,
							   &error);
		return retval;
	}

	gboolean retval;
	gchar *name;
	gboolean enable;
};

#endif
