/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "apt-worker-markpolicy.h"
#include "apt-worker-utils.h"
#include "error.h"
#include <apt-pkg/configuration.h>
#include <apt-pkg/debversion.h>
#include <cassert>
#include <cstring>

#include "debug.h"

myMarkPolicy::myMarkPolicy(Mode mode)
{
    setMode(mode);

    debug = _config->FindB("Debug::MarkPolicy", false);
}

void myMarkPolicy::setMode(Mode mode)
{
    switch (mode)
    {
    case AllowAll:
	nonAutoInstMustBeUserInstallable = false;
	nonAutoInstMustBeUserInstallableStops = false;
	forbidMarkOSMetaPackage = false;
	forbidMarkOSMetaPackageStops = false;
	forbidBreakVisiblePackage = false;
	forbidBreakVisiblePackageStops = false;
	break;

    case UserInstall:
	nonAutoInstMustBeUserInstallable = true;
	nonAutoInstMustBeUserInstallableStops = false;
	forbidMarkOSMetaPackage = true;
	forbidMarkOSMetaPackageStops = true;
	forbidBreakVisiblePackage = true;
	forbidBreakVisiblePackageStops = false;
	break;

    default:
	assert(!"Invalid myMarkPolicy::Mode");
    }

    breaksOSMetaPackage = false;
    installOSMetaPackageVer = VerIterator();
    deleteOSMetaPackageVer = VerIterator();
    breakVisiblePackageVer = VerIterator();
}

bool myMarkPolicy::isOSMetaPackage(const VerIterator &v)
{
    return is_os_meta_package(v);
}

bool myMarkPolicy::isVisible(const VerIterator &v)
{
    return is_user_package(v);
}

bool myMarkPolicy::isUserInstalled(const PkgIterator &p)
{
    pkgCache::VerIterator const& cur = p.CurrentVer();

    return isVisible(cur) && !isOSMetaPackage(cur);
}

bool myMarkPolicy::IsInstallOk(const PkgIterator &Pkg, bool AutoInst,
			       unsigned long Depth, bool FromUser)
{
    VerIterator cand = (*DCache)[Pkg].CandidateVerIter(*DCache);

    if (debug)
	DEBUG("%s (%s)", Pkg.Name(), cand.VerStr());
    
    DepIterator Dep = cand.DependsList();
    for (; Dep.end() == false; Dep++) {
        DepIterator Start = Dep;
        if ( Start->Type == pkgCache::Dep::DpkgBreaks ) {
            PkgIterator pkg = Start.TargetPkg();
            VerIterator ver = pkg.CurrentVer(); // Current version? Or candidate?
            if ( !ver.end() && isOSMetaPackage( ver ) ) {
                DEBUG("The package breaks OS meta package: %s (%s)", pkg.Name(), ver.VerStr());
                breaksOSMetaPackage = true;
                return false;
            }
        }
    }

    if (forbidMarkOSMetaPackage && isOSMetaPackage(cand))
    {
	DEBUG("Error: OS metapackage marked for Install: %s (%s)",
	    Pkg.Name(), cand.VerStr());
	installOSMetaPackageVer = cand;

	return !forbidMarkOSMetaPackageStops;
    }

    return true;
}

bool myMarkPolicy::IsDeleteOk(const PkgIterator &Pkg, bool AutoInst,
			      unsigned long Depth, bool FromUser)
{
    VerIterator cur = Pkg.CurrentVer();
    bool stop = false;

    if (debug)
	DEBUG("%s (%s)", Pkg.Name(), cur.VerStr());

    if (forbidMarkOSMetaPackage && isOSMetaPackage(cur))
    {
	DEBUG("Error: OS metapackage marked for Delete: %s (%s)",
	    Pkg.Name(), cur.VerStr());
	deleteOSMetaPackageVer = cur;
	stop = forbidMarkOSMetaPackageStops;
    }
    else if (forbidBreakVisiblePackage
	     && isUserInstalled(Pkg))
    {
	bool isDep = false;
	bool conflictsOrBreaks = false;

	for (DepIterator dep = Pkg.RevDependsList(); !dep.end(); dep++)
	{
	    if (!dep.Reverse())
		continue;

	    DEBUG("Target: %s, Parent: %s", dep.TargetPkg().Name(), dep.ParentPkg().Name());

	    if ((*DCache)[dep.ParentPkg()].Keep())
		continue;

	    isDep = true;
	    pkgCache::Dependency const& d = *dep;

	    DEBUG("%s %s %s",
		dep.ParentPkg().Name(),
		DCache->GetCache().DepType(d.Type),
		Pkg.Name());

	    if (d.Type == pkgCache::Dep::Conflicts
		|| d.Type == pkgCache::Dep::DpkgBreaks)
	    {
		conflictsOrBreaks = true;
		DEBUG("%s marked for Delete, but since parent %s %s it, it's ok",
		    Pkg.Name(), dep.ParentPkg().Name(),
		    DCache->GetCache().DepType(d.Type));
		break;
	    }
	}

	if (isDep && !conflictsOrBreaks)
	{
	    DEBUG("Error: User-installed package marked for Delete: %s (%s)",
		Pkg.Name(), cur.VerStr());
	    breakVisiblePackageVer = cur;
	    stop = forbidBreakVisiblePackageStops;
	}
    }

    return !stop;
}

void myMarkPolicy::error(GError **error) const
{
    if (*error)
	DEBUG("");

    if ( breaksOSMetaPackage ) {
        g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE_WITH_SYSTEM, "Package breaks/replaces OS meta package");
    } else if (!installOSMetaPackageVer.end()
	&& isOSMetaPackage(installOSMetaPackageVer))
    {
        pkgCache::PkgIterator pkg = installOSMetaPackageVer.ParentPkg();
        pkgCache::VerIterator cur = pkg.CurrentVer();
        if (cur.end())
        {
            g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_REQUIRES_SYSTEM_PACKAGE,
                        "Package requires installation of incompatible system upgrade '%s' version '%s'",
                        pkg.Name(),
                        installOSMetaPackageVer.VerStr());
        }
        else
        {
            int c = installOSMetaPackageVer.CompareVer(cur);
            if (c > 0)
            {
                g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE,
                            "Package requires system upgrade '%s' version '%s'",
                            pkg.Name(),
                            installOSMetaPackageVer.VerStr());
            }
            else
                g_set_error(error, PKGMGRD_ERROR, ERROR_NEED_DOWNGRADE,
                            "Package requires system metapackage '%s' version '%s'",
                            pkg.Name(),
                            installOSMetaPackageVer.VerStr());
        }
    }
    else if (!deleteOSMetaPackageVer.end())
    {
	g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE_WITH_SYSTEM,
		    "Package is not compatible with system package '%s' version '%s'",
		    deleteOSMetaPackageVer.ParentPkg().Name(),
		    deleteOSMetaPackageVer.VerStr());
    }
    else if (!breakVisiblePackageVer.end())
    {
	g_set_error(error, PKGMGRD_ERROR, ERROR_DEPENDENCIES_PREVENT_OPERATION,
		    "Would remove user-installed package: %s (%s)",
		    breakVisiblePackageVer.ParentPkg().Name(),
		    breakVisiblePackageVer.VerStr());
    }
}

pkgCache::VerIterator myMarkPolicy::illegalSysVer() const
{
    VerIterator v;
    if (v = installOSMetaPackageVer, !v.end())
	return v;
    else if (v = deleteOSMetaPackageVer, !v.end())
	return v;

    return v;
}
