/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <string.h>

#include "apt-worker-operation.h"
#include "debug.h"
#include "package.h"

OperationData*
operation_data_new ()
{
	OperationData *od = g_new0(OperationData, 1);
	od->current_bytes = -1;
	od->total_bytes = -1;
	od->operation_progress_pct = -1;
	od->error = "";
	od->send_signals = TRUE;
	od->send_dep_signals = TRUE;
	od->revert_install = FALSE;
	return od;
}

static void
__free_dep_list(GList *deps)
{
	GList *list = deps;
	while (list) {
		PackageVersion *p = (PackageVersion*)list->data;
		package_version_free (p);
		list = g_list_next(list);
	}
	g_list_free (deps);
}

void
operation_data_free (OperationData* od)
{
	if (!od) {
		return;
	}
	__free_dep_list(od->installed_user_deps);
	__free_dep_list(od->updated_user_deps);
	__free_dep_list(od->removed_user_deps);
	__free_dep_list(od->pending_pkgs);
	g_free (od->pkg);
	g_free (od->version);
	g_free (od->filename);
	g_free (od->domain);
	g_free ( od );
}

GList* copy_package_version_list (GList *list)
{
	GList* copy = NULL;
	while (list) {
		PackageVersion *p = (PackageVersion*)list->data;
		copy = g_list_append(copy, package_version_copy (p));
		list = g_list_next (list);
	}
	return copy;
}

OperationData*
operation_data_copy (OperationData* od)
{
	if (!od) {
		return NULL;
	}
	OperationData* copy = operation_data_new ();
	copy->pkg = g_strdup(od->pkg);
	copy->version = g_strdup (od->version);
	copy->filename = g_strdup (od->filename);
	copy->domain = g_strdup (od->domain);
	copy->type = od->type;
	copy->current_bytes = od->current_bytes;
	copy->total_bytes = od->total_bytes;
	copy->operation_progress_pct = od->operation_progress_pct;
	copy->error = od->error;
	copy->installed_user_deps = copy_package_version_list (od->installed_user_deps);
	copy->updated_user_deps = copy_package_version_list (od->updated_user_deps);
	copy->removed_user_deps = copy_package_version_list (od->removed_user_deps);
	copy->pending_pkgs = copy_package_version_list (od->pending_pkgs);
	copy->rejected_pkgs = copy_package_version_list (od->rejected_pkgs);
	copy->pkgflags = od->pkgflags;
	copy->send_signals = od->send_signals;
	copy->send_dep_signals = od->send_dep_signals;
	copy->update_failed_installs = od->update_failed_installs;
	copy->revert_install = od->revert_install;
	return copy;
}

gboolean
need_reboot (int pkgflags)
{
	return (pkgflags & (pkgflag_reboot | pkgflag_flash_and_reboot | pkgflag_reboot_update)) != 0;
}

gboolean
operation_need_reboot (OperationData* od)
{
	return need_reboot (od->pkgflags);
}

gboolean
operation_failed (const OperationData *od)
{
	return !(od->error == NULL || !strcmp(od->error, ""));
}

const gchar*
operation_type_to_string(OperationType type)
{
	switch (type) {
	case OPERATION_INSTALL:
		return "Install";
	case OPERATION_INSTALL_FILE:
		return "InstallFile";
	case OPERATION_UNINSTALL:
		return "Uninstall";
	case OPERATION_REFRESH:
		return "Refresh";
	case OPERATION_UPGRADE:
		return "Upgrade";
	case OPERATION_UPGRADE_ALL:
		return "UpgradeAll";
	case OPERATION_SYSTEM_UPGRADE:
		return "SystemUpgrade";
	case OPERATION_DOWNLOAD_SYSTEM_UPGRADE:
		return "DownloadSystemUpgrade";
	case OPERATION_RESTORE_INSTALL:
	case OPERATION_RESTORE_UPGRADE:
		return "Restore";
	default:
		return 0;
	}
}

PackageVersion*
package_version_new ()
{
	PackageVersion *pv = g_new0(PackageVersion, 1);
	return pv;
}

void
package_version_free (PackageVersion* pv)
{
	if (pv != NULL) {
		g_free (pv->pkg);
		g_free (pv->version);
                if (pv->error)
                  g_error_free (pv->error);
	}
}

PackageVersion*
package_version_copy (PackageVersion* pv)
{
	PackageVersion* copy = NULL;
	if (pv != NULL) {
		copy = package_version_new ();
		copy->pkg = g_strdup (pv->pkg);
		copy->version = g_strdup (pv->version);
		if (pv->error)
			copy->error = g_error_copy (pv->error);
	}
	return copy;
}
