/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "apt-worker-rescue.h"
#include "apt-worker-cache.h"
#include "apt-worker.h"
#include "catalogues.h"
#include "confutils.h"
#include "debug.h"
#include "xexp.h"
#include <cerrno>
#include <cstdlib>
#include <cstring>
#include <fcntl.h>
#include <unistd.h>
#include <sys/stat.h>
#include <apt-pkg/init.h>
#include <apt-pkg/error.h>

#define DBG if (verbosity >= 2) log_stderr
#define CURRENT_OPERATION_FILE "/var/lib/package-manager/current-operation"

static bool allow_reboot = true;
static bool only_rescue_with_download_dir = true;
static int pkgflags = 0;
extern const char *lc_messages;
extern int status_fd;
extern bool flag_collect_dpkg_output;


extern void
reboot_device (bool update_mode, int delay = 0);

extern bool
mark_named_package_for_install (const char *package, const char *version, const char *domain, gboolean only_user, gboolean only_user_system, gboolean upgrade, GError **error);

static bool
mark_for_install (const char *op, const char *package, int &error_code)
{
  GError *error = NULL;
  bool ok;

  error_code = 0;

  if (!strcmp (op, "Install"))
    ok = mark_named_package_for_install (package, NULL, NULL, TRUE, FALSE, FALSE, &error);
  else if (!strcmp (op, "Upgrade"))
    ok = mark_named_package_for_install (package, NULL, NULL, TRUE, FALSE, TRUE, &error);
  else if (!strcmp (op, "DownloadSystemUpgrade") || !strcmp (op, "SystemUpgrade"))
    ok = mark_named_package_for_install (package, NULL, NULL, TRUE, TRUE, TRUE, &error);
  else
    {
      fprintf (stderr, "error: invalid operation `%s'\n", op);
      return false;
    }

  if (!ok && error)
    {
      error_code = error->code;
      fprintf (stderr, "mark_named_package_for_install() failed: %d: %s\n",
	       error->code, error->message);
      g_error_free (error);
    }

  return ok;
}

static xexp *
read_operation_record ()
{
  struct stat buf;

  /* This check is not strictly necessary but it avoids a distracting
     complaint from xexp_read_file in the common case that the file
     doesn't exist.
  */
  if (stat (CURRENT_OPERATION_FILE, &buf))
    return NULL;

  return xexp_read_file (CURRENT_OPERATION_FILE);
}

static int
run_system (bool verbose, const char *fmt, ...)
{
  va_list ap;
  va_start (ap, fmt);
  char *cmd = g_strdup_vprintf (fmt, ap);
  va_end (ap);

  if (verbose)
    fprintf (stderr, "+ %s\n", cmd);
  int result = system (cmd);
  free (cmd);

  return result;
}

static int
rescue_operation_with_dir (const char *dir)
{
  GError *error = NULL;
  int result, tmp;

  fprintf (stderr, "Rescuing from %s\n", dir);
  result = operation (false, dir, false, &error, false, false, &tmp);
  fprintf (stderr, "Result code %d\n", result);
  if (error)
    fprintf (stderr, "Error: %d: %s\n", error->code, error->message);
  _error->DumpErrors ();
  return result;
}

static int
rescue_operation_with_dev (const char *dev)
{
  const char *dir = "/rescue";

  fprintf (stderr, "Rescuing from %s\n", dev);

  if (mkdir (dir, 0777) < 0
      && errno != EEXIST)
    log_stderr ("%s: %m", dir);

  run_system (true, "mount -t vfat '%s' '%s'", dev, dir);
  int result = rescue_operation_with_dir (dir);
  run_system (true, "umount '%s'", dir);

  if (rmdir (dir) < 0)
    log_stderr ("%s: %m", dir);

  return result;
}

static int
rescue_operation_with_devnode (int major, int minor)
{
  const char *node = "/dev.rescue";

  if (mknod (node, S_IFBLK | 0600, major << 8 | minor) < 0)
    log_stderr ("%s: %m", node);

  fprintf (stderr, "Rescuing from %d:%d\n", major, minor);

  int result = rescue_operation_with_dev (node);

  unlink (node);

  return result;
}

static const char* rescue_devs[] = {
  "/dev/mmcblk0p1", // Internal MMC ~/MyDocs
  "/dev/mmcblk1p1", // External MMC
  "/dev/mmcblk0p2", // Internal MMC /home
  "/dev/mmcblk0p3", // Internal MMC swap
  "/dev/mmcblk0",
  "/dev/mmcblk1",
  NULL
};

static int
rescue_with_all_devs ()
{
  int result = rescode_packages_not_found;

  for (int i = 0;
       (result == rescode_packages_not_found
	&& rescue_devs[i] != NULL);
       i++)
    {
      result = rescue_operation_with_dev (rescue_devs[i]);
    }

  return result;
}

static struct {
  int major, minor;
} rescue_devnodes[] = {
  { 254, 9 },
  { 254, 1 },
  { 254, 8 },
  { 254, 0 },
  { -1 }
};

static int
rescue_with_all_devnodes ()
{
  int result = rescode_packages_not_found;

  for (int i = 0;
       (result == rescode_packages_not_found
	&& rescue_devnodes[i].major != -1);
       i++)
    {
      result = rescue_operation_with_devnode (rescue_devnodes[i].major,
					      rescue_devnodes[i].minor);
    }

  return result;
}

static void
show_fb_text (int line, const char *text)
{
  run_system (false,
	      "/usr/bin/text2screen -s 2 -x 5 -y %d -B 0 -T 0xF000 -t '%s '",
	      400 + 20*line, text);
}

static void
show_fb_status (int percent)
{
  run_system (false,
	      "/usr/bin/text2screen -s 2 -x 5 -y %d -B 0 -T 0xF000 -t '%3d%% '",
	      440, percent);
}

static void
interpret_pmstatus (char *str)
{
  float percentage;
  const char *title;

  if (!strncmp (str, "pmstatus:", 9))
    {
      str += 9;
      str = strchr (str, ':');
      if (str == NULL)
	return;
      str += 1;
      percentage = atof (str);
      str = strchr (str, ':');
      if (str == NULL)
	title = "Working";
      else
	{
	  str += 1;
	  title = str;
	}

      show_fb_status ((int)percentage);
      // show_fb_text (2, title);
    }
}

static void
fork_progress_process ()
{
  int fds[2], child_pid;

  if (pipe (fds) < 0)
    {
      perror ("pipe");
      return;
    }

  if ((child_pid = fork ()) < 0)
    {
      perror ("fork");
      return;
    }

  if (child_pid == 0)
    {
      close (fds[1]);
      FILE *f = fdopen (fds[0], "r");
      if (f)
	{
	  char *line = NULL;
	  size_t len = 0;
	  ssize_t n;

	  while ((n = getline (&line, &len, f)) != -1)
	    {
	      if (n > 0 && line[n-1] == '\n')
		line[n-1] = '\0';

	      interpret_pmstatus (line);
	    }
	}
      exit (0);
    }
  else
    {
      close (fds[0]);
      status_fd = fds[1];
    }
}

/* MMC default mountpoints */
#define INTERNAL_MMC_MOUNTPOINT  "/home/user/MyDocs"
#define REMOVABLE_MMC_MOUNTPOINT "/media/mmc1"

static void
misc_init ()
{
  lc_messages = getenv ("LC_MESSAGES");
  DBG ("LC_MESSAGES %s", lc_messages);

  DBG ("OSSO_PRODUCT_HARDWARE %s", getenv ("OSSO_PRODUCT_HARDWARE"));

  load_system_settings ();
  read_domain_conf ();

  AptWorkerCache::Initialize ();

  cache_init (false);

// #ifdef HAVE_APT_TRUST_HOOK
//   apt_set_index_trust_level_for_package_hook (index_trust_level_for_package);
// #endif

  clean_temp_catalogues ();

  // initialize the MMC mount points with defaults
  setenv ("INTERNAL_MMC_MOUNTPOINT", INTERNAL_MMC_MOUNTPOINT, 1);
  setenv ("REMOVABLE_MMC_MOUNTPOINT", REMOVABLE_MMC_MOUNTPOINT, 1);
}

extern int
must_open_nonblock (const char *filename, int flags);

extern bool
must_mkfifo (const char *filename, int mode);

extern int
get_flags_union ();

static void
reboot_and_sleep (bool update)
{
  if (update)
    show_fb_text (2, "Done, rebooting to update mode...");
  else
    show_fb_text (2, "Done, rebooting...");

  fprintf (stderr, "Rebooting: update_mode=%d\n", update);
  reboot_device (update);

  fprintf (stderr, "Sleeping forever...\n");

  for (;;)
    sleep (60);
}

static void
do_rescue (const char *op, const char *package, const char *download_root,
	   bool erase_record)
{
  show_fb_text (0, "Rescuing software update.");
  show_fb_text (1, "Please do not interrupt.");

  must_mkfifo ("/tmp/apt-worker.status", 0600);
  must_open_nonblock ("/tmp/apt-worker.status", O_RDONLY);
  status_fd = must_open_nonblock ("/tmp/apt-worker.status", O_WRONLY);

  fork_progress_process ();

  fprintf (stderr, "Rescuing %s\n", package);

  /* This is just to clean the dpkg journal.  We let libapt-pkg
     configure the rest of the packages since we will get better
     progress reporting that way.
   */
  run_system (true, "dpkg --configure dpkg");

  misc_init ();

  /* Dpkg --reckless leaves many packages behind in the reinstreq
     state when interrupted.  We allow it to remove them anyway, which
     it might not be able to do cleanly since their list of files
     might not be up-to-date among other things, but that's still
     better than failing.
  */
  _config->Set ("DPkg::Options::", "--force-remove-reinstreq");

  // @todo Is this really necessary?
  AptWorkerCache::GetCurrent ()->init_cache_after_request = false;
  if (ensure_cache (false))
    {
      int error_code = 0;
      if (mark_for_install (op, package, error_code))
	{
	  pkgflags = get_flags_union ();

	  int result = rescue_operation_with_dir (download_root);

	  if (!only_rescue_with_download_dir)
	  {
	      if (result != rescode_success)
		  result = rescue_with_all_devs ();

	      if (result != rescode_success)
		  result = rescue_with_all_devnodes ();
	  }

	  /* If we get this far, we have done everything we can.  If
	     the rescue operation is interrupted before getting here,
	     we try again on the next boot.  This might lead to a
	     reboot loop, but the device is borked anyway and the user
	     should reflash it.
	   */
	  if (erase_record)
	    erase_operation_record ();

	  if (result == rescode_packages_not_found)
      {
        run_system (false, "/bin/umount /home");

	// Rescue can't possibly succeed without downloading more
	// packages.
	erase_operation_record ();
	return;
      }

	  if (result == rescode_success)
	    {
	      fprintf (stderr, "Success\n");
	      erase_operation_record ();
	      unlink ("/var/lib/package-manager/rescue-reboot-count");
	    }

	  if (result != rescode_success)
	    {
	      _system->UnLock();
	      show_fb_text (2, "Configuring all packages...");
	      run_system (true, "dpkg --configure -a --force-all");
	      _system->Lock();
	    }

	  run_system (false, "/bin/umount /home");

	  if (allow_reboot)
	    {
	      bool update = pkgflags & pkgflag_reboot_update;
	      reboot_and_sleep (update);
	    }
	  else
	    fprintf (stderr, "Reboot not allowed, exiting\n");
	}
      else
	{
	  if (error_code == ERROR_PACKAGE_ALREADY_NEWEST)
	    {
	      erase_operation_record ();

	      if (allow_reboot)
		reboot_and_sleep (false);
	    }
	}
    }
  else
    fprintf (stderr, "Failed to initialize package cache\n");
}

int
cmdline_rescue (char **argv)
{
  if (in_scratchbox ())
    allow_reboot = false;

  assume_network = true;
  flag_collect_dpkg_output = false;

  if (argv[1] == NULL)
    {
      xexp *record = read_operation_record ();

      if (record == NULL)
	{
	  fprintf (stderr, "apt-worker: Nothing to rescue.\n");
	  erase_operation_record ();
	  return 0;
	}

      const char *op = xexp_tag (record);
      if (g_strcmp0 (op, "SystemUpgrade") != 0)
	{
	  fprintf (stderr, "apt-worker: Operation %s does not require rescue.\n", op);
	  erase_operation_record ();
	  return 0;
	}

      const char *package = xexp_aref_text (record, "package");
      const char *download_root = xexp_aref_text (record, "download-root");

      do_rescue (op, package, download_root, false);
    }
  else
    do_rescue (argv[1], argv[2], argv[3], false);

  return 0;
}
