/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

// -*- Mode: C++; c-basic-offset: 8 -*-
#include <cassert>
#include <cerrno>
#include <sys/wait.h>
#include <stdlib.h>
#include <apt-pkg/configuration.h>
#include "apt-worker-cache.h"
#include "apt-worker-struct.h"
#include "apt-worker-thread.h"
#include "apt-worker-internal.h"
#include "debug.h"
#include "error.h"
#include "error-util.h"
#include "log.h"
#include "package-list.h"
#include "package-map.h"

static void waitpid_log(pid_t pid, const char *func, GError **error)
{
	int status;
	TEMP_FAILURE_RETRY(waitpid(pid, &status, NULL));

	if (WIFSIGNALED(status)) {
		DEBUG("%s: child terminated due to signal %d", func, WTERMSIG(status));
		g_set_error(error, PKGMGRD_ERROR, ERROR_INTERNAL, "%s: child terminated due to signal %d", func, WTERMSIG(status));
	}
}

#define WAITPID(pid, error) waitpid_log(pid, __func__, error)

static const char *pm_query_bin = 0;

static void ensure_pm_query_bin()
{
	if (!pm_query_bin)
		pm_query_bin =
		    strdup(_config->FindFile("Dir::Bin::pm-query").
			   c_str());
}

static void vexec_pm_query(const char *cmd, va_list ap)
{
	const char *argv[8] = {pm_query_bin, cmd, 0};
	int i = 2;
	const char *arg;

	for (;;) {
		arg = va_arg(ap, const char *);
		if (!arg) {
			argv[i] = 0;
			break;
		}
		argv[i++] = arg;
	}

	execv(pm_query_bin, (char *const *) argv);
}

static void exec_pm_query(const char *cmd, ...)
{
	va_list ap;
	va_start(ap, cmd);
	vexec_pm_query(cmd, ap);
	va_end(ap);
}

static GHashTable *
__forked_cmd_get_file_details_map (const char *filename,
				   GError **error)
{
	GHashTable *p = 0;

	ensure_pm_query_bin();
	cache_r_lock ();

	int fds[2] = {0};
	pipe (fds);
	pid_t pid = fork ();

	if (pid < 0)
	{
		/// @todo Handle error
		p = NULL;
	}
	else if (pid == 0)
	{
		close (fds[0]);
		dup2(fds[1], STDOUT_FILENO);
		exec_pm_query("show-file", filename, 0);
		exit(1);
	}
	else
	{
		close (fds[1]);
		FILE *f = fdopen (fds[0], "r");

		g_error_fread (error, f);
		package_hash_table_fread (&p, f);
		fclose (f);

		WAITPID (pid, error);
	}

	cache_r_unlock ();

	return p;
}

static GHashTable *
__forked_cmd_get_package_details_map (const char *package,
				      const char *version,
				      const char *domain,
				      gboolean installed,
				      GError **error)
{
	GHashTable *p = 0;

	ensure_pm_query_bin();
	cache_r_lock ();

	int fds[2] = {0};
	pipe (fds);
	pid_t pid = fork ();

	if (pid < 0)
	{
		/// @todo Handle error
		p = NULL;
	}
	else if (pid == 0)
	{
		close (fds[0]);
		dup2(fds[1], STDOUT_FILENO);
		if (installed)
			exec_pm_query("show-installed", package, 0);
		else
			exec_pm_query("show", package, version, domain, 0);
		exit(1);
	}
	else
	{
		close (fds[1]);
		FILE *f = fdopen (fds[0], "r");

		g_error_fread (error, f);
		package_hash_table_fread (&p, f);
		fclose (f);

		WAITPID (pid, error);
	}

	cache_r_unlock ();

	return p;
}

static void
__forked_cmd_get_package_list (const gchar *category,
			       PkgType type,
			       const gchar *pattern,
			       GList **packages,
			       PkgSort sort,
			       GError **error)
{
	ensure_pm_query_bin();
	cache_r_lock ();

	const char *cmd;
	switch (type) {
	case PKG_TYPE_INSTALLED:
		cmd = "list-installed";
		break;
	case PKG_TYPE_UPDATE:
		cmd = "list-updates";
		break;
	case PKG_TYPE_NOT_INSTALLED:
		cmd = "list-available";
		break;
	default:
		assert(!"Invalid command");
		break;
	}

	int fds[2] = {0};
	pipe (fds);
	pid_t pid = fork ();

	if (pid < 0)
	{
		/// @todo Handle error
	}
	else if (pid == 0)
	{
		close (fds[0]);
		dup2(fds[1], STDOUT_FILENO);

		if (category)
			exec_pm_query(cmd, category, 0);
		else
			exec_pm_query(cmd, 0);
		exit(1);
	}
	else
	{
		close (fds[1]);
		FILE *f = fdopen (fds[0], "r");

		g_error_fread (error, f);
		package_list_fread (packages, f);
		fclose (f);

		WAITPID (pid, error);
	}

	cache_r_unlock ();
}

static GList *
__forked_cmd_get_package_dep_details (const char *package,
				      const char *version,
				      GError **error)
{
	ensure_pm_query_bin();
	cache_r_lock ();

	int fds[2] = {0};
	pipe (fds);
	pid_t pid = fork ();

	GList *packages = NULL;
	// GList **pp = &packages;

	if (pid < 0)
	{
		/// @todo Handle error
	}
	else if (pid == 0)
	{
		close (fds[0]);
		dup2(fds[1], STDOUT_FILENO);

		exec_pm_query("show-deps", package, version, 0);
		exit (1);
	}
	else
	{
		close (fds[1]);
		FILE *f = fdopen (fds[0], "r");

		g_error_fread (error, f);
		package_hash_table_list_fread (&packages, f);
		fclose (f);

		WAITPID (pid, error);
	}

	cache_r_unlock ();

	return packages;
}

gchar* __forked_cmd_get_restore_status(GError **error)
{
	ensure_pm_query_bin();
	cache_r_lock ();

	gchar *status = 0;
	int fds[2] = {0};
	pipe (fds);
	pid_t pid = fork ();

	if (pid < 0)
	{
		/// @todo Handle error
	}
	else if (pid == 0)
	{
		close (fds[0]);
		dup2(fds[1], STDOUT_FILENO);

		exec_pm_query("restore-status", 0);
		exit(1);
	}
	else
	{
		close (fds[1]);
		FILE *f = fdopen (fds[0], "r");

		g_error_fread (error, f);

		char *line = 0;
		size_t n = 0;
		getline(&line, &n, f);
		int len = strlen(line);
		if (len > 0)
			status = g_strndup(line, len - 1);
		free(line);

		fclose (f);

		WAITPID (pid, error);
	}

	cache_r_unlock ();

	return status;
}

static gboolean __forked_cmd_get_restorable_packages(GList** packages, GError **error)
{
	ensure_pm_query_bin();
	cache_r_lock ();

	int fds[2] = {0};
	pipe (fds);
	pid_t pid = fork ();

	if (pid < 0)
	{
		/// @todo Handle error
	}
	else if (pid == 0)
	{
		close (fds[0]);
		dup2(fds[1], STDOUT_FILENO);

		exec_pm_query("list-restorable", 0);
		exit(1);
	}
	else
	{
		close (fds[1]);
		FILE *f = fdopen (fds[0], "r");

		g_error_fread (error, f);
		package_list_fread (packages, f);
		fclose (f);

		WAITPID (pid, error);
	}

	cache_r_unlock ();

	return *error == 0;
}

static GHashTable *
__forked_cmd_simulate_operation(const gchar* operation, const gchar* packageidentifier, const gchar* version, const char *domain, const gchar* flags, GError **error)
{
	cache_r_lock ();

	int fds[2] = {0};
	pipe (fds);
	pid_t pid = fork ();

	GHashTable *res = NULL;
	// GList **pp = &packages;

	if (pid < 0)
	{
		/// @todo Handle error
	}
	else if (pid == 0)
	{
		close (fds[0]);
		FILE *f = fdopen (fds[1], "w");

		res = cmd_simulate_operation (operation, packageidentifier, version, domain, flags, error);

		g_error_fwrite (*error, f);

		if (*error == 0)
		{
			package_hash_table_fwrite (res, f);
			// package_hash_table_list_fwrite (packages, stdout);
			fflush (f);
		}
		fclose (f);

		exit (error != 0);
	}
	else
	{
		close (fds[1]);
		FILE *f = fdopen (fds[0], "r");

		g_error_fread (error, f);
		package_hash_table_fread (&res, f);
		fclose (f);

		WAITPID (pid, error);
	}

	cache_r_unlock ();

	return res;	
}

static AptWorker apt_worker_funcs = {
    apt_worker_init,
    apt_worker_reset_state,
    __forked_cmd_get_package_list,
    cmd_get_file_details,
    __forked_cmd_get_file_details_map,
    __forked_cmd_get_package_details_map,
    __forked_cmd_get_package_dep_details,
    __forked_cmd_get_restore_status,
    __forked_cmd_get_restorable_packages,
    cmd_ignore_update_package,
    cmd_refresh_cache,
    cmd_get_current_operation_data,
    __forked_cmd_simulate_operation,
    cmd_install_file,
    cmd_install_package,
    cmd_upgrade_package,
    cmd_restore_package,
    cmd_ignore_restore_package,
    cmd_download_system_upgrade,
    cmd_upgrade_system,
    cmd_upgrade_all,
    cmd_uninstall_package,
    cmd_pkgmgr_fetch_files_of_installed_package,
    cmd_check_updates,
    cmd_cancel
// 	cmd_get_catalogues,
// 	cmd_add_catalogue,
// 	cmd_remove_catalogue,
// 	cmd_enable_catalogue
};

AptWorker *apt_worker_new ()
{
	return &apt_worker_funcs;
}
