/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

// -*- Mode: C++; c-basic-offset: 8 -*-
#include <stdio.h>
#include <unistd.h>

#include "debug.h"
#include "apt-worker-thread.h"

#define APT_WORKER_THREAD_STOP_CMD reinterpret_cast<gpointer>(-1)
#define IDLE_TIMER_TIMEOUT 120000

namespace {
struct GMutexLocker
{
	GMutexLocker(GMutex *m) : m(m) { g_mutex_lock(m); }
	~GMutexLocker() { g_mutex_unlock(m); }
	GMutex *m;
};
}

AptWorkerThread::AptWorkerThread (AptWorker *apt_worker,
				  GSourceFunc stopped_fn,
				  gpointer stopped_data,
				  DownloadProgressFunc download_progress_fn,
				  OperationStartedFunc operation_started_fn,
				  OperationProgressFunc operation_progress_fn,
				  OperationCompleteFunc operation_complete_fn,
				  PackageListUpdatedFunc package_list_updated_fn,
				  PackageIgnoredFunc package_ignored_fn,
				  gpointer signal_data,
				  gboolean sticky)
	: _apt_worker(apt_worker),
	  _stopped_fn(stopped_fn),
	  _stopped_data(stopped_data),
	  _download_progress_fn (download_progress_fn),
	  _operation_started_fn (operation_started_fn),
	  _operation_progress_fn (operation_progress_fn),
	  _operation_complete_fn (operation_complete_fn),
	  _package_list_updated_fn (package_list_updated_fn),
	  _package_ignored_fn (package_ignored_fn),
	  _signal_data (signal_data),
	  _cmd_worker (0),
	  _cmd_q (0),
	  _query_worker (0),
	  _query_q (0),
	  _stopping(FALSE),
	  _cmd_executing(false),
	  _query_executing(false),
	  _idle_timer_id(0),
	  _state_mutex(g_mutex_new()),
	  _sticky(sticky)
{
	g_assert (apt_worker != 0);

	_apt_worker->init (_download_progress_fn,
	           _operation_started_fn,
			   _operation_progress_fn,
			   _operation_complete_fn,
			   _package_list_updated_fn,
			   _package_ignored_fn,
			   _signal_data);
        start_idle_timer();
}

AptWorkerThread::~AptWorkerThread ()
{
	if (_cmd_worker != 0) {
		stop ();
 	        g_thread_join (_cmd_worker);
 	        g_thread_join (_query_worker);
	}
	if (_cmd_q)
		g_async_queue_unref (_cmd_q);
	if (_query_q)
		g_async_queue_unref (_query_q);
	g_mutex_free(_state_mutex);
}

static gboolean
idle_timer_timeout(gpointer data)
{
        DEBUG("");
        reinterpret_cast<AptWorkerThread *> (data)->stop();
        return FALSE; // Do not call this again.
}

void
AptWorkerThread::set_cmd_executing(bool b)
{
	GMutexLocker lock(_state_mutex);
	_cmd_executing = b;

	if (b)
		stop_idle_timer();
	else if (!is_busy())
		start_idle_timer();
}

void
AptWorkerThread::set_query_executing(bool b)
{
	GMutexLocker lock(_state_mutex);
	_query_executing = b;

	if (b)
		stop_idle_timer();
	else if (!is_busy())
		start_idle_timer();
}

void
AptWorkerThread::start_idle_timer()
{
        if ( _idle_timer_id ) {
                stop_idle_timer();
        }
        if ( ! _sticky ) {
                _idle_timer_id = g_timeout_add(IDLE_TIMER_TIMEOUT, idle_timer_timeout, this);
        }
}

void
AptWorkerThread::stop_idle_timer()
{
        if ( _idle_timer_id ) {
                g_source_remove( _idle_timer_id );
                _idle_timer_id = 0;
        }
}

static gboolean
apt_worker_callback (gpointer data)
{
	apt_worker_cmd *cmd = reinterpret_cast<apt_worker_cmd *> (data);
	cmd->callback (cmd);

	return FALSE;
}

void
AptWorkerThread::cmd_thread_func ()
{
	apt_worker_cmd *cmd;

	while ((cmd = reinterpret_cast<apt_worker_cmd *> (g_async_queue_pop (_cmd_q)))
	       != APT_WORKER_THREAD_STOP_CMD)
	{
		set_cmd_executing(true);
		cmd->success = cmd->execute ();
		g_idle_add (apt_worker_callback, cmd);
		set_cmd_executing(false);
	}
	if (_stopped_fn ) {
		g_idle_add (_stopped_fn, _stopped_data);
	}
}

void
AptWorkerThread::query_thread_func ()
{
	apt_worker_cmd *cmd;

	while ((cmd = reinterpret_cast<apt_worker_cmd *> (g_async_queue_pop (_query_q)))
	       != APT_WORKER_THREAD_STOP_CMD)
	{
		set_query_executing(true);
		cmd->success = cmd->execute ();
		g_idle_add (apt_worker_callback, cmd);
		set_query_executing(false);
	}
	// if (_stopped_fn ) {
	// 	g_idle_add (_stopped_fn, _stopped_data);
	// }
}

static gpointer
apt_worker_cmd_thread (gpointer data)
{
	reinterpret_cast<AptWorkerThread *> (data)->cmd_thread_func ();

	return 0;
}

static gpointer
apt_worker_query_thread (gpointer data)
{
	reinterpret_cast<AptWorkerThread *> (data)->query_thread_func ();

	return 0;
}

void
AptWorkerThread::start ()
{
	g_assert (_cmd_worker == 0);
	g_assert (_cmd_q == 0);
	g_assert (_query_worker == 0);
	g_assert (_query_q == 0);

	_stopping = FALSE;

	_cmd_q = g_async_queue_new ();
	_query_q = g_async_queue_new ();

	_cmd_worker = g_thread_create (apt_worker_cmd_thread,
				       this,
				       TRUE,
				       0);

	_query_worker = g_thread_create (apt_worker_query_thread,
					 this,
					 TRUE,
					 0);
}

gboolean
AptWorkerThread::is_stopping ()
{
	return _stopping;
}
void
AptWorkerThread::stop ()
{
	g_assert (_cmd_worker != 0);
	g_assert (_query_worker != 0);

	_stopping = TRUE;
	g_async_queue_push (_cmd_q, APT_WORKER_THREAD_STOP_CMD);
	g_async_queue_push (_query_q, APT_WORKER_THREAD_STOP_CMD);
// 	_cmd_worker = 0;
// 	_query_worker = 0;
}

void
AptWorkerThread::push_cmd (apt_worker_cmd *cmd,
			   apt_worker_cmd::callback_t callback)
{
	g_assert (_cmd_worker != 0);
	g_assert (callback != 0);

	cmd->apt_worker = _apt_worker;
	cmd->callback = callback;
	g_async_queue_push (_cmd_q, cmd);
}

void
AptWorkerThread::push_query (apt_worker_cmd *cmd,
			     apt_worker_cmd::callback_t callback)
{
	g_assert (_query_worker != 0);
	g_assert (callback != 0);

	cmd->apt_worker = _apt_worker;
	cmd->callback = callback;
	g_async_queue_push (_query_q, cmd);
}
