/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <errno.h>

#include <glib.h>
#include <conic.h>
#include <list>
#include <sstream>

#include <apt-pkg/sptr.h>
#include <apt-pkg/configuration.h>
#include <apt-pkg/error.h>

#include "apt-worker-cache.h"
#include "format-metadata.h"
#include "package.h"
#include "debug.h"
#include "error.h"
#include "confutils.h"
#include "apt-worker-utils.h"
#include "netconnection_defs.h"

/* Where we keep a journal of the current operation
 */
#define CURRENT_OPERATION_FILE "/var/lib/package-manager/current-operation"

#define AVAILABLE_UPDATES_FILE_NAME "available-updates"
#define AVAILABLE_UPDATES_FILE "/var/lib/package-manager/" AVAILABLE_UPDATES_FILE_NAME
#define CONNECTION_TIMEOUT 300000
const char *lc_messages;

extern bool flag_use_apt_algorithms;
DownloadProgressFunc download_progress_func = NULL;
OperationStartedFunc operation_started_func = NULL;
OperationProgressFunc operation_progress_func = NULL;
OperationCompleteFunc operation_complete_func = NULL;
PackageListUpdatedFunc package_list_updated_func = NULL;
PackageIgnoredFunc package_ignored_func = NULL;
gpointer signal_data = NULL;

/* Setting this to true will break the lock of the dpkg status
   database if necessary.
*/
bool flag_break_locks = false;

bool operation_started_signal_sent = false;

bool get_operation_started_signal_sent() {
    return operation_started_signal_sent;
}

struct flag_struct {
  const char *name;
  int flag;
} flag_names[] = {
  { "close-apps",       pkgflag_close_apps },
  { "suggest-backup",   pkgflag_suggest_backup },
  { "reboot",           pkgflag_reboot },
  { "system-update",    pkgflag_system_update },
  { "flash-and-reboot", pkgflag_flash_and_reboot },
  { "reboot-update",    pkgflag_reboot_update },
  { NULL,               0 }
};

static pkgCache::VerFileIterator
first_valid_file (const pkgCache::VerIterator &ver)
{
  pkgCache::VerFileIterator vf = ver.FileList();
  while (!vf.end() && (vf.File()->Flags & pkgCache::Flag::Ignore))
    vf++;
  return vf;
}

package_record::package_record (const pkgCache::VerIterator &ver)
  : Recs (*(AptWorkerCache::GetCurrent ()->cache)),
    P (Recs.Lookup (first_valid_file (ver)))
{
  const char *start, *stop;
  P.GetRec (start, stop);

  /* NOTE: pkTagSection::Scan only succeeds when the record ends in
           two newlines, but pkgRecords::Parser::GetRec does not
           include the second newline in its returned region.
           However, that second newline is always there, so we just
           pass one more character to Scan.
  */
  
  valid = section.Scan (start, stop-start+1);
}

bool
package_record::has (const char *tag)
{
  unsigned pos;
  return valid && section.Find (tag, pos);
}

bool
package_record::has_flag (const char *tag, const char *flag)
{
  string s = get_string (tag);
  return flag_string_contains (s.c_str(), flag);
}

string
package_record::get_string (const char *tag)
{
  if (valid)
    return section.FindS (tag);
  else
    return string ("");
}
 
int
all_white_space (const char *text)
{
  while (*text)
    if (!isspace (*text++))
      return 0;
  return 1;
}

char *
package_record::get (const char *tag)
{
  if (!valid)
    return NULL;

  string res = get_string (tag);
  if (all_white_space (res.c_str ()))
    return NULL;
  else
    return g_strdup (res.c_str());
}

int
package_record::get_int (const char *tag, int def)
{
  if (!valid)
    return def;

  return section.FindI (tag, def);
}

string
package_record::get_localized_string (const char *tag)
{
  if (lc_messages && *lc_messages)
    {
      char *locale_tag = g_strdup_printf ("%s-%s", tag, lc_messages);

      if (has (locale_tag))
	{
	  string res = get_string (locale_tag);
	  g_free (locale_tag);
	  return res;
	}
      g_free (locale_tag);
    }

  return get_string (tag);
}

string
package_record::get_localized_multiline_string (const char *tag)
{
  string s = get_localized_string (tag);
  gchar *fs = format_multiline_field (s.c_str ());
  s = fs;
  g_free (fs);
  return s;
}

gchar *
escape_for_shell (const gchar *str)
{
  gchar buf[2000];
  gchar *p = buf;
  /* Enclose the string in single quotes and escape single quotes.
   */

  *p++ = '\'';
  while (*str)
    {
      if (p >= buf+sizeof(buf)-6)
	return NULL;

      if (*str == '\'')
	{
	  // Don't you love bourne shell syntax?
	  *p++ = '\'';
	  *p++ = '\\';
	  *p++ = '\'';
	  *p++ = '\'';
	  str++;
	}
      else
	*p++ = *str++;
    }
  *p++ = '\'';
  *p++ = '\0';
  return g_strdup (buf);
}
void
save_operation_record (OperationData *od, const char *download_root)
{
	const gchar *op = operation_type_to_string(od->type);
	xexp *record = xexp_list_new (op);
	xexp_aset_text (record, "package", od->pkg);
	xexp_aset_text (record, "download-root", download_root);
	xexp_write_file (CURRENT_OPERATION_FILE, record);
	xexp_free (record);
}

void
erase_operation_record ()
{
  unlink (CURRENT_OPERATION_FILE);
}

// static xexp *
// read_operation_record ()
// {
//   struct stat buf;

//   /* This check is not strictly necessary but it avoids a distracting
//      complaint from xexp_read_file in the common case that the file
//      doesn't exist.
//   */
//   if (stat (CURRENT_OPERATION_FILE, &buf))
//     return NULL;

//   return xexp_read_file (CURRENT_OPERATION_FILE);
// }

/** MANAGEMENT OF FILE WITH INFO ABOUT AVAILABLE UPDATES */

bool
read_available_updates_file (std::map<string, string> &U)
{
  U.clear();

  if (!ensure_cache (false))
    return false;

  // AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

  FILE *in = fopen (AVAILABLE_UPDATES_FILE, "r");
  if (!in)
    return false;

  GError *err = NULL;
  xexp *updates = xexp_read (in, &err);

  for (xexp *pkg = xexp_first (updates); pkg; pkg = xexp_rest (pkg))
    {
      if (pkg)
	{
	  const char *name = xexp_aref_text (pkg, "name");
	  const char *ver = xexp_aref_text (pkg, "version");
	  if (name && ver)
	    U[name] = ver;
	}
    }

  xexp_free (updates);
  fclose (in);

  return true;
}

extern bool
is_update_package_ignored(pkgCache::PkgIterator pkg, pkgCache::VerIterator candidate);

void
write_available_updates_file ()
{
  if (!ensure_cache (false))
    return;

  xexp *x_updates = xexp_list_new ("updates");

  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
    {
      /* This duplicates the logic that determines which packages
	 would be shown in the "Check for Updates" view in blue-pill
	 mode.

	 Namely, a package is shown when it has both a installed and
	 available version, the available version is newer than the
	 installed version, and the available version is user-visible.

	 Packages are classified as "OS", "Certified", and "Other".  A
	 package belongs to the "OS" class when it has the
	 system-update flag set.  Otherwise, it belongs to the
	 "Certified" class when it is in a certified domain.
	 Otherwise it belongs to the "Other" class.  (The update
	 notifier presents the "Certified" class as "Nokia".)

	 XXX - it would be good to not duplicate so much logic, of
	       course.
      */

      pkgCache::VerIterator installed = pkg.CurrentVer ();
      pkgCache::VerIterator candidate = cache[pkg].CandidateVerIter(cache);

      if (!candidate.end ()
	  && !installed.end()
	  && installed.CompareVer (candidate) < 0
	  && is_user_package (candidate)
	)
	{
	  if (is_update_package_ignored (pkg, candidate))
	    continue;

	  xexp *x_pkg = NULL;
// 	  int domain_index = awc->cache->extra_info[pkg->ID].cur_domain;

	  x_pkg = xexp_list_new ("package");

	  const char *type;

          
	  if (is_os_meta_package (candidate))
	    type = "os";
// 	  else if (domain_is_certified(domain_index))
// 	    x_pkg = xexp_text_new ("certified", pkg.Name());
	  else
	    type = "other";

	  xexp_cons (x_pkg, xexp_text_new ("type", type));
	  xexp_cons (x_pkg, xexp_text_new ("version", candidate.VerStr()));
	  xexp_cons (x_pkg, xexp_text_new ("name", pkg.Name()));
	  xexp_cons (x_updates, x_pkg);
	}
    }

  xexp_write_file (AVAILABLE_UPDATES_FILE, x_updates);

  if (x_updates)
    xexp_free (x_updates);
}

bool
is_user_section (const char *section, const char *end)
{
  if (section == NULL)
    return false;

#if ENABLE_OLD_MAEMO_SECTION_TEST
  if (end-section > 6 && !strncmp (section, "maemo/", 6))
    return true;
#endif
  
  return end-section > 5 && !strncmp (section, "user/", 5);
}

bool
is_user_package (const pkgCache::VerIterator &ver)
{
  return (ver->Flags & pkgCache::Flag::Visible);
}

bool
is_user_package (const pkgTagSection &section)
{
  const char *start, *end;
  return ((section.Find ("Section", start, end)
           && is_user_section (start, end))
          || (section.Find ("Maemo-Flags", start, end)
              && flag_string_contains (start, end - start, "visible")));
}

bool
is_os_meta_package (const pkgCache::VerIterator &ver)
{
  return (ver->Flags & pkgCache::Flag::OSMetaPackage);
}

bool
is_user_system_section (const char *section, const char *end)
{
  if (section == NULL)
    return false;

  return end-section == 11 && !strncmp (section, "user/system", 11);
}

bool
is_user_system_package (const pkgCache::VerIterator &ver)
{
  const char *section = ver.Section ();

  if (section == NULL)
    return false;

  return is_user_system_section (section, section + strlen (section));
}

/* Determine whether a package was installed automatically to satisfy
   a dependency.
*/
bool
is_auto_package (pkgCache::PkgIterator &pkg)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  return (cache[pkg].Flags & pkgCache::Flag::Auto) != 0;
}

/* Determine whether a package is related to the current operation.
*/
bool
is_related (pkgCache::PkgIterator &pkg)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  //pkgDepCache &cache = *(awc->cache);
  // if (flag_use_apt_algorithms)
  //   return !cache[pkg].Keep() || (cache[pkg].iFlags & pkgDepCache::ReInstall);
  // else
  return awc->cache->extra_info[pkg->ID].related;
}

/* Determine whether a package is broken in some way and needs repair.
*/
bool
is_broken (pkgCache::PkgIterator &pkg)
{
  // Apt does all the work for us...
  return (pkg->Flags & pkgCache::Flag::Broken);
}

// bool
// is_ssu_dependency (pkgCache::PkgIterator pkg)
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &cache = *(awc->cache);
//   pkgCache::VerIterator verdpkg = cache[pkg].CandidateVerIter (cache);

//   // Return false if we do not have an SSU package available
//   if (ssu_packages == NULL)
//     return false;

//   // Non-user packages are not part of SSU
//   if (verdpkg.end () || is_user_package (verdpkg))
//     return false;

//   // Find out whether it's a dependency for a SSU package
//   for (pkgCache::DepIterator Dep = pkg.RevDependsList ();
//        Dep.end () != true;
//        Dep++)
//     {
//       const char *depname = Dep.ParentPkg().Name();

//       /* Look through the list of packages with the 'system-update' flag */
//       for (guint i = 0; i < ssu_packages->len; i++)
//         {
//           gchar *ssu_pkg = g_array_index (ssu_packages, gchar*, i);
//           if (!strcmp (depname, ssu_pkg))
//             return true;
//         }
//     }

//   // not a ssu package if reached
//   return false;
// }

/* Determine whether PKG replaces TARGET.
 */
bool
package_replaces (pkgCache::PkgIterator &pkg,
		  pkgCache::PkgIterator &target)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  pkgCache::DepIterator Dep = cache[pkg].InstVerIter(cache).DependsList();
  for (; Dep.end() != true; Dep++)
    {
      if (Dep->Type == pkgCache::Dep::Replaces)
	{
	  SPtrArray<pkgCache::Version *> List = Dep.AllTargets();
	  for (pkgCache::Version **I = List; *I != 0; I++)
	    {
	      pkgCache::VerIterator Ver(cache,*I);
	      pkgCache::PkgIterator Pkg = Ver.ParentPkg();

	      if (Pkg == target)
		return true;
	    }
	}
    }
  return false;
}

int
get_flags (package_record &rec)
{
  int flags = 0;
  char *flag_string = rec.get ("Maemo-Flags");
  char *ptr = flag_string, *tok;
  while ((tok = strsep (&ptr, ",")))
    {
      for (int i = 0; flag_names[i].name != NULL; i++)
	if (tokens_equal (flag_names[i].name, tok))
	  {
	    flags |= flag_names[i].flag;
	    break;
	  }
    }
  g_free (flag_string);

  return flags;
}

template <typename T>
static string join_strings (const T &L, const string &sep)
{
  std::ostringstream out;
  typename T::const_iterator i = L.begin();

  if (i != L.end())
    {
      out << *i;

      for (++i; i != L.end(); ++i)
	out << sep << *i;
    }

  return out.str();
}

string make_flags_str (int flags)
{
  std::list<string> S;

  for (flag_struct *s = flag_names; s->name != NULL; ++s)
    if (s->flag & flags)
      S.push_back(s->name);

  return join_strings (S, ", ");
}

void
send_download_status (const gchar* operation, const gchar* packagename, const gchar* version, int already, int total, int min_change)
{
	// DEBUG("%d/%d (min change: %d)", already, total, min_change);
	static int last_already;
	static int last_total;

	if (already == -1
		|| already < last_already
		|| already >= last_already + min_change
		|| total != last_total)
	{
		last_already = already;
		last_total = total;
		if (download_progress_func && total != 0) {
			download_progress_func (operation, packagename, version, already, total, signal_data);
		}
	}
}

void send_download_progress_signal(int current, int total, int min_change)
{
	OperationData *od = cache_get_current_operation();
	if (!od)
		return;
	od->current_bytes = current;
	od->total_bytes = total;

	if (od->send_signals)
          send_download_status(operation_type_to_string(od->type),
                               (od->type == OPERATION_INSTALL_FILE)? od->filename : od->pkg,
                               od->version,
                               current, total, min_change);
}

void send_operation_started_signal()
{
    DEBUG("");
    operation_started_signal_sent = true;

    if ( operation_started_func ) {
        OperationData* od = cache_get_current_operation();
	if (!od)
	    return;
        if (od->send_signals) {
            if ( od->type == OPERATION_INSTALL_FILE ) {
                operation_started_func(operation_type_to_string(od->type), od->filename, od->version, signal_data);
            } else {
                operation_started_func(operation_type_to_string(od->type), od->pkg, od->version, signal_data);
            }
        }
    }
}

void
send_operation_progress_signal (int percentage)
{
	// DEBUG("%d\r", percentage);

	static int last_percentage;
	if (percentage < 0) {
		last_percentage = -1;
	} else if (percentage > last_percentage)
	{
		last_percentage = percentage;
		OperationData* od = cache_get_current_operation();
		if (!od)
			return;

		if (operation_progress_func && percentage >= 0 && od != NULL && od->send_signals) {
			g_assert (od);
			od->operation_progress_pct = percentage;
			// for (int i = 0; i < 1000; ++i)
            if ( od->type == OPERATION_INSTALL_FILE ) {
                operation_progress_func (operation_type_to_string(od->type), od->filename, od->version, percentage, signal_data);
			} else {
                operation_progress_func (operation_type_to_string(od->type), od->pkg, od->version, percentage, signal_data);
		    }
		}
	}
}

void
send_package_list_updated_signal (bool update_changed)
{
	if (package_list_updated_func) {
		DEBUG("update_changed=%d", update_changed ? 1 : 0);
		package_list_updated_func (update_changed, signal_data);
	}
}

void
send_package_ignored_signal(const gchar* type, const gchar* package, const gchar* version)
{
	if (package_ignored_func) {
		DEBUG("type: %s Package: %s (%s)", type, package, version);
		package_ignored_func(type, package, version, signal_data);
	}
}

static void
__send_op_dep_complete (OperationData *od)
{
	if (operation_complete_func && od->send_signals) {
		operation_complete_func (operation_type_to_string(od->type),
					 od->pkg, od->version, od->error, operation_need_reboot (od), signal_data);
	}
}
static void
__send_op_dep_complete_signals(GList *list, OperationType type, OperationType main_type)
{
	OperationData* od = operation_data_new ();
	while (list) {
		PackageVersion *pv = static_cast<PackageVersion*>(list->data);
		od->pkg = g_strdup(pv->pkg);
		od->version = g_strdup(pv->version);
		od->type = type;
		if (type == OPERATION_UNINSTALL && main_type == OPERATION_SYSTEM_UPGRADE) {
			od->type = OPERATION_INSTALL; // yes, really
			od->error = pkgmgr_error_to_dbus_error(ERROR_REMOVED_BY_SYSTEM_UPGRADE);
		}

		__send_op_dep_complete (od);

		g_free(od->pkg);
		od->pkg = 0;
		g_free(od->version);
		od->version = 0;

		list = g_list_next(list);
	}
	operation_data_free (od);
}
static gboolean
__mainloop_send_op_complete_signal (gpointer data)
{
	OperationData *od = static_cast<OperationData*> (data);
	if (operation_complete_func && od->send_signals) {
		if (od->send_dep_signals && !operation_failed(od)) {
			__send_op_dep_complete_signals(od->installed_user_deps, OPERATION_INSTALL, od->type);
			__send_op_dep_complete_signals(od->updated_user_deps, OPERATION_UPGRADE, od->type);
			__send_op_dep_complete_signals(od->removed_user_deps, OPERATION_UNINSTALL, od->type);
		}

        if ( od->type == OPERATION_INSTALL_FILE ) {
            operation_complete_func (operation_type_to_string(od->type), od->filename, od->version, od->error, operation_need_reboot (od), signal_data);
        } else {
            operation_complete_func (operation_type_to_string(od->type), od->pkg, od->version, od->error, operation_need_reboot (od), signal_data);
		}
	}
	operation_data_free (od);
	return FALSE;
}
void
send_operation_complete_signal (OperationData *od)
{
    operation_started_signal_sent = false;
	g_assert (od);
	DEBUG("%s %s: %s", operation_type_to_string(od->type),
		od->pkg, od->error);
	if (od->send_signals)
		g_idle_add(__mainloop_send_op_complete_signal, operation_data_copy (od));
}


static void
__dpkg_status(const gchar* str)
{
	// Handles only install (for installing standalone package)
	if (g_str_has_prefix (str, "processing")) {
		gint percentage = 0;
		const char *end = NULL;
		str = strchr (str, ':');
		if (str == NULL)
			return;
		str += 1;
		end = strchr (str, ':');
		if (end) {
			if (g_str_has_prefix (str+1, "install")) {
				percentage = 0;
			} else if (g_str_has_prefix (str+1, "configure")) {
				percentage = 50;
			}
		}
		send_operation_progress_signal(percentage);
	} else if (g_str_has_prefix (str, "status:")) {
		if (g_str_has_suffix(str, ": installed")) {
			send_operation_progress_signal(100);
		}
	}
}
static void
__interpret_pmstatus (char *str)
{
  // DEBUG ("dpkg working: '%s'\r", str);
  if (!strncmp (str, "pmstatus:", 9)) {
	  gdouble percentage = 0;
	  char *end = NULL;
	  str += 9;
	  str = strchr (str, ':');
	  if (str == NULL)
		  return;
	  str += 1;
	  end = strchr (str, ':');
	  if (end) {
		  percentage = g_ascii_strtod(str, NULL);
		  DEBUG ("dpkg: '%d': '%s'\r", (gint)percentage, end+1);
		  send_operation_progress_signal((int)percentage);
	  }
/*       (op_general, (int)percentage, 100); */
  } else {
	  __dpkg_status (str);
  }
}

static GString *pmstatus_line;

static gboolean
__read_pmstatus (GIOChannel *channel, GIOCondition cond, gpointer data)
{
  char buf[256], *line_end;
  int n, fd = g_io_channel_unix_get_fd (channel);

  n = read (fd, buf, 256);
  if (n > 0)
    {
      g_string_append_len (pmstatus_line, buf, n);
      while ((line_end = strchr (pmstatus_line->str, '\n')))
	{
	  *line_end = '\0';
	  __interpret_pmstatus (pmstatus_line->str);
	  g_string_erase (pmstatus_line, 0, line_end - pmstatus_line->str + 1);
	}
      return TRUE;
    }
  else
    {
      g_io_channel_shutdown (channel, 0, NULL);
      return FALSE;
    }
}

void
status_ind_setup_from_fd (int fd)
{
	DEBUG("fd: %d", fd);
	pmstatus_line = g_string_new ("");
	GIOChannel *channel = g_io_channel_unix_new (fd);
	g_io_add_watch (channel, GIOCondition(G_IO_IN | G_IO_HUP | G_IO_ERR),
			__read_pmstatus, NULL);
	g_io_channel_unref (channel);
}

void
free_dpkg_status_ind()
{
        if ( pmstatus_line == 0 ) {
                return;
        }
        g_string_free(pmstatus_line, TRUE);
        pmstatus_line = 0;
}

char *
get_command_stdout (const char *cmd, bool with_empty_last_line, int *status_ptr)
{
  FILE *f = popen (cmd, "r");

  if (f)
    {
      const size_t incr = 2000;
      char *record = NULL;
      size_t size = 0;

      do
	{
	  // increase buffer and try to fill it, leaving room for the
	  // trailing newlines and nul.
	  // XXX - do it properly.

	  char *new_record = new char[size + incr + 3];
	  if (record)
	    {
	      memcpy (new_record, record, size);
	      delete [] record;
	    }
	  record = new_record;

	  size += fread (record + size, 1, incr, f);
	}
      while (!feof (f));

      int status = pclose (f);
      if (status_ptr)
        *status_ptr = status;
      else if (status != 0)
	{
	  delete [] record;
	  return NULL;
	}

      if (with_empty_last_line)
        {
          record[size++] = '\n';
          record[size++] = '\n';
        }
      record[size] = '\0';

      return record;
    }
  return NULL;
}

static char *
get_deb_record (const char *filename)
{
  char *esc_filename = escape_for_shell (filename);
  if (esc_filename == NULL) {
    return NULL;
  }

  char *cmd = g_strdup_printf ("/usr/bin/dpkg-deb -f %s", esc_filename);
  char *record = get_command_stdout (cmd, true);

  g_free (cmd);
  g_free (esc_filename);
  return record;
}

gchar*
read_tag_section(const gchar *filename, pkgTagSection *section, GError **error)
{
	if (!filename || !section) {
		  g_set_error(error, PKGMGRD_ERROR, ERROR_INTERNAL,
			  "Cannot read package details from '%s'!", filename);
		  return NULL;
	}
	if (!g_file_test(filename, G_FILE_TEST_EXISTS)) {
		g_set_error(error, PKGMGRD_ERROR, ERROR_FILE_NOT_FOUND,
			"Cannot find file '%s'!",filename);
		return NULL;
	}
	char *record = get_deb_record (filename);
	if (record == NULL || !section->Scan (record, strlen (record))) {
		g_set_error(error, PKGMGRD_ERROR, ERROR_INVALID_FILE,
			"Cannot read package details from '%s'!", filename);
		delete[] record;
		return NULL;
	}
	return record;
}

const gchar*
user_section_to_category(const gchar* section)
{
	if ( g_strcmp0( "user/desktop", section) == 0) {
		return "Desktop";
	} else if ( g_strcmp0( "user/development", section) == 0) {
		return "Programming";
	} else if ( g_strcmp0( "user/education", section) == 0) {
		return "Education";
	} else if ( g_strcmp0( "user/games", section) == 0) {
		return "Games";
	} else if ( g_strcmp0( "user/graphics", section) == 0) {
		return "Graphics";
	} else if ( g_strcmp0( "user/multimedia", section) == 0) {
		return "Multimedia";
	} else if ( g_strcmp0( "user/navigation", section) == 0) {
		return "Navigation";
	} else if ( g_strcmp0( "user/network", section) == 0) {
		return "InternetAndNetworking";
	} else if ( g_strcmp0( "user/office", section) == 0) {
		return "Office";
	} else if ( g_strcmp0( "user/science", section) == 0) {
		return "Science";
	} else if ( g_strcmp0( "user/system", section) == 0) {
		return "System";
	} else if ( g_strcmp0( "user/utilities", section) == 0) {
		return "Accessories";
	} else {
		return "Other";
	}
}

// void
// cmd_third_party_policy_check ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgCache::PkgIterator pkg;
//   pkgCache::VerIterator ver;
//   const char *package = request.decode_string_in_place ();
//   const char *version = request.decode_string_in_place ();
//   third_party_policy_status policy_status = third_party_compatible;

//   if (find_package_version (awc->cache, pkg, ver, package, version))
//     {
//       pkgDepCache &cache = *(awc->cache);
//       pkgCache::VerIterator candidate = cache[pkg].CandidateVerIter (cache);
//       package_record rec (candidate);
//       int flags = get_flags (rec);

//       // Apply policy to user packages only skip system update
//       // meta-packages that are not installed
//       if (is_user_package (candidate) && !candidate.end ()
//           && !(flags & pkgflag_system_update))
//         {
//           for (pkgCache::DepIterator Dep = candidate.DependsList ();
//                Dep.end () != true;
//                Dep++)
//             {
//               pkgCache::PkgIterator dpkg = Dep.TargetPkg ();
//               pkgCache::VerIterator verdpkg =
//                 cache[dpkg].CandidateVerIter (cache);

//               // Check whether SSU metapackage is dependant on this
//               if (!is_ssu_dependency (dpkg))
//                 continue;

//               int op = Dep->CompareOp & 0x0F;

//               if (Dep->Type == pkgCache::Dep::Depends)
//                 {
//                   if (op == pkgCache::Dep::NoOp
//                       || op == pkgCache::Dep::GreaterEq
//                       || op == pkgCache::Dep::Greater)
//                     continue;

//                   log_stderr ("%s breaks 3rd party dependencies policy:",
//                               pkg.Name ());
//                   policy_status = third_party_incompatible;
//                   break;
//                 }
//               else if (Dep->Type == pkgCache::Dep::Conflicts)
//                 {
//                   if (op == pkgCache::Dep::Less
//                       || op == pkgCache::Dep::LessEq
//                       || op == pkgCache::Dep::Equals)
//                     continue;

//                   log_stderr ("%s breaks 3rd party conflicts policy",
//                               pkg.Name ());
//                   policy_status = third_party_incompatible;
//                   break;
//                 }
//             }
//         }
//     }

//   // return result
//   response.encode_int (policy_status);
// }

/* Get a lock as with GetLock from libapt-pkg, breaking it if needed
   and allowed by flag_break_locks.

   We do this so that the users can not lock themselves out.  We break
   locks instead of not locking since noisily breaking a lock is
   better than silently corrupting stuff.
 */
int
ForceLock (string File, bool Errors)
{
  int lock_fd = GetLock (File, false);
  if (lock_fd >= 0)
    return lock_fd;

  if (flag_break_locks)
    {
      int res = unlink (File.c_str ());
      if (res < 0 && errno != ENOENT)
	log_stderr ("Can't remove %s: %m", File.c_str ());
      else if (res == 0)
	log_stderr ("Forcing %s", File.c_str ());
    }

  return GetLock (File, Errors);
}


FieldIterator &
FieldIterator::operator++ ()
{
  if (i < section.Count ())
    ++i;

  return *this;
}

string
FieldIterator::key () const
{
  const char *s0, *s1;
  section.Get (s0, s1, i);

  const char *colon = (const char *) memchr (s0, ':', s1 - s0);
  if (colon)
    return string (s0, colon - s0);
  else
    return string ();
}

static const char *
find_end_of_field (const char *begin, const char *end)
{
  const char *p = begin;
  const char *newl;

  do {
    newl = (const char *) memchr (p, '\n', end - p);
    if (! newl)
      return end;
    p = newl + 1;
  }
  while (*p == ' ');

  return newl;
}

string
FieldIterator::value () const
{
  const char *s0, *s1;
  section.Get (s0, s1, i);

  const char *colon = (const char *) memchr (s0, ':', s1 - s0);
  if (colon)
    {
      const char *v;
      if (*(colon + 1) == ' ')
	v = colon + 2;
      else // '\n'
	v = colon + 1;

      const char *eof = find_end_of_field (v, s1);
      if (eof)
	return string (v, eof - v);
    }

  return string ();
}

bool
in_scratchbox ()
{
  return getenv ("_SBOX_DIR") != NULL;
}

/* Our own version of debSystem.  We override the Lock member function
   to be able to break locks and to avoid failing when dpkg has left a
   journal.
*/

class mydebSystem : public debSystem
{
  // For locking support
  int LockFD;
  unsigned LockCount;

public:

  virtual signed Score (Configuration const &Cnf);
  virtual bool Lock ();
  virtual bool UnLock (bool NoErrors);
  
  mydebSystem ();
};

mydebSystem::mydebSystem ()
  : debSystem ()
{
   Label = "handsfree Debian dpkg interface";
}

signed
mydebSystem::Score (Configuration const &Cnf)
{
  return debSystem::Score (Cnf) + 10;  // Pick me, pick me, pick me!
}

bool
mydebSystem::Lock ()
{
  /* This is a modified copy of debSystem::Lock, which in turn is a
     copy of the behavior of dpkg.
  */

  // Disable file locking
  if (_config->FindB("Debug::NoLocking",false) == true || LockCount > 1)
    {
      LockCount++;
      return true;
    }

  // Create the lockfile, breaking the lock if requested.
  string AdminDir = flNotFile(_config->Find("Dir::State::status"));
  LockFD = ForceLock(AdminDir + "lock");
  if (LockFD == -1)
    {
      if (errno == EACCES || errno == EAGAIN)
	return _error->Error("Unable to lock the administration directory (%s), "
			     "is another process using it?",AdminDir.c_str());
      else
	return _error->Error("Unable to lock the administration directory (%s), "
			     "are you root?",AdminDir.c_str());
    }

  LockCount++;
      
  return true;
}

// System::UnLock - Drop a lock						/*{{{*/
// ---------------------------------------------------------------------
/* */
bool mydebSystem::UnLock(bool NoErrors)
{
   if (LockCount == 0 && NoErrors == true)
      return false;
   
   if (LockCount < 1)
      return _error->Error("Not locked");
   if (--LockCount == 0)
   {
      close(LockFD);
      LockCount = 0;
   }
   
   return true;
}
									/*}}}*/
mydebSystem mydebsystem;

static void
run_apt_desktop (const char *command, GList *args)
{
  string cmd = string("apt-desktop ") + command;
  for (GList *a = args; a; a = a->next)
    cmd = cmd + " " + (const char *)a->data;
  g_debug ("apt-desktop...");
  system (cmd.c_str());
}

void
add_interesting_packages (GList *packages)
{
  run_apt_desktop ("add-interesting", packages);
}

void
set_interesting_packages (GList *packages)
{
  GList *args = g_list_prepend (packages, (void *)"--had-error");
  run_apt_desktop ("set-interesting", args);
  g_list_free_1 (args);
}


/* Network
 */

#if 0
static ConIcConnection *connection_object = NULL;
static GMainLoop *connection_loop = NULL;
static GMainContext *connection_context = NULL;
static bool connection_session = false;
static bool connection_success = false;

static void
cancel_current_operation ()
{
  g_debug ("conic: cancelling");

  OperationData *op = cache_get_current_operation();
  if (op)
    op->cancel_flag = true;
}

static void
iap_callback (ConIcConnection *connection,
	      ConIcConnectionEvent *event,
	      gpointer user_data)
{
  bool success = false;
  static gchar *iap = NULL;

  g_debug ("iap_callback");

  ConIcConnectionError con_err = con_ic_connection_event_get_error (event);
  ConIcConnectionStatus con_state = con_ic_connection_event_get_status (event);
  const gchar *event_iap_id = con_ic_event_get_iap_id (CON_IC_EVENT (event));

  switch (con_err)
    {
    case CON_IC_CONNECTION_ERROR_NONE:
      break;
    case CON_IC_CONNECTION_ERROR_INVALID_IAP:
      g_debug ("conic: IAP is invalid");
      break;
    case CON_IC_CONNECTION_ERROR_CONNECTION_FAILED:
      g_debug ("conic: connection failed");
      break;
    case CON_IC_CONNECTION_ERROR_USER_CANCELED:
      g_debug ("conic: user cancelled");
      break;
    default:
      g_debug ("conic: unknown error");
      break;
    }

  switch (con_state)
    {
    case CON_IC_STATUS_CONNECTED:{
      g_debug ("connected to %s", event_iap_id);

      /* if IAP changed we shall break the download */
      if (g_strcmp0  (iap, event_iap_id) != 0)
        {
          if (iap)
            {
              g_free (iap);
              iap = NULL;
              cancel_current_operation ();
            }

          iap = g_strdup (event_iap_id);
        }

      success = true;
      break;
    }

    case CON_IC_STATUS_DISCONNECTING:
      g_debug ("disconnecting from %s", event_iap_id);
      return;

    case CON_IC_STATUS_DISCONNECTED:{
      g_debug ("disconnected from %s", event_iap_id);

      /* If the IAP is the same then the current connection
       * was disconnected */
      if (iap && event_iap_id && g_strcmp0 (iap, event_iap_id) == 0)
        {
          g_free (iap);
          iap = NULL;
        }

      if (connection_session)
        cancel_current_operation ();
      break;
    }

    default:
      g_debug ("conic: unknown event");
      return;
    }

  if (connection_loop)
    {
      connection_success = success;
      GMainLoop *loop = connection_loop;
      connection_loop = NULL;
      g_main_loop_quit (loop);
    }
}

gboolean
connection_timeout (gpointer data)
{
    if (connection_loop)
    {
        g_debug ("connection loop timeout");
        connection_success = false;
        GMainLoop *loop = connection_loop;
        connection_loop = NULL;
        g_main_loop_quit (loop);
    }
    return FALSE;
}

bool assume_network = false;

bool
connect_network (bool automatic)
{
  if (assume_network || in_scratchbox ())
    return true;

  if (connection_object == NULL)
    {
      connection_object = con_ic_connection_new ();
      g_signal_connect (connection_object, "connection-event",
                        G_CALLBACK (iap_callback), NULL);
    }

  connection_context = g_main_context_new ();
  connection_loop = g_main_loop_new (connection_context, FALSE);

  if (!con_ic_connection_connect (connection_object,
                                  (automatic
                                   ? CON_IC_CONNECT_FLAG_AUTOMATICALLY_TRIGGERED
                                   : CON_IC_CONNECT_FLAG_NONE)))
    {
      g_main_loop_unref (connection_loop);
      connection_loop = NULL;
      return false;
    }

  connection_session = true;

  g_debug ("conic: waiting for reply");
  GMainLoop *loop = connection_loop;
  g_timeout_add(CONNECTION_TIMEOUT, connection_timeout, NULL); // backup mechanism to make sure that the process doesn't get stuck to connection loop
  g_main_loop_run (loop);
  g_main_loop_unref (loop);
  g_main_context_unref (connection_context);
  connection_context = NULL;

  g_debug ("connected");

  return connection_success;
}

void
disconnect_network ()
{
  /* We don't disconnect explicitly anymore.  Disconnection requests
     can come at any time and have been observed to come even after
     the next connection request.  Instead of disconnecting, we rely
     on the idle timeout terminating the whole process, which also
     terminates our network session.
  */

#ifdef LIBCONIC_HAS_NO_RACES
  if (connection_object && connection_session)
    {
      con_ic_connection_disconnect (connection_object);
      connection_session = false;
    }
#endif
}
#endif

bool assume_network = FALSE;
static gint network_write_fd;
static bool connection_ok = false;

static void
cancel_current_operation (bool download_failed=false)
{
  g_debug ("conic: cancelling");
  g_debug("awu, cancel_current_operation()");

  OperationData *op = cache_get_current_operation();
  if (op) {
    op->cancel_flag = true;
    if(download_failed) {
        op->download_failed_flag = true;
    }
  }
}

static gboolean
read_connection_message (GIOChannel *channel, GIOCondition cond, gpointer data)
{
    g_debug ("read_message");
    g_debug("awu, read_connection_message()");
    char buf[MAX_BUF_SIZE];
    int n, fd = g_io_channel_unix_get_fd (channel);

    n = read (fd, buf, MAX_BUF_SIZE);
    if (n > 0) {
        g_debug ("buf %s", buf);
        char msg[50];
        sprintf(msg, "awu, read_connection_message(), buf: %s", buf);
        g_debug(msg);
        if(strcmp(buf, DISCONNECTED_STR) == 0) {
            g_debug("awu, read_connection_message(), DISCONNECTED_STR");
            connection_ok = false;
        } else if(strcmp(buf, DISCONNECTING_STR) == 0) {
            g_debug("awu, read_connection_message(), DISCONNECTING_STR");
            connection_ok = false;
            cancel_current_operation(true);
        } else if(strcmp(buf, CONNECTED_STR) == 0) {
            g_debug("awu, read_connection_message(), CONNECTED_STR");
            connection_ok = true;
        }
        return TRUE;
    } else {
        g_io_channel_shutdown (channel, 0, NULL);
        return FALSE;
    }
}

static void
handle_netconnection_exit (GPid pid, gint status, gpointer data)
{
}

static bool
launch_netconnection(bool automatic)
{
    g_debug ("launch_netconnection");
    g_debug("awu,  launch_netconnection()");

    pid_t pid;
    int out_fd[2];
    int in_fd[2];
    int status_fd[2];

    if ( (pipe(out_fd) < 0) || pipe(in_fd) < 0 || pipe(status_fd) < 0) {
        g_debug ("launch_netconnection, pipe problem");
        return false;
    }
    if ( (pid = fork()) < 0 ) {
        g_debug ("launch_netconnection, fork problem");
        return false;
    } else  if (pid == 0) { // child
        close(out_fd[1]); // close write of out_fd
        close(in_fd[0]); // close read of in_fd
        close(status_fd[0]); // close read of status_fd

        char request_fd_str[5];
        sprintf(request_fd_str, "%d", out_fd[0]);
        char response_fd_str[5];
        sprintf(response_fd_str, "%d", in_fd[1]);
        char status_fd_str[5];
        sprintf(status_fd_str, "%d", status_fd[1]);

        char *args[] = { (char*)"netconnection", (char*)"-q", request_fd_str, (char*)"-r", response_fd_str, (char*)"-s", status_fd_str, (char*)"-a", (char *)0};
        if(!automatic) {
            args[7] = (char *)0;
        }
        if ( execv("/usr/lib/package-manager/netconnection", args )) {
            g_debug ("launch_netconnection, execv problem");
            return false;
        }
        exit(1);
    } else {
        network_write_fd = out_fd[1];
        close(out_fd[0]); // close read of out_fd
        close(in_fd[1]); // close write of in_fd
        close(status_fd[1]); // close write of status_fd

	g_child_watch_add (pid, handle_netconnection_exit, 0);

        // start listening status pipe
        GIOChannel *channel = g_io_channel_unix_new (status_fd[0]);
        g_io_add_watch (channel, GIOCondition(G_IO_IN | G_IO_HUP | G_IO_ERR),
                        read_connection_message, NULL);
        g_io_channel_unref (channel);
        char response[MAX_BUF_SIZE];
        read(in_fd[0], response, MAX_BUF_SIZE);
        if(strcmp(response, CONNECTED_STR) == 0) {
            connection_ok = true;
        }
    }
    return true;
}

bool
connect_network (bool automatic)
{
  g_debug ("connect_network");
  g_debug("awu, connect_network()");
  if (assume_network || in_scratchbox () || connection_ok)
    return true;

  if(!launch_netconnection(automatic)) {
      g_debug ("launch_netconnection failed");
      g_debug("awu, launch_netconnection() failed");
      return false;
  }
  return connection_ok;
}

void
disconnect_network ()
{
    g_debug("awu, disconnect_network()");
  if (assume_network || !connection_ok)
    return;

  write(network_write_fd, DISCONNECT_STR, strlen(DISCONNECT_STR) + 1);
}

static bool
is_flag_delim (int c)
{
  return !isalnum (c);
}

bool
flag_string_contains (const char *flag_string,
                      int flag_string_len,
                      const char *flag)
{
  if (flag_string == NULL)
    return false;

  const char *pos = (const char *)memmem (flag_string, flag_string_len,
                                          flag, strlen (flag));
  if (pos == NULL)
    return false;

  const char *end = pos + strlen (flag);

  return ((pos == flag_string || is_flag_delim (pos[-1]))
          && (end >= flag_string + flag_string_len || is_flag_delim (end[0])));
}

bool
flag_string_contains (const char *flag_string,
                      const char *flag)
{
  if (flag_string == NULL)
    return false;

  return flag_string_contains (flag_string, strlen (flag_string),
                               flag);
}

bool
is_downloadable (const pkgCache::VerIterator &ver)
{
  if (ver.Downloadable())
    return true;
  for (pkgCache::VerIterator v = ver.ParentPkg().VersionList(); !v.end(); v++)
    {
      if (strcmp (v.VerStr(), ver.VerStr()) == 0
          && v.Downloadable())
        return true;
    }
  return false;
}

