/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#ifndef APT_WORKER_UTILS_H
#define APT_WORKER_UTILS_H

#include <apt-pkg/tagfile.h>
#include <apt-pkg/pkgrecords.h>
#include <map>

#include "log.h"
#include "xexp.h"
#include "apt-worker-operation.h"

/** GENERAL UTILITIES
 */

#define log_stderr(fmt, ...) g_debug("apt-worker: " fmt, ## __VA_ARGS__)

/* Getting the package record in a nicely parsable form.
 */

struct package_record {
  pkgRecords Recs;
  pkgRecords::Parser &P;
  pkgTagSection section;
  bool valid;

  package_record (const pkgCache::VerIterator &ver);

  bool has (const char *tag);
  bool has_flag (const char *tag, const char *flag);
  string get_string (const char *tag);
  char *get (const char *tag);
  int get_int (const char *tag, int def);

  string get_localized_string (const char *tag);
  string get_localized_multiline_string (const char *tag);
};

typedef void (*DownloadProgressFunc) (const gchar* operation, const gchar* packagename, const gchar* version, int already, int total, gpointer data);

typedef void (*OperationStartedFunc) (const gchar* op, const gchar* pkg, const gchar* version, gpointer data);

typedef void (*OperationProgressFunc) (const gchar* op, const gchar* pkg, const gchar* version, int percentage, gpointer data);

typedef void (*OperationCompleteFunc) (const gchar* op, gchar* pkg, const gchar* version, const gchar* error, gboolean need_reboot, gpointer data);

typedef void (*PackageListUpdatedFunc) (gboolean updates_changed, gpointer data);

typedef void (*PackageIgnoredFunc) (const gchar* type, const gchar* package, const gchar* version, gpointer data);

// void
// log_stderr (const gchar *fmt, ...);

char *
escape_for_shell (const char *str);

void
save_operation_record (OperationData *od, const char *download_root);

void
erase_operation_record ();

// xexp *read_operation_record ();

bool
get_operation_started_signal_sent();

bool
read_available_updates_file (std::map<string, string> &updates);

void
write_available_updates_file ();

bool
is_user_section (const char *section, const char *end);

bool
is_user_package (const pkgCache::VerIterator &ver);

bool
is_user_package (const pkgTagSection &section);

bool
is_os_meta_package (const pkgCache::VerIterator &ver);

bool
is_user_system_section (const char *section, const char *end);

bool
is_user_system_package (const pkgCache::VerIterator &ver);

bool
is_auto_package (pkgCache::PkgIterator &pkg);

bool
is_related (pkgCache::PkgIterator &pkg);

bool
is_broken (pkgCache::PkgIterator &pkg);

bool
package_replaces (pkgCache::PkgIterator &pkg,
		pkgCache::PkgIterator &target);

int
get_flags (package_record &rec);

string
make_flags_str (int flags);

void
send_download_status (const gchar* operation, const gchar* packagename, const gchar* packageversion, int already, int total, int min_change);
void
send_download_progress_signal(int current, int total, int min_change);
void
send_operation_started_signal();
void
send_operation_progress_signal (int percentage);
void
send_operation_complete_signal (OperationData *od);
void
send_package_list_updated_signal (bool update_changed);
void
send_package_ignored_signal(const gchar* type, const gchar* package, const gchar* version);

void status_ind_setup_from_fd (int fd);

void free_dpkg_status_ind();

gchar *
get_command_stdout (const char *cmd, bool with_empty_last_line = false, int *status_ptr = NULL);

gchar*
read_tag_section(const gchar *filename, pkgTagSection *section, GError **error);

const gchar*
user_section_to_category(const gchar* section);

int
ForceLock (string File, bool Errors = true);

/**
 * Package tag field iterator class.
 */
struct FieldIterator
{
  FieldIterator (pkgTagSection &s) : section (s), i (0) { }
  FieldIterator &operator++ ();
  string key () const;
  string value () const;
  bool end () const { return i >= section.Count (); }

  pkgTagSection &section;
  unsigned i;
};

bool
in_scratchbox ();

void add_interesting_packages (GList *packages);
void set_interesting_packages (GList *packages);

gboolean connection_timeout (gpointer data);

extern bool assume_network;
extern bool protect_user_packages_from_autoremoval;

bool connect_network (bool automatic);
void disconnect_network ();

bool flag_string_contains (const char *flag_string, const char *flag);
bool flag_string_contains (const char *flag_string, int flag_string_len, const char *flag);

bool is_downloadable (const pkgCache::VerIterator &ver);

#endif /* APT_WORKER_UTILS_H */
