/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "archivecleaner.h"
#include <apt-pkg/configuration.h>
#include <apt-pkg/error.h>
#include <apt-pkg/debversion.h>
#include <apt-pkg/strutl.h>
#include <dirent.h>
#include <glib.h>
#include <unistd.h>

#include "debug.h"

#define _(x) (x)

using namespace std;

static bool Fetchable(pkgCache &Cache, const pkgCache::PkgIterator &P, const string &Ver)
{
   if (P.end())
      return false;

   const bool CleanInstalled = true;
   pkgCache::VerIterator V = P.VersionList();
   for (; V.end() == false; V++)
   {
      // See if we can fetch this version at all
      bool IsFetchable = false;
      for (pkgCache::VerFileIterator J = V.FileList(); 
	   J.end() == false; J++)
      {
	 if (CleanInstalled == true &&
	     (J.File()->Flags & pkgCache::Flag::NotSource) != 0)
	    continue;
	 IsFetchable = true;
	 break;
      }
	    
      // See if this verison matches the file
      if (IsFetchable == true && Ver == V.VerStr())
	 return true;
   }
	 
   // We found a match, keep the file
   return false;
}

static bool InstalledOrOlder(const pkgCache::PkgIterator &P, const string &Ver)
{
   pkgCache::VerIterator Cur = P.CurrentVer();
   if (Cur.end())
      return false;

   int C = debVS.CmpVersion(Ver.c_str(), Cur.VerStr());

   return C < 0 || (C == 0 && !(P->Flags & pkgCache::Flag::Broken));
}

bool ArchiveCleaner::Go(string Dir,pkgCache &Cache)
{
   bool CleanInstalled = _config->FindB("APT::Clean-Installed",true);
   string MyArch = _config->Find("APT::Architecture");
      
   DIR *D = opendir(Dir.c_str());
   if (D == 0)
      return _error->Errno("opendir",_("Unable to read %s"),Dir.c_str());

   string StartDir = SafeGetCWD();
   if (chdir(Dir.c_str()) != 0)
   {
      closedir(D);
      return _error->Errno("chdir",_("Unable to change to %s"),Dir.c_str());
   }
   
   for (struct dirent *Dir = readdir(D); Dir != 0; Dir = readdir(D))
   {
      // Skip some files..
      if (strcmp(Dir->d_name,"lock") == 0 ||
	  strcmp(Dir->d_name,"partial") == 0 ||
	  strcmp(Dir->d_name,".") == 0 ||
	  strcmp(Dir->d_name,"..") == 0)
	 continue;

      struct stat St;
      if (stat(Dir->d_name,&St) != 0)
      {
	 chdir(StartDir.c_str());
	 closedir(D);
	 return _error->Errno("stat",_("Unable to stat %s."),Dir->d_name);
      }
      
      // Grab the package name
      const char *I = Dir->d_name;
      for (; *I != 0 && *I != '_';I++);
      if (*I != '_')
	 continue;
      string Pkg = DeQuoteString(string(Dir->d_name,I-Dir->d_name));

      // Grab the version
      const char *Start = I + 1;
      for (I = Start; *I != 0 && *I != '_';I++);
      if (*I != '_')
	 continue;
      string Ver = DeQuoteString(string(Start,I-Start));
  
      // Grab the arch
      Start = I + 1;
      for (I = Start; *I != 0 && *I != '.' ;I++);
      if (*I != '.')
	 continue;
      string Arch = DeQuoteString(string(Start,I-Start));
      
      if (Arch != "all" && Arch != MyArch)
	 continue;

      pkgCache::PkgIterator P = Cache.FindPkg(Pkg);
      if (P.end()
	  || !Fetchable(Cache, P, Ver)
	  || InstalledOrOlder(P, Ver))
      {
	 Erase(Dir->d_name,Pkg,Ver,St);
      }
   };
   
   chdir(StartDir.c_str());
   closedir(D);
   return true;   
}

void ArchiveCleaner::Erase(const char *File,string Pkg,string Ver,struct stat & St)
{
   if (_config->FindB("Debug::ArchiveCleaner"))
      DEBUG("Del %s %s [%sB]\n", Pkg.c_str(), Ver.c_str(),
	      SizeToStr (St.st_size).c_str());

   unlink(File);      
}
