/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <unistd.h>
#include <errno.h>

#include "debug.h"
#include "confutils.h"
#include "catalogue.h"
#include "catalogues.h"

#define FAILED_CATALOGUES_FILE "/var/lib/package-manager/failed-catalogues"

GList*
catalogues_get_all ()
{
	xexp *catalogues = read_catalogues ();
	GList *list = NULL;
	if (!catalogues) {
		DEBUG("No catalogues!");
		return NULL;
	}

	for (xexp *cat = xexp_first (catalogues); cat; cat = xexp_rest (cat)) {
		Catalogue *c = catalogue_new ();
		c->name = g_strdup (xexp_aref_text (cat, "name"));
		c->uri = g_strdup(xexp_aref_text (cat, "uri"));
		c->enabled = !xexp_aref_bool (cat, "disabled");
		c->essential = xexp_aref_bool (cat, "essential");
		c->dist = g_strdup(xexp_aref_text (cat, "dist"));
		if (!c->dist) {
			c->dist = g_strdup(default_distribution);
		}
		c->components = g_strdup(xexp_aref_text (cat, "components"));
		list = g_list_append(list, c);
	}
	xexp_free (catalogues);
	return list;
}

void
save_failed_catalogues (xexp *catalogues)
{
  xexp *failed_catalogues = xexp_list_new ("catalogues");

  /* Save ONLY catalogues with errors */
  for (xexp *cat = xexp_first (catalogues); cat; cat = xexp_rest (cat))
    {
      xexp *errors_item = xexp_aref (cat, "errors");

      /* If containing errors details, add it to the list */
      if (errors_item)
	xexp_append_1 (failed_catalogues, xexp_copy (cat));
    }

  if (xexp_length (failed_catalogues) > 0)
    xexp_write_file (FAILED_CATALOGUES_FILE, failed_catalogues);
  else
    clean_failed_catalogues ();

  xexp_free (failed_catalogues);
}
void
clean_failed_catalogues ()
{
  if (unlink (FAILED_CATALOGUES_FILE) < 0 && errno != ENOENT)
    DEBUG ("error unlinking %s: %m", FAILED_CATALOGUES_FILE);
}

void
clean_temp_catalogues ()
{
  if (unlink (TEMP_APT_SOURCE_LIST) < 0 && errno != ENOENT)
    DEBUG ("error unlinking %s: %m", TEMP_APT_SOURCE_LIST);
}

// xexp *
// load_failed_catalogues ()
// {
//   struct stat buf;
//   int stat_result;
//   xexp *failed_catalogues = NULL;

//   /* Check if there are problems reading the file */
//   stat_result = stat (FAILED_CATALOGUES_FILE, &buf);
//   if (!stat_result)
//     {
//       failed_catalogues = xexp_read_file (FAILED_CATALOGUES_FILE);
//       if (xexp_length (failed_catalogues) <= 0)
// 	{
// 	  /* Return NULL and clean failed catalogues file if there
// 	     are no errors contained inside of it (quite odd) */
// 	  xexp_free (failed_catalogues);
// 	  failed_catalogues = NULL;

// 	  clean_failed_catalogues ();
// 	}
//     }
//   else if (errno != ENOENT)
//     log_stderr ("error reading file %s: %m", FAILED_CATALOGUES_FILE);

//   return failed_catalogues;
// }

// static xexp *
// merge_catalogues_with_errors (xexp *catalogues)
// {
//   xexp *failed_catalogues = NULL;

//   /* First check if there's an errors file after the last refresh */
//   failed_catalogues = load_failed_catalogues ();

//   /* Add errors (if present) for each catalogue */
//   if (failed_catalogues != NULL)
//     {
//       const char *uri = NULL, *f_uri = NULL;
//       const char *dist = NULL, *f_dist = NULL;
//       const char *comp = NULL, *f_comp = NULL;

//       for (xexp *cat = xexp_first (catalogues); cat; cat = xexp_rest (cat))
// 	{
// 	  uri = xexp_aref_text (cat, "uri");
// 	  dist = xexp_aref_text (cat, "dist");
// 	  comp = xexp_aref_text (cat, "components");

// 	  /* Look for errors for this catalogue */
// 	  for (xexp *f_cat = xexp_first (failed_catalogues); f_cat; f_cat = xexp_rest (f_cat))
// 	    {
// 	      f_uri = xexp_aref_text (f_cat, "uri");
// 	      f_dist = xexp_aref_text (f_cat, "dist");
// 	      f_comp = xexp_aref_text (f_cat, "components");

// 	      /* If matches the uri (mandatory field), check the other ones */
// 	      if (!strcmp (uri, f_uri))
// 		{
// 		  xexp *errors_item = NULL;

// 		  /* 'dist' fields must be the same, or NULL, to continue */
// 		  if ((dist == NULL || f_dist == NULL || strcmp (dist, f_dist)) &&
// 		      !(dist == NULL && f_dist == NULL))
// 		    continue;

// 		  /* 'dist' fields must be the same, or NULL, to continue */
// 		  if ((comp == NULL || f_comp == NULL || strcmp (comp, f_comp)) &&
// 		      !(comp == NULL && f_comp == NULL))
// 		    continue;

// 		  /* If reached, errors would be about the same
// 		     catalogue found in the configuration file */
// 		  errors_item = xexp_aref (f_cat, "errors");
// 		  if (errors_item != NULL)
// 		    {
// 		      /* Append a copy of errors to the catalogues xexp */
// 		      xexp_append_1 (cat, xexp_copy (errors_item));

// 		      /* Delete the xexp from errors file, in order
// 			 not to check it again in later checks */
// 		      xexp_del (failed_catalogues, f_cat);
// 		    }
// 		}
// 	    }
// 	}

//       /* Free the catalogues errors xexp */
//       xexp_free (failed_catalogues);
//     }

//   return catalogues;
// }

// static xexp *
// map_catalogue_error_details (xexp *x)
// {
//   xexp *mapped_x = xexp_copy (x);
//   xexp *errors_item = xexp_aref (mapped_x, "errors");

//   /* If containing errors details, remove errors */
//   if (errors_item)
//     xexp_del (mapped_x, errors_item);

//   /* Return the mapped xexp */
//   return mapped_x;
// }

void
reset_catalogue_errors (xexp *catalogues)
{
  /* Clean the catalogues errors */
  clean_failed_catalogues ();

  if (catalogues == NULL)
    return;

  for (xexp *c = xexp_first (catalogues); c; c = xexp_rest (c))
    xexp_adel (c, "errors");
}

/**
 * Match two URIs ignoring trailing slash.
 *
 * assert (uri_match ("http://a/b/c/", "http://a/b/c/"));
 * assert (uri_match ("http://a/b/c", "http://a/b/c"));
 * assert (uri_match ("http://a/b/c/", "http://a/b/c"));
 * assert (uri_match ("http://a/b/c", "http://a/b/c/"));
 */
static bool
uri_match (const gchar *uri1, const gchar *uri2)
{
	if (uri1 && uri2) {
		int len1 = strlen (uri1);
		if (g_str_has_suffix (uri1, "/"))
			--len1;

		int len2 = strlen (uri2);
		if (g_str_has_suffix (uri2, "/"))
			--len2;

		return len1 == len2 && ! strncmp (uri1, uri2, len1);
	}

	return false;
}

gchar*
find_catalogue_by_info (const char* p_uri,
                        const char* p_dist,
                        const char* p_comp,
			gboolean find_disabled)
{
  gchar *catname = NULL;
  xexp *catalogues = read_catalogues ();
//   DEBUG("find_catalogue_by_info: %s %s %s", p_uri, p_dist, p_comp);
  if (!catalogues) {
	  DEBUG("No catalogues!");
    return NULL;
  }

  for (xexp *cat = xexp_first (catalogues); cat; cat = xexp_rest (cat))
    {
      const gchar *uri = xexp_aref_text (cat, "uri");
      bool enabled = !xexp_aref_bool (cat, "disabled");

      if ((find_disabled || enabled) && uri_match (uri, p_uri))
        {
          const gchar *dist = xexp_aref_text (cat, "dist");
          if (!dist)
            dist = default_distribution;

          if (!g_strcmp0 (dist, p_dist))
            {
              if (p_comp != NULL && p_comp[0] != '\0')
                {
                  const gchar *comp = xexp_aref_text (cat, "components");
                  gchar **comps = g_strsplit_set (comp, " \t\n", -1);
		  gchar **p_comps = g_strsplit_set (p_comp, " \t\n", -1);

                  if (comps != NULL)
                    {
                      for (gint i = 0; comps[i]; i++)
                        {
			   for (gint j = 0; p_comps[j]; j++)
			     {
			       if (!g_strcmp0 (comps[i], p_comps[j]))
			         {
				   g_strfreev (comps);
				   catname = g_strdup (xexp_aref_text (cat, "name"));
				   goto done;
				 }
			     }
			}
		      g_strfreev (comps);
		      g_strfreev (p_comps);
                    }
                }
              else
                {
                  // it has no components defined
                  catname = g_strdup (xexp_aref_text (cat, "name"));
                  goto done;
                }
            }
        }
    }

 done:
  xexp_free (catalogues);
  return catname;
}

GList *
find_catalogues_for_item_desc (xexp *catalogues, string desc_uri)
{
  /* This is a hack to associate error messages produced during
     downloading with a specific catalogue so that a good error report
     can be shown to the user.

     DESC_URI is matched against all the catalogues and we return the
     first hit.

     DESC_URI matches a catalogue if it is of the form

        URI/dists/DIST/<no-more-slashes>

     or

        URI/dists/DIST/COMP/<rest-with-slashes>

     or

        URI/DIST<rest-with-slashes>

     or

        URIDIST (when dist is only a '/')

     where URI and DIST are the respective elements of the catalogue,
     and COMP is one of the components of the catalogue.

     XXX - This is not the right thing to do, of course.  Apt-pkg
           should offer a way to easily associate user level objects
           with acquire items.
  */

  if (catalogues == NULL)
    return NULL;

  GList *cat_glist = NULL;

  const char *match_uri = desc_uri.c_str ();

  for (xexp *cat = xexp_first (catalogues); cat; cat = xexp_rest (cat))
    {
      char *uri = g_strdup (xexp_aref_text (cat, "uri"));
      const char *dist = xexp_aref_text (cat, "dist");
      const char *comp_element = xexp_aref_text (cat, "components");
      gchar **comps = (comp_element
		       ? g_strsplit_set (comp_element, " \t\n", 0)
		       : NULL);
      char *pfx = NULL;

      if (dist == NULL)
	dist = default_distribution;

      while (uri[0] && uri[strlen(uri)-1] == '/')
	uri[strlen(uri)-1] = '\0';

      if (dist[0] && dist[strlen(dist)-1] == '/')
	{
	  /* A simple repository without components
	   */

	  if (dist[0] != '/')
	    pfx = g_strconcat (uri, "/", dist, NULL);
	  else /* dist can be only '/' */
	    pfx = g_strconcat (uri, dist, NULL);

	  if (g_str_has_prefix (match_uri, pfx))
	    goto found_it;
	}
      else
	{
	  /* A repository with components
	   */

	  pfx = g_strconcat (uri, "/dists/", dist, "/", NULL);
	  if (!g_str_has_prefix (match_uri, pfx))
	    goto try_next;

	  const char *rest = match_uri + strlen (pfx);

	  if (!strchr (rest, '/'))
	    goto found_it;

	  if (comps)
	    {
	      for (int i = 0; comps[i]; i++)
		{
		  gchar *comp = comps[i];

		  if (comp[0] == '\0')
		    continue;

		  if (g_str_has_prefix (rest, comp)
		      && rest[strlen(comp)] == '/')
		    goto found_it;
		}
	    }
	}

    try_next:
      g_strfreev (comps);
      g_free (pfx);
      g_free (uri);
      continue;

    found_it:
      g_strfreev (comps);
      g_free (pfx);
      g_free (uri);

      /* Append found item to the list */
      cat_glist = g_list_append (cat_glist, cat);
    }

  return cat_glist;
}

gboolean
update_sources_list (xexp *catalogues)
{
  gboolean success = FALSE;

  /* Write the new sources list to disk */
  success = (write_user_catalogues (catalogues)
	     && write_sources_list (CATALOGUE_APT_SOURCE, catalogues));

  return success;
}

