/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <string.h>
#include <stdlib.h>
#include <sys/stat.h>

#include <apt-pkg/pkgcache.h>
#include <apt-pkg/debmetaindex.h>

#include "debug.h"
#include "domain.h"
#include "confutils.h"
#include "apt-worker-utils.h"

/* Package source domains.

   Package sources are classified into domains, depending on which key
   was used to sign their Release file.  When upgrading a package, the
   update must come from the same domain as the installed package, or
   from one that 'dominates' it.

   Domains are identified by symbolic names.  The defining property of
   a domain is the list of keys.  Any source that is signed by one of
   these keys is associated with that domain.

   In addition to the explicitly defined domains (see below), there
   are two implicitly defined domains, "signed" and "unsigned".  The
   "signed" domain includes all sources that are signed but are not
   associated with any of the explicit domains.  The "unsigned" domain
   includes all the unsigned sources.

   The explicitly defined domains can be declared to be "certified".
   When a package is installed or upgraded from such a domain, the
   does not need to agree to a legal disclaimer.

   No explicit domain dominates any other explicit domain.  All trust
   domains dominate the "signed" and "unsigned" domains, and "signed"
   dominates "unsigned".  Thus, the dominance rules are hard coded.

   Domains are also used when installing a package for the first time:
   when a given package version is available from more than one
   source, sources belonging to any explicit domain or the "signed"
   domain are preferred over the "unsigned" domain.  You can influence
   which source is used via the normal apt source priorities.

   One domain is designated as the default domain (usually the Nokia
   certified domain).  When storing and reporting information about
   domains, only the non-default domains are explicitly mentioned.
   This saves some space and time when most of the packages are in the
   default domain, which they usually are.
*/

/* Inside this process, domains are identified with small integers.
 */

static xexp *domain_conf = NULL;

static domain_info *domains = NULL;
static int domains_number = 0;
static time_t domains_last_modified = -1;

static domain_t domain_unknown;

static time_t
file_last_modified (const char *file_name)
{
  struct stat buf;

  if (stat (file_name, &buf) == -1)
    {
	    DEBUG ("error retriving file info: %s", file_name);
      return -1;
    }

  return buf.st_mtime;
}

void
domain_conf_free ()
{
  for (int i = 0; i < domains_number; ++i)
    xexp_free (domains[i].conf);

  delete [] domains;
  domains = 0;

  xexp_free (domain_conf);
  domain_conf = 0;
}

void
read_domain_conf ()
{
  domain_conf_free ();

  domain_conf = read_domains ();

  int n_domains = 1;
  if (domain_conf)
    n_domains += xexp_length (domain_conf);

  domains = new domain_info[n_domains];
  int i = 0;

  /* Setup the 'unknown' domain.  This domain is normally also defined
     in the external configuration files, but we make sure that we
     always have it since we fall back to it in certain error cases.
   */
  domain_unknown = i++;
  domains[domain_unknown].name = "";
  domains[domain_unknown].conf = xexp_list_new ("domain");
  domains[domain_unknown].trust_level = 0;
  domains[domain_unknown].is_certified = false;

  if (domain_conf)
    {
      while (xexp *d = xexp_pop (domain_conf))
	{
	  if (!xexp_is (d, "domain"))
	    continue;

          const char *name = xexp_aref_text (d, "name");
          if (name == NULL)
            continue;

          int j;
          for (j = 0; j < i; j++)
            if (strcmp (domains[j].name, name) == 0)
              break;
          
          if (j == i)
            {
              domains[i].name = name;
              domains[i].conf = d;
              i += 1;
            }
          else
            xexp_append (domains[j].conf, d);
	}
    }

  for (int j = 0; j < i; j++)
    {
      xexp *d = domains[j].conf;
      domains[j].trust_level = xexp_aref_int (d, "trust-level", 0);
      domains[j].is_certified = xexp_aref_bool (d, "certified");
    }

  domains_number = i;
  domains_last_modified = file_last_modified (PACKAGE_DOMAINS);
}

domain_t
find_domain_by_tag (const char *tag, const char *val,
                    bool (*hit) (const char *conf, const char *val))
{
  for (int i = 0; i < domains_number; i++)
    {
      if (domains[i].conf == NULL)
	continue;

      for (xexp *x = xexp_aref (domains[i].conf, tag);
	   x;
	   x = xexp_aref_rest (x, tag))
	{
	  if (xexp_is_text (x) && hit (xexp_text (x), val))
	    return i;
	}
    }

  return DOMAIN_INVALID;
}

static bool
conf_is_prefix (const char *conf, const char *val)
{
  return g_str_has_prefix (val, conf);
}

static bool
conf_has_suffix (const char *conf, const char *val)
{
  return g_str_has_suffix (conf, val);
}

domain_t
find_domain_by_key (const char *key)
{
  return find_domain_by_tag ("key", key, conf_has_suffix);
}

domain_t
find_domain_by_uri (const char *uri)
{
  return find_domain_by_tag ("uri", uri, conf_is_prefix);
}

domain_t
find_domain_by_name (const char *name)
{
  for (int i = 0; i < domains_number; i++)
    {
      if (strcmp (domains[i].name, name) == 0)
        return i;
    }

  return DOMAIN_INVALID;
}

static pkgSourceList *cur_sources;

void
set_sources_for_get_domain (pkgSourceList *sources)
{
  cur_sources = sources;
}

static debReleaseIndex *
find_deb_meta_index (pkgIndexFile *index)
{
  if (cur_sources == NULL)
    return NULL;

  for (pkgSourceList::const_iterator I = cur_sources->begin();
       I != cur_sources->end(); I++)
    {
      vector<pkgIndexFile *> *Indexes = (*I)->GetIndexFiles();
      for (vector<pkgIndexFile *>::const_iterator J = Indexes->begin();
	   J != Indexes->end(); J++)
	{
	  if ((*J) == index)
	    {
	      if (strcmp ((*I)->GetType(), "deb") == 0)
		return (debReleaseIndex *)(*I);
	      else
		return NULL;
	    }
	}
    }
  
  return NULL;
}

#define VALIDSIG "VALIDSIG"
#define GOODSIG  "GOODSIG"

static char *
get_meta_info_key (debReleaseIndex *meta)
{
  struct stat buf;
  string gpg_file = meta->MetaIndexFile ("Release.gpg");
  if (stat (gpg_file.c_str(), &buf) == -1)
    return NULL;

  string info_file = meta->MetaIndexFile ("Release.gpg.info");
  char *key = NULL;

  FILE *f = fopen (info_file.c_str(), "r");
  if (f)
    {
      char *line = NULL;
      size_t len = 0;
      ssize_t n;

      while ((n = getline (&line, &len, f)) != -1)
	{
	  if (n > 0 && line[n-1] == '\n')
	    line[n-1] = '\0';

	  bool is_valid =  g_str_has_prefix (line, VALIDSIG);
	  if (is_valid || g_str_has_prefix (line, GOODSIG))
	    {
	      int dis = is_valid ? strlen (VALIDSIG) : strlen (GOODSIG);
	      key = g_strchug (g_strdup (line + dis)); // *Cough*
	      break;
	    }
	}

      free (line);
      fclose (f);
    }

  return key;
}

domain_t
get_domain (pkgIndexFile *index)
{
  debReleaseIndex *meta = find_deb_meta_index (index);
  if (meta)
    return get_domain (meta);

  return domain_unknown;
}

domain_t
get_domain (metaIndex *meta)
{
  if (strcmp (meta->GetType(), "deb") != 0)
    return domain_unknown;

  domain_t d = find_domain_by_uri (meta->GetURI().c_str());
  if (d != DOMAIN_INVALID)
    return d;

  char *key = get_meta_info_key ((debReleaseIndex *)meta);
  if (key)
    {
      domain_t d = find_domain_by_key (key);
      g_free (key);
      if (d != DOMAIN_INVALID)
        return d;
    }

  return domain_unknown;
}

const gchar*
domain_get_name(int n)
{
  return domains[n].name;
}

int
domain_get_count()
{
  return domains_number;
}

int
domain_get_trust_level(int n)
{
  return domains[n].trust_level;
}

bool
domain_is_certified(int n)
{
  return domains[n].is_certified;
}

bool
domain_dominates_or_is_equal (int a, int b)
{
  return domains[a].trust_level >= domains[b].trust_level;
}

domain_t
get_file_domain (const char *filename)
{
  char *esc_filename = escape_for_shell (filename);
  if (esc_filename == NULL) {
    return NULL;
  }

  char *cmd = g_strdup_printf ("/usr/bin/aegis-origin %s", esc_filename);
  int status = 0;
  char *origin = get_command_stdout (cmd, false, &status);

  if (origin)
    {
      int l = strlen (origin);
      while (l > 0 && isspace (origin[l-1]))
        l--;
      origin[l] = '\0';
      g_debug ("aegis-origin says: %s", origin);
    }
  else
    origin = g_strdup ("");

  // aegis-origin returns "Unknown" for unknown origins, but the real
  // name of the origin as used in the domain configurations is
  // actually the empty string.

  if (strcmp (origin, "Unknown") == 0)
    {
      free (origin);
      origin = g_strdup ("");
    }
  else if (status != 0)
    {
      // Running aegis-origin has failed and the output was not
      // "Unknown", which means that something bad has happened.
      // XXX - we might want to fail in this case.
      g_debug ("aegis-origin has failed!\n");
      free (origin);
      origin = g_strdup ("");
    }

  domain_t d = find_domain_by_name (origin);
  if (d == DOMAIN_INVALID)
    {
      g_warning ("Origin '%s' is not known!", origin);
      d = find_domain_by_name ("");
    }

  free (origin);

  g_debug ("File origin: %s \"%s\"", filename, domain_get_name (d));

  g_free (cmd);
  g_free (esc_filename);
  return d;
}
