/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "failed-installs.h"
#include "apt-worker-cache.h"
#include "error.h"
#include "glib-util.h"
#include <cstring>
#include <sstream>

FailedInstalls::FailedInstalls (const char *filename)
  : filename (filename),
    installs (0)
{
}

FailedInstalls::~FailedInstalls ()
{
  reset ();
}

void FailedInstalls::reset ()
{
  if (installs)
    {
      xexp_free (installs);
      installs = 0;
    }
}

bool FailedInstalls::read ()
{
  reset ();

  installs = xexp_read_file (filename);
  if (installs == NULL)
    installs = xexp_list_new ("failed-installs");

  return true;
}

bool FailedInstalls::write ()
{
  if (installs == NULL)
    return false;

  gint retval = xexp_write_file (filename, installs);

  return retval == 1;
}

static xexp *find_failed_install (xexp *fif, const char *package)
{
  if (!fif || !package)
    return NULL;

  if (!xexp_is (fif, "failed-installs"))
    return NULL;

  for (xexp *i = xexp_first (fif); i; i = xexp_rest (i))
    {
      if (!xexp_is (i, "package"))
	continue;

      const char *p = xexp_aref_text (i, "name");

      if (!p)
	continue;

      if (!strcmp (package, p))
	return i;
    }

  return NULL;
}

const char* FailedInstalls::failed_version (const char *package)
{
  xexp *i = find_failed_install (installs, package);
  if (!i)
    return NULL;

  return xexp_aref_text (i, "version");
}

const char* FailedInstalls::failed_domain (const char *package)
{
  xexp *i = find_failed_install (installs, package);
  if (!i)
    return NULL;

  return xexp_aref_text (i, "domain");
}

static xexp *find_failed_install (xexp *fif, const char *package, const char *version)
{
  if (!fif || !package || !version)
    return NULL;

  for (xexp *i = xexp_first (fif); i; i = xexp_rest (i))
    {
      if (!xexp_is (i, "package"))
	continue;

      const char *p = xexp_aref_text (i, "name");
      const char *v = xexp_aref_text (i, "version");

      if (!p || !v)
	continue;

      if (!strcmp (package, p) && !strcmp (version, v))
	return i;
    }

  return NULL;
}

static xexp *find_failed_file_install (xexp *fif, const char *filename)
{
  if (!fif || !filename)
    return NULL;

  for (xexp *i = xexp_first (fif); i; i = xexp_rest (i))
    {
      if (!xexp_is (i, "package"))
        continue;

      const char *f = xexp_aref_text (i, "filename");

      if (!f)
        continue;

      if ( !strcmp (filename, f) )
        return i;
    }

  return NULL;
}

void FailedInstalls::store (const char *package, const char *version,
                            const char *fname, const char *domain, GError *error)
{
  xexp *i = find_failed_install (installs, package);
  if (i != NULL)
    xexp_del (installs, i);

  i = xexp_list_new ("package");
  xexp_aset_text (i, "name", package);
  xexp_aset_text (i, "version", version);
  if (fname)
    xexp_aset_text (i, "filename", fname);
  if (domain)
    xexp_aset_text (i, "domain", domain);
  xexp_aset_text (i, "error-code", pkgmgr_error_to_dbus_error ((PkgmgrdError) error->code));
  // gchar *msg = xml_escape (error->message);
  const char *msg = error->message;
  xexp_aset_text (i, "error-msg", msg);
  // g_free (msg);
  xexp_append_1 (installs, i);
}

bool FailedInstalls::clear (const char *package)
{
  xexp *i = find_failed_install (installs, package);
  if (i == NULL)
    return false;

  xexp_del (installs, i);

  return true;
}

bool FailedInstalls::clear_by_filename (const char *filename)
{
  xexp *i = find_failed_file_install (installs, filename);
  if (i == NULL)
    return false;

  xexp_del (installs, i);

  return true;
}

bool FailedInstalls::clear (const char *package, const char *version)
{
  xexp *i = find_failed_install (installs, package, version);
  if (i == NULL)
    return false;

  xexp_del (installs, i);

  return true;
}

bool FailedInstalls::contains (const char *package)
{
  return find_failed_install (installs, package) != NULL;
}

// bool is_failed_install (const char *package, const char *version)
// {
//   return find_failed_install (package, version) != NULL;
// }

static bool get_failed_install_data (xexp *i, GHashTable *p)
{
  if (i == NULL)
    return false;

  const char *error_code = xexp_aref_text (i, "error-code");
  const char *error_msg = xexp_aref_text (i, "error-msg");
  const char *filename = xexp_aref_text (i, "filename");

  if (error_code)
    g_hash_table_insert_string_value (p, "ErrorName", error_code);
  if (error_msg)
    {
      // gchar *msg = xml_unescape (error_msg);
      const char *msg = error_msg;
      g_hash_table_insert_string_value (p, "ErrorMessage", msg);
      // g_free (msg);
    }
  if ( filename && strcmp(filename, "") != 0 ) {
      g_hash_table_insert_boolean_value(p, "FlagLocalFile", TRUE);
      g_hash_table_insert_string_value(p, "Filename", filename);
  }

  return true;
}

bool FailedInstalls::get_data (const char *package, const char *version, const char *&error_code, const char *&error_msg, const char *&filename, const char *&domain)
{
  xexp *i = find_failed_install (installs, package, version);
  if (i == NULL)
    return false;

  error_code = xexp_aref_text (i, "error-code");
  error_msg = xexp_aref_text (i, "error-msg");
  filename = xexp_aref_text (i, "filename");
  domain = xexp_aref_text (i, "domain");

  return true;
}

bool FailedInstalls::get_data (const char *package, const char *version, GHashTable *p)
{
  xexp *i = find_failed_install (installs, package, version);
  if (i == NULL)
    return false;

  return get_failed_install_data (i, p);
}

GList *FailedInstalls::get_packages ()
{
  GList *packages = NULL;
  for (xexp *i = xexp_first (installs); i; i = xexp_rest (i))
    if (xexp_is (i, "package"))
      {
        const char *n = xexp_aref_text (i, "name");
        if (n)
          packages = g_list_append (packages, (void *)n);
      }
  return packages;
}
