/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */


#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <unistd.h>
#include <stdio.h>

#include "inst-pkg-ts-storage.h"

#include "debug.h"


PkgTsStorage::PkgTsStorage()
{
    storage = NULL;
}

PkgTsStorage::~PkgTsStorage()
{
    clearStorage();
}

void clearList(GList* list)
{
    GList *i = list;
    while ( i ) {
        g_free( ((package_info *)i->data)->package );
        g_free( ((package_info *)i->data)->version );
        g_free( ((package_info *)i->data)->filename );
        g_free(i->data);
        i = g_list_next(i);
    }
    g_list_free( list );
}

bool PkgTsStorage::clearStorage()
{
    bool changed = storage != 0;
    clearList( storage );
    storage = NULL;
    return changed;
}

gconstpointer PkgTsStorage::findPackage(const gchar* package)
{
    GList* i = storage;
    gchar* pkg = NULL;
    while ( i ) {
        pkg = ((package_info *)i->data)->package;
        if ( strcmp(pkg, package) == 0 ) {
            return i->data;
        }
        i = g_list_next(i);
    }

    return NULL;
}

gconstpointer PkgTsStorage::findPackage(const gchar* package, const gchar* version)
{
    GList* i = storage;
    gchar* pkg = NULL;
    gchar* ver = NULL;
    while ( i ) {
        pkg = ((package_info *)i->data)->package;
        ver = ((package_info *)i->data)->version;
        if ( strcmp(pkg, package) == 0 && strcmp(ver, version) == 0 ) {
            return i->data;
        }
        i = g_list_next(i);
    }

    return NULL;
}

gconstpointer PkgTsStorage::findFile(const gchar* filename)
{
    GList* i = storage;
    gchar* f = NULL;
    while ( i ) {
        f = ((package_info *)i->data)->filename;
        if ( strcmp(f, filename) == 0 ) {
            return i->data;
        }
        i = g_list_next(i);
    }

    return NULL;
}

gconstpointer PkgTsStorage::findFile(const gchar* filename, const gchar* version)
{
    GList* i = storage;
    gchar* f = NULL;
    gchar* ver = NULL;
    while ( i ) {
        f = ((package_info *)i->data)->filename;
        ver = ((package_info *)i->data)->version;
        if ( strcmp(f, filename) == 0 && strcmp(ver, version) == 0 ) {
            return i->data;
        }
        i = g_list_next(i);
    }

    return NULL;
}

GList* PkgTsStorage::copyStorage()
{
    GList* i = storage;
    GList* copy = NULL;
    package_info* item = NULL;
    package_info* data = NULL;
    while ( i ) {
        item = (package_info *)i->data;
        data = (package_info *)g_malloc(sizeof(package_info));
        data->package = g_strdup(item->package);
        data->version = g_strdup(item->version);
        data->filename = g_strdup(item->filename);
        data->timestamp = item->timestamp;
        copy = g_list_append(copy, data);
        i = g_list_next(i);
    }

    return copy;
}

bool PkgTsStorage::addPackage(const gchar* package, const gchar* version, time_t timestamp)
{
    gconstpointer p = findPackage(package, version);
    package_info* pkg_info = NULL;
    if ( p == NULL ) {
        pkg_info = (package_info *) g_malloc(sizeof(package_info));

        if ( pkg_info == NULL ) {
            DEBUG("Can not allocate needed memory for the package info");
            return false;
        }
    
        pkg_info->package = g_strdup(package);
        pkg_info->version = g_strdup(version);
        pkg_info->filename = g_strdup("");

        storage = g_list_append(storage, pkg_info);
    } else {
        pkg_info = (package_info *)p;
    }
    pkg_info->timestamp = timestamp;

    return (storage != NULL);
}

bool PkgTsStorage::addFile(const gchar* package, const gchar* version, const gchar* filename, time_t timestamp)
{
    gconstpointer p = findPackage(package, version);
    package_info* pkg_info = NULL;
    if ( p == NULL ) {
        pkg_info = (package_info *) g_malloc(sizeof(package_info));

        if ( pkg_info == NULL ) {
            DEBUG("Can not allocate needed memory for the package info");
            return false;
        }
    
        pkg_info->package = g_strdup(package);
        pkg_info->version = g_strdup(version);
        pkg_info->filename = g_strdup(filename);

        storage = g_list_append(storage, pkg_info);
    } else {
        pkg_info = (package_info *)p;
    }
    pkg_info->timestamp = timestamp;

    return (storage != NULL);
}

bool PkgTsStorage::removePackage(const gchar* package)
{
    if ( storage == NULL ) {
        return false;
    }
    
    gconstpointer p = findPackage(package);
    if ( p == NULL ) {
        return false;
    }
    
    while ( p != NULL ) {
        package_info* pkg = (package_info *)p;
        g_free(pkg->package);
        g_free(pkg->version);
        g_free(pkg->filename);
        g_free(pkg);
        storage = g_list_remove(storage, pkg);

        p = findPackage(package);
    }

    return true;
}

bool PkgTsStorage::removePackage(const gchar* package, const gchar* version)
{
    if ( storage == NULL ) {
        return false;
    }
    
    gconstpointer p = findPackage(package, version);
    if ( p == NULL ) {
        return false;
    }

    package_info* pkg = (package_info *)p;
    g_free(pkg->package);
    g_free(pkg->version);
    g_free(pkg->filename);
    g_free(pkg);
    storage = g_list_remove(storage, pkg);

    return true;
}

bool PkgTsStorage::removeFile(const gchar* filename)
{
    if ( storage == NULL ) {
        return false;
    }
    
    gconstpointer p = findFile(filename);
    if ( p == NULL ) {
        return false;
    }
    
    while ( p != NULL ) {
        package_info* pkg = (package_info *)p;
        g_free(pkg->package);
        g_free(pkg->version);
        g_free(pkg->filename);
        g_free(pkg);
        storage = g_list_remove(storage, pkg);

        p = findFile(filename);
    }

    return true;
}

bool PkgTsStorage::removeFile(const gchar* filename, const gchar* version)
{
    if ( storage == NULL ) {
        return false;
    }
    
    gconstpointer p = findFile(filename, version);
    if ( p == NULL ) {
        return false;
    }

    package_info* pkg = (package_info *)p;
    g_free(pkg->package);
    g_free(pkg->version);
    g_free(pkg->filename);
    g_free(pkg);
    storage = g_list_remove(storage, pkg);

    return true;
}

gchar* PkgTsStorage::packageToString(package_info* pkg)
{
    return g_strdup_printf("%s v=%s %ld %s\n", pkg->package, pkg->version, pkg->timestamp, pkg->filename);
}

package_info* PkgTsStorage::stringToPackage(const gchar* package, bool *ok)
{
    *ok = true;

    gchar* str = strtok( g_strdup(package), " ");
    if ( str == NULL ) {
        *ok = false;
        return NULL;
    }
    gchar* version = strtok(NULL, " ");
    gchar* time = strtok(NULL, " ");
    gchar* filename = strtok(NULL, "\n");
    
    if ( version == NULL || time == NULL ) {
        g_free( str );
        *ok = false;
        return NULL;
    }
    
    package_info* result = (package_info *)g_malloc(sizeof(package_info));
    result->package = str;
    strtok(version, "=");
    result->version = g_strdup( strtok(NULL, "=") );
    
    if ( result->version == NULL ) {
        result->version = g_strdup("");
    }

    long int t = -1;
    if ( sscanf(time, "%ld", &t) == EOF || t == -1 ) {
        *ok = false;
    }
    result->timestamp = t;
    
    if ( filename ) {
        result->filename = g_strdup(filename);
    } else {
        result->filename = g_strdup("");
    }

    return result;
}

bool PkgTsStorage::saveToFile(const gchar* filename)
{
    FILE *f = fopen (filename, "w");

    if (f == NULL) {
        return false;
    }

    GList* i = storage;
    gchar* str = NULL;
    while ( i ) {
        str = packageToString( (package_info *)i->data );
        if ( fputs(str, f) == EOF ) {
            DEBUG("Something went wrong!\n");
            fclose(f); // Something went wrong :'(
            return false;
        }
        i = g_list_next(i);
        g_free( str );
    }

    fclose(f);
    return true;
}

bool PkgTsStorage::readFromFile(const gchar* filename)
{
    if ( storage != NULL ) {
        clearStorage();
        storage = NULL;
    }
    
    FILE *f = fopen (filename, "r");
    if (f == NULL) {
        if (errno != ENOENT)
            DEBUG("File '%s' can not be opened", filename);
        return false;
    }

    char str[1024];
    bool ok = false;
    while ( fgets( str, 1024, f ) != NULL ) {
        package_info* pkg = stringToPackage(str, &ok);
        if ( !ok ) {
            break;
        }

        if ( pkg != NULL ) {
            storage = g_list_append(storage, pkg);
        }
    }

    fclose(f);

    return ( storage != NULL && ok );
}

time_t PkgTsStorage::getPackageTimestamp(const gchar* package, bool *ok)
{
    gconstpointer p = findPackage(package);
    if ( p == NULL ) {
        *ok = false;
        return 0;
    }
    
    *ok = true;
    return ((package_info *)p)->timestamp;
}

time_t PkgTsStorage::getPackageTimestamp(const gchar* package, const gchar* version, bool *ok)
{
    gconstpointer p = findPackage(package, version);
    if ( p == NULL ) {
        *ok = false;
        return 0;
    }
    
    *ok = true;
    return ((package_info *)p)->timestamp;
}

gchar* PkgTsStorage::getPackageFilename(const gchar* package, bool *ok)
{
    gconstpointer p = findPackage(package);
    if ( p == NULL ) {
        *ok = false;
        return 0;
    }
    
    *ok = true;
    return g_strdup( ((package_info *)p)->filename );
}

gchar* PkgTsStorage::getPackageFilename(const gchar* package, const gchar* version, bool *ok)
{
    gconstpointer p = findPackage(package, version);
    if ( p == NULL ) {
        *ok = false;
        return 0;
    }
    
    *ok = true;
    return g_strdup( ((package_info *)p)->filename );
}

time_t PkgTsStorage::getFileTimestamp(const gchar* filename, bool *ok)
{
    gconstpointer p = findFile(filename);
    if ( p == NULL ) {
        *ok = false;
        return 0;
    }
    
    *ok = true;
    return ((package_info *)p)->timestamp;
}

time_t PkgTsStorage::getFileTimestamp(const gchar* filename, const gchar* version, bool *ok)
{
    gconstpointer p = findFile(filename, version);
    if ( p == NULL ) {
        *ok = false;
        return 0;
    }
    
    *ok = true;
    return ((package_info *)p)->timestamp;
}

gint comparePkgs(gconstpointer a, gconstpointer b)
{
    if ( ((package_info *)a)->timestamp < ((package_info *)b)->timestamp ) {
        return -1;
    } else if ( ((package_info *)a)->timestamp > ((package_info *)b)->timestamp ) {
        return 1;
    }

    return 0;
}

GList *PkgTsStorage::sortedList()
{
    if ( storage == NULL ) {
        return NULL;
    }
    
    storage = g_list_sort(storage, comparePkgs);

    return copyStorage();
}


