/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "log.h"
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <glib.h>
#include <unistd.h>

FILE *log_out = NULL;
FILE *log_err = NULL;
static bool log_debug_flag = false;
static bool log_enabled = true;

static char *exe_name ()
{
  static char buf[256] = {0};
  if (!buf[0])
    {
      if (readlink ("/proc/self/exe", buf, sizeof (buf)) == -1)
	abort ();

      char *slash = strrchr (buf, '/');
      if (slash)
	{
	  memmove (buf, slash + 1, strlen (slash) + 1);
	}
    }

  return buf;
}

static void log_handler (const gchar *log_domain,
			 GLogLevelFlags log_level,
			 const gchar *message,
			 gpointer unused_data)
{
  if (!log_enabled)
    return;

  FILE *out;
  if (log_level & (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL))
    out = log_err;
  else
    out = log_out;

  bool debug = (log_level & G_LOG_LEVEL_DEBUG) != 0;
  if (debug && !log_debug_flag)
    return;

  char tbuf[40] = {0};
  {
    time_t ts = time (0);
    tm t;
    localtime_r (&ts, &t);
    strftime (tbuf, sizeof (tbuf), "%F %T", &t);
  }

  fprintf (out, "%s %s[%i]: %s%s\n",
	   tbuf, exe_name (), getpid (), (debug ? "DEBUG: " : ""), message);
  fflush (out);

  if (log_level & G_LOG_LEVEL_ERROR)
    {
      fputs ("aborting...\n", out);
      abort ();
    }
}

int init_log_output (const char *out, const char *err)
{
  assert (!log_out);

  log_out = fopen (out, "a+");
  if (!log_out)
    return -1;

  log_err = fopen (err, "a+");
  if (!log_err)
    {
      fclose (log_out);
      return -1;
    }

  // Redirect stdout, stderr to log_out, log_err

  fflush (stdout);
  fflush (stderr);
  dup2 (fileno (log_out), STDOUT_FILENO);
  dup2 (fileno (log_err), STDERR_FILENO);

  g_log_set_handler (NULL,
		     GLogLevelFlags (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL | G_LOG_LEVEL_WARNING | G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO | G_LOG_LEVEL_DEBUG | G_LOG_FLAG_FATAL | G_LOG_FLAG_RECURSION),
		     log_handler, NULL);

  return 0;
}

void enable_debug_output (int b)
{
  log_debug_flag = b != 0;
}

void enable_logging (int b)
{
  log_enabled = b;
}
