/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <unistd.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sstream>
#include <fstream>
#include <apt-pkg/configuration.h>
#include <apt-pkg/error.h>

#include <glib.h>

#include "dbus.h"
#include "debug.h"
#include "apt-worker-cache.h"
#include "apt-worker-internal.h"
#include "apt-worker-rescue.h"
#include "apt-worker-utils.h"
#include "domain.h"

#define LOCK_FILE   "/var/lock/LCK..pkgmgrd.lock"
#define CONFIG_FILE "/etc/package-manager/config"

using namespace std;

static int lock_file = 0;

static gboolean no_root_check = FALSE;
static gboolean verbose_flag = FALSE;
static gint debug_level = 0;
static gboolean sticky = FALSE;
static gchar *log_file = NULL;
static GMainLoop *loop = NULL;
int verbosity = 1;

static gboolean
set_apt_option (const gchar *option_name,
		const gchar *value,
		gpointer data,
		GError **error)
{
  char *f = strdupa (value);
  const char *key = strtok (f, "=");
  const char *val = strtok (0, "=");
  if (!val)
    val = "1";

  _config->Set (key, val);

  g_message ("Set %s", value);

  return TRUE;
}

static GOptionEntry entries[] =
{
  { "debug", 'd', 0, G_OPTION_ARG_INT, &debug_level, "Debug level", NULL },
  { "option", 'o', 0, G_OPTION_ARG_CALLBACK, (void *) &set_apt_option, "Set configuration option", NULL },
  { "no-root-check", 'n', 0, G_OPTION_ARG_NONE, &no_root_check, "Don't check whether we are running as root", NULL },
  { "log-file", 'l', 0, G_OPTION_ARG_FILENAME, &log_file, "Log file", NULL },
  { "verbose", 'v', 0, G_OPTION_ARG_NONE, &verbose_flag, "Increase verbosity", NULL },
  { "sticky", 's', 0, G_OPTION_ARG_NONE, &sticky, "Do not use idle timer", NULL },
  { NULL }
};

/**
 * @brief Gets the free and cached memory.
 *
 * @param pointer to unsigned long which contains the amount of free memory while this function returns
 * @param pointer to unsigned long which contains the amount of cached memory while this function returns
 * @return boolean value True, if succeeded
*/
bool get_free_memory(unsigned long *free_mem, unsigned long *cached)
{
	bool free_mem_ok = false;
	bool cached_ok = false;
	string line;
	string str;
	unsigned long value_mem_free = 0;
	unsigned long value_cached = 0;
	ifstream inFile("/proc/meminfo");

	if (inFile) {
		while (!inFile.eof()) {
			getline(inFile, line);
			istringstream inStr(line);
			inStr >> str;

			if (str.compare("MemFree:") == 0) {
				inStr >> value_mem_free;
				free_mem_ok = true;
			} else if (str.compare("Cached:") == 0) {
				inStr >> value_cached;
				cached_ok = true;
			}
			if ( free_mem_ok && cached_ok ) {
				break; // No need to continue
			}
		} // while eof
		inFile.close();
	}

	*free_mem = value_mem_free;
	*cached = value_cached;
	return ( free_mem_ok && cached_ok );
}

static void unlink_lockfile ()
{
	close(lock_file);
	unlink(LOCK_FILE);
}

static PkgMgr *pkgmgr = NULL;
static gboolean
__quit(gpointer data)
{
	if (loop && g_main_loop_is_running(loop))
		g_main_loop_quit( loop );
	if (pkgmgr) {
		pkgmgr_unref (pkgmgr);
	}
	unlink_lockfile ();
	return FALSE;
}

/**
 * @brief Custom signal handler to SIGHUP,SIGTERM,SIGQUIT and SIGINT.
 *
 * - SIGTEMR/SIGQUIT/SIGINT quits main loop
 *
 * @param sig signal to handle
 */
static void
signal_handler(int sig)
{
        switch (sig) {
        case SIGTERM:
        case SIGQUIT:
        case SIGINT: {
		signal(SIGTERM, SIG_DFL);
		signal(SIGQUIT, SIG_DFL);
		signal(SIGINT, SIG_DFL);

		g_message ("Received signal %d (%s), shutting down...",
			   sig, strsignal (sig));

		if (pkgmgr)
			deinit_dbus (pkgmgr);
		else
			raise (sig);

// 		g_print ("Exiting...(%d)\n", sig);
		break;
        }
	case SIGSEGV:
		g_message ("Received signal %d (%s)",
			   sig, strsignal (sig));
		signal(SIGSEGV, SIG_DFL);
		raise(sig);
		break;
	default:
		g_error ("Got unhandled signal: %d\n", sig);
		break;
        }
}

static void
read_configuration ()
{
  GError *error = NULL;

  GKeyFile *conf = g_key_file_new ();
  if (!g_key_file_load_from_file (conf, CONFIG_FILE, G_KEY_FILE_NONE, &error))
    {
      if (error->code != G_KEY_FILE_ERROR_NOT_FOUND)
        g_message ("Can't read configuration file %s: %s",
                   CONFIG_FILE, error->message);
      g_error_free (error);
      return;
    }

  assume_network = g_key_file_get_boolean (conf, "Main", "assumeNetwork", NULL);
  g_message ("ASSUME NETWORK %d", assume_network);

  protect_user_packages_from_autoremoval =
    g_key_file_get_boolean (conf, "Main", "protectUserPackagesFromAutoremoval", NULL);

  gchar *f = g_key_file_get_string (conf, "Main", "logFile", NULL);
  if (f)
    {
      log_file = f;
      g_message ("logFile=%s", log_file);
    }

  verbose_flag = g_key_file_get_boolean (conf, "Main", "verbosity", NULL);

  g_key_file_free (conf);
}

static void
setup_stuff()
{
	gchar str[100];
        signal(SIGTERM, signal_handler);
        signal(SIGQUIT, signal_handler);
        signal(SIGINT, signal_handler);
        signal(SIGSEGV, signal_handler);

        lock_file = open(LOCK_FILE, O_RDWR|O_CREAT, 0640);

        /* can't open lock file*/
        if (lock_file < 0) {
                g_printerr("Can't open lock file\n");
                exit(1);
        }

        /* can't lock */
        if (lockf(lock_file, F_TLOCK, 0) < 0) {
                g_printerr("Can't set F_TLOCK to the lock file.\n");
                g_printerr("Daemon is probably already running!\n");
                exit(1);
        }

        /* record program ID to lockfile */
        sprintf(str, "%d\n", getpid());
        write(lock_file, str, strlen(str));

        // redirect stdin to /dev/null
        int fd = open ("/dev/null", O_RDONLY);
        if (fd >= 0 && fd != STDIN_FILENO)
          dup2 (fd, STDIN_FILENO);
}

int main(int argc, char** argv)
{
	GError *error = NULL;
	GOptionContext *context;

	/* When launched from D-Bus the environment contains only
	 * variables DBUS_STARTER_ADDRESS and
	 * DBUS_STARTER_BUS_TYPE. dpkg fails if PATH is unset. */
	setenv("PATH", "/usr/local/sbin:/usr/local/bin:/usr/sbin:/usr/bin:/sbin:/bin:/usr/bin/X11", 0);
	setenv("USER", "root", 0);
	setenv("HOME", "/root", 0);
	setenv("TMPDIR", "/var/tmp", 0);
	unsetenv("AEGIS_FIXED_ORIGIN");

	g_type_init();
	g_thread_init(NULL);
	dbus_g_thread_init ();
	AptWorkerCache::Initialize ();

        read_configuration();

	context = g_option_context_new ("-");
	g_option_context_add_main_entries (context, entries, NULL);
	if (!g_option_context_parse (context, &argc, &argv, &error)) {
		g_printerr("%s\n", error->message);
		return 1;
	}
	if (!no_root_check) {
		if (getuid() != 0) {
			g_printerr("I'm not root. Refusing to run. (use --no-root-check or -n to override)\n");
			return 1;
		}
	}
	if (verbose_flag)
		++verbosity;
	if (debug_level >= 1) {
		_config->Set("Debug::Algorithms", true);
		_config->Set("Debug::pkgDepCache::Marker", true);
		_config->Set("Debug::pkgProblemResolver", true);
	}

	setup_stuff();

	if (argc > 1 && !strcmp (argv[1], "rescue")) {
		if (!log_file)
			log_file = g_strdup("/dev/stdout");
		init_log_output(log_file, log_file);
		enable_debug_output(TRUE);
		return cmdline_rescue (argv + 1);
	}

	pkgmgr = init_dbus(&__quit, NULL);
	if (!pkgmgr) {
		unlink_lockfile ();
		return 1;
	}

	loop = g_main_loop_new (NULL, FALSE);
	if (!loop) {
		g_printerr("Failed to create main loop!\n");
		unlink_lockfile ();
		return 1;
	}

	if (!log_file)
		log_file = g_strdup("/dev/null");

	if (init_log_output(log_file, log_file) == -1) {
		g_printerr("Failed to open log file '%s'", log_file);
		return 1;
	}
	if (verbose_flag)
		enable_debug_output(TRUE);

	pkgmgr_start(pkgmgr, &__quit, NULL, sticky);

	g_message ("Running");
	g_main_loop_run (loop);
	g_message ("Exit");

	GMainContext *ctx = g_main_context_default ();
	g_main_context_unref (ctx);
	g_main_loop_unref (loop);
	g_main_context_unref (ctx);
	apt_worker_free ();
	// dbus_connection_unref(dbus_g_bus_get(DBUS_BUS_SYSTEM, NULL));
	if ( log_file ) {
	        g_free( log_file );
	}

	delete _config;
	delete _error;

	return 0;
}
