/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "markpolicydepcache.h"
#include <apt-pkg/algorithms.h>
#include <apt-pkg/configuration.h>
#include <apt-pkg/strutl.h>
#include <apt-pkg/sptr.h>

#define NEEDS_UPDATE (1<<15)

MarkPolicy::~MarkPolicy()
{
    if (DCache)
	DCache->SetMarkPolicy(0);
}

bool MarkPolicy::IsInstallOk(const PkgIterator &Pkg, bool AutoInst,
			     unsigned long Depth, bool FromUser)
{
    return DCache->IsInstallOk(Pkg, AutoInst, Depth, FromUser);
}

bool MarkPolicy::IsDeleteOk(const PkgIterator &Pkg, bool AutoInst,
			    unsigned long Depth, bool FromUser)
{
    return DCache->IsDeleteOk(Pkg, AutoInst, Depth, FromUser);
}

MarkPolicyDepCache::MarkPolicyDepCache(pkgCache *Cache, Policy *Plcy, MarkPolicy *markPlcy)
  : pkgDepCache(Cache, Plcy), markPolicy(markPlcy)
{
  // DebugMarker = false; 
  UpdatesNeeded = 0;
}

MarkPolicyDepCache::~MarkPolicyDepCache()
{
}

void MarkPolicyDepCache::SetMarkPolicy(MarkPolicy *Plcy)
{
    if (markPolicy)
	markPolicy->DCache = 0;

    markPolicy = Plcy;

    if (markPolicy)
	markPolicy->DCache = this;
}

static bool 
ConfigValueInSubTree(const char* SubTree, const char *needle)
{
   Configuration::Item const *Opts;
   Opts = _config->Tree(SubTree);
   if (Opts != 0 && Opts->Child != 0)
   {
      Opts = Opts->Child;
      for (; Opts != 0; Opts = Opts->Next)
      {
	 if (Opts->Value.empty() == true)
	    continue;
	 if (strcmp(needle, Opts->Value.c_str()) == 0)
	    return true;
      }
   }
   return false;
}

bool MarkPolicyDepCache::MarkInstall1(PkgIterator const &Pkg,bool AutoInst,
                                      unsigned long Depth, bool FromUser,
                                      bool ForceImportantDeps,
                                      bool StopOnError)
{
  bool res = MarkInstall_p (Pkg, AutoInst, Depth, FromUser, ForceImportantDeps, StopOnError);
  DoUpdates ();
  return res;
}

bool MarkPolicyDepCache::MarkDelete1(PkgIterator const &Pkg, bool rPurge,
                                      unsigned long Depth, bool FromUser)
{
  bool res = MarkDelete_p (Pkg, rPurge, Depth, FromUser);
  DoUpdates ();
  return res;
}

void MarkPolicyDepCache::NeedsUpdate(const PkgIterator &Pkg)
{
  PkgState[Pkg->ID].iFlags |= NEEDS_UPDATE;
  UpdatesNeeded += 1;
}

void MarkPolicyDepCache::DoUpdates()
{
  if (UpdatesNeeded > 10)
    {
      for (PkgIterator I = PkgBegin(); I.end() != true; I++)
        PkgState[I->ID].iFlags &= ~NEEDS_UPDATE;
      Update (NULL);
    }
  else
    {
      for (PkgIterator I = PkgBegin(); I.end() != true; I++)
        if (PkgState[I->ID].iFlags & NEEDS_UPDATE)
          {
            PkgState[I->ID].iFlags &= ~NEEDS_UPDATE;
            Update (I);
          }
    }
  UpdatesNeeded = 0;
}

bool MarkPolicyDepCache::MarkInstall_p(PkgIterator const &Pkg,bool AutoInst,
                                       unsigned long Depth, bool FromUser,
                                       bool ForceImportantDeps,
                                       bool StopOnError)
{
   if (Depth > 100)
      return true;
   
   // Simplifies other routines.
   if (Pkg.end() == true)
      return true;
   
   ActionGroup group(*this);

   /* Check that it is not already marked for install and that it can be 
      installed */
   StateCache &P = PkgState[Pkg->ID];
   P.iFlags &= ~AutoKept;
   if ((P.InstPolicyBroken() == false && P.InstBroken() == false) && 
       (P.Mode == ModeInstall ||
	P.CandidateVer == (Version *)Pkg.CurrentVer()))
   {
      if (P.CandidateVer == (Version *)Pkg.CurrentVer() && P.InstallVer == 0)
	 MarkKeep(Pkg, false, FromUser, Depth+1);
      return true;
   }

   // See if there is even any possible installation candidate
   if (P.CandidateVer == 0)
      return true;
   // We dont even try to install virtual packages..
   if (Pkg->VersionList == 0)
      return true;

   // check if we are allowed to install the package
   if (IsInstallOk(Pkg,AutoInst,Depth,FromUser) == false)
      return false;

   /* Target the candidate version and remove the autoflag. We reset the
      autoflag below if this was called recursively. Otherwise the user
      should have the ability to de-auto a package by changing its state */
   RemoveSizes(Pkg);
   RemoveStates(Pkg);
   
   P.Mode = ModeInstall;
   P.InstallVer = P.CandidateVer;

   if(FromUser)
     {
       // Set it to manual if it's a new install or cancelling the
       // removal of a garbage package.
       if(P.Status == 2 || (!Pkg.CurrentVer().end() && !P.Marked))
	 P.Flags &= ~Flag::Auto;
     }
   else
     {
       // Set it to auto if this is a new install.
       if(P.Status == 2)
	 P.Flags |= Flag::Auto;
     }
   if (P.CandidateVer == (Version *)Pkg.CurrentVer())
      P.Mode = ModeKeep;
       
   AddStates(Pkg);
   NeedsUpdate(Pkg);
   AddSizes(Pkg);

   if (AutoInst == false)
      return true;

   if (DebugMarker == true)
      std::clog << OutputInDepth(Depth) << "MarkInstall " << Pkg << " FU=" << FromUser << std::endl;

   DepIterator Dep = P.InstVerIter(*this).DependsList();
   for (; Dep.end() != true;)
   {
      // Grok or groups
      DepIterator Start = Dep;
      bool Result = true;
      unsigned Ors = 0;
      for (bool LastOR = true; Dep.end() == false && LastOR == true; Dep++,Ors++)
      {
	 LastOR = (Dep->CompareOp & Dep::Or) == Dep::Or;

	 if ((DepState[Dep->ID] & DepInstall) == DepInstall)
	    Result = false;
      }
      
      // Dep is satisfied okay.
      if (Result == false)
	 continue;

      /* Check if this dep should be consider for install. If it is a user
         defined important dep and we are installed a new package then 
	 it will be installed. Otherwise we only check for important
         deps that have changed from the installed version
      */
      if (IsImportantDep(Start) == false)
	 continue;
      
      /* Check if any ImportantDep() (but not Critical) were added
       * since we installed the package.  Also check for deps that
       * were satisfied in the past: for instance, if a version
       * restriction in a Recommends was tightened, upgrading the
       * package should follow that Recommends rather than causing the
       * dependency to be removed. (bug #470115)
       */
      bool isNewImportantDep = false;
      bool isPreviouslySatisfiedImportantDep = false;
      if(!ForceImportantDeps && !Start.IsCritical())
      {
	 bool found=false;
	 VerIterator instVer = Pkg.CurrentVer();
	 if(!instVer.end())
	 {
	   for (DepIterator D = instVer.DependsList(); D.end() != true; D++)
	     {
	       //FIXME: deal better with or-groups(?)
	       DepIterator LocalStart = D;

	       if(IsImportantDep(D) && !D.IsCritical() &&
		  Start.TargetPkg() == D.TargetPkg())
		 {
		   if(!isPreviouslySatisfiedImportantDep)
		     {
		       DepIterator D2 = D;
		       while((D2->CompareOp & Dep::Or) != 0)
			 ++D2;

		       isPreviouslySatisfiedImportantDep =
			 (((*this)[D2] & DepGNow) != 0);
		     }

		   found=true;
		 }
	     }
	    // this is a new dep if it was not found to be already
	    // an important dep of the installed package
	    isNewImportantDep = !found;
	 }
      }
      if(isNewImportantDep)
	 if(DebugAutoInstall == true)
	    std::clog << OutputInDepth(Depth) << "new important dependency: "
		      << Start.TargetPkg().Name() << std::endl;
      if(isPreviouslySatisfiedImportantDep)
	if(DebugAutoInstall == true)
	  std::clog << OutputInDepth(Depth) << "previously satisfied important dependency on "
		    << Start.TargetPkg().Name() << std::endl;

      // skip important deps if the package is already installed
      if (Pkg->CurrentVer != 0 && Start.IsCritical() == false 
	  && !isNewImportantDep && !isPreviouslySatisfiedImportantDep
	  && !ForceImportantDeps)
	 continue;
      
      /* If we are in an or group locate the first or that can 
         succeed. We have already cached this.. */
      for (; Ors > 1 && (DepState[Start->ID] & DepCVer) != DepCVer; Ors--)
	 Start++;

      /* This bit is for processing the possibilty of an install/upgrade
         fixing the problem */
      SPtrArray<Version *> List = Start.AllTargets();
      if (Start->Type != Dep::DpkgBreaks &&
	  (DepState[Start->ID] & DepCVer) == DepCVer)
      {
	 // Right, find the best version to install..
	 Version **Cur = List;
	 PkgIterator P = Start.TargetPkg();
	 PkgIterator InstPkg(*Cache,0);
	 
	 // See if there are direct matches (at the start of the list)
	 for (; *Cur != 0 && (*Cur)->ParentPkg == P.Index(); Cur++)
	 {
	    PkgIterator Pkg(*Cache,Cache->PkgP + (*Cur)->ParentPkg);
	    if (PkgState[Pkg->ID].CandidateVer != *Cur)
	       continue;
	    InstPkg = Pkg;
	    break;
	 }

	 // Select the highest priority providing package
	 if (InstPkg.end() == true)
	 {
	    pkgPrioSortList(*Cache,Cur);
	    for (; *Cur != 0; Cur++)
	    {
	       PkgIterator Pkg(*Cache,Cache->PkgP + (*Cur)->ParentPkg);
	       if (PkgState[Pkg->ID].CandidateVer != *Cur)
		  continue;
	       InstPkg = Pkg;
	       break;
	    }
	 }
	 
	 if (InstPkg.end() == false)
	 {
	    if(DebugAutoInstall == true)
	       std::clog << OutputInDepth(Depth) << "Installing " << InstPkg.Name()
			 << " as " << Start.DepType() << " of " << Pkg.Name()
			 << std::endl;
 	    // now check if we should consider it an automatic dependency or not
 	    if(Pkg.Section() && ConfigValueInSubTree("APT::Never-MarkAuto-Sections", Pkg.Section()))
 	    {
	       if(DebugAutoInstall == true)
		  std::clog << OutputInDepth(Depth) << "Setting NOT as auto-installed (direct "
                            << Start.DepType() << " of pkg in APT::Never-MarkAuto-Sections)" << std::endl;
 	       if (!MarkInstall_p(InstPkg,true,Depth + 1, true, false, StopOnError)
		   && StopOnError)
		  return false;
 	    }
 	    else 
 	    {
 	       // mark automatic dependency
	       if (!MarkInstall_p(InstPkg,true,Depth + 1, false, ForceImportantDeps, StopOnError)
		   && StopOnError)
		   return false;

 	       // Set the autoflag, after MarkInstall because MarkInstall unsets it
 	       if (P->CurrentVer == 0)
 		  PkgState[InstPkg->ID].Flags |= Flag::Auto;
 	    }
	 }
	 continue;
      }

      /* For conflicts we just de-install the package and mark as auto,
         Conflicts may not have or groups.  For dpkg's Breaks we try to
         upgrade the package. */
      if (Start->Type == Dep::Conflicts || Start->Type == Dep::Obsoletes ||
	  Start->Type == Dep::DpkgBreaks)
      {
	 for (Version **I = List; *I != 0; I++)
	 {
	    VerIterator Ver(*this,*I);
	    PkgIterator Pkg = Ver.ParentPkg();

	    if (Start->Type != Dep::DpkgBreaks)
	    {
		if (!IsDeleteOk(Pkg))
		   return false;

		if (!MarkDelete_p(Pkg,false,Depth + 1, false)
		    && StopOnError)
		    return false;
	    }
	    else if (PkgState[Pkg->ID].CandidateVer != *I)
	       if (!MarkInstall_p(Pkg,true,Depth + 1, false, ForceImportantDeps, StopOnError)
		   && StopOnError)
		  return false;
	 }
	 continue;
      }      
   }

   return true;
}

bool MarkPolicyDepCache::MarkDelete_p(PkgIterator const &Pkg, bool rPurge,
                                      unsigned long Depth, bool FromUser)
{
   // Simplifies other routines.
   if (Pkg.end() == true)
      return true;

   ActionGroup group(*this);

   // Check that it is not already marked for delete
   StateCache &P = PkgState[Pkg->ID];
   P.iFlags &= ~(AutoKept | Purge);
   if (rPurge == true)
      P.iFlags |= Purge;
   
   if ((P.Mode == ModeDelete || P.InstallVer == 0) && 
       (Pkg.Purge() == true || rPurge == false))
      return true;
   
   // We dont even try to delete virtual packages..
   if (Pkg->VersionList == 0)
      return true;

   // check if we are allowed to install the package
   if (IsDeleteOk(Pkg,rPurge,Depth,FromUser) == false)
      return false;

   if (DebugMarker == true)
      std::clog << OutputInDepth(Depth) << "MarkDelete " << Pkg << " FU=" << FromUser << std::endl;

   RemoveSizes(Pkg);
   RemoveStates(Pkg);
   
   if (Pkg->CurrentVer == 0 && (Pkg.Purge() == true || rPurge == false))
      P.Mode = ModeKeep;
   else
      P.Mode = ModeDelete;
   P.InstallVer = 0;

   AddStates(Pkg);   
   NeedsUpdate(Pkg);
   AddSizes(Pkg);

   return true;
}

bool MarkPolicyDepCache::IsInstallOk(const PkgIterator &Pkg,bool AutoInst,
				     unsigned long Depth, bool FromUser)
{
    if (markPolicy)
	return markPolicy->IsInstallOk(Pkg, AutoInst, Depth, FromUser);

    return pkgDepCache::IsInstallOk(Pkg, AutoInst, Depth, FromUser);
}

bool MarkPolicyDepCache::IsDeleteOk(const PkgIterator &Pkg, bool AutoInst,
				    unsigned long Depth, bool FromUser)
{
    if (markPolicy)
	return markPolicy->IsDeleteOk(Pkg, AutoInst, Depth, FromUser);

    return pkgDepCache::IsDeleteOk(Pkg, AutoInst, Depth, FromUser);
}
