/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "netconnection.h"
#include <apt-pkg/configuration.h>
#include <cstdio>
#include <string.h>

#include "netconnection_defs.h"

static void log_handler (const gchar *log_domain,
                         GLogLevelFlags log_level,
                         const gchar *message,
                         gpointer unused_data)
{
  fprintf (stderr, "%s\n", message);
}


NetConnection::NetConnection (GMainLoop *loop, gint response_fd, gint status_fd)
// NetConnection::NetConnection (GMainLoop *loop)
    : debug (_config->FindB ("Debug::NetConnection")),
      state (Disconnected),
      // status_mutex (g_mutex_new()),
      // status_change_cond (g_cond_new()),
      status (StatusDisconnected),
      connection (0),
      event_loop (loop),
      response_fd(response_fd),
      status_fd(status_fd),
      send_response(false)
{
    g_log_set_handler (NULL,
                       GLogLevelFlags (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL
                                       | G_LOG_LEVEL_WARNING |
                                       G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO |
                                       G_LOG_LEVEL_DEBUG | G_LOG_FLAG_FATAL |
                                       G_LOG_FLAG_RECURSION), log_handler,
                       NULL);
}

NetConnection::~NetConnection ()
{
  if (connection)
    g_object_unref (connection);  
}

void NetConnection::connect (bool automatic)
{
  g_debug("NetConnection::connect");
  if (status == StatusConnected || send_response)
    return;

  send_response = true;
  connection = con_ic_connection_new ();
  g_signal_connect (connection, "connection-event",
		    G_CALLBACK (static_connection_event), this);

  transit (Connecting);

  if (!con_ic_connection_connect (connection,
                                  (automatic
                                   ? CON_IC_CONNECT_FLAG_AUTOMATICALLY_TRIGGERED
                                   : CON_IC_CONNECT_FLAG_NONE)))
    {
      g_debug ("con_ic_connection_connect() failed");
      transit (Disconnected);
    }
}

void NetConnection::disconnect ()
{
  g_debug ("NetConnection::disconnect()");

  if (!_config->FindB ("PM::wait-disconnect-reply"))
    {
      transit (Disconnected);
      return;
    }  

  if (!con_ic_connection_disconnect (connection))
    {
      g_debug ("con_ic_connection_disconnect() failed");
      transit (Disconnected);
    }  
}  

void NetConnection::transit (State st)
{
  state = st;

  if (state == Connected)
    {
      g_debug ("NetConnection::transit(), status is connected");
      if(send_response) {
          g_debug ("NetConnection::transit(), send response CONNECTED_STR");
        send_response = false;
        write(response_fd, CONNECTED_STR, strlen(CONNECTED_STR) + 1);
      } else {
          g_debug ("NetConnection::transit(), send status CONNECTED_STR");
        write(status_fd, CONNECTED_STR, strlen(CONNECTED_STR) + 1);
      }
    }

  else if (state == Disconnected || state == Disconnecting)
    {
      if(state == Disconnected) {
        g_debug ("NetConnection::transit(), send status DISCONNECTED_STR");
        if(send_response) {
            g_debug ("NetConnection::transit(), send response CONNECTED_STR");
            send_response = false;
            write(response_fd, DISCONNECTED_STR, strlen(DISCONNECTED_STR) + 1);
        } else {
          write(status_fd, DISCONNECTED_STR, strlen(DISCONNECTED_STR) + 1 );
        }
    }
      else {
          g_debug ("NetConnection::transit(), send status DISCONNECTING_STR");
        write(status_fd, DISCONNECTING_STR, strlen(DISCONNECTING_STR) + 1 );
    }
      g_object_unref (connection);
      connection = 0;
      if (g_main_loop_is_running (event_loop))
        g_main_loop_quit (event_loop);
    }  
}

void NetConnection::set_connection_status (ConnectionStatus st)
{
  // g_mutex_lock (status_mutex);
  if (st != status)
    {
      status = st;
      // g_cond_signal (status_change_cond);
    }
  // g_mutex_unlock (status_mutex);
}

void NetConnection::connection_event (ConIcConnection *connection,
				      ConIcConnectionEvent *event)
{
    g_debug ("NetConnection::connection_event()");
  ConIcConnectionError con_err = con_ic_connection_event_get_error (event);
  ConIcConnectionStatus con_state = con_ic_connection_event_get_status (event);
  const gchar *event_iap_id = con_ic_event_get_iap_id (CON_IC_EVENT (event));

  switch (con_err)
    {
    case CON_IC_CONNECTION_ERROR_NONE:
      break;
    case CON_IC_CONNECTION_ERROR_INVALID_IAP:
      g_debug("conic: IAP is invalid");
      break;
    case CON_IC_CONNECTION_ERROR_CONNECTION_FAILED:
      g_debug("conic: connection failed");
      break;
    case CON_IC_CONNECTION_ERROR_USER_CANCELED:
      g_debug("conic: user cancelled");
      break;
    default:
      g_debug("conic: unknown error");
      break;
    }

  switch (con_state)
    {
    case CON_IC_STATUS_CONNECTED:
      g_debug("conic: connected to %s", event_iap_id);

      /* if IAP changed we shall break the download */
      if (!iap_id.empty () && iap_id != event_iap_id)
	{
	  if (!iap_id.empty ())
	    {
	      iap_id.clear ();
	      set_connection_status (StatusCanceled);
	      transit (Disconnected);
	    }
	  iap_id = event_iap_id;
	}
      else
	{          
	  iap_id = event_iap_id;
	  set_connection_status (StatusConnected);
	  transit (Connected);
	}

      break;

    case CON_IC_STATUS_DISCONNECTING:
      // IAP was disconnected
      g_debug("conic: disconnecting from %s", event_iap_id);

      if (iap_id == event_iap_id)
	{
	  if (!_config->FindB ("PM::cancel-on-disconnecting"))
	    {
              g_debug("conic: Cancel on disconnecting signal disabled");
	      break;
	    }          

	  iap_id.clear ();
	  set_connection_status (StatusDisconnected);
          transit (Disconnecting);
	}

      break;

    case CON_IC_STATUS_DISCONNECTED:
      g_debug("conic: disconnected from %s", event_iap_id);

      if (!event_iap_id)
	{
	  if (iap_id.empty ())
	    {
              g_debug("conic: connection canceled");
	      iap_id.clear ();
	      set_connection_status (StatusCanceled);
	      transit (Disconnected);
	    }
	}
      else
	{
	  /* If the IAP is the same then the current connection
	   * was disconnected */
	  if (!iap_id.empty () && event_iap_id && iap_id == event_iap_id)
	    {
	      iap_id.clear ();
	      set_connection_status (StatusDisconnected);
	      transit (Disconnected);
	    }
	}

      break;

    default:
      g_debug("conic: unknown event");
      break;
    }
}

void NetConnection::static_connection_event (ConIcConnection *connection,
					     ConIcConnectionEvent *event,
					     gpointer user_data)
{
  static_cast<NetConnection *> (user_data)->connection_event (connection, event);
}
