/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <apt-pkg/configuration.h>
#include <apt-pkg/init.h>
#include <cassert>
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <string.h>
#include <conic.h>
#include <unistd.h>
#include "netconnection.h"
#include "netconnection_defs.h"

static gboolean automatic = FALSE;
static gint request_fd, response_fd, status_fd;
static NetConnection *net_connection = NULL;
static GMainLoop *loop = NULL;

static GOptionEntry entries[] =
{
  { "request_fd", 'q', 0, G_OPTION_ARG_INT, &request_fd, "Request_fd", NULL },
  { "response_fd", 'r', 0, G_OPTION_ARG_INT, &response_fd, "Response_fd", NULL },
  { "status_fd", 's', 0, G_OPTION_ARG_INT, &status_fd, "Status_fd", NULL },
  { "automatic", 'a', 0, G_OPTION_ARG_NONE, &automatic, "Automatic netconnection", NULL },
  { NULL }
};

static void log_handler (const gchar *log_domain,
                         GLogLevelFlags log_level,
                         const gchar *message,
                         gpointer unused_data)
{
  fprintf (stderr, "%s\n", message);
}


// launch net connection
static gboolean connect (gpointer data)
{
    g_debug("main, connect");
    if(net_connection) {
        net_connection->connect(automatic);
    }
    return FALSE;
}

static bool
dbus_connection_init (GMainContext *context)
{
  DBusConnection *conn;
  DBusError err;

  dbus_error_init (&err);
  conn = dbus_bus_get (DBUS_BUS_SYSTEM, &err);
  if (!conn)
    {
      dbus_error_free (&err);
      return false;
    }

  dbus_connection_setup_with_g_main (conn, context);

  return true;
}

static gboolean
read_request (GIOChannel *channel, GIOCondition cond, gpointer data)
{
    g_debug("main, read_request");
    char buf[MAX_BUF_SIZE];
    int n, fd = g_io_channel_unix_get_fd (channel);

    n = read (fd, buf, MAX_BUF_SIZE);
    if (n > 0) {
        g_debug("main, read_request, buf: %s", buf);
        if(strcmp(buf, DISCONNECT_STR) == 0) {
            if(net_connection) {
                net_connection->disconnect();
            }
        }
        return TRUE;
    } else {
        g_io_channel_shutdown (channel, 0, NULL);
        return FALSE;
    }
}

// main
int main (int argc, char *argv[])
{
    g_debug("main");

    g_log_set_handler (NULL,
                       GLogLevelFlags (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL
                                       | G_LOG_LEVEL_WARNING |
                                       G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO |
                                       G_LOG_LEVEL_DEBUG | G_LOG_FLAG_FATAL |
                                       G_LOG_FLAG_RECURSION), log_handler,
                       NULL);


    // check that process is not already running
    if(loop) {
        g_debug("main, mainloop exists");
        return 1;
    } else {
        loop = g_main_loop_new(NULL, FALSE); // init main loop
    }

    // parse the arguments
    GError *error = NULL;
    GOptionContext *context;
    context = g_option_context_new ("");
    g_option_context_add_main_entries (context, entries, NULL);
    if (!g_option_context_parse (context, &argc, &argv, &error)) {
        g_debug("main, parse failed");
        g_main_loop_unref(loop);
        return 1;
    }

    g_debug("main, parsed args, request_fd: %d, response_fd: %d, status_fd: %d, automatic: %d", request_fd, response_fd, status_fd, (int)automatic);

    // set config
    _config->Set ("quiet", 0);
    _config->Set ("Debug::NetConnection", true);
    _config->Set ("PM::cancel-on-disconnecting", 1);
    _config->Set ("PM::wait-disconnect-reply", 1);

    // required inittings
    g_type_init ();
    assert (dbus_connection_init (g_main_loop_get_context(loop)));

    // start listening request pipe
    GIOChannel *channel = g_io_channel_unix_new (request_fd);
    g_io_add_watch (channel, GIOCondition(G_IO_IN | G_IO_HUP | G_IO_ERR),
                    read_request, NULL);

    // init net_connection and delay the connecting
    net_connection = new NetConnection(loop, response_fd, status_fd);
    //net_connection->connect(automatic);
    g_timeout_add(50, connect, NULL);

    // start main loop
    g_main_loop_run (loop);
    g_main_loop_unref (loop);

    delete net_connection;

    g_debug("main, end");

    return 0;
}
