/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <string.h>

#include <dbus/dbus.h>
#include <dbus/dbus-glib.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <cassert>

#include "package.h"
#include "catalogue.h"
#include "debug.h"
#include "error.h"

#include "pkgmgr-dbus.h"
#include "pkgmgr-dbus-glue.h"
#include "dbus-marshal.h"
#include "pkgmgr-if.h"

#include "apt-worker.h"
#include "glib-util.h"
#include "format-metadata.h"

G_DEFINE_TYPE(PkgMgr, pkgmgr, G_TYPE_OBJECT)


static void
pkgmgr_finalize(GObject *obj)
{
	DEBUG("");
	PkgMgr *self = PKGMGR (obj);

	delete self->apt_worker_thread;
}

static void
pkgmgr_class_init(PkgMgrClass *klass)
{
	GObjectClass *gobject_class = G_OBJECT_CLASS (klass);
	gobject_class->finalize = pkgmgr_finalize;

	klass->download_progress =
		g_signal_new("download_progress",
			G_OBJECT_CLASS_TYPE (klass),
			G_SIGNAL_RUN_LAST,
			0, NULL, NULL, __marshaller_VOID__STRING_STRING_STRING_INT_INT,
			G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT, G_TYPE_INT);
	klass->operation_started =
		g_signal_new("operation_started",
			G_OBJECT_CLASS_TYPE (klass),
			G_SIGNAL_RUN_LAST,
			0, NULL, NULL, __marshaller_VOID__STRING_STRING_STRING,
			G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
	klass->operation_progress =
		g_signal_new("operation_progress",
			G_OBJECT_CLASS_TYPE (klass),
			G_SIGNAL_RUN_LAST,
			0, NULL, NULL, __marshaller_VOID__STRING_STRING_STRING_INT,
			G_TYPE_NONE, 4, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_INT);
	klass->operation_complete =
		g_signal_new("operation_complete",
			G_OBJECT_CLASS_TYPE (klass),
			G_SIGNAL_RUN_LAST,
			0, NULL, NULL, __marshaller_VOID__STRING_STRING_STRING_STRING_BOOL,
			G_TYPE_NONE, 5, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_BOOLEAN);

	klass->package_list_updated =
		g_signal_new("package_list_updated",
			G_OBJECT_CLASS_TYPE (klass),
			G_SIGNAL_RUN_LAST,
			0, NULL, NULL, __marshaller_VOID__BOOL,
			G_TYPE_NONE, 1, G_TYPE_BOOLEAN);

    klass->package_ignored =
		g_signal_new("package_ignored",
			G_OBJECT_CLASS_TYPE (klass),
			G_SIGNAL_RUN_LAST,
			0, NULL, NULL, __marshaller_VOID__STRING_STRING_STRING,
			G_TYPE_NONE, 3, G_TYPE_STRING, G_TYPE_STRING, G_TYPE_STRING);
}
static void
__download_progress_signal (const gchar* operation, const gchar* packagename, const gchar* version, int already, int total, gpointer data)
{
	PkgMgr* pkgmgr = static_cast<PkgMgr*>(data);

	DEBUG("%d/%d", already, total);
	g_signal_emit (pkgmgr,
		PKGMGR_GET_CLASS(pkgmgr)->download_progress,
		0, operation, packagename, version, already, total);
}
static void
__operation_started_signal (const gchar* op, const gchar *pkg, const gchar *ver, gpointer data)
{
	PkgMgr* pkgmgr = static_cast<PkgMgr*>(data);

	DEBUG("");
	g_signal_emit (pkgmgr,
		PKGMGR_GET_CLASS(pkgmgr)->operation_started,
		0, op, pkg, ver);
}
static void
__operation_progress_signal (const gchar* op, const gchar *pkg, const gchar *version, int percentage, gpointer data)
{
	PkgMgr* pkgmgr = static_cast<PkgMgr*>(data);

	DEBUG("%d %%\r", percentage);
	g_signal_emit (pkgmgr,
		PKGMGR_GET_CLASS(pkgmgr)->operation_progress,
		0, op, pkg, version, percentage);
}
static void
__operation_complete_signal (const gchar* op, gchar* pkg, const gchar *version, const gchar* error, gboolean need_reboot, gpointer data)
{
	PkgMgr* pkgmgr = static_cast<PkgMgr*>(data);

	DEBUG("%s %s: %s\r", op, pkg, error);
	g_signal_emit (pkgmgr,
		PKGMGR_GET_CLASS(pkgmgr)->operation_complete,
		0, op, pkg, version, error, need_reboot);
// 	g_free (pkg);
}

static void
__package_list_updated_signal (gboolean updates_changed, gpointer data)
{
	PkgMgr* pkgmgr = static_cast<PkgMgr*>(data);

	DEBUG("%s\r", __PRETTY_FUNCTION__);
	g_signal_emit (pkgmgr,
		       PKGMGR_GET_CLASS(pkgmgr)->package_list_updated,
		       0, updates_changed);
}

static void
__package_ignored_signal (const gchar* type, const gchar* package, const gchar* version, gpointer data)
{
	PkgMgr* pkgmgr = static_cast<PkgMgr*>(data);

	DEBUG("%s\r", __PRETTY_FUNCTION__);
	g_signal_emit (pkgmgr,
		       PKGMGR_GET_CLASS(pkgmgr)->package_ignored,
		       0, type, package, version);
}

static void
pkgmgr_init(PkgMgr *pkgmgr)
{
	dbus_g_object_type_install_info (PKGMGR_TYPE,
					&dbus_glib_pkgmgr_object_info);


// 	apt_worker_init (&__download_progress_signal, pkgmgr);
}

static void
pkgmgr_error (DBusGMethodInvocation* method_context,
		const gchar *error_name, const gchar *message)
{
	DEBUG("Name: %s: %s", error_name, message);
	g_return_if_fail(method_context);
	GError *err = NULL;
	DBusError error = {0,};
	error.name = error_name;
	error.message = message;
	dbus_set_g_error(&err, &error);
	dbus_g_method_return_error(method_context, err);
	g_error_free( err );
}
static void
pkgmgr_error (DBusGMethodInvocation* method_context,
		GError *error)
{
    assert( error );
	const gchar *msg = "";
	if (error && error->message) {
		msg = error->message;
	}
	pkgmgr_error (method_context, pkgmgr_error_to_dbus_error(PkgmgrdError(error->code)), msg);
}

static gboolean
__service_shutdown(PkgMgr *pkgmgr, DBusGMethodInvocation *method_context)
{
	gboolean stopping = !pkgmgr->apt_worker_thread || pkgmgr->apt_worker_thread->is_stopping();
	if (stopping) {
		DEBUG("We are stopping...");
		GError *err = NULL;
		g_set_error (&err, PKGMGRD_ERROR, ERROR_SERVICE_SHUTDOWN, "Service is shutting down manually");
		pkgmgr_error (method_context, err);
		g_error_free (err);
		err = NULL;
	}
	return stopping;
}
// static void
//  __free_package(gpointer data, gpointer user_data)
// {
// 	Package *pkg = static_cast<Package*> (data);
// 	package_free(pkg);
// }
static void __insert_value_string(GHashTable *map, const gchar *key, const gchar *s)
{
	if (s) {
		GValue* val = g_new0(GValue,1);
		g_value_init(val, G_TYPE_STRING);
                gchar *utf8_s = format_utf8(s);
                g_value_set_string(val, utf8_s);
		g_hash_table_insert (map, g_strdup(key), val);
                g_free(utf8_s);
// 		DEBUG("%s: %s", key, s);
	}
}
static void __insert_value_int64(GHashTable *map, const gchar *key, gint64 i)
{
	GValue* val = g_new0(GValue,1);
	g_value_init(val, G_TYPE_INT64);
	g_value_set_int64(val, i);
	g_hash_table_insert (map, g_strdup(key), val);
// 	DEBUG("%s: %"G_GINT64_FORMAT, key, i);
}
// static void __insert_value_int(GHashTable *map, const gchar *key, gint i)
// {
// 	GValue* val = g_new0(GValue,1);
// 	g_value_init(val, G_TYPE_INT);
// 	g_value_set_int(val, i);
// 	g_hash_table_insert (map, g_strdup(key), val);
// // 	DEBUG("%s: %d", key, i);
// }
static void __insert_value_boolean(GHashTable *map, const gchar *key, gboolean b)
{
	GValue* val = g_new0(GValue,1);
	g_value_init(val, G_TYPE_BOOLEAN);
	g_value_set_boolean(val, b);
	g_hash_table_insert (map, g_strdup(key), val);
// 	DEBUG("%s: %d", key, i);
}

void __copy(gpointer key, gpointer value, gpointer user_data)
{
//    DEBUG("key: %s\n", static_cast<gchar*>(key));
    GHashTable *to = static_cast<GHashTable*>(user_data);
    GValue *gvalue = static_cast<GValue*>(value);

    if(G_VALUE_HOLDS_STRING(gvalue)) {
      gchar *oldvalue = g_value_dup_string(gvalue);
      gchar *newvalue = format_utf8(oldvalue);
      g_hash_table_insert_string_value(to, g_strdup(static_cast<gchar*>(key)), newvalue);
      g_free (newvalue);
      g_free (oldvalue);
    } else {
      GValue* to_value = g_new0 (GValue, 1);
      g_value_init (to_value, G_VALUE_TYPE(value));
      g_value_copy(static_cast<GValue*>(value), to_value);
      g_hash_table_insert(to, g_strdup(static_cast<gchar*>(key)), to_value);
    }
}

static void
__insert_from_hash_table(GHashTable *from, GHashTable *to)
{
    g_hash_table_foreach(from, __copy, to);
}

static void
__pack_package_to_map(Package *pkg, GHashTable *map)
{
	__insert_value_string(map, "Name", pkg->name);
	__insert_value_boolean(map, "IsBroken", pkg->is_broken);
	__insert_value_string(map, "Version", pkg->version);
	__insert_value_string(map, "Description", pkg->description);
	__insert_value_string(map, "DisplayName", pkg->display_name);
	__insert_value_string(map, "Category", pkg->category);
	__insert_value_string(map, "Repository", pkg->repository);
	__insert_value_string(map, "IconData", pkg->icon_data);
	__insert_value_string(map, "Type", package_type_to_string(pkg->type));
	__insert_value_int64(map, "InstalledSize", pkg->installed_size);
	__insert_value_int64(map, "Size", pkg->download_size);
    if ( pkg->filename && strcmp(pkg->filename, "") != 0 ) {
    	__insert_value_string(map, "Filename", pkg->filename);
    }
    if ( pkg->installed_timestamp != 0 )
    	__insert_value_int64(map, "InstalledTimestamp", pkg->installed_timestamp);
    if ( pkg->available_timestamp != 0 )
    	__insert_value_int64(map, "AvailableTimestamp", pkg->available_timestamp);
	if ( pkg->flag_reboot )
    	__insert_value_boolean(map, "FlagReboot", pkg->flag_reboot);
	if ( pkg->flag_ignored )
    	__insert_value_boolean(map, "FlagIgnored", pkg->flag_ignored);
    if ( pkg->flag_system_update )
    	__insert_value_boolean(map, "FlagSystemUpdate", pkg->flag_system_update);
        if ( pkg->flag_local_file )
                __insert_value_boolean(map, "FlagLocalFile", pkg->flag_local_file);
	if (pkg->sys_upgrade_name)
		__insert_value_string(map, "DependsSystemUpgradeName", pkg->sys_upgrade_name);
	if (pkg->sys_upgrade_version)
		__insert_value_string(map, "DependsSystemUpgradeVersion", pkg->sys_upgrade_version);
	if (pkg->error_name)
		__insert_value_string(map, "ErrorName", pkg->error_name);
	if (pkg->error_message)
		__insert_value_string(map, "ErrorMessage", pkg->error_message);
	if (pkg->extra_fields) {
    	__insert_from_hash_table(pkg->extra_fields, map);
    }
}
static GPtrArray*
__pack_package_list_to_array_map(GList *pkgs)
{
	GPtrArray *packages = g_ptr_array_new ();
	GList *list = pkgs;
	while(list) {
		Package *pkg = (Package*)list->data;
		GHashTable* map = g_hash_table_new(g_str_hash, g_str_equal);
		__pack_package_to_map(pkg, map);
		g_ptr_array_add (packages, map);
		list = g_list_next(list);
	}

	return packages;
}
// static void
// __pack_catalogue_to_map(Catalogue *cat, GHashTable *map)
// {
// 	__insert_value_string(map, "Name", cat->name);
// 	__insert_value_string(map, "URI", cat->uri);
// 	__insert_value_string(map, "Dist", cat->dist);
// 	__insert_value_string(map, "Components", cat->components);
// 	__insert_value_boolean(map, "Enabled", cat->enabled);
// 	__insert_value_boolean(map, "Essential", cat->essential);
// }
// static GPtrArray*
// __pack_catalogue_list_to_array_map(GList *cats)
// {
// 	GPtrArray *catalogues = g_ptr_array_new ();
// 	GList *list = cats;
// 	while(list) {
// 		Catalogue *cat = (Catalogue*)list->data;
// 		GHashTable* map = g_hash_table_new(g_str_hash, g_str_equal);
// 		__pack_catalogue_to_map(cat, map);
// 		g_ptr_array_add (catalogues, map);
// 		list = g_list_next(list);
// 	}
// 	return catalogues;
// }
static gboolean
__free_map_value (gpointer key,
	gpointer value,
	gpointer user_data)
{
	gchar *k = static_cast<gchar*>(key);
	GValue *v = static_cast<GValue*>(value);
	g_free (k);
	g_value_unset (v);
	g_free (v);
	return TRUE;
}
static void
__free_map(GHashTable *map)
{
	g_hash_table_foreach_steal (map, __free_map_value, NULL);
	g_hash_table_destroy(map);
}
static void
__free_map_array(GPtrArray* map_array)
{
	for(guint i=0; i < map_array->len; i++) {
		__free_map(static_cast<GHashTable*>(map_array->pdata[i]));
	}
	g_ptr_array_free(map_array, TRUE);
}
/**
 */
static void
pkgmgr_fetch_packages_reply (DBusGMethodInvocation *method_context,
				GList *pkgs)
{
	g_return_if_fail(method_context);

	GPtrArray *packages = __pack_package_list_to_array_map (pkgs);
	dbus_g_method_return(method_context, packages);
	__free_map_array(packages);
}

static void
pkgmgr_fetch_packages_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_package_list *cmd =
		static_cast<apt_worker_cmd_get_package_list *> (cmd_base);

	if (cmd->error) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
		pkgmgr_fetch_packages_reply (method_context, cmd->packages);
// 		g_list_foreach (cmd->packages, __free_package, NULL);
// 		g_list_free (cmd->packages);
	}

	delete cmd;
}

static gboolean check_parameter_str(const gchar *p,
				    DBusGMethodInvocation *method_context)
{
	if (strchr(p, '%')) {
		GError *error = NULL;
		g_set_error(&error, PKGMGRD_ERROR, ERROR_INVALID_ARGUMENT,
			    "Invalid character '%%' detected");
		pkgmgr_error(method_context, error);
		return FALSE;
	}
	return TRUE;
}

gboolean pkgmgr_fetch_packages(PkgMgr *obj,
		const gchar *category,
		DBusGMethodInvocation *method_context)
{
	DEBUG("category: %s\n" ,category);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(category, method_context))
		return TRUE;
// 	gchar *sender = dbus_g_method_get_sender (method_context);
// 	DEBUG ("sender: %s", sender);
// 	g_free (sender);
	const gchar *category_real = category;
	if (g_strcmp0("", category) == 0 ) {
		category_real = NULL;
	}
	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_package_list (category_real,
						     PKG_TYPE_NOT_INSTALLED,
						     NULL,
						     PKG_SORT_NAME);
	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_packages_cb);

	return TRUE;
}

/**
 */
static void
pkgmgr_fetch_available_updates_reply (DBusGMethodInvocation *method_context,
					GList *pkgs)
{
	g_return_if_fail(method_context);

	GPtrArray *packages = __pack_package_list_to_array_map (pkgs);
	dbus_g_method_return(method_context, packages);

	__free_map_array(packages);
}

static void
pkgmgr_fetch_available_updates_list_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_package_list *cmd =
		static_cast<apt_worker_cmd_get_package_list *> (cmd_base);

	if (cmd->error) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
		pkgmgr_fetch_available_updates_reply(method_context, cmd->packages);
// 		g_list_foreach (cmd->packages, __free_package, NULL);
// 		g_list_free (cmd->packages);
	}

	delete cmd;
}

/*
static void
pkgmgr_fetch_available_updates_refresh_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_batch_cmd *batch_cmd =
		static_cast<apt_worker_batch_cmd *> (cmd_base);
	GList *cmds = batch_cmd->cmds;
	apt_worker_cmd_check_updates *u_cmd =
		static_cast<apt_worker_cmd_check_updates *> (cmds->data);
	cmds = g_list_next(cmds);
	apt_worker_cmd_get_package_list *list_cmd =
			static_cast<apt_worker_cmd_get_package_list *> (cmds->data);

	if (!u_cmd->success) {
		GError* error = NULL;
		g_set_error (&error, PKGMGRD_ERROR, ERROR_CHECK_UPDATES_FAILED, "Failed to check updates");
		pkgmgr_error (method_context, error);
	} else {
		pkgmgr_fetch_available_updates_list_cb(list_cmd);
		cmds = g_list_remove (cmds, list_cmd);
	}

	delete batch_cmd;
}
*/

gboolean pkgmgr_fetch_available_updates(PkgMgr *obj,
					const gchar *category,
					DBusGMethodInvocation *method_context)
{
	DEBUG("category: %s\n" ,category);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(category, method_context))
		return TRUE;
	const gchar *category_real = category;
	if (g_strcmp0("", category) == 0 ) {
		category_real = NULL;
	}
	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_package_list (
			category_real,
			PKG_TYPE_UPDATE,
			NULL,
			PKG_SORT_NAME);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_available_updates_list_cb);

	return TRUE;
}


/**
 */
static void
pkgmgr_fetch_installed_reply (DBusGMethodInvocation *method_context,
				GList *pkgs)
{
	g_return_if_fail(method_context);

	GPtrArray *packages = __pack_package_list_to_array_map (pkgs);
	dbus_g_method_return(method_context, packages);
	__free_map_array(packages);
}

static void
pkgmgr_fetch_installed_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_package_list *cmd =
		static_cast<apt_worker_cmd_get_package_list *> (cmd_base);

	if (cmd->error) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
		pkgmgr_fetch_installed_reply (method_context, cmd->packages);
	}

	delete cmd;
}

gboolean pkgmgr_fetch_installed(PkgMgr *obj,
				DBusGMethodInvocation *method_context)
{
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}

	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_package_list(
			NULL,
			PKG_TYPE_INSTALLED,
			NULL,
			PKG_SORT_NAME);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_installed_cb);

	return TRUE;
}

static void
pkgmgr_void_reply (DBusGMethodInvocation *method_context)
{
    DEBUG("");
    g_return_if_fail(method_context);
    dbus_g_method_return(method_context);
}

static void
pkgmgr_install_file_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_install_file *cmd =
		static_cast<apt_worker_cmd_install_file *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_install_file (PkgMgr *obj,
			const gchar *filename,
			DBusGMethodInvocation *method_context)
{
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}

	apt_worker_cmd *cmd =
		new apt_worker_cmd_install_file (filename);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_install_file_cb);

	return TRUE;
}

static void
pkgmgr_format_utf8_map(GHashTable *pkgmap)
{
        DEBUG("");
        GHashTableIter iter;
        gpointer key, value;
        g_hash_table_iter_init (&iter, pkgmap);
        while (g_hash_table_iter_next (&iter, &key, &value))
        {
                GValue *gvalue = (GValue *)value;
                if(G_VALUE_HOLDS_STRING(gvalue)) {
                        gchar *oldvalue = g_value_dup_string(gvalue);
                        gchar *newvalue = format_utf8(oldvalue);
                        if (g_strcmp0(oldvalue, newvalue) != 0)
                            g_hash_table_replace_string_value(pkgmap, (char *)key, newvalue);
                        else
                            g_free(newvalue);
                        g_free(oldvalue);
                }
        }
}

static void
pkgmgr_fetch_package_data_file_reply (DBusGMethodInvocation *method_context,
				      GHashTable *pkgmap)
{
	DEBUG("");
	g_return_if_fail (method_context);
        pkgmgr_format_utf8_map(pkgmap);
	dbus_g_method_return (method_context, pkgmap);
}

static void
pkgmgr_fetch_package_data_file_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_file_details_map *cmd =
		static_cast<apt_worker_cmd_get_file_details_map *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_fetch_package_data_file_reply (method_context, cmd->retval);
	}

	delete cmd;
}

gboolean pkgmgr_fetch_package_data_file (PkgMgr *obj,
			const gchar *filename,
			DBusGMethodInvocation *method_context)
{
	DEBUG("filename: %s", filename);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}

	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_file_details_map (filename);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_package_data_file_cb);

	return TRUE;
}

static void
pkgmgr_fetch_package_data_reply (DBusGMethodInvocation *method_context,
				 GHashTable *pkgmap)
{
	DEBUG("");
	g_return_if_fail(method_context);
// 	GValue entry = { 0, };
	// GHashTable* map = g_hash_table_new(g_str_hash, g_str_equal);
	// __pack_package_to_map(pkg, map);
// 	dbus_g_method_return(method_context, g_value_get_boxed(&entry));
        pkgmgr_format_utf8_map(pkgmap);
	dbus_g_method_return(method_context, pkgmap);
	// __free_map(map);
// 	package_free(pkg);
}

static void
pkgmgr_fetch_package_data_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_package_details_map *cmd =
		static_cast<apt_worker_cmd_get_package_details_map *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_fetch_package_data_reply (method_context, cmd->retval);
	}

	delete cmd;
}

gboolean pkgmgr_fetch_package_data (PkgMgr *obj,
				const gchar *packagename,
				const gchar *version,
				DBusGMethodInvocation *method_context)
{
	DEBUG("package: %s (%s)", packagename, version);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_package_details_map (
			packagename,
			version);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_package_data_cb);

	return TRUE;
}

gboolean pkgmgr_fetch_package_data_installed (PkgMgr *obj,
					const gchar *packagename,
					DBusGMethodInvocation *method_context)
{
	DEBUG("package: %s", packagename);
	if (!check_parameter_str(packagename, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_package_details_map (
			packagename,
			TRUE);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_package_data_cb);

	return TRUE;
}

static void
pkgmgr_fetch_package_dependency_data_reply (DBusGMethodInvocation *method_context,
					    GList *pkgs)
{
	DEBUG("");

	GPtrArray *packages = g_ptr_array_new ();

	for (GList *i = pkgs; i != NULL; i = g_list_next (i))
		g_ptr_array_add (packages, i->data);

	g_return_if_fail (method_context);
	dbus_g_method_return (method_context, packages);
	g_ptr_array_free (packages, TRUE);
}

static void
pkgmgr_fetch_package_dependency_data_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_package_dep_details *cmd =
		static_cast<apt_worker_cmd_get_package_dep_details *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_fetch_package_dependency_data_reply (method_context, cmd->retval);
	}

	delete cmd;
}

gboolean pkgmgr_fetch_package_dependency_data (PkgMgr *obj,
					       const gchar *packagename,
					       const gchar *version,
					       DBusGMethodInvocation *method_context)
{
	DEBUG("package: %s", packagename);
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_package_dep_details (
			packagename,
			version);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_package_dependency_data_cb);

	return TRUE;
}

static void
pkgmgr_fetch_restore_status_reply (DBusGMethodInvocation *method_context,
					    gchar *data)
{
	DEBUG(" Status: %s", data);

	g_return_if_fail (method_context);
	dbus_g_method_return (method_context, data);
	DEBUG(" out");
}

static void
pkgmgr_fetch_restore_status_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_restore_status *cmd =
		static_cast<apt_worker_cmd_get_restore_status *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_fetch_restore_status_reply (method_context, cmd->retval);
	}

	delete cmd;
}

gboolean pkgmgr_fetch_restore_status (PkgMgr *obj,
					       DBusGMethodInvocation *method_context)
{
	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_restore_status ();

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_restore_status_cb);

	return TRUE;
}

static void
pkgmgr_fetch_restorable_packages_reply (DBusGMethodInvocation *method_context,
					    GList *data)
{
    g_return_if_fail(method_context);

    GPtrArray *packages = __pack_package_list_to_array_map (data);
    dbus_g_method_return(method_context, packages);
    __free_map_array(packages);
}

static void
pkgmgr_fetch_restorable_packages_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_restorable_packages *cmd =
		static_cast<apt_worker_cmd_get_restorable_packages *> (cmd_base);

	if ( cmd->error ) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_fetch_restorable_packages_reply (method_context, cmd->retval);
	}

	delete cmd;
}

gboolean pkgmgr_fetch_restorable_packages (PkgMgr *obj,
					       DBusGMethodInvocation *method_context)
{
	DEBUG("");

	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_restorable_packages ();

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_restorable_packages_cb);

	return TRUE;
}

static void
pkgmgr_string_variant_map_reply (DBusGMethodInvocation *method_context,
				 GHashTable *map)
{
	DEBUG("");
	g_return_if_fail (method_context);
        pkgmgr_format_utf8_map(map);
	dbus_g_method_return (method_context, map);
}

static void
pkgmgr_simulate_operation_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_simulate_operation *cmd =
		static_cast<apt_worker_cmd_simulate_operation *> (cmd_base);

	if ( cmd->error ) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_string_variant_map_reply (method_context, cmd->retval);
	}

	delete cmd;
}

gboolean pkgmgr_simulate_operation(PkgMgr *obj,
			const gchar *operation,
			const gchar *packageidentifier,
			const gchar *version,
			const gchar *flags,
			DBusGMethodInvocation *method_context)
{
    DEBUG(" Operation: %s, pkg: %s (%s), flags: %s", operation, packageidentifier, version, flags);
	if (!check_parameter_str(operation, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;
	if (!check_parameter_str(flags, method_context))
		return TRUE;
    apt_worker_cmd *cmd =
		new apt_worker_cmd_simulate_operation (operation, packageidentifier, version, flags);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_simulate_operation_cb);

	return TRUE;
}

static void
pkgmgr_restore_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_restore_package *cmd =
		static_cast<apt_worker_cmd_restore_package *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_restore (PkgMgr *obj,
			const gchar *packagename,
			const gchar *version,
			DBusGMethodInvocation *method_context)
{
    DEBUG("package %s (%s)", packagename, version);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_restore_package (
			packagename,
			version);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_restore_cb);

	return TRUE;
}

static void
pkgmgr_ignore_restore_package_reply (DBusGMethodInvocation *method_context)
{
	DEBUG("");
	g_return_if_fail(method_context);
	dbus_g_method_return(method_context);
}

static void
pkgmgr_ignore_restore_package_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_ignore_restore_package *cmd =
		static_cast<apt_worker_cmd_ignore_restore_package *> (cmd_base);

/*	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {*/
 		pkgmgr_ignore_restore_package_reply (method_context);
//	}

	delete cmd;
}

gboolean pkgmgr_ignore_restore_package (PkgMgr *obj,
				const gchar *packagename,
				DBusGMethodInvocation *method_context)
{
    DEBUG("package: %s", packagename);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_ignore_restore_package (
			packagename);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_ignore_restore_package_cb);

	return TRUE;
}

static void
pkgmgr_uninstall_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_uninstall_package *cmd =
		static_cast<apt_worker_cmd_uninstall_package *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_uninstall (PkgMgr *obj,
			const gchar *packagename,
			DBusGMethodInvocation *method_context)
{
	DEBUG("packagename: %s", packagename);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_uninstall_package (packagename);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_uninstall_cb);

	return TRUE;
}

static void
pkgmgr_ignore_update_package_reply (DBusGMethodInvocation *method_context)
{
	DEBUG("");
	g_return_if_fail(method_context);
// 	GValue entry = { 0, };
	// GHashTable* map = g_hash_table_new(g_str_hash, g_str_equal);
	// __pack_package_to_map(pkg, map);
// 	dbus_g_method_return(method_context, g_value_get_boxed(&entry));
	dbus_g_method_return(method_context);
	// __free_map(map);
// 	package_free(pkg);
}

static void
pkgmgr_ignore_update_package_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_ignore_update_package *cmd =
		static_cast<apt_worker_cmd_ignore_update_package *> (cmd_base);

/*	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {*/
 		pkgmgr_ignore_update_package_reply (method_context);
//	}

	delete cmd;
}

gboolean pkgmgr_ignore_update_package (PkgMgr *obj,
				const gchar *packagename,
				const gchar *version,
				DBusGMethodInvocation *method_context)
{
	DEBUG("package: %s (%s)", packagename, version);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_ignore_update_package (
			packagename,
			version);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_ignore_update_package_cb);

	return TRUE;
}


static void
pkgmgr_refresh_cache_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_refresh_cache *cmd =
		static_cast<apt_worker_cmd_refresh_cache *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_refresh_cache (PkgMgr *obj,
			gboolean clearignored,
			DBusGMethodInvocation *method_context)
{
	DEBUG("clearignored: %s", clearignored?"true":"false");
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}

	apt_worker_cmd *cmd =
		new apt_worker_cmd_refresh_cache( NULL, clearignored?"clearignored":"" );

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_refresh_cache_cb);

	return TRUE;
}

gboolean pkgmgr_refresh_package_cache (PkgMgr *obj,
                        const gchar* only_domain,
			const gchar* flags,
			DBusGMethodInvocation *method_context)
{
	DEBUG("flags: %s", flags);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}

	apt_worker_cmd *cmd =
		new apt_worker_cmd_refresh_cache( only_domain, flags );

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_refresh_cache_cb);

	return TRUE;
}

static void
pkgmgr_fetch_files_of_installed_package_reply (DBusGMethodInvocation *method_context, GList *files)

{
    int len =  g_list_length(files)+1;

    gchar **buffer = (gchar**) g_malloc(len*sizeof(gchar *));

    GList *l = files;
    int i = 0;
    while(l) {
	buffer[i] = (gchar*)l->data;
	i++;
	l = g_list_next(l);
    }
    buffer[i] = 0;
    
    DEBUG("pkgmgr_fetch_files_of_installed_package_reply()");
    g_return_if_fail(method_context);
    
    
    dbus_g_method_return(method_context,buffer);
    g_strfreev(buffer);
}

static void
pkgmgr_fetch_files_of_installed_package_cb (apt_worker_cmd *cmd_base)
{
	DEBUG("pkgmgr_fetch_files_of_installed_package_cb()");

        DBusGMethodInvocation * method_context =
                reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
        apt_worker_cmd_pkgmgr_fetch_files_of_installed_package *cmd =
                static_cast<apt_worker_cmd_pkgmgr_fetch_files_of_installed_package *> (cmd_base);

        if (! cmd->retval) {
                pkgmgr_error (method_context, cmd->error);
        }
        else {
	    pkgmgr_fetch_files_of_installed_package_reply (method_context, cmd->files);
        }

        delete cmd;
}

gboolean pkgmgr_fetch_files_of_installed_package(PkgMgr *obj,
                        const gchar *packagename,
                        DBusGMethodInvocation *method_context)
{
        DEBUG("packagename: %s", packagename);
        if (__service_shutdown(obj, method_context)) {
                return TRUE;
        }
	if (!check_parameter_str(packagename, method_context))
		return TRUE;

        apt_worker_cmd *cmd =
                new apt_worker_cmd_pkgmgr_fetch_files_of_installed_package (packagename);

        cmd->user_data = method_context;
        obj->apt_worker_thread->push_cmd (cmd, pkgmgr_fetch_files_of_installed_package_cb);

        return TRUE;
}

static void
pkgmgr_install_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_install_package *cmd =
		static_cast<apt_worker_cmd_install_package *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_install (PkgMgr *obj,
			const gchar *packagename,
			const gchar *version,
			DBusGMethodInvocation *method_context)
{
	DEBUG("packagename=%s, version=%s", packagename, version);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_install_package (
			packagename,
			version,
			NULL);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_install_cb);

	return TRUE;
}

static void
pkgmgr_upgrade_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_upgrade_package *cmd =
		static_cast<apt_worker_cmd_upgrade_package *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_upgrade (PkgMgr *obj,
			const gchar *packagename,
			const gchar *version,
			DBusGMethodInvocation *method_context)
{
	DEBUG("packagename: %s", packagename);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_upgrade_package (
			packagename,
			version,
			NULL);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_upgrade_cb);

	return TRUE;
}

static void
pkgmgr_upgrade_all_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_upgrade_all *cmd =
		static_cast<apt_worker_cmd_upgrade_all *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_upgrade_all (PkgMgr *obj,
			     gboolean allow_system_upgrade,
			     DBusGMethodInvocation *method_context)
{
// 	DEBUG("packagename: %s", packagename);
    if (__service_shutdown(obj, method_context)) {
	return TRUE;
    }

    apt_worker_cmd *cmd =
	new apt_worker_cmd_upgrade_all (allow_system_upgrade);

    cmd->user_data = method_context;
    obj->apt_worker_thread->push_cmd (cmd, pkgmgr_upgrade_all_cb);

    return TRUE;
}

static void
pkgmgr_download_system_upgrade_cb (apt_worker_cmd *cmd_base)
{
	DEBUG("%s\n", __PRETTY_FUNCTION__);

	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_download_system_upgrade *cmd =
		static_cast<apt_worker_cmd_download_system_upgrade *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_download_system_upgrade (PkgMgr *obj,
					 const gchar *packagename,
					 const gchar *version,
					 DBusGMethodInvocation *method_context)
{
	DEBUG("packagename: %s", packagename);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_download_system_upgrade (
			packagename,
			version,
			NULL);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_download_system_upgrade_cb);

	return TRUE;
}

static void
pkgmgr_upgrade_system_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_upgrade_system *cmd =
		static_cast<apt_worker_cmd_upgrade_system *> (cmd_base);

	if (! cmd->retval) {
		pkgmgr_error (method_context, cmd->error);
	}
	else {
 		pkgmgr_void_reply (method_context);
	}

	delete cmd;
}

gboolean pkgmgr_upgrade_system (PkgMgr *obj,
				const gchar *packagename,
				const gchar *version,
				DBusGMethodInvocation *method_context)
{
	DEBUG("packagename: %s, version: %s", packagename, version);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(packagename, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	apt_worker_cmd *cmd =
		new apt_worker_cmd_upgrade_system (
			packagename,
			version,
			NULL);

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_upgrade_system_cb);

	return TRUE;
}

static void
pkgmgr_fetch_current_operation_data_cb (apt_worker_cmd *cmd_base)
{
	DBusGMethodInvocation * method_context =
		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
	apt_worker_cmd_get_current_operation_data *cmd =
		static_cast<apt_worker_cmd_get_current_operation_data *> (cmd_base);

	pkgmgr_string_variant_map_reply (method_context, cmd->retval);

	delete cmd;
}

gboolean pkgmgr_fetch_current_operation_data (PkgMgr *obj,
					      DBusGMethodInvocation *method_context)
{
	DEBUG("");
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}

	apt_worker_cmd *cmd =
		new apt_worker_cmd_get_current_operation_data ();

	cmd->user_data = method_context;
	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_current_operation_data_cb);

	return TRUE;
}

gboolean pkgmgr_cancel (PkgMgr *obj,
			const gchar *operation,
			const gchar *package,
			const gchar *version,
			DBusGMethodInvocation *method_context)
{
	DEBUG("operation: %s, package: %s, version: %s", operation, package, version);
	if (__service_shutdown(obj, method_context)) {
		return TRUE;
	}
	if (!check_parameter_str(operation, method_context))
		return TRUE;
	if (!check_parameter_str(version, method_context))
		return TRUE;

	GError *error = NULL;
	cmd_cancel (operation, package, version, &error);
	if (error)
		pkgmgr_error (method_context, error);
	else
 		pkgmgr_void_reply (method_context);

	return TRUE;
}

/**
 */
// static void
// pkgmgr_fetch_catalogues_reply (DBusGMethodInvocation *method_context,
// 				GList *cats)
// {
// 	DEBUG("");
// 	g_return_if_fail(method_context);

// 	GPtrArray *catalogues = __pack_catalogue_list_to_array_map (cats);
// 	dbus_g_method_return(method_context, catalogues);
// 	__free_map_array(catalogues);
// }

// static void
// pkgmgr_fetch_catalogues_cb (apt_worker_cmd *cmd_base)
// {
// 	DBusGMethodInvocation * method_context =
// 		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
// 	apt_worker_cmd_get_catalogues *cmd =
// 		static_cast<apt_worker_cmd_get_catalogues *> (cmd_base);

// 	if (cmd->error) {
// 		pkgmgr_error (method_context, cmd->error);
// 	}
// 	else {
// 		pkgmgr_fetch_catalogues_reply (method_context, cmd->catalogues);
// 	}

// 	delete cmd;
// }

// gboolean pkgmgr_fetch_catalogues(PkgMgr *obj,
// 				DBusGMethodInvocation *method_context)
// {
// 	DEBUG("");
// 	if (__service_shutdown(obj, method_context)) {
// 		return TRUE;
// 	}

// 	apt_worker_cmd *cmd =
// 		new apt_worker_cmd_get_catalogues ();

// 	cmd->user_data = method_context;
// 	obj->apt_worker_thread->push_query (cmd, pkgmgr_fetch_catalogues_cb);

// 	return TRUE;
// }

// static void
// pkgmgr_add_catalogue_reply (DBusGMethodInvocation *method_context)
// {
// 	DEBUG("");
// 	g_return_if_fail(method_context);
// 	dbus_g_method_return(method_context);
// }

// static void
// pkgmgr_add_catalogue_cb (apt_worker_cmd *cmd_base)
// {
// 	DBusGMethodInvocation * method_context =
// 		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
// 	apt_worker_cmd_add_catalogue *cmd =
// 		static_cast<apt_worker_cmd_add_catalogue *> (cmd_base);

// 	if (cmd->error) {
// 		pkgmgr_error (method_context, cmd->error);
// 	}
// 	else {
// 		pkgmgr_add_catalogue_reply (method_context);
// 	}

// 	delete cmd;
// }

// gboolean pkgmgr_add_catalogue(PkgMgr *obj,
// 			      const gchar *name,
// 			      const gchar *uri,
// 			      const gchar *distribution,
// 			      const gchar *components,
// 			      DBusGMethodInvocation *method_context)
// {
// 	DEBUG("");
// 	if (__service_shutdown(obj, method_context)) {
// 		return TRUE;
// 	}

// 	apt_worker_cmd *cmd =
// 		new apt_worker_cmd_add_catalogue (
// 			name,
// 			uri,
// 			distribution,
// 			components);

// 	cmd->user_data = method_context;
// 	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_add_catalogue_cb);

// 	return TRUE;
// }

// static void
// pkgmgr_remove_catalogue_reply (DBusGMethodInvocation *method_context)
// {
// 	DEBUG("");
// 	g_return_if_fail(method_context);
// 	dbus_g_method_return(method_context);
// }

// static void
// pkgmgr_remove_catalogue_cb (apt_worker_cmd *cmd_base)
// {
// 	DBusGMethodInvocation * method_context =
// 		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
// 	apt_worker_cmd_remove_catalogue *cmd =
// 		static_cast<apt_worker_cmd_remove_catalogue *> (cmd_base);

// 	if (cmd->error) {
// 		pkgmgr_error (method_context, cmd->error);
// 	}
// 	else {
// 		pkgmgr_remove_catalogue_reply (method_context);
// 	}

// 	delete cmd;
// }

// gboolean pkgmgr_remove_catalogue (PkgMgr *obj,
// 				  const gchar *name,
// 				  DBusGMethodInvocation *method_context)
// {
// 	DEBUG("");
// 	if (__service_shutdown(obj, method_context)) {
// 		return TRUE;
// 	}

// 	apt_worker_cmd *cmd =
// 		new apt_worker_cmd_remove_catalogue (name);

// 	cmd->user_data = method_context;
// 	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_remove_catalogue_cb);

// 	return TRUE;
// }

// static void
// pkgmgr_enable_catalogue_reply (DBusGMethodInvocation *method_context)
// {
// 	DEBUG("");
// 	g_return_if_fail(method_context);
// 	dbus_g_method_return(method_context);
// }

// static void
// pkgmgr_enable_catalogue_cb (apt_worker_cmd *cmd_base)
// {
// 	DBusGMethodInvocation * method_context =
// 		reinterpret_cast<DBusGMethodInvocation *> (cmd_base->user_data);
// 	apt_worker_cmd_enable_catalogue *cmd =
// 		static_cast<apt_worker_cmd_enable_catalogue *> (cmd_base);

// 	if (cmd->error) {
// 		pkgmgr_error (method_context, cmd->error);
// 	}
// 	else {
// 		pkgmgr_enable_catalogue_reply (method_context);
// 	}

// 	delete cmd;
// }

// gboolean pkgmgr_enable_catalogue (PkgMgr *obj,
// 				  const gchar *name,
// 				  gboolean enable,
// 				  DBusGMethodInvocation *method_context)
// {
// 	DEBUG("");
// 	if (__service_shutdown(obj, method_context)) {
// 		return TRUE;
// 	}

// 	apt_worker_cmd *cmd =
// 		new apt_worker_cmd_enable_catalogue (name, enable);

// 	cmd->user_data = method_context;
// 	obj->apt_worker_thread->push_cmd (cmd, pkgmgr_enable_catalogue_cb);

// 	return TRUE;
// }

PkgMgr *pkgmgr_new ()
{
	PkgMgr *obj = static_cast<PkgMgr*> (g_object_new (PKGMGR_TYPE, NULL));
	return obj;
}

void pkgmgr_start(PkgMgr *obj, GSourceFunc quit_fn, gpointer quit_data, gboolean sticky)
{
	AptWorker *apt_worker = apt_worker_new ();
	obj->apt_worker_thread = new AptWorkerThread (
		apt_worker,
		quit_fn,
		quit_data,
		&__download_progress_signal,
		&__operation_started_signal,
		&__operation_progress_signal,
		&__operation_complete_signal,
		&__package_list_updated_signal,
		&__package_ignored_signal,
		obj,
		sticky);
	obj->apt_worker_thread->start ();
}

void pkgmgr_stop(PkgMgr *obj)
{
	obj->apt_worker_thread->stop ();
}

void pkgmgr_unref (PkgMgr *obj)
{
	DEBUG("");
	g_object_unref(obj);
}
