/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

// -*- c-file-style: "gnu" -*-
/*!
  @file pm-query.cc
  @brief pkgmgrd query helper
 */

#include "apt-worker.h"
#include "apt-worker-cache.h"
#include "apt-worker-internal.h"
#include "confutils.h"
#include "error-util.h"
#include "log.h"
#include "package-list.h"
#include "package-map.h"
#include <apt-pkg/configuration.h>
#include <glib-object.h>
// #include <glib.h>

int verbosity = 0;

static void
print_usage ()
{
  const char *usage =
    "usage: pm-query COMMAND [OPTION...]\n"
    "\n"
    "Commands:\n"
    "  list-installed\n"
    "  list-updates [CATEGORY]\n"
    "  list-available [CATEGORY]\n"
    "  list-restorable\n"
    "  restore-status\n"
    "  show PACKAGE [VERSION] [DOMAIN]\n"
    "  show-installed PACKAGE\n"
    "  show-deps PACKAGE [VERSION]\n"
    "  show-file FILENAME\n"
    "\n";

  fprintf (stderr, usage);
}

static void
log_handler (const gchar *log_domain,
	     GLogLevelFlags log_level,
	     const gchar *message,
	     gpointer unused_data)
{
}

static void
init ()
{
  g_type_init ();

  g_log_set_handler (NULL,
		     GLogLevelFlags (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL
				     | G_LOG_LEVEL_WARNING |
				     G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO |
				     G_LOG_LEVEL_DEBUG | G_LOG_FLAG_FATAL |
				     G_LOG_FLAG_RECURSION), log_handler,
		     NULL);

  load_system_settings ();
  read_domain_conf ();
  AptWorkerCache::Initialize ();

  _config->Set ("Dir::Cache::pkgcache",
		_config->Find ("Dir::Cache::pkgcache-read"));
  _config->Set ("Dir::State::status",
		_config->Find ("Dir::State::status-read"));
  _config->Set ("Dir::Aegis::Restok::conf",
		_config->Find ("Dir::Aegis::Restok::conf-read"));

  cache_init (false, false);
}

static int
pm_query_list (int argc, char *argv[])
{
  const char *cmd = argv[0];
  const char *category = argv[1];
  GList *P = 0;
  GError *error = 0;
  PkgType type;

  if (!strcmp (cmd, "list-installed"))
    type = PKG_TYPE_INSTALLED;
  else if (!strcmp (cmd, "list-updates"))
    type = PKG_TYPE_UPDATE;
  else if (!strcmp (cmd, "list-available"))
    type = PKG_TYPE_NOT_INSTALLED;
  else
    {
      fprintf (stderr, "Invalid command: %s\n", cmd);
      return 2;
    }

  cmd_get_package_list (category, type, 0, &P, PKG_SORT_NAME,
			&error);

  g_error_fwrite (error, stdout);
  if (!error)
    package_list_fwrite (P, stdout);

  return 0;
}

static int
pm_query_restore_status (int argc, char *argv[])
{
  GError *error = 0;
  gchar *status;

  status = cmd_get_restore_status (&error);

  g_error_fwrite (error, stdout);
  if (!error)
    puts (status);

  return 0;
}

static int
pm_query_list_restorable (int argc, char *argv[])
{
  GList *P = 0;
  GError *error = 0;

  cmd_get_restorable_packages (&P, &error);

  g_error_fwrite (error, stdout);
  if (!error)
    package_list_fwrite (P, stdout);

  return 0;
}

static int
pm_query_show (int argc, char *argv[])
{
  const char *cmd = argv[0];
  const char *package = 0, *version = 0, *domain = 0;
  bool installed = false;
  GHashTable *p = 0;
  GError *error = 0;

  if (argc == 1)
    {
      print_usage ();
      exit (2);
    }

  if (argc > 1)
    package = argv[1];
  if (argc > 2)
    version = argv[2];
  if (argc > 3)
    domain = argv[3];

  if (!strcmp (cmd, "show-installed"))
    installed = true;

  p = cmd_get_package_details_map (package, version, domain, installed, &error);

  g_error_fwrite (error, stdout);
  if (!error)
    package_hash_table_fwrite (p, stdout);

  return 0;
}

static int
pm_query_show_deps (int argc, char *argv[])
{
  const char *package = 0, *version = 0;
  GList *P = 0;
  GError *error = 0;

  if (argc == 1)
    {
      print_usage ();
      exit (2);
    }

  if (argc > 1)
    package = argv[1];
  if (argc > 2)
    version = argv[2];

  P = cmd_get_package_dep_details (package, version, &error);

  g_error_fwrite (error, stdout);
  if (!error)
    package_hash_table_list_fwrite (P, stdout);

  return 0;
}

static int
pm_query_show_file (int argc, char *argv[])
{
  const char *filename = 0;
  GHashTable *p = 0;
  GError *error = 0;

  if (argc != 2)
    {
      print_usage ();
      exit (2);
    }

  filename = argv[1];

  p = cmd_get_file_details_map (filename, &error);

  g_error_fwrite (error, stdout);
  if (!error)
    package_hash_table_fwrite (p, stdout);

  return 0;
}

struct Command
{
  const char *name;
  int (*func) (int, char **);
  int flags;
};

const static struct Command commands[] = {
  {"list-available", pm_query_list},
  {"list-installed", pm_query_list},
  {"list-restorable", pm_query_list_restorable},
  {"list-updates", pm_query_list},
  {"restore-status", pm_query_restore_status},
  {"show", pm_query_show},
  {"show-deps", pm_query_show_deps},
  {"show-file", pm_query_show_file},
  {"show-installed", pm_query_show},
  {0}
};

static void
handle_internal_command (int argc, char *argv[])
{
  const char *const cmd = argv[0];

  for (const Command * p = commands; p->name; ++p)
    {
      if (strcmp (cmd, p->name))
	continue;

      init ();
      exit (p->func (argc, argv));
    }

  fprintf (stderr, "No internal command: %s\n", cmd);
  exit(2);
}

int main (int argc, char *argv[])
{
  if (argc == 1)
    {
      print_usage ();
      exit (2);
    }

  --argc;
  ++argv;

  handle_internal_command (argc, argv);

  return 0;
}
