#!/bin/sh
#
# pm-rescue script.
#
# If there is an unfinished operation after the boot, this script is
# executed.  It enables an operative environment for the rescue
# command in pkgmgrd.

LOGDIR1=/home/user/.package-manager/rescue
LOGDIR2=/var/log/rescue
LOGBASE=pm-rescue
REBOOTCOUNTFILE=/var/lib/package-manager/rescue-reboot-count
CURRENTOPERATIONFILE=/var/lib/package-manager/current-operation

if [ ! -s $CURRENTOPERATIONFILE ];
then
    exit 0
fi

read_conf_num() {
    n=$(grep "^$1[ \t]*=" /etc/package-manager/config | sed "s/^$1[ \t]*=[ \t]*\([0-9]*\).*$/\1/")
    if [ -z "$n" ]; then
	n=0
    fi
    echo $n
}

# 0 == infinite
MAXREBOOTCOUNT=$(read_conf_num maxRescueAttempts)

log() {
    echo "rescue: $@"
}

fbtext() {
    row=$1
    y=`expr 400 + $row \* 20`
    shift
    /usr/bin/text2screen -s 2 -x 5 -y $y -B 0 -T 0xF000 -t "$@   "
    echo "$@" >/dev/stderr
}

init_udev() {
    udevd --daemon
    udevadm trigger
    udevadm settle
}

# Copied from /etc/init/timeconfig.conf
init_hwclock() {
    TIMED_DATA_DIR="/var/cache/timed"
    LOST_FILE="$TIMED_DATA_DIR/SYSTEM_TIME_LOST"
    THRESHOLD="2010-06-30 23:59:59" # up to June 2010
    FALLBACK="2010-07-01 12:00:00" # July 2010

    # RTC to system time
    hwclock -s -u

    # system time value and threshold
    syst_value=$(date -u +%s)
    thrs_value=$(date +%s -u -d "$THRESHOLD")

    # if the system time is too old, it's invalid and LOST_FILE is created
    # if the LOST_FILE is not removed by timed, system time is still invalid

    if [ "$syst_value" -le "$thrs_value" ] || [ -f "$LOST_FILE" ] ; then
	echo "System time not valid !"
	echo    " Using fall back value : $FALLBACK"
	date -u -s "$FALLBACK" # Required by SUW
	hwclock -u -w
	# The timed directory is needed (for LOST_FILE)
	test -d "$TIMED_DATA_DIR" || mkdir -m 755 -p "$TIMED_DATA_DIR"
	echo "$syst_value" >> "$LOST_FILE"
    fi
}

status() {
    echo -n "rescue: $@"
}

try() {
    if $@; then
	echo "Done"
    else
	echo "ERROR"
    fi
}

# Copied from /etc/init/aegis.conf:
init_aegis() {
    # Mount securityfs if it is not already mounted
    if [ -d /sys/kernel/security ]; then
        if ! grep -q securityfs /proc/mounts; then
	    status "Mounting securityfs... "
            try mount -t securityfs securityfs /sys/kernel/security
        fi
    fi

    # Load Aegis runtime policy 
    if [ -d /sys/kernel/security/credp -a -x /usr/bin/aegis-loader ]; then
	status "Loading aegis... "
        try /usr/bin/aegis-loader
    fi
    if [ -x /usr/bin/libcreds-audit ]; then
	status "libcreds-audit... "
        try /usr/bin/libcreds-audit
    fi

    if [ -e /etc/enable_security ]; then
	status "aegis-session --clear... "
	try /usr/bin/aegis-session --clear
    fi
    status "aegis-session --set TMPDIR=/var/tmp"
    try /usr/bin/aegis-session --set TMPDIR=/var/tmp
}

init_dsme() {
    # Source necessary environment variables
    #source /etc/osso-af-init/af-defines.sh

    # Force USER state in DSME (ACT_DEAD isn't good for rescuing...)
    export BOOTSTATE="USER"
    touch /tmp/$BOOTSTATE
    echo $BOOTSTATE > /tmp/STATE
    source /etc/resource_limits.conf

    export DSME_INIT=1

    SYSINFOCLIENT=/usr/bin/sysinfoclient
    [ -x $SYSINFOCLIENT ] \
        && $SYSINFOCLIENT -p /device/rd-mode | grep -q enabled \
        && export DSME_RD_FLAGS=`$SYSINFOCLIENT -p /device/rd-flags`

    # Start DSME to avoid device hanging/restarting...
    /sbin/dsme -p /lib/dsme/libstartup.so &
    # wait for dsme (like in dsme upstart script)
    /sbin/waitfordsme
}

init_network() {
    ifup lo
}

init_dbus() {
    echo "Starting dbus"

    mkdir -p /var/run/dbus
    chown messagebus:messagebus /var/run/dbus
    rm -f /var/run/dbus/pid
    
    eval `/usr/bin/aegis-session --sh-syntax --with-ulimit`
    /usr/bin/dbus-daemon --system --fork
}

init_rescue() {
    init_hwclock
    init_aegis
    init_dbus
    init_dsme
    init_network
}

rotate_logs() {
    # Rotate old logs
    for i in `seq 4 -1 1`; do
	j=`expr $i + 1`
	if [ -f $LOGBASE.$i.log ]; then
	    rm -f $LOGBASE.$j.log
	    mv $LOGBASE.$i.log $LOGBASE.$j.log
	fi
    done

    if [ -f $LOG ]; then
	mv $LOG $LOGBASE.1.log
    fi
}

sleep_forever() {
    while [ 1 ]; do
	/bin/sleep 60
    done
}

reboot_and_sleep_forever() {
    local disableReboot=$(read_conf_num disableReboot)
    if [ $disableReboot -eq 0 ]; then
	/sbin/reboot
	sleep_forever
    else
	pkill dsme
    fi
}

enter_malf() {
    test -f $CURRENTOPERATIONFILE && rm $CURRENTOPERATIONFILE
    test -f $REBOOTCOUNTFILE && rm $REBOOTCOUNTFILE
    sync

    echo 1 >/var/lib/package-manager/malf
    /usr/sbin/enter_malf SECURITY rescue OS update rescue failed too many times
    sleep_forever
}

check_reboot_count() {
    if [ "$MAXREBOOTCOUNT" -eq 0 ]; then
	return
    fi

    REBOOTCOUNT=0
    if [ -s $REBOOTCOUNTFILE ]; then
	read REBOOTCOUNT <$REBOOTCOUNTFILE
	echo `expr $REBOOTCOUNT + 1` >$REBOOTCOUNTFILE
    else
	echo 1 >$REBOOTCOUNTFILE
    fi

    if [ "$REBOOTCOUNT" -ge "$MAXREBOOTCOUNT" ]; then
	fbtext 0 "Maximum rescue count exceeded."

	enter_malf
    fi
}

init_log() {
    mount /home
    mount /home/user/MyDocs

    if [ -d /home/user ]; then
	LOGDIR=$LOGDIR1
	mkdir -p $LOGDIR
	chown -R user:users /home/user/.package-manager
    else
	LOGDIR=$LOGDIR2
	mkdir -p $LOGDIR
    fi

    LOG=$LOGDIR/$LOGBASE.log

    log "Log will be written to $LOG"

    rotate_logs
    touch $LOG
}

rescue() {
    # Log mounts
    mount

    # Log trustdb.gpg size, for if it's 0, dpkg will reject all packages
    ls -l /root/.gnupg/trustdb.gpg

    f=/var/lib/aegis/restok/restok.conf.lock
    if [ -f $f ]; then
	log "$f exists, removing"
	rm $f
    fi

    log "Setting up..."
    init_rescue

    log "Rescuing..."
    /usr/bin/pkgmgrd -l /dev/stdout rescue

    reboot_and_sleep_forever
}

log "Running $0..."
log "Initializing udev..."
init_udev
check_reboot_count
log "Initializing log..."
init_log

rescue 2>&1 | tee -a $LOG
