/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#ifndef PROBLEMRESOLVER_H
#define PROBLEMRESOLVER_H

#include <apt-pkg/acquire.h>
#include <apt-pkg/packagemanager.h>
#include "markpolicydepcache.h"

/*!
 * Package dependency problem solver.
 *
 * This class uses the MarkPolicyDepCache to be able to stop problem
 * resolving early if MarkPolicy::IsInstallOk() or
 * MarkPolicy::IsDeleteOk() fails.
 *
 * This is class is a copy of libapt-pkg pkgProblemResolver class.  It
 * replaces some calls of pkgDepCache::MarkInstall() and
 * pkgDepCache::MarkDelete() to MarkPolicyDepCache::MarkInstall1() and
 * MarkPolicyDepCache::MarkInstall2().
 */
class ProblemResolver
{
   MarkPolicyDepCache &Cache;
   typedef pkgCache::PkgIterator PkgIterator;
   typedef pkgCache::VerIterator VerIterator;
   typedef pkgCache::DepIterator DepIterator;
   typedef pkgCache::PrvIterator PrvIterator;
   typedef pkgCache::Version Version;
   typedef pkgCache::Package Package;
   
   enum Flags {Protected = (1 << 0), PreInstalled = (1 << 1),
               Upgradable = (1 << 2), ReInstateTried = (1 << 3),
               ToRemove = (1 << 4), ProtectedByKeep = (1 << 5)};
   signed short *Scores;
   unsigned char *Flags;
   bool Debug;
   
   // Sort stuff
   static ProblemResolver *This;
   static int ScoreSort(const void *a,const void *b);

   struct PackageKill
   {
      PkgIterator Pkg;
      DepIterator Dep;
   };

   void MakeScores();
   bool DoUpgrade(pkgCache::PkgIterator Pkg);
   
   public:
   
   inline void Protect(pkgCache::PkgIterator Pkg) {Flags[Pkg->ID] |= Protected;};
   inline void Remove(pkgCache::PkgIterator Pkg) {Flags[Pkg->ID] |= ToRemove;};
   inline void Clear(pkgCache::PkgIterator Pkg) {Flags[Pkg->ID] &= ~(Protected | ToRemove);};

   /*!
    * Try to intelligently resolve problems by installing and removing
    * packages.
    *
    * If StopOnError == true, return early when MarkInstall1() or
    * MarkDelete1() returns false.  This allows for faster detection
    * of errors if the caller is only interested on whether
    * pkgDepCache::IsInstallOk() or pkgDepCache::IsDeleteOk() failed.
    */
   bool Resolve(bool BrokenFix = false, bool StopOnError = false);
   
   // Try to resolve problems only by using keep
   bool ResolveByKeep();

   // Install all protected packages   
   void InstallProtect();   
   
   ProblemResolver(MarkPolicyDepCache *Cache);
   ~ProblemResolver();
};

#endif
