/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#ifndef PACKAGEMANAGERPENDINGCALLWATCHER_H
#define PACKAGEMANAGERPENDINGCALLWATCHER_H

#include <QString>
#include <QtDBus/QtDBus>
#include "packagemanager.h"

class PackageManager;

/*!
  \class PackageManagerPendingCallWatcher
  \brief Wrapper class for D-Bus transactions to Package Manager interfaces

  PackageManagerPendingCallWatcher pointer is returned when a D-Bus method is called 
  through PackageManger class. The calling side should connect to the dbusError signal
  provided by this class.
*/
class PackageManagerPendingCallWatcher : public QObject
{
    Q_OBJECT

public:
    /*!
        \brief Constructs Package Manager D-Bus transaction wrapper
        \param packageidentifier Package name or file name with path.
        \param parent Owning component.
    */
    PackageManagerPendingCallWatcher( const QString& packageidentifier, QObject * parent = 0 );

    /*!
        \brief Destructs the object
    */
    ~PackageManagerPendingCallWatcher();

    /*!
        \brief Returns package identifier or an empty string
    */
    QString packageidentifier();

    /*!
        \brief Returns an error code for the call
    */
    PackageManager::ErrorCode errorCode();

    /*!
        \brief Returns a D-Bus error name or an empty string
    */
    QString errorName();

    /*!
        \brief Returns a D-Bus error message or an empty string
    */
    QString errorMessage();

    /*!
        \brief Returns true if the call is finished
    */
    bool isFinished();

public Q_SLOTS: // Move to private class
    void callFinished( QDBusPendingCallWatcher* self ); 

protected:
    QString _packageidentifier;
    QString _errorName;
    QString _errorMessage;
    QDBusError::ErrorType _errorType;
    bool isError;
    bool finished;

signals:
    void callFinished( PackageManagerPendingCallWatcher* watcher );
    void dbusError( PackageManagerPendingCallWatcher* watcher );
};


/*!
  \class PackageManagerPendingFetchWatcher
  \brief Wrapper class for data fetches of Package Manager interface

  Special case of PackageManagerPendingCallWatcher used when data is fetched. In addition
  to methods and error signal provided by the base class, this class also provides 
  signals where the data is carried.
  PackageManagerPendingFetchWatcher pointer is returned when a D-Bus method is called 
  through PackageManager class. The calling side should connect to the dbusError signal
  provided by this class.
*/
class PackageManagerPendingFetchWatcher : public PackageManagerPendingCallWatcher
{
    Q_OBJECT
public:
    /*!
        \brief Constructs the object
        \param packageidentifier Package name or file name with path.
        \param parent Owning component.
    */
    PackageManagerPendingFetchWatcher( const QString& packageidentifier, QObject * parent = 0 );

    /*!
        \brief Destructs the object
    */
    ~PackageManagerPendingFetchWatcher();

public Q_SLOTS: // Move to private class
    void fetchOneFinishedSlot( QDBusPendingCallWatcher* self ); 
    void fetchManyFinishedSlot( QDBusPendingCallWatcher* self );
    void fetchStringListFinishedSlot( QDBusPendingCallWatcher* self );
    void fetchStringFinishedSlot( QDBusPendingCallWatcher* self );

signals:
    /*!
        \brief A signal that is emitted when data fetch for list of packages is finished

        Delivers a list of package meta-data. Each QMap represents meta-data for a single
        package, containing key-value pairs such as "Name" and the name identifier of the
        package. Because of the long list of packages this signal might contain, only a
        subset of meta-data is returned, including:
        - Name
        - Type
        - Version
        - Size
        - Installed size
        - Category
        - Repository
        - Display name (if available)
        - Localized display names (all available)
        - Description (if available)
        - Icon data (if available)
        - Install timestamp (for installation) or availability timestamp (for updates)
        - Error name and error message (in case of error)
        - Is broken flag (if the package installation is broken somehow)
        - System upgrade name and version (if package depends on the system upgrade.
          These are deprecated)
    */
    void dataFetched(const QList< QMap<QString, QVariant> >& data);

    /*!
        \brief This signal is emitted when data fetch for a single package is finished
        or a data of a current operation is fetched or a simulation is finished.

        Delivers a hashmap containing all known meta-data fields known to the queried
        package. Contains key-value pairs i.e. "Name" and the name identifier of the
        package or "Reboot" and the reboot flag.
    */
   void dataFetched(const QMap<QString, QVariant>& data);

    /*!
        \brief A signal that is emitted when a list of files contained by a package is
        finished

        Delivers a list of files (with a full path) contained in a single, installed
        package.
    */
    void dataFetched(const QStringList& fileList);
    
    /*!
        \brief A signal that is emitted when a status of a restore is fetched.

        Delivers a status which can be one of the following:
         - NoRestorableContent
         - NewRestorePerformed
         - RestorableContentAvailable
        
    */
    void dataFetched(const QString& data);
};

#endif
