/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "glib-util.h"
#include "serialize.h"
#include <glib-object.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

static gboolean
_free_map_value (gpointer key,
		 gpointer value,
		 gpointer user_data)
{
  gchar *k = (gchar *) key;
  GValue *v = (GValue *) value;
  g_free (k);
  g_value_unset (v);
  g_free (v);
  return TRUE;
}

void
g_hash_table_free_values (GHashTable *map)
{
  g_hash_table_foreach_steal (map, _free_map_value, NULL);
}

static void
_insert_value (GHashTable *map, const gchar *key, GValue *val)
{
  g_hash_table_insert (map, g_strdup (key), val);
}

static void
_replace_value (GHashTable *map, const gchar *key, GValue *val)
{
  g_hash_table_replace (map, g_strdup (key), val);
}

void
g_hash_table_insert_string_value (GHashTable *map, const gchar *key, const gchar *s)
{
  GValue *val;

  if (s)
    {
      val = g_new0 (GValue, 1);
      g_value_init (val, G_TYPE_STRING);
      g_value_set_string (val, s);
      _insert_value (map, key, val);
    }
}

void
g_hash_table_replace_string_value (GHashTable *map, const gchar *key, const gchar *s)
{
    GValue *oldval = g_hash_table_lookup (map, key);
    if (!s || !oldval || !G_VALUE_HOLDS_STRING (oldval))
        return;

    GValue *newval = g_new0 (GValue, 1);
    g_value_init (newval, G_TYPE_STRING);
    g_value_set_string (newval, s);
    _replace_value (map, key, newval);
    g_free (oldval);
}

void
g_hash_table_insert_int_value (GHashTable *map, const gchar *key, gint i)
{
  GValue *val = g_new0 (GValue, 1);
  g_value_init (val, G_TYPE_INT);
  g_value_set_int (val, i);
  _insert_value (map, key, val);
}

void
g_hash_table_insert_int64_value (GHashTable *map, const gchar *key, gint64 i)
{
  GValue *val = g_new0 (GValue, 1);
  g_value_init (val, G_TYPE_INT64);
  g_value_set_int64 (val, i);
  _insert_value (map, key, val);
}

void
g_hash_table_insert_boolean_value (GHashTable *map, const gchar *key, gboolean b)
{
  GValue *val = g_new0 (GValue, 1);
  g_value_init(val, G_TYPE_BOOLEAN);
  g_value_set_boolean (val, b);
  _insert_value (map, key, val);
}

const gchar *
g_hash_table_get_string_value (GHashTable *map, const gchar *key)
{
  GValue *val = g_hash_table_lookup (map, key);
  if (!val)
    return NULL;

  assert (G_VALUE_HOLDS_STRING (val));

  return g_value_get_string (val);
}

gint64
g_hash_table_get_int64_value (GHashTable *map, const gchar *key)
{
  GValue* val = g_hash_table_lookup (map, key);
  if (!val)
    return 0;

  assert (G_VALUE_HOLDS_INT64 (val));

  return g_value_get_int64 (val);
}

gint
g_hash_table_get_int_value (GHashTable *map, const gchar *key)
{
  GValue *val = g_hash_table_lookup (map, key);
  if (!val)
    return 0;

  assert (G_VALUE_HOLDS_INT (val));

  return g_value_get_int (val);
}

gboolean
g_hash_table_get_boolean_value (GHashTable *map, const gchar *key)
{
  GValue *val = g_hash_table_lookup (map, key);
  if (!val)
    return FALSE;

  assert (G_VALUE_HOLDS_BOOLEAN (val));

  return g_value_get_boolean (val);
}

static int count_chars (const char *str, char c)
{
  int n = 0;

  while (*str)
    if (*str++ == c)
      ++n;

  return n;
}

gchar **strsplitv (const gchar *sc, char sep)
{
  gchar *s = g_strdup (sc);
  int n = count_chars (s, sep) + 1;
  gchar **strv = (gchar **) g_malloc ((n + 1) * sizeof (gchar *));

  const char sepstr[2] = { sep, 0 };
  char *t = strtok (s, sepstr);
  int i = 0;
  strv[i++] = g_strdup (t);

  while ((t = strtok (0, sepstr)))
    strv[i++] = g_strdup (t);

  strv[i] = 0;

  g_free (s);

  return strv;
}
