/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "package-map.h"
#include "glib-util.h"
#include "serialize.h"
#include <assert.h>
#include <stdlib.h>
#include <string.h>

void package_hash_table_free (GHashTable *p)
{
  if (p)
    {
      g_hash_table_free_values (p);
      g_hash_table_destroy (p);
    }
}

static Field _known_fields[] = {
  { "Name", G_TYPE_STRING, 0 },
  { "Version", G_TYPE_STRING, 0 },
  { "Category", G_TYPE_STRING, 0 },
  { "DisplayName", G_TYPE_STRING, 0 },
  { "Description", G_TYPE_STRING, 0 },
  { "IconData", G_TYPE_STRING, 0 },
  { "Repository", G_TYPE_STRING, 0 },
  { "Type", G_TYPE_STRING, 0 },
  { "IsBroken", G_TYPE_BOOLEAN, 0 },
  { "Filename", G_TYPE_STRING, 0 },
  { "FlagSystemUpdate", G_TYPE_BOOLEAN, 0 },
  { "FlagIgnored", G_TYPE_BOOLEAN, 0 },
  { "FlagReboot", G_TYPE_BOOLEAN, 0 },
  { "FlagTrusted", G_TYPE_BOOLEAN, 0 },
  { "FlagLocalFile", G_TYPE_BOOLEAN, 0 },
  { "InstalledSize", G_TYPE_INT64, 0 },
  { "InstalledTimestamp", G_TYPE_INT64, 0 },
  { "AvailableTimestamp", G_TYPE_INT64, 0 },
  { "TotalInstalledSize", G_TYPE_INT64, 0 },
  { "DownloadSize", G_TYPE_INT64, 0 },
  { "Size", G_TYPE_INT64, 0 },
  { 0, 0 }
};

static int
_find_field_type (const char *key)
{
  int i;

  for (i = 0; _known_fields[i].key != NULL; ++i)
    {
      if (!strcmp (_known_fields[i].key, key))
	return _known_fields[i].type;
    }

  static const char *strv[] = {
    "InstallDeps",
    "InstallUserDeps",
    "UpgradeDeps",
    "UpgradeUserDeps",
    "RemovePkgs",
    "RemoveUserPkgs",
    "RevertDeps",
    "RevertUserDeps",
    0
  }, **k;

  for (k = strv; *k; ++k)
    if (!strcmp (key, *k))
      return G_TYPE_STRV;

  return G_TYPE_STRING;
}

static void
_set_value (GValue *val, int type, const char *value)
{
  gchar *str;

  switch (type)
  {
  case G_TYPE_INT:
    g_value_init (val, type);
    g_value_set_int (val, atoi (value));
    return;

  case G_TYPE_INT64:
    g_value_init (val, type);
    g_value_set_int64 (val, atoll (value));
    return;

  case G_TYPE_BOOLEAN:
    g_value_init (val, type);
    g_value_set_boolean (val, atoi (value) != 0 ? TRUE : FALSE);
    return;

  case G_TYPE_STRING:
    g_value_init (val, type);
    str = decode_strdup (value, strlen (value));
    g_value_set_string (val, str);
    g_free (str);
    return;

  default:
    break;
  }

  if (type == G_TYPE_STRV)
    {
      g_value_init (val, type);
      str = decode_strdup (value, strlen (value));
      g_value_take_boxed (val, strsplitv (str, ','));
      g_free (str);
    }
  else {
    assert(!"Invalid GValue type");
  }
}

gboolean
_read_field (FILE *f, char **line, size_t *len,
	     Field **fieldpp)
{
  Field *field;
  char *colon;
  const char *value;
  int n;

  if ((n = getline (line, len, f)) != -1)
    {
      (*line)[n - 1] = '\0';
      colon = strstr (*line, ": ");
      if (colon != NULL)
	{
	  *colon = '\0';
	  value = colon + 2;

	  field = g_new0 (Field, 1);
	  field->key = g_strdup (*line);
	  field->type = _find_field_type (field->key);
	  field->value = g_new0 (GValue, 1);
	  _set_value (field->value, field->type, value);
	  *fieldpp = field;

	  return TRUE;
	}
    }

  return FALSE;
}

gboolean
package_hash_table_fread (GHashTable **pp, FILE *f)
{
  char *line = NULL;
  size_t len = 0;
  GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);
  Field *field = NULL;

  while (_read_field (f, &line, &len, &field))
    {
      g_hash_table_insert (map, field->key, field->value);
      g_free (field);
      field = NULL;
    }

  free (line);

  if (g_hash_table_size (map) > 0)
    {
      *pp = map;
      return TRUE;
    }
  else
    {
      package_hash_table_free (map);
      return FALSE;
    }
}

gboolean
_write_value (FILE *f, const char *key, GValue *val)
{
  int type = _find_field_type ((const char *) key);

  fprintf (f, "%s: ", (char *) key);

  switch (type)
  {
  case G_TYPE_INT:
    return put_int (f, g_value_get_int (val));

  case G_TYPE_INT64:
    return put_int64 (f, g_value_get_int64 (val));

  case G_TYPE_BOOLEAN:
    return put_boolean (f, g_value_get_boolean (val));

  case G_TYPE_STRING:
    return put_str (f, g_value_get_string (val));

  default:
    break;
  }

  if (type == G_TYPE_STRV)
    return put_strv (f, (gchar **) g_value_get_boxed (val));

  return FALSE;
}

gboolean
package_hash_table_fwrite (GHashTable *map, FILE *f)
{
  GHashTableIter iter;
  gpointer key, value;

  g_hash_table_iter_init (&iter, map);

  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      if (!_write_value (f, (const char *) key, (GValue *) value)) {
	    return FALSE;
      }
    }

  return TRUE;
}

/* gboolean */
/* package_hash_table_ptr_array_fread (GPtrArray **pp, FILE *f) */
/* { */
/*   GPtrArray *P = g_ptr_array_new (); */
/*   GHashTable *p; */

/*   while (package_hash_table_fread (&p, f)) */
/*     g_ptr_array_add (P, p); */

/*   if (P->len > 0) */
/*     { */
/*       *pp = P; */
/*       return TRUE; */
/*     } */
/*   else */
/*     { */
/*       g_ptr_array_free (P, TRUE); */
/*       return FALSE; */
/*     } */
/* } */

/* gboolean */
/* package_hash_table_ptr_array_fwrite (GPtrArray *p, FILE *f) */
/* { */
/*   int i; */

/*   if (!p) */
/*     return FALSE; */

/*   for (i = 0; i < p->len; ++i) */
/*     package_hash_table_fwrite (g_ptr_array_index (p, i), f); */

/*   return TRUE; */
/* } */

gboolean
package_hash_table_list_fread (GList **pp, FILE *f)
{
  GList *l = NULL;
  GHashTable *p;

  while (package_hash_table_fread (&p, f))
    l = g_list_append (l, p);

  if (l)
    {
      *pp = l;
      return TRUE;
    }
  else
    {
      return FALSE;
    }
}

gboolean
package_hash_table_list_fwrite (GList *p, FILE *f)
{
  GList *l;

  if (!p)
    return FALSE;

  for (l = g_list_first (p); l != NULL; l = g_list_next (l))
    {
      package_hash_table_fwrite ((GHashTable *) l->data, f);
      fprintf (f, "\n");
    }

  return TRUE;
}
