/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "serialize.h"
#include "package-map.h"
#include "glib-util.h"
#include <cassert>
#include <cstdlib>
#include <cstring>

gchar *decode_strdup (const char *str, size_t len)
{
  const char *p = str;
  gchar *s = (gchar *) g_malloc (len + 1);
  gchar *p1 = s;

  if (s)
  {
    s[0] = '\0';

    while (*p)
    {
      const char *newl = strstr (p, "\\n");
      if (newl != NULL)
      {
	memcpy (p1, p, newl - p);
	p1 += newl - p;
	strcpy (p1, "\n");
	++p1;

	p = newl + 2;
      }
      else
      {
	strcpy (p1, p);
	break;
      }
    }
  }
  return s;
}

gboolean
get_str (FILE *f, char **line, size_t *len, gchar **p)
{
  int n;

  if ((n = getline (line, len, f)) != -1)
  {
    assert(n > 0);
    (*line)[n - 1] = '\0';
    *p = decode_strdup (*line, n - 1);

    return TRUE;
  }
  else
    return FALSE;
}

gboolean
get_strv (FILE *f, char **line, size_t *len, gchar ***v)
{
  gchar *s = 0;
  gboolean ok = get_str (f, line, len, &s);
  if (!ok)
    return FALSE;

  *v = strsplitv (s, ',');

  return TRUE;
}

gboolean
get_int (FILE *f, char **line, size_t *len, gint *p)
{
  if ((getline (line, len, f)) != -1)
  {
    *p = atoi (*line);

    return TRUE;
  }
  else
    return FALSE;
}

gboolean
get_boolean (FILE *f, char **line, size_t *len, gboolean *p)
{
  int i;
  gboolean ok = get_int (f, line, len, &i);
  *p = i != 0 ? TRUE : FALSE;
  return ok;
}

gboolean
get_int64 (FILE *f, char **line, size_t *len, gint64 *p)
{
  if ((getline (line, len, f)) != -1)
  {
    *p = atoll (*line);

    return TRUE;
  }
  else
    return FALSE;
}

gboolean
get_hash_table(FILE *f, char **line, size_t *len, GHashTable **p)
{
  gint cnt = 0;
  gboolean ok = get_int(f, line, len, &cnt);
  if ( ! ok ) {
    g_print("%s not ok\n", __PRETTY_FUNCTION__);
    return FALSE;
  }

  GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);
  Field *field = NULL;
//  g_print("%s cnt: %d\n", __PRETTY_FUNCTION__, cnt);

  for ( gint i = 0; i < cnt ; i++ ) {
    if (_read_field (f, line, len, &field)) {
      g_hash_table_insert (map, field->key, field->value);
      g_free (field);
      field = NULL;
    } else {
      package_hash_table_free(map);
      return FALSE;
    }
  }

  *p = map;
  return TRUE;
}

gboolean
put_str (FILE *f, const char *s)
{
  char *p = (char *) s;
  char *bs;

  if (s != NULL)
  {
    while (*p)
    {
      bs = strchr (p, '\n');
      if (bs != NULL)
      {
	fwrite (p, bs - p, 1, f);
	fputs ("\\n", f);
	p = bs + 1;
      }
      else
      {
	fputs (p, f);
	break;
      }
    }
  }

  fputc ('\n', f);

  return TRUE;
}

gboolean
put_strv (FILE *f, gchar **v)
{
  gchar *s = g_strjoinv (",", v);
  put_str (f, s);
  g_free (s);

  return TRUE;
}

gboolean
put_int (FILE *f, gint i)
{
  return fprintf (f, "%d\n", i) != -1;
}

gboolean
put_boolean (FILE *f, gboolean b)
{
  return put_int (f, b != FALSE ? 1 : 0);
}

gboolean
put_int64 (FILE *f, gint64 v)
{
  return fprintf (f, "%Ld\n", v) != -1;
}

gboolean
put_hash_table (FILE *f, GHashTable *v)
{
  guint cnt = 0;
  if ( v ) {
    cnt = g_hash_table_size(v);
  }
  GHashTableIter iter;
  gpointer key, value;
  gboolean ok = put_int(f, cnt);
  if ( !ok ) {
    return FALSE;
  }
  if ( cnt == 0 ) {
    return TRUE;
  }

//  g_print("%s cnt: %d\n", __PRETTY_FUNCTION__, cnt);

  g_hash_table_iter_init (&iter, v);

  while (g_hash_table_iter_next (&iter, &key, &value))
    {
      if (!_write_value (f, (const char *) key, (GValue *) value)) {
	    return FALSE;
      }
    }

  return TRUE;
}

