/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

// -*- Mode: C++; c-basic-offset: 8 -*-
#include <check.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <dbus/dbus-glib-bindings.h>
#include <glib-object.h>
#include "apt-worker-thread.h"
#include "check-common.h"

static AptWorkerThread *awt = 0;
static GMainLoop *loop = 0;
static bool command_finished = false;
static FILE *debug_file = 0;
static char *debug_buf[0x1000] = {0};
static int QUERY_SLEEP_S = 1;

static gpointer
main_loop_thread (gpointer p)
{
	g_main_loop_run ((GMainLoop *) p);

	return NULL;
}

static void
command_finished_cb (apt_worker_cmd *cmd)
{
	puts (__PRETTY_FUNCTION__);

	if (cmd->user_data == (gpointer) 1)
		command_finished = true;
}

static void
ready_queries (AptWorkerThread &t);

static int n_running_queries = 0;

static void
query_finished_cb (apt_worker_cmd *cmd)
{
	printf ("%s: n queries == %d\n", __PRETTY_FUNCTION__, n_running_queries);

	if (--n_running_queries == 0)
	{
		if (command_finished)
		{
			puts ("last command finished, stopping");

			awt->stop ();
			g_main_loop_quit (loop);
		}
		else
		{
			sleep(QUERY_SLEEP_S);
			ready_queries (*awt);
		}
	}
}

static void
ready_queries (AptWorkerThread &t)
{
	apt_worker_cmd *c;

	puts (__PRETTY_FUNCTION__);

	n_running_queries = 3;

	c = new apt_worker_cmd_get_package_list (NULL, // category
						 PKG_TYPE_NOT_INSTALLED,
						 NULL, // pattern
						 PKG_SORT_INSTALL_TIME);
	t.push_query (c, query_finished_cb);

	c = new apt_worker_cmd_get_package_list (NULL, // category
						 PKG_TYPE_INSTALLED,
						 NULL, // pattern
						 PKG_SORT_INSTALL_TIME);
	t.push_query (c, query_finished_cb);

	c = new apt_worker_cmd_get_package_list (NULL, // category
						 PKG_TYPE_UPDATE,
						 NULL, // pattern
						 PKG_SORT_INSTALL_TIME);
	t.push_query (c, query_finished_cb);
}

static void
download_progress_func (const gchar* operation, const gchar* packagename, const gchar *version, int already, int total, gpointer data)
{
	fprintf (debug_file, "%s: operation=%s, packagename=%s, version=%s, already=%d, total=%d\n",
		 __func__, operation, packagename, version, already, total);
}

static void
operation_started_func (const gchar* op, const gchar* pkg, const gchar* version, gpointer data)
{
	fprintf (debug_file, "%s: op=%s, pkg=%s, ver=%s\n",
		 __func__, op, pkg, version);
}

static void
operation_progress_func (const gchar* op, const gchar* pkg, const gchar *version, int percentage, gpointer data)
{
	fprintf (debug_file, "%s: op=%s, pkg=%s, version=%s, percentage=%d\n",
		 __func__, op, pkg, version, percentage);
}

static void
operation_complete_func (const gchar* op, gchar* pkg, const gchar *version, const gchar* error, gboolean need_reboot, gpointer data)
{
	fprintf (debug_file, "%s: op=%s, pkg=%s, version=%s, error=%s, need_reboot=%s\n",
		 __func__, op, pkg, version, error, (need_reboot?"true":"false"));
}

static void
package_list_updated_func (gboolean updates_changed, gpointer data)
{
	fprintf (debug_file, "%s: updates_changed=%s\n", __func__, (updates_changed?"true":"false"));
}

static void
package_ignored_func (const gchar* type, const gchar* package, const gchar* version, gpointer data)
{
	fprintf (debug_file, "%s: type=%s Package: %s (%s)\n", __func__, type, package, version);
}

/**
 * Test queries ran in parallel to package installation.
 */
START_TEST (test_parallel_operations)
{
	puts (__PRETTY_FUNCTION__);

	g_type_init ();
	g_thread_init (NULL);
	dbus_g_thread_init ();
	loop = g_main_loop_new (NULL, FALSE);

	AptWorkerThread t (apt_worker_new (),
			   NULL,
			   NULL,
			   download_progress_func,
			   operation_started_func,
			   operation_progress_func,
			   operation_complete_func,
			   package_list_updated_func,
			   package_ignored_func,
			   NULL);
	awt = &t;

	GThread *main_loop_thread_;
	apt_worker_cmd *c;

	main_loop_thread_ = g_thread_create (main_loop_thread,
					     loop,
					     TRUE,
					     0);

	debug_file = fmemopen (debug_buf, sizeof (debug_buf), "w");

	t.start ();

	ready_queries (t);

	for (int i = 0; i < 10; ++i)
	{
		c = new apt_worker_cmd_install_package ("hpm-nonexisting-package",
							NULL,
							NULL);
		t.push_cmd (c, command_finished_cb);
	}

// 	c = new apt_worker_cmd_install_package ("hpm-nonexisting-package",
// 						NULL);
// 	c->user_data = (gpointer) 1;
// 	t.push_cmd (c, command_finished_cb);

	c = new apt_worker_cmd_install_package ("hpm-long-install",
						NULL,
						NULL);
	t.push_cmd (c, command_finished_cb);

	c = new apt_worker_cmd_uninstall_package ("hpm-long-install");
	c->user_data = (gpointer) 1;
	t.push_cmd (c, command_finished_cb);

	g_thread_join (main_loop_thread_);
	fclose (debug_file);
}
END_TEST

Suite *
AptWorkerThread_long_test_suite ()
{
	Suite *s = suite_create ("AptWorkerThread_long");
	TCase *tc;

	tc = tcase_create ("parallel");
	tcase_set_timeout(tc, 180);
	tcase_add_test (tc, test_parallel_operations);
	suite_add_tcase (s, tc);

	return s;
}

int main (int argc, char *argv[])
{
	tc_timeout_s = 30;

	get_options (argc, argv);

	fork_mode = 0;

	return run_suite (AptWorkerThread_long_test_suite ())
		? EXIT_SUCCESS : EXIT_FAILURE;
}
