/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

// -*- Mode: C++; c-basic-offset: 8 -*-
#include <check.h>
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <glib-object.h>
#include "apt-worker-thread.h"
#include "mock-apt-worker.h"
#include "check-common.h"

static AptWorkerThread *w = 0;
static MockAptWorker *mock = 0;
static GMainLoop *loop = NULL;
static apt_worker_cmd *cmd_done = 0;

static void cmd_done_cb (apt_worker_cmd *cmd)
{
	cmd_done = cmd;
	g_main_loop_quit (loop);
}

static void setup ()
{
	cmd_done = 0;
	g_thread_init (0);
	loop = g_main_loop_new (NULL, FALSE);
	mock = reinterpret_cast<MockAptWorker *>(
		mock_apt_worker_new ());
	fail_if (mock == 0);
	w = new AptWorkerThread (&mock->base, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	fail_if (w == 0);
 	w->start ();
}

static void teardown ()
{
	delete w;
	g_main_loop_unref (loop);
}

START_TEST (test_apt_worker_cmd_get_package_list)
{
	apt_worker_cmd_get_package_list *cmd =
		new apt_worker_cmd_get_package_list ("category", // category
						     PKG_TYPE_NOT_INSTALLED,
						     "pattern", // pattern
						     PKG_SORT_INSTALL_TIME);

	mock->retval_packages = g_list_append (NULL, package_new ());
	w->push_cmd (cmd, cmd_done_cb);
	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (! strcmp (cmd->category, mock->category));
	fail_unless (cmd->type == mock->type);
	fail_unless (! strcmp (cmd->pattern, mock->pattern));
	fail_unless (cmd->packages == mock->retval_packages);
	fail_unless (! strcmp ("mock_cmd_get_package_list", cmd->error->message));

	delete cmd;
}
END_TEST

START_TEST (test_apt_worker_cmd_get_file_details)
{
	apt_worker_cmd_get_file_details *cmd =
		new apt_worker_cmd_get_file_details ("filename");

	w->push_cmd (cmd, cmd_done_cb);
	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (! strcmp (cmd->filename, mock->filename));
	fail_unless (! strcmp ("mock_cmd_get_file_details", cmd->error->message));
}
END_TEST

START_TEST (test_apt_worker_cmd_get_package_details)
{
	apt_worker_cmd_get_package_details_map *cmd =
		new apt_worker_cmd_get_package_details_map ("package",
							    "version");

	mock->retval_pkg_list_info = (Package *) 42;
	w->push_cmd (cmd, cmd_done_cb);

	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (! strcmp (cmd->package, mock->package));
	fail_unless (! strcmp (cmd->version, mock->version));
//	fail_unless (cmd->retval == mock->retval_pkg_list_info);
//	fail_unless (! strcmp ("mock_cmd_get_package_details",
//			       cmd->error->message));
}
END_TEST

START_TEST (test_apt_worker_cmd_install_file)
{
	apt_worker_cmd_install_file *cmd =
		new apt_worker_cmd_install_file ("filename");

	cmd->retval = TRUE;
	w->push_cmd (cmd, cmd_done_cb);
	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (cmd->retval == mock->retval_boolean);
	fail_unless (! strcmp ("mock_cmd_install_file",
			       cmd->error->message));
}
END_TEST

START_TEST (test_apt_worker_cmd_uninstall_package)
{
	apt_worker_cmd_uninstall_package *cmd =
		new apt_worker_cmd_uninstall_package ("package");

	cmd->retval = TRUE;
	w->push_cmd (cmd, cmd_done_cb);

	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (! strcmp (cmd->package, mock->package));
	fail_unless (cmd->retval == mock->retval_boolean);
	fail_unless (! strcmp ("mock_cmd_uninstall_package",
			       cmd->error->message));
}
END_TEST

START_TEST (test_apt_worker_cmd_install_package)
{
	apt_worker_cmd_install_package *cmd =
		new apt_worker_cmd_install_package ("package",
						    NULL,
						    "alt_download_root");

	cmd->retval = TRUE;
	w->push_cmd (cmd, cmd_done_cb);

	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (! strcmp (cmd->package, mock->package));
	fail_unless (! strcmp (cmd->alt_download_root, mock->alt_download_root));
	fail_unless (cmd->retval == mock->retval_boolean);
	fail_unless (! strcmp ("mock_cmd_install_package",
			       cmd->error->message));
}
END_TEST

START_TEST (test_apt_worker_cmd_check_updates)
{
	apt_worker_cmd_check_updates *cmd =
		new apt_worker_cmd_check_updates (TRUE);

	cmd->retval = rescode_success;
	w->push_cmd (cmd, cmd_done_cb);

	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (cmd->retval == mock->retval_result_code);
}
END_TEST

START_TEST (test_apt_worker_cmd_get_catalogues)
{
	apt_worker_cmd_get_catalogues *cmd =
		new apt_worker_cmd_get_catalogues ();

	cmd->retval = TRUE;
	mock->retval_catalogues = (GList*) 42;
	w->push_cmd (cmd, cmd_done_cb);

	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	fail_unless (cmd->retval == mock->retval_boolean);
	fail_unless (! strcmp ("mock_cmd_get_catalogues",
			       cmd->error->message));
}
END_TEST

START_TEST (test_apt_worker_batch_cmd)
{
	apt_worker_cmd_check_updates *u_cmd =
		new apt_worker_cmd_check_updates (TRUE);
	apt_worker_cmd_get_package_list *list_cmd =
		new apt_worker_cmd_get_package_list ("category", // category
						     PKG_TYPE_NOT_INSTALLED,
						     "pattern", // pattern
						     PKG_SORT_INSTALL_TIME);
	GList *cmds = NULL;
	cmds = g_list_append(cmds, u_cmd);
	cmds = g_list_append(cmds, list_cmd);
	apt_worker_batch_cmd *cmd =
		new apt_worker_batch_cmd (cmds);

	mock->retval_packages = g_list_append (NULL, package_new ());

	u_cmd->retval = rescode_success;
	w->push_cmd (cmd, cmd_done_cb);

	g_main_loop_run (loop);

	fail_unless (cmd == cmd_done);
	// check_updates
	fail_unless (u_cmd->retval == mock->retval_result_code);

	// get_package_list
	fail_unless (! strcmp (list_cmd->category, mock->category));
	fail_unless (list_cmd->type == mock->type);
	fail_unless (! strcmp (list_cmd->pattern, mock->pattern));
	fail_unless (list_cmd->packages == mock->retval_packages);
	fail_unless (! strcmp ("mock_cmd_get_package_list", list_cmd->error->message));
	delete cmd;
}
END_TEST

Suite *
AptWorkerThread_test_suite ()
{
	Suite *s = suite_create ("AptWorkerThread");
	TCase *tc;

	tc = tcase_create ("apt_worker_cmd");
	tcase_set_timeout(tc, tc_timeout_s);
	tcase_add_checked_fixture (tc, setup, teardown);
	tcase_add_test (tc, test_apt_worker_cmd_check_updates);
	tcase_add_test (tc, test_apt_worker_cmd_get_package_list);
	tcase_add_test (tc, test_apt_worker_cmd_get_file_details);
	tcase_add_test (tc, test_apt_worker_cmd_get_package_details);
	tcase_add_test (tc, test_apt_worker_cmd_install_file);
	tcase_add_test (tc, test_apt_worker_cmd_install_package);
	tcase_add_test (tc, test_apt_worker_cmd_uninstall_package);
	tcase_add_test (tc, test_apt_worker_cmd_get_catalogues);
	tcase_add_test (tc, test_apt_worker_batch_cmd);
	suite_add_tcase (s, tc);

	return s;
}

int main (int argc, char *argv[])
{
	tc_timeout_s = 30;

	get_options (argc, argv);

	return run_suite (AptWorkerThread_test_suite ())
		? EXIT_SUCCESS : EXIT_FAILURE;
}
