/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

// -*- Mode: C++; c-basic-offset: 8 -*-
#include "apt-worker-struct.h"
#include "apt-worker-internal.h"
#include "mock-apt-worker.h"
#include "check-common.h"
#include "catalogue.h"
#include <cstdlib>
#include <cstdio>
#include <cstring>
#include <set>
#include <glib-object.h>

static const char *package = "usertestpkg";
static const char *nonuser_package = "nonusertestpkg";
static const char *not_found_package = "packagenotfound";
static const char *pkg_filename = "usertestpkg_0.1.0-1_"DEB_HOST_ARCH".deb";
static const char *update_pkg_filename = "usertestpkg_0.1.1-1_"DEB_HOST_ARCH".deb";
static const char *nonuser_pkg_filename = "nonusertestpkg_0.1.0-1_"DEB_HOST_ARCH".deb";
static const char *not_found_filename = "/file/not/found";
static const char *pkg_version = "0.1.0-1";
static const char *update_pkg_version = "0.1.1-1";
static const char *alt_download_root = "";
static gboolean with_status = FALSE;
static const gchar *catalogue_name = "Test1";
static const gchar *catalogue_uri = "file:///tmp/pkgmgr-test/Test";
// static const gchar *catalogue_uri = "file:///tmp/TestCatalogue";
static const gchar *catalogue_dist = "default";
static const gchar *catalogue_components = "main user";
static const gchar *essential_catalogue_name = "TestEssential";

static gchar *pkg_filepath = 0;
static gchar *update_pkg_filepath = 0;
static gchar *nonuser_pkg_filepath = 0;

static GError *error = 0;
// static AptWorker *apt_worker[3] = { 0 };
static AptWorker *apt_worker = 0;

static void disable_output ()
{
	if (verbosity == 0) {
		freopen("/dev/null", "w", stdout);
		freopen("/dev/null", "w", stderr);
	}
}

static void enable_output ()
{
	if (verbosity == 0) {
		freopen("/dev/stdout", "w", stdout);
		freopen("/dev/stdout", "w", stderr);
	}
}

static void setup_install ()
{
	// fail_unless (init_test_install ());
}

static void teardown_install ()
{
}

static void setup_update ()
{
	// fail_unless (init_test_update ());
}

static void teardown_update ()
{
}

static void setup ()
{
	disable_output ();
	g_type_init ();
	g_thread_init (NULL);
	error = 0;
// 	apt_worker = mock_apt_worker_new (NULL, NULL);
	apt_worker = apt_worker_new ();
	fail_if (apt_worker == 0);
	apt_worker->init (NULL, NULL, NULL, NULL, NULL, NULL, NULL);
	pkg_filepath = g_strdup_printf ("%s/%s", test_dir, pkg_filename);
	update_pkg_filepath = g_strdup_printf ("%s/%s", test_dir, update_pkg_filename);
	nonuser_pkg_filepath = g_strdup_printf ("%s/%s", test_dir, nonuser_pkg_filename);

// 	MockAptWorker *mock = reinterpret_cast<MockAptWorker *>(apt_worker);
	// Success retvals:
// 	mock->error = 0;
// 	mock->pkg_list_info = package_new();
// 	mock->boolean = TRUE;
// 	mock->result_code = rescode_success;
}

static void teardown ()
{
	g_free (pkg_filepath);
	pkg_filepath = 0;
	g_free (update_pkg_filepath);
	update_pkg_filepath = 0;
	g_free (nonuser_pkg_filepath);
	nonuser_pkg_filepath = 0;

	enable_output ();
}

// Catalogue test cases:

START_TEST (test_cmd_get_catalogues)
{
	GList *catalogues = NULL;
	GError *error = NULL;
	gboolean ret = apt_worker->cmd_get_catalogues (&catalogues, &error);

	fail_unless (ret == TRUE);
	fail_unless (error == NULL);
	fail_unless (catalogues != NULL);
	GList *list = catalogues;
	gboolean found = FALSE;
	while (list) {
		Catalogue* c = static_cast<Catalogue*> (list->data);
		if (!found && g_strcmp0(c->name, catalogue_name) == 0) {
			found = TRUE;
		}
		catalogue_free (c);
		list = g_list_next (list);
	}
	fail_unless (found);
	g_list_free (catalogues);
}
END_TEST

static Catalogue *find_catalogue (const gchar *name)
{
	GList *catalogues = NULL;
	GError *error = 0;
	gboolean ret;

	ret = apt_worker->cmd_get_catalogues (&catalogues, &error);
	if (ret && catalogues != NULL) {
		fail_if_GError (error);

		GList *list = catalogues;
		Catalogue *found = NULL;

		while (list) {
			Catalogue *c = reinterpret_cast<Catalogue *> (list->data);
			if (! found && ! g_strcmp0 (c->name, name))
				found = c;
			else
				catalogue_free (c);
			list = g_list_next (list);
		}
		g_list_free (catalogues);

		return found;
	}
	else
		return NULL;
}

START_TEST (test_cmd_add_catalogue)
{
	gboolean ret;

	ret = apt_worker->cmd_add_catalogue (
		catalogue_name,
		catalogue_uri,
		catalogue_dist,
		catalogue_components,
		&error);

	fail_if_GError (error);
	fail_unless (ret == TRUE);
	Catalogue *cat = find_catalogue (catalogue_name);
	fail_if (cat == 0);
	fail_unless (! g_strcmp0 (cat->uri, catalogue_uri));
	fail_unless (! g_strcmp0 (cat->dist, catalogue_dist));
	fail_unless (! g_strcmp0 (cat->components, catalogue_components));
	fail_unless (cat->enabled); /* enabled by default */
	catalogue_free (cat);
}
END_TEST

START_TEST (test_cmd_add_catalogue__info_already_exists)
{
	gboolean ret;

	ret = apt_worker->cmd_add_catalogue (
		"UniqueName",
		catalogue_uri,
		catalogue_dist,
		catalogue_components,
		&error);

	fail_if (error == 0);
	fail_unless (ret == FALSE);
}
END_TEST

START_TEST (test_cmd_add_catalogue__name_already_exists)
{
	gboolean ret;

	ret = apt_worker->cmd_add_catalogue (
		catalogue_name,
		"file:///tmp/unique/path",
		"somedist",
		"comp1 comp2",
		&error);

	fail_if (error == 0);
	fail_unless (ret == FALSE);
}
END_TEST

START_TEST (test_cmd_enable_catalogue)
{
	gboolean ret;

	ret = apt_worker->cmd_enable_catalogue (
		catalogue_name, TRUE,
		&error);

	fail_if_GError (error);
	fail_unless (ret == TRUE);
	Catalogue *cat = find_catalogue (catalogue_name);
	fail_if (cat == 0);
	fail_unless (! g_strcmp0 (cat->uri, catalogue_uri));
	fail_unless (! g_strcmp0 (cat->dist, catalogue_dist));
	fail_unless (! g_strcmp0 (cat->components, catalogue_components));
	fail_unless (cat->enabled);
	catalogue_free (cat);
}
END_TEST

START_TEST (test_cmd_disable_catalogue)
{
	gboolean ret;

	ret = apt_worker->cmd_enable_catalogue (
		catalogue_name, FALSE,
		&error);

	fail_if_GError (error);
	fail_unless (ret == TRUE);
	Catalogue *cat = find_catalogue (catalogue_name);
	fail_if (cat == 0);
	fail_unless (! g_strcmp0 (cat->uri, catalogue_uri));
	fail_unless (! g_strcmp0 (cat->dist, catalogue_dist));
	fail_unless (! g_strcmp0 (cat->components, catalogue_components));
	fail_if (cat->enabled);
	catalogue_free (cat);
}
END_TEST

START_TEST (test_cmd_disable_catalogue__essential)
{
	gboolean ret;

	ret = apt_worker->cmd_enable_catalogue (
		essential_catalogue_name, FALSE,
		&error);

	fail_if (error == 0);
	fail_unless (ret == FALSE);
	Catalogue *cat = find_catalogue (essential_catalogue_name);
	fail_if (cat == 0);
	fail_unless (cat->essential);
	catalogue_free (cat);
}
END_TEST

START_TEST (test_cmd_remove_catalogue)
{
	gboolean ret;

	ret = apt_worker->cmd_remove_catalogue (catalogue_name, &error);

	fail_if_GError (error);
	fail_unless (ret == TRUE);
	fail_unless (find_catalogue (catalogue_name) == 0);
}
END_TEST

START_TEST (test_cmd_remove_catalogue__essential)
{
	gboolean ret;

	ret = apt_worker->cmd_remove_catalogue (
		essential_catalogue_name,
		&error);

	fail_if (error == 0);
	fail_unless (ret == FALSE);
	Catalogue *cat = find_catalogue (essential_catalogue_name);
	fail_if (cat == 0);
	fail_unless (cat->essential);
	catalogue_free (cat);
}
END_TEST

// Package install test cases:

START_TEST (test_cmd_get_package_list)
{
	PkgType type = PKG_TYPE_NOT_INSTALLED;
	PkgSort sort = PKG_SORT_INSTALL_TIME;
	const gchar *category = "";
	const gchar *pattern = "";
	GList *packages = 0;

	apt_worker->cmd_get_package_list (category,
					  type,
					  pattern,
					  &packages,
					  sort,
					  &error);

	fail_unless (error == 0);
}
END_TEST

START_TEST (test_cmd_get_file_details)
{
	Package *ret;

	ret = apt_worker->cmd_get_file_details (pkg_filepath, &error);

	fail_if (ret == 0);
	fail_unless (error == 0);
}
END_TEST

START_TEST (test_cmd_get_file_details__not_found)
{
	Package *ret;

	ret = apt_worker->cmd_get_file_details (not_found_filename, &error);

	fail_unless (ret == 0);
	fail_if (error == 0);
}
END_TEST

START_TEST (test_cmd_get_file_details__not_installed)
{
	Package *ret;

	ret = apt_worker->cmd_get_file_details (pkg_filepath, &error);

	fail_if (ret == 0);
	fail_unless (error == 0);
}
END_TEST

START_TEST (test_cmd_get_package_details)
{
	GHashTable *ret;

	ret = apt_worker->cmd_get_package_details_map ("hpm-user1",
						       pkg_version,
                                                       NULL,
						       FALSE,
						       &error);

	fail_if_GError (error);
	fail_if (ret == 0);
	// fail_unless (! strcmp (ret->name, package));
	// fail_unless (! strcmp (ret->version, pkg_version));
}
END_TEST

START_TEST (test_cmd_get_package_details__installed)
{
	GHashTable *ret;

	ret = apt_worker->cmd_get_package_details_map (package,
                                                       NULL,
                                                       NULL,
                                                       TRUE,
                                                       &error);

	fail_if (ret == 0);
	fail_unless (error == 0);
	//fail_unless (! strcmp (ret->name, package));
	//fail_unless (! strcmp (ret->version, pkg_version));
}
END_TEST

START_TEST (test_cmd_get_package_details__nonuser)
{
	GHashTable *ret;

	ret = apt_worker->cmd_get_package_details_map ("apt",
                                                       pkg_version,
                                                       NULL,
                                                       FALSE,
                                                       &error);

	fail_unless (ret == 0);
	fail_if (error == 0);
}
END_TEST

START_TEST (test_cmd_get_package_details__not_found)
{
	GHashTable *ret;

	ret = apt_worker->cmd_get_package_details_map (not_found_package,
                                                       pkg_version,
                                                       NULL,
                                                       FALSE,
                                                       &error);

	fail_unless (ret == 0);
	fail_if (error == 0);
}
END_TEST

START_TEST (test_cmd_install_file)
{
	gboolean ret;

	ret = apt_worker->cmd_install_file (pkg_filepath,
					    &error);

	fail_unless (ret);
	fail_unless (error == 0);
}
END_TEST

START_TEST (test_cmd_install_file__nonuser_package)
{
	gboolean ret;

	ret = apt_worker->cmd_install_file (nonuser_pkg_filepath,
					    &error);

	fail_unless (ret == FALSE);
	fail_unless (error != 0);
}
END_TEST

START_TEST (test_cmd_install_file__file_not_found)
{
	gboolean ret;

	ret = apt_worker->cmd_install_file (not_found_filename,
					    &error);

	fail_unless (ret == FALSE);
	fail_unless (error != 0);
}
END_TEST

START_TEST (test_cmd_install_package)
{
	gboolean ret;

	ret = apt_worker->cmd_install_package (package,
					       NULL,
                                               NULL,
					       alt_download_root,
					       &error);
	
	fail_unless (ret == TRUE);
	fail_unless (error == 0);
}
END_TEST

START_TEST (test_cmd_install_package__nonuser)
{
	gboolean ret;

	ret = apt_worker->cmd_install_package (nonuser_package,
					       NULL,
                                               NULL,
					       alt_download_root,
					       &error);
	
	fail_unless (ret == FALSE);
	fail_unless (error != 0);
}
END_TEST

START_TEST (test_cmd_install_package__not_found)
{
	gboolean ret;

	ret = apt_worker->cmd_install_package (package,
					       NULL,
                                               NULL,
					       alt_download_root,
					       &error);
	
	fail_unless (ret == FALSE);
	fail_unless (error != 0);
}
END_TEST

START_TEST (test_cmd_uninstall_package)
{
	gboolean ret;

	ret = apt_worker->cmd_uninstall_package (package,
						 &error);

	fail_unless (ret == TRUE);
	fail_unless (error == 0);
}
END_TEST

START_TEST (test_cmd_uninstall_package__not_found)
{
	gboolean ret;

	ret = apt_worker->cmd_uninstall_package (not_found_package,
						 &error);

	fail_unless (ret == FALSE);
	fail_unless (error != 0);
}
END_TEST

START_TEST (test_cmd_check_updates)
{
	ResultCode ret;

	ret = apt_worker->cmd_check_updates (with_status, NULL, false);

	fail_unless (ret == rescode_success || ret == rescode_partial_success);
}
END_TEST

// Update test cases

START_TEST (test_cmd_fetch_updates)
{
	GError* error = NULL;
	GList *list = NULL;

	apt_worker->cmd_get_package_list (NULL,
					PKG_TYPE_UPDATE,
					NULL,
					&list,
					PKG_SORT_NAME,
					&error);

	fail_unless (error == NULL);
	fail_if (list == NULL);
	GList *tmp = list;
	gboolean found = FALSE;
	while( tmp ) {
		Package *pkg = static_cast<Package*>(tmp->data);
		if (strcmp(package, pkg->name) == 0) {
			found = TRUE;
			break;
		}
		tmp = g_list_next(tmp);
	}
	fail_unless (found);
}
END_TEST
START_TEST (test_cmd_upgrade_file)
{
	gboolean ret;

	ret = apt_worker->cmd_install_file (update_pkg_filepath,
					    &error);

	fail_unless (ret);
	fail_unless (error == 0);
}
END_TEST

START_TEST (test_cmd_get_package_details__0_1_1)
{
	GHashTable *ret;

	ret = apt_worker->cmd_get_package_details_map (package,
                                                       update_pkg_version,
                                                       NULL,
                                                       FALSE,
                                                       &error);

	fail_if (ret == 0);
	fail_unless (error == 0);

	GValue *val = (GValue*)g_hash_table_lookup (ret, "Name");
	fail_if ( val == 0 );
	const gchar* name = g_value_get_string (val);
	fail_unless (! strcmp (name, package));
	
	val = (GValue*)g_hash_table_lookup (ret, "Version");
	fail_if ( val == 0 );
	const gchar* ver = g_value_get_string (val);
	fail_unless (! strcmp (ver, update_pkg_version));
}
END_TEST

START_TEST (test_cmd_get_package_details__latest)
{
	GHashTable *ret;

	ret = apt_worker->cmd_get_package_details_map (package,
                                                       NULL,
                                                       NULL,
                                                       FALSE,
                                                       &error);

	fail_if (ret == 0);
	fail_unless (error == 0);

	GValue *val = (GValue*)g_hash_table_lookup (ret, "Name");
	fail_if ( val == 0 );
	const gchar* name = g_value_get_string (val);
	fail_unless (! strcmp (name, package));
	
	val = (GValue*)g_hash_table_lookup (ret, "Version");
	fail_if ( val == 0 );
	const gchar* ver = g_value_get_string (val);
	fail_unless (! strcmp (ver, update_pkg_version));
}
END_TEST

START_TEST (test_cmd_upgrade_package)
{
	gboolean ret;

	ret = apt_worker->cmd_upgrade_package (package,
					       NULL,
                                               NULL,
					       alt_download_root,
					       &error);

	fail_unless (ret == TRUE);
	fail_unless (error == 0);
}
END_TEST

using namespace std;

static bool match_test_filter (set<string> const& F, string const& test)
{
	for (set<string>::const_iterator i = F.begin ();
	     i != F.end ();
	     ++i)
	{
		// printf ("%s/%s\n", i->c_str (), test.c_str ());
		if (test.find (*i) != string::npos)
		{
			// puts ((*i).c_str ());
			return true;
		}
	}
	return false;
}

#define ADD_TEST(name) \
	if (match_test_filter (test_filters, #name)) \
		tcase_add_test (tc, name);

Suite *
apt_worker_test_suite (set<string> const& test_filters)
{
	Suite *s = suite_create ("apt-worker");
	TCase *tc;

	tc = tcase_create ("catalogue");
	tcase_set_timeout(tc, tc_timeout_s);
	tcase_add_unchecked_fixture (tc, setup_install, teardown_install);
	tcase_add_checked_fixture (tc, setup, teardown);
	ADD_TEST (test_cmd_add_catalogue);
	ADD_TEST (test_cmd_add_catalogue__info_already_exists);
	ADD_TEST (test_cmd_add_catalogue__name_already_exists);
	ADD_TEST (test_cmd_get_catalogues);
	ADD_TEST (test_cmd_enable_catalogue); // already enabled
	ADD_TEST (test_cmd_disable_catalogue);
	ADD_TEST (test_cmd_enable_catalogue);
	ADD_TEST (test_cmd_remove_catalogue);
	ADD_TEST (test_cmd_disable_catalogue__essential);
	ADD_TEST (test_cmd_remove_catalogue__essential);
	suite_add_tcase (s, tc);

	tc = tcase_create ("install");
	tcase_set_timeout(tc, tc_timeout_s);
	tcase_add_unchecked_fixture (tc, setup_install, teardown_install);
	tcase_add_checked_fixture (tc, setup, teardown);

	// No packages in testrepo
	ADD_TEST (test_cmd_install_package__not_found);
	ADD_TEST (test_cmd_uninstall_package__not_found);
	ADD_TEST (test_cmd_check_updates); /* update package list */

	// usertestpkg, nonusertestpkg in testrepo
	ADD_TEST (test_cmd_get_package_list);
	ADD_TEST (test_cmd_get_package_details);
	ADD_TEST (test_cmd_get_package_details__nonuser);
	ADD_TEST (test_cmd_get_package_details__not_found);

	ADD_TEST (test_cmd_get_file_details__not_found);
	ADD_TEST (test_cmd_get_file_details__not_installed);
	ADD_TEST (test_cmd_install_file);
	ADD_TEST (test_cmd_get_file_details);
	ADD_TEST (test_cmd_install_file__nonuser_package);
	ADD_TEST (test_cmd_uninstall_package);
	ADD_TEST (test_cmd_install_package);
	ADD_TEST (test_cmd_install_file__file_not_found);
	ADD_TEST (test_cmd_get_package_details__installed);
	ADD_TEST (test_cmd_install_package__nonuser);
	ADD_TEST (test_cmd_install_package__not_found);
	ADD_TEST (test_cmd_uninstall_package);
	ADD_TEST (test_cmd_uninstall_package__not_found);
	suite_add_tcase (s, tc);

	tc = tcase_create ("upgrade-file");
	tcase_set_timeout(tc, tc_timeout_s);
	tcase_add_unchecked_fixture (tc, setup_update, teardown_update);
	tcase_add_checked_fixture (tc, setup, teardown);

	ADD_TEST (test_cmd_get_package_details);
	ADD_TEST (test_cmd_get_package_details__installed);
	ADD_TEST (test_cmd_check_updates);
	ADD_TEST (test_cmd_upgrade_file);
	ADD_TEST (test_cmd_get_package_details__0_1_1);
	ADD_TEST (test_cmd_get_package_details__latest);
	ADD_TEST (test_cmd_uninstall_package);
	suite_add_tcase (s, tc);

	tc = tcase_create ("upgrade-package");
	tcase_set_timeout(tc, tc_timeout_s);
	tcase_add_unchecked_fixture (tc, setup_update, teardown_update);
	tcase_add_checked_fixture (tc, setup, teardown);

	ADD_TEST (test_cmd_get_package_details);
	ADD_TEST (test_cmd_get_package_details__installed);
	ADD_TEST (test_cmd_check_updates);
	ADD_TEST (test_cmd_fetch_updates);
	ADD_TEST (test_cmd_upgrade_package);
	ADD_TEST (test_cmd_get_package_details__0_1_1);
	ADD_TEST (test_cmd_get_package_details__latest);
	ADD_TEST (test_cmd_uninstall_package);
	suite_add_tcase (s, tc);

	return s;
}

int main (int argc, char *argv[])
{
	tc_timeout_s = 30;

	get_options (argc, argv);

	fork_mode = 0;

	set<string> test_filters;
	for (; optind < argc; ++optind)
	{
		// puts (argv[optind]);
		test_filters.insert (argv[optind]);
	}

	return run_suite (apt_worker_test_suite (test_filters))
		? EXIT_SUCCESS : EXIT_FAILURE;
}
