/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "check-common.h"
#include <getopt.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <glib.h>

int verbosity = 0;
int fork_mode = 1;
print_output output_level = CK_NORMAL;
int tc_timeout_s = 10;
const char *test_dir = 0;

bool run_suite (Suite *suite)
{
	int n_failed;
	SRunner *srunner;

	srunner = srunner_create (suite);
	srunner_set_fork_status (srunner, fork_mode ? CK_FORK : CK_NOFORK);
	srunner_run_all (srunner, output_level);
	n_failed = srunner_ntests_failed (srunner);
	srunner_free (srunner);

	return n_failed == 0;
}

/*!
 * Similar to \a system function, but redirects output to optional \a
 * out stream.
 */
static bool fsystem (const char *cmd, FILE *out = 0)
{
	char *preload = getenv ("SBOX_PRELOAD");
	gchar *sbox_cmd;

	if (preload) {
		// In Scratchbox, LD_PRELOAD needs to be set for
		// correct operations.
		sbox_cmd = g_strdup_printf ("LD_PRELOAD=%s %s",
					    "/scratchbox/tools/lib/libsb.so.0",
					   cmd);
	}
	else
		sbox_cmd = g_strdup (cmd);

	FILE *r = popen (sbox_cmd, "r");
	g_free (sbox_cmd);

	int ret;

	for (;;) {
		ret = fgetc (r);
		if (ret == EOF)
			break;
		if (out)
			fputc (ret, out);
	}

	ret = pclose (r);
	return WIFEXITED (ret) ? WEXITSTATUS (ret) == 0 : false;
// 	return system (cmd) == 0;
}

bool init_test_install ()
{
	return fsystem ("./init-test-install.sh");
}

bool init_test_update ()
{
	return fsystem ("./init-test-update.sh");
}

static void usage (char *argv0)
{
	const char *fmt =
		"usage: %s [OPTION...] TEST-DIR\n\n"
		"Options:\n"
		"  -p, --print-output=LEVEL  Check output level [0-3] (default: %d)\n"
		"  -t, --case-timeout=SEC    Test case timeout in seconds (default: %d)\n"
		"  -v, --verbose             Increase verbosity\n"
		;

	fprintf (stderr,
		 fmt,
		 argv0,
		 (int) output_level,
		 tc_timeout_s);

	exit (EXIT_FAILURE);
}

void get_options (int argc, char *argv[])
{
	int c;

	for (;;) {
		static struct option long_options[] = {
			{ "verbose", no_argument, 0, 'v' },
			{ "fork-mode", no_argument, &fork_mode, 1 },
			{ "print-output", required_argument, 0, 'p' },
			{ "case-timeout", required_argument, 0, 't' },
			{ "test-dir", required_argument, 0, 'd' },
			{ 0, 0, 0, 0 }
		};

		int option_index = 0;

		c = getopt_long (argc, argv, "vfd:o:p:t:",
				 long_options, &option_index);
		if (c == -1)
			break;

		switch (c) {

		case 'p':
			output_level = (print_output) atoi (optarg);
			if (output_level < 0 || output_level > CK_VERBOSE)
				usage (argv[0]);
			break;

		case 'd':
			test_dir = optarg;
			break;

		case 't':
			tc_timeout_s = atoi (optarg);
			if (tc_timeout_s < 0)
				usage (argv[0]);
			break;

		case 'v':
			++verbosity;
			break;
		case '?':
		default:
			perror(optarg);
			usage (argv[0]);
		}
	}

	if (verbosity > 0) {
		puts ("Options:");
		printf ("fork-mode: %d\n", fork_mode);
		printf ("print-output: %d\n", (int) output_level);
		printf ("case-timeout: %d\n", tc_timeout_s);
	}

	if (optind >= argc) {
		usage (argv[0]);
	}

	test_dir = argv[optind++];

// 	if (chdir (test_dir) < 0) {
// 		perror (test_dir);
// 		exit (EXIT_FAILURE);
// 	}
}
