/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

/* -*- Mode: C++; c-file-style: "gnu" -*- */
#include "check-common.h"
#include "apt-worker.h"
#include "apt-worker-internal.h"
#include "apt-worker-cache.h"
#include "apt-worker-struct.h"
#include "glib-util.h"
#include "package-map.h"
#include <cassert>
#include <cstdio>
#include <cstdlib>
#include <cstring>
#include <glib-object.h>

static GHashTable *
_make_package_map ()
{
  GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert_string_value (map, "Name", "Name value\n Continued");
  // g_hash_table_insert_int_value (map, "Type", G_MAXINT);
  g_hash_table_insert_int64_value (map, "InstalledSize", G_MAXINT64);
  return map;
}

START_TEST (test_package_map_read_write)
{
  GHashTable *map1 = _make_package_map ();
  GHashTable *map2 = NULL;
  static char buf1[0x1000];
  static char buf2[0x1000];
  FILE *f;
  gboolean ok;
  const gchar *str1, *str2;

  f = fmemopen (buf1, sizeof (buf1), "w");
  ok = package_hash_table_fwrite (map1, f);
  fail_unless (ok);
  fclose (f);

  f = fmemopen (buf1, sizeof (buf1), "r");
  ok = package_hash_table_fread (&map2, f);
  fail_unless (ok);
  fclose (f);

  fail_unless (g_hash_table_size (map1) == g_hash_table_size (map2));
  str1 = g_hash_table_get_string_value (map1, "Name");
  str2 = g_hash_table_get_string_value (map2, "Name");
  fail_unless (!strcmp (str1, str2));
  // fail_unless (g_hash_table_get_int_value (map1, "Type") ==
  // 	       g_hash_table_get_int_value (map2, "Type"));
  fail_unless (g_hash_table_get_int64_value (map1, "InstalledSize") ==
	       g_hash_table_get_int64_value (map2, "InstalledSize"));

  f = fmemopen (buf2, sizeof (buf2), "w");
  ok = package_hash_table_fwrite (map2, f);
  fail_unless (ok);
  fclose (f);

  fail_unless (!strcmp (buf1, buf2));

  // printf ("buf1: '%s'\n", buf1);
  // printf ("buf2: '%s'\n", buf1);

  package_hash_table_free (map1);
  package_hash_table_free (map2);
}
END_TEST

START_TEST (test_get_all_package_fields)
{
  const char *package = "hpm-user1";
  char buf1[0x10000] = {0};
  char buf2[0x10000] = {0};
  FILE *f1, *f2;

  apt_worker_init (0, 0, 0, 0, 0, 0, 0);
  cache_init (false);
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  // pkgCache &cache = *(awc->cache);
  pkgDepCache &dep_cache = *(awc->cache);
  pkgCache::PkgIterator pkg = dep_cache.FindPkg (package);
  pkgCache::VerIterator cur = pkg.CurrentVer ();
  assert (!cur.end ());
  package_record r (cur);

  f1 = fmemopen (buf1, sizeof (buf1), "w");
  f2 = fmemopen (buf2, sizeof (buf2), "w");

  {
    const char *s0, *s1;
    r.P.GetRec (s0, s1);
    fwrite (s0, s1 - s0, 1, f1);
    // fwrite (s0, s1 - s0, 1, stdout);
  }

  for (FieldIterator i (r.section); !i.end (); ++i)
    {
      fprintf (f2, "%s: %s\n",
	       i.key ().c_str (),
	       i.value ().c_str ());
      // fprintf (stdout, "%s: %s\n",
      // 	       i.key ().c_str (),
      // 	       i.value ().c_str ());
    }

  fail_unless (!strcmp (buf1, buf2));

  fclose (f1);
  fclose (f2);
}
END_TEST

START_TEST (test_cmd_get_package_details)
{
  GHashTable *map;
  AptWorker *apt_worker;
  GError *error = NULL;

  // disable_output ();
  g_type_init ();
  g_thread_init (NULL);
  apt_worker = apt_worker_new ();
  apt_worker->init (NULL, NULL, NULL, NULL, NULL, NULL, NULL);

  const char *name = "hpm-user1";
  const char *version = "0.1.0-1";
  map = apt_worker->cmd_get_package_details_map (name,
						 version,
                                                 NULL,
						 FALSE,
						 &error);

  fail_if_GError (error);
  fail_if (map == 0);

  const gchar *name2, *version2;
  name2 = g_hash_table_get_string_value (map, "Name");
  fail_unless (!strcmp (name2, name));
  version2 = g_hash_table_get_string_value (map, "Version");
  fail_unless (!strcmp (version2, version));
}
END_TEST

Suite *
apt_worker_test_suite ()
{
  Suite *s = suite_create ("daemon");
  TCase *tc;

  tc = tcase_create ("fields");
  tcase_set_timeout(tc, tc_timeout_s);
  // tcase_add_unchecked_fixture (tc, setup_install, teardown_install);
  // tcase_add_checked_fixture (tc, setup, teardown);
  tcase_add_test (tc, test_package_map_read_write);
  tcase_add_test (tc, test_get_all_package_fields);
  tcase_add_test (tc, test_cmd_get_package_details);
  suite_add_tcase (s, tc);

  return s;
}

int main (int argc, char *argv[])
{
  tc_timeout_s = 30;

  g_type_init ();

  get_options (argc, argv);

  fork_mode = 0;

  return run_suite (apt_worker_test_suite ())
    ? EXIT_SUCCESS : EXIT_FAILURE;
}
