/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

/* -*- Mode: C++; c-file-style: "gnu" -*- */
#include "check-common.h"
#include "inst-pkg-ts-storage.h"
//#include "apt-worker.h"
//#include "apt-worker-internal.h"
//#include "apt-worker-cache.h"
//#include "apt-worker-struct.h"
//#include "glib-util.h"
//#include "package-map.h"
//#include <cassert>
#include <glib/gstdio.h>
#include <cstdlib>
//#include <cstring>
#include <glib-object.h>
#include <time.h>

#include "debug.h"

const gchar* test_package_1 = "test-package-1";
const gchar* test_package_2 = "test-package-2";
const gchar* test_package_3 = "test-package-3";
const gchar* test_package_4 = "test-package-4";
const gchar* test_file_1 = "test-file-1";
const gchar* test_file_2 = "test-file-2";
const gchar* test_file_3 = "test-file-3";

const gchar* test_package_version_1 = "version-1";
const gchar* test_package_version_2 = "version-2";
const gchar* test_package_version_3 = "version-3";
const gchar* test_package_version_4 = "version-4";
const gchar* test_file_version_1 = "file-version-1";
const gchar* test_file_version_2 = "file-version-2";
const gchar* test_file_version_3 = "file-version-3";

const gchar* test_package_filename_1 = "";
const gchar* test_package_filename_2 = "";
const gchar* test_package_filename_3 = "";
const gchar* test_package_filename_4 = "";
const gchar* test_file_filename_1 = "filename-1";
const gchar* test_file_filename_2 = "filename-2";
const gchar* test_file_filename_3 = "filename-3";

const gchar* test_filename = "inst-pkg-ts-storage-test-file";
const gchar* test_non_existing_filename = "inst-pkg-ts-storage-test-non-existing-file";
const gchar* test_corrupted_filename = "inst-pkg-ts-storage-test-corrupted-file";
const gchar* test_data = "test-package-1 1268639498\ntest-package-2 <corruption>\n";

static PkgTsStorage *pkgTsStorage = 0;

static void setup ()
{
    pkgTsStorage = new PkgTsStorage();
    fail_if( pkgTsStorage == NULL );
}

static void teardown ()
{
    delete pkgTsStorage;
    pkgTsStorage = NULL;

    // Clean up a test case.
    if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) {
        g_remove(test_filename);
        fail_if(g_file_test(test_filename, G_FILE_TEST_EXISTS), "The test case could not be cleaned up!");
    }
    if ( g_file_test(test_corrupted_filename, G_FILE_TEST_EXISTS) ) {
        g_remove(test_corrupted_filename);
        fail_if(g_file_test(test_corrupted_filename, G_FILE_TEST_EXISTS), "The test case could not be cleaned up!");
    }
}

// A helper function to check the package data.
bool checkPkgInfo(const package_info* pkg, const gchar* name, const gchar* version, const gchar* filename, const time_t timestamp)
{
    if ( strcmp(pkg->package, name) != 0 ) {
        return false;
    }
    if ( strcmp(pkg->version, version) != 0 ) {
        return false;
    }
    if ( pkg->timestamp != timestamp ) {
        return false;
    }
    if ( strcmp(pkg->filename, filename) != 0 ) {
        return false;
    }
    
    return true;
}

START_TEST (test_add_package)
{
    DEBUG("");

    GList *list = pkgTsStorage->sortedList();
    if ( list != NULL ) {
        fail();
        clearList( list );
        return;
    }
    
    time_t now = time(NULL);

    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, now), "Test package could not be added!" );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    int cnt = g_list_length(list);
    if ( cnt == 1 ) {
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, now) ); // Verify that the data is correct.
    } else {
        fail("Data corrupted!"); // There must be exactly one item
    }

    clearList( list );
    DEBUG("out");
}
END_TEST

START_TEST (test_check_sorting)
{
    DEBUG("");

    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_4, test_package_version_4, time3 ) ); // To have a better branch coverage.
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );

    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 4 ); // There must be exactly four items
    
    gpointer data = g_list_nth_data( list, 0 ); // Get the first item
    fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify that the data is the first one.
    
    data = g_list_nth_data( list, 1 ); // Get the second item
    fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify that the data is the second one.
    
    data = g_list_nth_data( list, 2 ); // Get the third item
    fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify that the data is the third one.

    data = g_list_nth_data( list, 3 ); // Get the fourth item
    fail_unless( checkPkgInfo((package_info *)data, test_package_4, test_package_version_4, test_package_filename_4, time3) ); // Verify that the data is the fourth one.
        
    g_list_free( list );

    DEBUG("out");
}
END_TEST

START_TEST (test_check_added_timestamps_and_filenames)
{
    DEBUG("");

    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;
    time_t time4 = time3 + 1200;
    time_t time5 = time4 + 1500;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );
    fail_unless( pkgTsStorage->addFile(test_file_2, test_file_version_2, test_file_filename_2, time5 ) );
    fail_unless( pkgTsStorage->addFile(test_file_1, test_file_version_1, test_file_filename_1, time4 ) );

    bool ok = true;
    time_t ts = pkgTsStorage->getPackageTimestamp(test_package_4, &ok); // There should not be a package with this name!
    fail_if(ok);
    fail_unless( ts == 0 );
    
    ok = true;
    ts = pkgTsStorage->getPackageTimestamp(test_package_3, test_package_version_1, &ok); // There should not be a package with this name and version!
    fail_if(ok);
    fail_unless( ts == 0 );
    
    ok = false;
    ts = pkgTsStorage->getPackageTimestamp(test_package_3, &ok);
    fail_unless(ok);
    fail_unless( ts == time3 );
    
    ok = false;
    ts = pkgTsStorage->getPackageTimestamp(test_package_3, test_package_version_3, &ok);
    fail_unless(ok);
    fail_unless( ts == time3 );
    
    ok = false;
    ts = pkgTsStorage->getPackageTimestamp(test_package_2, &ok);
    fail_unless(ok);
    fail_unless( ts == time2 );
    
    ok = true;
    ts = pkgTsStorage->getPackageTimestamp(test_package_2, test_package_version_3, &ok); // There should not be a package with this name and version!
    fail_if(ok);
    fail_unless( ts == 0 );
    
    ok = false;
    ts = pkgTsStorage->getPackageTimestamp(test_package_2, test_package_version_2, &ok);
    fail_unless(ok);
    fail_unless( ts == time2 );
    
    ok = false;
    ts = pkgTsStorage->getPackageTimestamp(test_package_1, &ok);
    fail_unless(ok);
    fail_unless( ts == time1 );
    
    ok = false;
    ts = pkgTsStorage->getPackageTimestamp(test_package_1, test_package_version_1, &ok);
    fail_unless(ok);
    fail_unless( ts == time1 );
    
    ok = false;
    ts = pkgTsStorage->getFileTimestamp(test_file_filename_2, &ok);
    fail_unless(ok);
    fail_unless( ts == time5 );
    
    ok = false;
    ts = pkgTsStorage->getFileTimestamp(test_file_filename_1, &ok);
    fail_unless(ok);
    fail_unless( ts == time4 );
    
    ok = false;
    gchar* fn = pkgTsStorage->getPackageFilename(test_file_2, &ok);
    fail_unless(ok);
    fail_unless( strcmp(fn, test_file_filename_2) == 0 );
    
    ok = false;
    fn = pkgTsStorage->getPackageFilename(test_file_1, test_file_version_1, &ok);
    fail_unless(ok);
    fail_unless( strcmp(fn, test_file_filename_1) == 0 );
    
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time3 ) ); // This should update the timestamp of the already existing package.
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 5 );
    g_list_free( list );
    
    ok = false;
    ts = pkgTsStorage->getPackageTimestamp(test_package_2, test_package_version_2, &ok);
    fail_unless(ok);
    fail_unless( ts == time3 );
    
    fail_unless( pkgTsStorage->addFile(test_file_2, test_file_version_2, test_file_filename_2, time3 ) ); // This should update the timestamp of the already existing file.
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 5 );
    g_list_free( list );

    ok = false;
    fn = pkgTsStorage->getPackageFilename(test_file_2, &ok);
    fail_unless(ok);
    fail_unless( strcmp(fn, test_file_filename_2) == 0 );

    ok = false;
    ts = pkgTsStorage->getFileTimestamp(test_file_filename_2, &ok);
    fail_unless(ok);
    fail_unless( ts == time3 );
    
    DEBUG("out");
}
END_TEST

START_TEST (test_remove_package)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;
    time_t time4 = time3 + 1000;
    time_t time5 = time4 + 1000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_4, test_package_version_4, time4 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_3, time5 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 5 ); // There must be exactly five items
    g_list_free( list );
    
    fail_unless( pkgTsStorage->removePackage(test_package_2) );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 4 ) { // There must be exactly four items
        gpointer data = g_list_nth_data( list, 0 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify data.

        data = g_list_nth_data( list, 1 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify data.
    
        data = g_list_nth_data( list, 2 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_4, test_package_version_4, test_package_filename_4, time4) ); // Verify data.
    
        data = g_list_nth_data( list, 3 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_3, test_package_filename_1, time5) ); // Verify data.
    } else {
        fail("Data corrupted!"); // There must be exactly four items otherwise it is a failure.
    }
    
    g_list_free( list );
    
    fail_unless( pkgTsStorage->removePackage(test_package_1) ); // This should remove two packages with the same name.
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }

    if ( g_list_length(list) == 2 ) { // There must be exactly two items
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify that the data is the second one.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_4, test_package_version_4, test_package_filename_4, time4) ); // Verify that the data is the third one.
    } else {
        fail("Data corrupted!"); // There must be exactly two items otherwise it is a failure.
    }

    fail_unless( pkgTsStorage->removePackage(test_package_3) );
    fail_unless( pkgTsStorage->removePackage(test_package_4) );

    list = pkgTsStorage->sortedList();
    fail_unless( list == NULL );

    DEBUG("out");
}
END_TEST

START_TEST (test_remove_package_with_version)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;
    time_t time4 = time3 + 1000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_4, test_package_version_4, time4 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_3, time3 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 5 ); // There must be exactly five items
    g_list_free( list );
    
    fail_unless( pkgTsStorage->removePackage(test_package_1, test_package_version_3) );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 4 ) { // There must be exactly four items
        gpointer data = g_list_nth_data( list, 0 ); //
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify that the data is the first one.
    
        data = g_list_nth_data( list, 1 ); //
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify that the data is the second one.
        
        data = g_list_nth_data( list, 2 ); //
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify that the data is the second one.
    
        data = g_list_nth_data( list, 3 ); //
        fail_unless( checkPkgInfo((package_info *)data, test_package_4, test_package_version_4, test_package_filename_4, time4) ); // Verify that the data is the third one.
    } else {
        fail("Data corrupted!"); // There must be exactly four items otherwise it is a failure.
    }
    
    g_list_free( list );
    
    fail_unless( pkgTsStorage->removePackage(test_package_1) );
    fail_unless( pkgTsStorage->removePackage(test_package_2) );
    fail_unless( pkgTsStorage->removePackage(test_package_3) );
    fail_unless( pkgTsStorage->removePackage(test_package_4) );

    list = pkgTsStorage->sortedList();
    fail_unless( list == NULL );

    DEBUG("out");
}
END_TEST

START_TEST (test_clear_storage)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;
    time_t time4 = time3 + 1000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_4, test_package_version_4, time4 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );
    fail_unless( pkgTsStorage->addFile(test_file_1, test_file_version_1, test_file_filename_1, time1 ) );
    fail_unless( pkgTsStorage->addFile(test_file_2, test_file_version_2, test_file_filename_2, time2 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 6 ); // There must be exactly four items
    g_list_free( list );
    
    pkgTsStorage->clearStorage();

    list = pkgTsStorage->sortedList();
    fail_unless( list == NULL );
    
    DEBUG("out");
}
END_TEST

START_TEST (test_save_read_file)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;
    time_t time4 = time3 + 1000;
    time_t time5 = time4 + 1500;
    time_t time6 = time5 + 2000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_4, test_package_version_4, time4 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );
    fail_unless( pkgTsStorage->addFile(test_file_1, test_file_version_1, test_file_filename_1, time6 ) );
    fail_unless( pkgTsStorage->addFile(test_file_2, test_file_version_2, test_file_filename_2, time5 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail("List sorting returned null list");
        return;
    }
    fail_unless( g_list_length(list) == 6 ); // There must be exactly six items
    g_list_free( list );

    if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // If the file exists try to remove it.
        g_remove(test_filename);
        if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // If the file still exists then the test can not be run anymore.
            fail();
            return;
        }
    }

    if ( ! pkgTsStorage->saveToFile(test_filename) ) {
        fail("The file could not be saved!");
        return;
    }

    if ( ! g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // Check that it really saved the file. If it doesn't exist it is a fail.
        fail("Saved file does not exist!");
        return;
    }
    
    delete pkgTsStorage; // Make sure that it really is cleaned
    pkgTsStorage = new PkgTsStorage();

    if ( ! pkgTsStorage->readFromFile(test_filename) ) {
        fail("The file could not be read!");
        return;
    }
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 6 ) { // There must be exactly six items
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify that the data is the first one.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify that the data is the second one.
    
        data = g_list_nth_data( list, 2 ); // Get the third item
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify that the data is the third one.
        
        data = g_list_nth_data( list, 3 ); // Get the fourth item
        fail_unless( checkPkgInfo((package_info *)data, test_package_4, test_package_version_4, test_package_filename_4, time4) ); // Verify that the data is the fourth one.
        
        data = g_list_nth_data( list, 4 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_2, test_file_version_2, test_file_filename_2, time5) );
    
        data = g_list_nth_data( list, 5 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_1, test_file_version_1, test_file_filename_1, time6) );
    } else {
        fail(); // There must be exactly six items otherwise it is a failure.
    }
    
    g_list_free( list );
    list = NULL;

    // TODO: This test does not check the syntax of the storage file.

    // Read it twice and check it again.
    if ( ! pkgTsStorage->readFromFile(test_filename) ) {
        fail("The file could not be read twice!");
        return;
    }
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 6 ) { // There must be exactly six items
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify that the data is the first one.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify that the data is the second one.
    
        data = g_list_nth_data( list, 2 ); // Get the third item
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify that the data is the third one.
        
        data = g_list_nth_data( list, 3 ); // Get the fourth item
        fail_unless( checkPkgInfo((package_info *)data, test_package_4, test_package_version_4, test_package_filename_4, time4) ); // Verify that the data is the fourth one.
        
        data = g_list_nth_data( list, 4 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_2, test_file_version_2, test_file_filename_2, time5) );
    
        data = g_list_nth_data( list, 5 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_1, test_file_version_1, test_file_filename_1, time6) );
    } else {
        fail("Data corrupted!"); // There must be exactly four items otherwise it is a failure.
    }
    
    g_list_free( list );

    if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // Clean up the test case.
        g_remove(test_filename);
    }
}
END_TEST

START_TEST (test_save_read_file_empty_version)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;
    time_t time4 = time3 + 1000;
    time_t time5 = time4 + 1000;
    time_t time6 = time5 + 1000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, "", time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, "", time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_4, "", time4 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_2, "", time2 ) );
    fail_unless( pkgTsStorage->addFile(test_file_1, "", test_file_filename_1, time5 ) );
    fail_unless( pkgTsStorage->addFile(test_file_2, "", test_file_filename_2, time6 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail("List sorting returned null list");
        return;
    }
    fail_unless( g_list_length(list) == 6 ); // There must be exactly six items
    g_list_free( list );

    if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // If the file exists try to remove it.
        g_remove(test_filename);
        if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // If the file still exists then the test can not be run anymore.
            fail();
            return;
        }
    }

    if ( ! pkgTsStorage->saveToFile(test_filename) ) {
        fail("The file could not be saved!");
        return;
    }

    if ( ! g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // Check that it really saved the file. If it doesn't exist it is a fail.
        fail("Saved file does not exist!");
        return;
    }
    
    delete pkgTsStorage; // Make sure that it really is cleaned
    pkgTsStorage = new PkgTsStorage();

    if ( ! pkgTsStorage->readFromFile(test_filename) ) {
        fail("The file could not be read!");
        return;
    }
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 6 ) { // There must be exactly six items
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, "", "", time1) ); // Verify that the data is the first one.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, "", "", time2) ); // Verify that the data is the second one.
    
        data = g_list_nth_data( list, 2 ); // Get the third item
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, "", "", time3) ); // Verify that the data is the third one.
        
        data = g_list_nth_data( list, 3 ); // Get the fourth item
        fail_unless( checkPkgInfo((package_info *)data, test_package_4, "", "", time4) ); // Verify that the data is the fourth one.

        data = g_list_nth_data( list, 4 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_1, "", test_file_filename_1, time5) );
    
        data = g_list_nth_data( list, 5 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_2, "", test_file_filename_2, time6) );
    } else {
        fail(); // There must be exactly four items otherwise it is a failure.
    }
    
    g_list_free( list );
    list = NULL;

    // TODO: This test does not check the syntax of the storage file.

    // Read it twice and check it again.
    if ( ! pkgTsStorage->readFromFile(test_filename) ) {
        fail("The file could not be read twice!");
        return;
    }
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 6 ) {
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, "", "", time1) ); // Verify that the data is the first one.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, "", "", time2) ); // Verify that the data is the second one.
    
        data = g_list_nth_data( list, 2 ); // Get the third item
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, "", "", time3) ); // Verify that the data is the third one.
        
        data = g_list_nth_data( list, 3 ); // Get the fourth item
        fail_unless( checkPkgInfo((package_info *)data, test_package_4, "", "", time4) ); // Verify that the data is the fourth one.
        
        data = g_list_nth_data( list, 4 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_1, "", test_file_filename_1, time5) );
    
        data = g_list_nth_data( list, 5 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_2, "", test_file_filename_2, time6) );
    } else {
        fail("Data corrupted!"); // There must be exactly four items otherwise it is a failure.
    }
    
    g_list_free( list );

    if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // Clean up the test case.
        g_remove(test_filename);
    }
}
END_TEST

START_TEST (test_try_to_read_not_existing_file)
{
    DEBUG("");

    // First, make sure that the file does not exist.
    if ( g_file_test(test_non_existing_filename, G_FILE_TEST_EXISTS) ) { // If the file exists try to remove it.
        g_remove(test_non_existing_filename);
        if ( g_file_test(test_non_existing_filename, G_FILE_TEST_EXISTS) ) { // If the file still exists then the test can not be run anymore.
            fail();
            return;
        }
    }
    
    // Then just try to read it. If it can not be read the test passes.
    fail_if( pkgTsStorage->readFromFile(test_non_existing_filename) );
}
END_TEST

START_TEST (test_try_to_remove_non_existing_package)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time3 = time1 + 1000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 2 ); // There must be exactly two items
    g_list_free( list );
    
    fail_if( pkgTsStorage->removePackage(test_package_2) );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 2 ) { // There must be exactly two items
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify that the data is correct.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify that the data is correct.
    } else {
        fail("Data corrupted!"); // There must be exactly two items otherwise it is a failure.
    }
    
    g_list_free( list );

    DEBUG("out");
}
END_TEST

START_TEST (test_try_to_remove_from_empty_storage)
{
    fail_if( pkgTsStorage->removePackage(test_package_2) );
    fail_if( pkgTsStorage->removePackage(test_package_1, test_package_version_1) );
    
    fail_if( pkgTsStorage->removeFile(test_file_2) );
    fail_if( pkgTsStorage->removeFile(test_file_1, test_file_version_1) );
}
END_TEST

START_TEST (test_save_empty_storage)
{
    // 1. Create a test file.
    FILE *f = fopen (test_filename, "w");

    if (f == NULL) {
        fail("Could not create a test data!");
        return;
    }
    
    if ( fputs(test_data, f) == EOF ) {
        fail("Could not write a test data to a file!");
        fclose(f); // Something went wrong :'(
        return;
    }
    fclose(f);
    f = NULL;
    
    // 2. Save an empty storage.
    fail_unless( pkgTsStorage->saveToFile(test_filename) );

    // 3. Last check that the file is empty.
    f = fopen (test_filename, "r");
    if (f == NULL) {
        fail("Can not verify that the file is empty since the file can not be opened.");
        return;
    }

    fseek( f, 0, SEEK_END );
    if ( ftell( f ) != 0 ) { // The file must be empty aka 0 as its size.
        fail("File not empty!");
    }

    fclose(f);
}
END_TEST

START_TEST (test_try_to_read_corrupted_storage)
{
    // Create the corrupted test file.
    FILE *f = fopen (test_corrupted_filename, "w");

    if (f == NULL) {
        fail("Could not create a test data!");
        return;
    }
    
    if ( fputs(test_data, f) == EOF ) {
        fail("Could not write test data to a file!");
        fclose(f); // Something went wrong :'(
        return;
    }
    
    fclose(f);
    
    // Then just try to read it. If it can not be read the test passes.
    fail_if( pkgTsStorage->readFromFile(test_corrupted_filename) );
    
    // TODO: Should there be a part of the data which is not corrupted? Or should it be just empty?
}
END_TEST

START_TEST (test_compination_of_add_save_read_add_save_read_remove_save_read)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;

    // 1. Add
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail("List sorting returned null list");
        return;
    }
    fail_unless( g_list_length(list) == 1 ); // There must be exactly one item
    g_list_free( list );

    if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // If the file exists try to remove it.
        g_remove(test_filename);
        if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // If the file still exists then the test can not be run anymore.
            fail();
            return;
        }
    }

    // 2. Save
    if ( ! pkgTsStorage->saveToFile(test_filename) ) {
        fail("The file could not be saved!");
        return;
    }

    if ( ! g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // Check that it really saved the file. If it doesn't exist it is a fail.
        fail("Saved file does not exist!");
        return;
    }
    
    delete pkgTsStorage; // Make sure that it really is cleaned
    pkgTsStorage = new PkgTsStorage();

    // 3. Read
    if ( ! pkgTsStorage->readFromFile(test_filename) ) {
        fail("The file could not be read!");
        return;
    }
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail("List null!");
        return;
    }
    
    if ( g_list_length(list) == 1 ) { // There must be exactly one item
        gpointer data = g_list_nth_data( list, 0 ); // Get the item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify that the data is the first one.
    } else {
        fail(); // There must be exactly one item otherwise it is a failure.
    }
    
    g_list_free( list );
    list = NULL;

    // 4. Add
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail("List sorting returned null list");
        return;
    }
    fail_unless( g_list_length(list) == 2 ); // There must be exactly two items.
    g_list_free( list );

    // 5. Save
    if ( ! pkgTsStorage->saveToFile(test_filename) ) {
        fail("The file could not be saved!");
        return;
    }

    if ( ! g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // Check that it really saved the file. If it doesn't exist it is a fail.
        fail("Saved file does not exist!");
        return;
    }
    
    delete pkgTsStorage; // Make sure that it really is cleaned
    pkgTsStorage = new PkgTsStorage();
    
    // 6. Read
    if ( ! pkgTsStorage->readFromFile(test_filename) ) {
        fail("The file could not be read!");
        return;
    }
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 2 ) { // There must be exactly two items
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify that the data is the first one.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify that the data is the second one.
    } else {
        fail("Data corrupted!"); // There must be exactly two items otherwise it is a failure.
    }
    
    g_list_free( list );
    
    // 7. Remove
    fail_unless( pkgTsStorage->removePackage(test_package_1) );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 1 ) { // There must be exactly one item
        gpointer data = g_list_nth_data( list, 0 ); // Get the only item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify that the data is correct.
    } else {
        fail("Data corrupted!"); // There must be exactly one item otherwise it is a failure.
    }
    
    g_list_free( list );
    list = NULL;
    
    // 8. Save
    if ( ! pkgTsStorage->saveToFile(test_filename) ) {
        fail("The file could not be saved!");
        return;
    }

    if ( ! g_file_test(test_filename, G_FILE_TEST_EXISTS) ) { // Check that it really saved the file. If it doesn't exist it is a fail.
        fail("Saved file does not exist!");
        return;
    }
    
    delete pkgTsStorage; // Make sure that it really is not the same instance.
    pkgTsStorage = new PkgTsStorage();
    
    // 9. Read
    if ( ! pkgTsStorage->readFromFile(test_filename) ) {
        fail("The file could not be read!");
        return;
    }
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 1 ) { // There must be exactly one item
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify that the data is correct.
    } else {
        fail("Data corrupted!"); // There must be exactly one item otherwise it is a failure.
    }
    
    g_list_free( list );
    
    // 10. Clean up
    if ( g_file_test(test_filename, G_FILE_TEST_EXISTS) ) {
        g_remove(test_filename);
    }
}
END_TEST

START_TEST (test_add_file)
{
    DEBUG("");

    GList *list = pkgTsStorage->sortedList();
    if ( list != NULL ) {
        fail();
        clearList( list );
        return;
    }
    
    time_t now = time(NULL);

    fail_unless( pkgTsStorage->addFile(test_file_1, test_file_version_1, test_file_filename_1, now), "Test file could not be added!" );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    int cnt = g_list_length(list);
    if ( cnt == 1 ) {
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_file_1, test_file_version_1, test_file_filename_1, now) ); // Verify that the data is correct.
    } else {
        fail("Data corrupted!"); // There must be exactly one item
    }

    clearList( list );
    DEBUG("out");
}
END_TEST

START_TEST (test_remove_file)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 +  500;
    time_t time3 = time2 + 1000;
    time_t time4 = time3 + 1000;
    time_t time5 = time4 + 1000;
    time_t time6 = time5 + 1000;

    fail_unless( pkgTsStorage->addPackage(test_package_3, test_package_version_3, time3 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_1, test_package_version_1, time1 ) );
    fail_unless( pkgTsStorage->addPackage(test_package_2, test_package_version_2, time2 ) );
    fail_unless( pkgTsStorage->addFile(test_file_2, test_file_version_2, test_file_filename_2, time5 ) );
    fail_unless( pkgTsStorage->addFile(test_file_1, test_file_version_1, test_file_filename_1, time4 ) );
    fail_unless( pkgTsStorage->addFile(test_file_3, test_file_version_3, test_file_filename_3, time6 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 6 );
    g_list_free( list );
    
    fail_unless( pkgTsStorage->removeFile(test_file_filename_1, test_file_version_1) );

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 5 ) {
        gpointer data = g_list_nth_data( list, 0 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) ); // Verify data.

        data = g_list_nth_data( list, 1 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) ); // Verify data.

        data = g_list_nth_data( list, 2 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify data.
    
        data = g_list_nth_data( list, 3 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_2, test_file_version_2, test_file_filename_2, time5) ); // Verify data.
        
        data = g_list_nth_data( list, 4 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_3, test_file_version_3, test_file_filename_3, time6) ); // Verify data.
    } else {
        fail("Data corrupted!");
    }
    
    g_list_free( list );
    
    fail_unless( pkgTsStorage->removeFile(test_file_filename_2) );
    
    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }

    if ( g_list_length(list) == 4 ) {
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_package_1, test_package_version_1, test_package_filename_1, time1) );
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_package_2, test_package_version_2, test_package_filename_2, time2) );

        data = g_list_nth_data( list, 2 );
        fail_unless( checkPkgInfo((package_info *)data, test_package_3, test_package_version_3, test_package_filename_3, time3) ); // Verify data.
    
        data = g_list_nth_data( list, 3 );
        fail_unless( checkPkgInfo((package_info *)data, test_file_3, test_file_version_3, test_file_filename_3, time6) ); // Verify data.
    } else {
        fail("Data corrupted!");
    }

    fail_unless( pkgTsStorage->removePackage(test_package_1) );
    fail_unless( pkgTsStorage->removePackage(test_package_2) );
    fail_unless( pkgTsStorage->removePackage(test_package_3) );
    fail_unless( pkgTsStorage->removeFile(test_file_filename_3) );

    list = pkgTsStorage->sortedList();
    fail_unless( list == NULL );

    DEBUG("out");
}
END_TEST

START_TEST (test_try_to_remove_non_existing_file)
{
    DEBUG("");
    
    time_t time1 = time(NULL);
    time_t time2 = time1 + 1000;

    fail_unless( pkgTsStorage->addFile(test_file_2, test_file_version_2, test_file_filename_2, time2 ) );
    fail_unless( pkgTsStorage->addFile(test_file_1, test_file_version_1, test_file_filename_1, time1 ) );
    
    GList *list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    fail_unless( g_list_length(list) == 2 ); // There must be exactly two items
    g_list_free( list );
    
    fail_if( pkgTsStorage->removeFile(test_package_4) ); // Try to remove non-existing file.

    list = pkgTsStorage->sortedList();
    if ( list == NULL ) {
        fail();
        return;
    }
    
    if ( g_list_length(list) == 2 ) { // There must be exactly two items
        gpointer data = g_list_nth_data( list, 0 ); // Get the first item
        fail_unless( checkPkgInfo((package_info *)data, test_file_1, test_file_version_1, test_file_filename_1, time1) ); // Verify that the data is correct.
    
        data = g_list_nth_data( list, 1 ); // Get the second item
        fail_unless( checkPkgInfo((package_info *)data, test_file_2, test_file_version_2, test_file_filename_2, time2) ); // Verify that the data is correct.
    } else {
        fail("Data corrupted!"); // There must be exactly two items otherwise it is a failure.
    }
    
    g_list_free( list );

    DEBUG("out");
}
END_TEST

Suite *
inst_ts_storage_test_suite ()
{
    Suite *s = suite_create ("inst-ts-storage");
    TCase *tc = tcase_create ("inst-ts-storage");

    tcase_set_timeout(tc, tc_timeout_s);
//    tcase_add_unchecked_fixture (tc, setup_install, teardown_install);
    tcase_add_checked_fixture (tc, setup, teardown);

    tcase_add_test (tc, test_add_package);
    tcase_add_test (tc, test_check_sorting);
    tcase_add_test (tc, test_check_added_timestamps_and_filenames);
    tcase_add_test (tc, test_remove_package);
    tcase_add_test (tc, test_clear_storage);
    tcase_add_test (tc, test_save_read_file);
    tcase_add_test (tc, test_save_read_file_empty_version);

    tcase_add_test (tc, test_try_to_read_not_existing_file);
    tcase_add_test (tc, test_try_to_remove_non_existing_package);
    tcase_add_test (tc, test_try_to_remove_from_empty_storage);
    tcase_add_test (tc, test_save_empty_storage);
    tcase_add_test (tc, test_try_to_read_corrupted_storage);

    tcase_add_test (tc, test_compination_of_add_save_read_add_save_read_remove_save_read);

    tcase_add_test (tc, test_add_file);
    tcase_add_test (tc, test_remove_file);
    tcase_add_test (tc, test_try_to_remove_non_existing_file);

    suite_add_tcase (s, tc);

    return s;
}


int main (int argc, char *argv[])
{
  tc_timeout_s = 30;

  g_type_init ();

  get_options (argc, argv);

  fork_mode = 0;

  return run_suite (inst_ts_storage_test_suite ())
    ? EXIT_SUCCESS : EXIT_FAILURE;
}

