/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <check.h>
#include "check-common.h"
#include "package-map.h"
#include "serialize.h"
#include <cstdlib>
#include <glib.h>
#include <glib-object.h>

static gchar **make_strv (size_t s)
{
  gchar **strv = (gchar **) g_malloc ((s + 1) * sizeof (gchar *));
  size_t i;

  for (i = 0; i < s; ++i)
    strv[i] = g_strdup_printf ("strv %u", i);

  strv[i] = 0;

  return strv;
}

static bool strv_equals (gchar **v0, gchar **v1)
{
  for (; *v0; v0++, v1++)
    {
      if (g_strcmp0 (*v0, *v1))
	return false;
    }

  if (*v1)
    return false;

  return true;
}

START_TEST (test_hash_table_strv)
{
  gchar **sv0 = make_strv (3);

  GValue *val = g_new0 (GValue, 1);
  g_value_init (val, G_TYPE_STRV);
  g_value_take_boxed (val, sv0);

  GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);
  g_hash_table_insert (map, g_strdup ("key"), val);
  gpointer svp = g_hash_table_lookup (map, "key");
  GValue *val1 = (GValue *) svp;
  fail_unless (val == val1);

  gchar **sv1 = (gchar **) g_value_get_boxed (val1);

  g_value_unset (val);
}
END_TEST

const int N = 0x1000;
static char buf[N] = {0};

START_TEST (test_strv_serialize)
{
  // GValue *val = g_new0 (GValue, 1);
  // g_value_init (val, G_TYPE_STRV);

  // g_value_take_boxed (val, sv0);

  // gchar **sv1 = (gchar **) g_value_get_boxed (val);
  // fail_unless (strv_equals (sv0, sv1));

  FILE *wf = fmemopen (buf, sizeof (buf), "w");
  FILE *rf = fmemopen (buf, sizeof (buf), "r");

  // Serialize
  gchar **sv0 = make_strv (3);
  fail_unless (put_strv (wf, sv0));
  fflush (wf);

  // Deserialize
  char *line = 0;
  size_t len = 0;
  gchar **sv = 0;
  fail_unless (get_strv (rf, &line, &len, &sv));
  free (line);

  fail_unless (strv_equals (sv0, sv));

  g_strfreev (sv0);
  g_strfreev (sv);

  fclose (wf);
  fclose (rf);
}
END_TEST

void
g_hash_table_insert_strv_value (GHashTable *map, const gchar *key, gchar **sv)
{
  GValue *val;

  if (sv)
    {
      val = g_new0 (GValue, 1);
      g_value_init (val, G_TYPE_STRV);
      g_value_take_boxed (val, sv);
      g_hash_table_insert (map, g_strdup (key), val);
    }
}

START_TEST (test_package_map_serialize)
{
  FILE *wf = fmemopen (buf, sizeof (buf), "w");
  FILE *rf = fmemopen (buf, sizeof (buf), "r");
  GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);

  gchar **sv0 = make_strv (3);
  g_hash_table_insert_strv_value (map, "InstallDeps", sv0);

  // Serialize
  fail_unless (package_hash_table_fwrite (map, wf));
  fflush (wf);

  // Deserialize
  GHashTable *map1 = 0;
  fail_unless (package_hash_table_fread (&map1, rf));
  gpointer svp = g_hash_table_lookup (map, "InstallDeps");
  fail_unless (svp != 0);

  GValue *v1 = (GValue *) svp;
  gchar **sv1 = (gchar **) g_value_get_boxed (v1);
  fail_unless (strv_equals (sv0, sv1));

  g_hash_table_unref (map);
  fclose (wf);
  fclose (rf);
}
END_TEST

static Suite *
create_suite ()
{
  Suite *s = suite_create ("package-map");
  TCase *tc;

  tc = tcase_create ("package-map");
  // tcase_set_timeout (tc, 180);
  tcase_add_test (tc, test_strv_serialize);
  tcase_add_test (tc, test_package_map_serialize);
  suite_add_tcase (s, tc);

  return s;
}

int main (int argc, char *argv[])
{
  g_type_init ();
  setenv ("G_DEBUG", "fatal_warnings", 1);

  fork_mode = 0;

  return run_suite (create_suite ()) ? EXIT_SUCCESS : EXIT_FAILURE;
}
