/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <QtTest/QtTest>
#include <QDBusError>
#include <memory>
// #include "test-libpkgmgr-dbus-qt.h"
#include "packagemanager.h"
#include "packagemanagerpendingcallwatcher.h"
#include "com.nokia.package_manager.h"

#define DUMP_SIGNALS 0
#define ERROR_PREFIX "com.nokia.package_manager.Error."
#define PKGMGR_DBUS_SERVICE "com.nokia.package_manager"
#define PKGMGR_DBUS_PATH "/com/nokia/package_manager"
#define PKGMGR_DBUS_INTERFACE "com.nokia.package_manager"

using std::auto_ptr;

class TestPkgmgrDbus : public QObject
{
  Q_OBJECT

  void reset()
  {
    system("rm /var/cache/apt/archives/hpm-*.deb 2>/dev/null");

    user1_name = "hpm-user1";
    user1_version = "0.1.0-1";
    user1_updateversion = "0.1.1-1";
    user1_filename = QString ("%1_%2_%3.deb")
      .arg (user1_name)
      .arg (user1_version)
      .arg (DEB_HOST_ARCH);

    _list_dataFetched = ListDataFetched();
    _stringList_dataFetched = StringListDataFetched();
    _string_dataFetched = StringDataFetched();
    _dataFetched = DataFetched();
    _error = Error();
    _operationStarted = OperationStarted();
    _operationComplete = OperationComplete();
    _installDone = InstallDone();
    _uninstallDone = UninstallDone();
    _downloadProgress.clear();
    _operationProgress.clear();
    _packageListUpdated = UpgradesAvailable();

    pkgmgr.reset(new PackageManager);
    pkgmgr->connect(pkgmgr.get(), SIGNAL(operationStarted(const QString &, const QString &, const QString &)),
    		    this, SLOT(operationStarted(const QString &, const QString &, const QString &)));
    pkgmgr->connect(pkgmgr.get(), SIGNAL(operationComplete(const QString &, const QString &, const QString &, const QString &, const bool)),
    		    this, SLOT(operationComplete(const QString &, const QString &, const QString &, const QString &, const bool)));
    pkgmgr->connect(pkgmgr.get(), SIGNAL(downloadProgress(const QString &, const QString &, const QString &, int, int)),
    		    this, SLOT(downloadProgress(const QString &, const QString &, const QString &, int, int)));
    pkgmgr->connect(pkgmgr.get(), SIGNAL(operationProgress(const QString &, const QString &, const QString &, int)),
    		    this, SLOT(operationProgress(const QString &, const QString &, const QString &, int)));
    pkgmgr->connect(pkgmgr.get(), SIGNAL(packageListUpdated(const bool)),
    		    this, SLOT(packageListUpdated(const bool)));
  }

  static void dumpPkgAttribute(QMap<QString, QVariant> const& pkg,
			       QString const& key)
  {
    QMap<QString, QVariant>::const_iterator i = pkg.find(key);
    if (i != pkg.end())
      qDebug() << i.key() << ": " << i.value();
  }

  static void verifyPackageInfo(QMap<QString, QVariant> const& pkg)
  {
    QList<QString> Type;
    Type << "NotInstalled"
	 << "Installed"
	 << "Update"
	 << "OlderThanInstalled"
	 << "Unknown";

//     QList<QString> keys = pkg.keys();
//     QString key;
//     foreach (key, keys) {
//       qDebug() << key;
//     }

    QVERIFY(pkg.contains("Type"));
    QVERIFY(pkg.find("Type").value().type() == QVariant::String);
    QVERIFY(Type.contains(pkg.find("Type").value().toString()));

    dumpPkgAttribute(pkg, "InstallStatus");
  }

  void verifyOperationProgress(QString const& op, QString const& pkg)
  {
    QVERIFY(_operationProgress.size() > 0);

    QList<OperationProgress>::const_iterator i = _operationProgress.begin();
    if (i != _operationProgress.end()) {
      OperationProgress p = *i;

      QCOMPARE(p.op, op);
      QCOMPARE(p.pkg, pkg);
      QCOMPARE(p.already, 0);

      for (++i; i != _operationProgress.end(); ++i) {
	QCOMPARE(i->op, op);
	QCOMPARE(i->pkg, pkg);
	QVERIFY(i->already >= p.already);
      }
    }
  }

  void dumpSignalCounts()
  {
    qDebug() << _list_dataFetched.n_signals
	     << _dataFetched.n_signals
	     << _operationStarted.n_signals
	     << _operationComplete.n_signals
	     << _installDone.n_signals
	     << _uninstallDone.n_signals
	     << _downloadProgress.size()
	     << _operationProgress.size();
  }

  void runUntil(const char *pkgmgr_signal)
  {
    pkgmgr->connect(pkgmgr.get(), pkgmgr_signal,
		    qApp, SLOT(quit()));
  }

  void _uninstall(const QString &pkgname)
  {
    reset();

    PackageManagerPendingCallWatcher *w = pkgmgr->uninstall(pkgname);
    wait(w);

    verifyOperationProgress("Uninstall", pkgname);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 1);
    QCOMPARE(_operationComplete.n_signals, 1);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QVERIFY(_operationProgress.size() >= 2);
  }

  QString makeFilename(const QString &package, const QString &version) const
  {
    return QString("%1_%2_%3.deb").arg(package).arg(version).arg(DEB_HOST_ARCH);
  }

  QString fullError(const QString &suffix) const
  {
    return QString("%1%2").arg(ERROR_PREFIX).arg(suffix);
  }

  QString addDir(const QString &pkgname) const
  {
    return pool_dir + pkgname;
  }

  QString repoFilename(const QString &pkgname, const QString &version, const QString &section) const
  {
    return QString("%1/pool/%2/%3").arg(testrepo_dir).arg(section).arg(makeFilename(pkgname, version));
  }

  void connectSignals(PackageManagerPendingCallWatcher *w)
  {
    w->connect(w, SIGNAL(dbusError(PackageManagerPendingCallWatcher *)),
	       this, SLOT(dbusError(PackageManagerPendingCallWatcher *)));
  }

  void connectSignals(PackageManagerPendingFetchWatcher *w)
  {
    connectSignals(static_cast<PackageManagerPendingCallWatcher *>(w));

    w->connect(w, SIGNAL(dataFetched(const QList< QMap<QString, QVariant> >&)),
	       this, SLOT(list_dataFetched(const QList< QMap<QString, QVariant> >&)));
    w->connect(w, SIGNAL(dataFetched(const QMap<QString, QVariant>&)),
	       this, SLOT(dataFetched(const QMap<QString, QVariant>&)));
    w->connect(w, SIGNAL(dataFetched(const QStringList&)),
	       this, SLOT(stringList_dataFetched(const QStringList&)));
	w->connect(w, SIGNAL(dataFetched(const QString&)),
	       this, SLOT(string_dataFetched(const QString&)));
  }

  void waitCallFinished(PackageManagerPendingCallWatcher *w, int timeout_ms = 0)
  {
    QEventLoop l;
    w->connect(w, SIGNAL(callFinished(PackageManagerPendingCallWatcher *)),
	       &l, SLOT(quit()));

    wait(l, timeout_ms);
  }

  void wait(PackageManagerPendingCallWatcher *w, int timeout_ms = 0)
  {
    QEventLoop l;
    pkgmgr->connect(pkgmgr.get(), SIGNAL(operationComplete(const QString &, const QString &, const QString &, const QString &, const bool)),
		    &l, SLOT(quit()));
    w->connect(w, SIGNAL(dbusError(PackageManagerPendingCallWatcher *)),
	       &l, SLOT(quit()));

    wait(l, timeout_ms);
  }

  void waitDownloadProgress(PackageManagerPendingCallWatcher *w, int timeout_ms = 0)
  {
    QEventLoop l;
    l.connect(pkgmgr.get(), SIGNAL(downloadProgress(const QString &, const QString &, const QString &, int, int)),
		    &l, SLOT(quit()));
    l.connect(w, SIGNAL(dbusError(PackageManagerPendingCallWatcher *)),
	      &l, SLOT(quit()));

    wait(l, timeout_ms);
  }

  void wait(PackageManagerPendingFetchWatcher *w, int timeout_ms = 0)
  {
    QEventLoop l;
    w->connect(w, SIGNAL(dbusError(PackageManagerPendingCallWatcher *)),
	       &l, SLOT(quit()));
    w->connect(w, SIGNAL(dataFetched(const QList< QMap<QString, QVariant> >&)),
	       &l, SLOT(quit()));
    w->connect(w, SIGNAL(dataFetched(const QMap<QString, QVariant>&)),
	       &l, SLOT(quit()));
    w->connect(w, SIGNAL(dataFetched(const QStringList&)),
	       &l, SLOT(quit()));
	w->connect(w, SIGNAL(dataFetched(const QString&)),
	       &l, SLOT(quit()));

    wait(l, timeout_ms);
  }

  void wait(QEventLoop &l, int timeout_ms)
  {
    QTimer t;
    if (timeout_ms > 0) {
      t.connect(&t, SIGNAL(timeout()),
		&l, SLOT(quit()));
      t.setSingleShot(true);
      t.start(timeout_ms);
    }
    l.exec();
    if (timeout_ms > 0) {
      QVERIFY(t.isActive()); // check no timeout
      t.stop();
    }

    l.processEvents();
  }

private Q_SLOTS:
  void initTestCase()
  {
    testrepo_dir = getenv("HPM_TESTREPO_DIR");
    QVERIFY(! testrepo_dir.isEmpty());
    pool_dir = testrepo_dir + "/pool/user/desktop/";

    reset();
  }

  void cleanupTestCase()
  {
  }

  void fetchPackages_ok()
  {
    reset();

    QString category = "";
    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchPackages(category);
    connectSignals(w);
    wait(w);

    QCOMPARE(_error.name, QString());

    QCOMPARE(_list_dataFetched.n_signals, 1);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);

    QVERIFY(_list_dataFetched.list.size() > 0);

    QMap<QString, QVariant> pkg;
    foreach (pkg, _list_dataFetched.list) {
      verifyPackageInfo(pkg);
    }
  }

  void fetchAvailableUpdates_ok()
  {
    reset();

    QString category = "";
    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchAvailableUpdates(category);
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 1);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
    QCOMPARE(_packageListUpdated.n_signals, 0);
  }

  void fetchInstalled_ok()
  {
    reset();

    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchInstalled();
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 1);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }

  void fetchPackageDataFile_ok()
  {
    reset();

    QString filename = addDir(user1_filename);
    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchPackageDataFile(filename);
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 1);
    QCOMPARE(_stringList_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }

  void fetchPackageDataFile_fail_data()
  {
    QTest::addColumn<QString>("filename");
    QTest::addColumn<QString>("error");

    QTest::newRow("empty filename")
      << QString("") << "InvalidFile";
    QTest::newRow("invalid file")
      << QString("/etc/group") << "InvalidFile";
    QTest::newRow("file not found")
      << QString("/file/not/found") << "FileNotFound";
    QTest::newRow("non-user package")
      << repoFilename("hpm-nonuser", "0.1.0-1", "main/misc") << "PackageNotUser";
  }

  void fetchPackageDataFile_fail()
  {
    reset();

    QFETCH(QString, filename);
    QFETCH(QString, error);

    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchPackageDataFile(filename);
    connectSignals(w);

    while (_error.n_signals == 0)
      QCoreApplication::processEvents();

    QCOMPARE(_error.name, fullError(error));

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 1);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }

  void fetchPackageDataRemote_ok()
  {
    reset();

    QString package = "hpm-user2";
    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchPackageDataRemote(package, user1_version);
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 1);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);

    verifyPackageInfo(_dataFetched.package);
    QCOMPARE(_dataFetched.package.find("Name").value().toString(), package);
    QCOMPARE(_dataFetched.package.find("Version").value().toString(), user1_version);
  }

  void fetchPackageDataInstalled_ok()
  {
    reset();

    QString package = "hpm-user-files";
    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchPackageDataInstalled(package);
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 1);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);

    verifyPackageInfo(_dataFetched.package);
    QCOMPARE(_dataFetched.package.find("Name").value().toString(), package);
    QCOMPARE(_dataFetched.package.find("Version").value().toString(), user1_version);
  }

  void fetchPackageDataInstalled_fail_data()
  {
    QTest::addColumn<QString>("package");
    QTest::addColumn<QString>("error");

    QTest::newRow("package not found")
      << QString("nonexistingpackage") << "PackageNotFound";
    QTest::newRow("package not installed")
      << QString("hpm-user1") << "PackageNotInstalled";
  }

  void fetchPackageDataInstalled_fail()
  {
    reset();

    QFETCH(QString, package);
    QFETCH(QString, error);

    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchPackageDataInstalled(package);
    connectSignals(w);
    wait(w);

    QCOMPARE(_error.name, fullError(error));

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 1);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }

  void fetchPackageDependencyData_ok()
  {
    reset();

    QString package = "hpm-user-valid-deps";
    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchPackageDependencyData(package);
    connectSignals(w);

    while (_list_dataFetched.n_signals == 0) {
        QCoreApplication::processEvents();
    }

    QCOMPARE(_list_dataFetched.n_signals, 1);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }

  void fetchFilesOfInstalledPackage_ok()
  {
    reset();

    QString package = "hpm-user-files";
    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchFilesOfInstalledPackage(package);
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_stringList_dataFetched.n_signals, 1);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }

  void fetch_current_operation_data__no_operation()
  {
    reset();

    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchCurrentOperationData();
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 1);
    QCOMPARE(_stringList_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);

    QCOMPARE(_dataFetched.package.size(), 0);
  }
  
  void fetchRestoreStatus()
  {
    reset();

    PackageManagerPendingFetchWatcher *w = pkgmgr->fetchRestoreStatus();
    connectSignals(w);
    wait(w);

    QFileInfo restore_info ("/var/lib/package-manager/installed-packages.restored");
    QFileInfo restore_info_tmp ("/home/user/.package-manager/installed-packages.restored");
    QString expected;

    if ((restore_info.isFile() && restore_info.size() > 0)
        || (restore_info_tmp.isFile() && restore_info_tmp.size() > 0))
      expected = "RestorableContentAvailable";
    else
      expected = "NoRestorableContent";

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_stringList_dataFetched.n_signals, 0);
    QCOMPARE(_string_dataFetched.n_signals, 1);
    QCOMPARE(_string_dataFetched.str, expected);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }
  
  void simulate_upgrade_ok()
  {
    reset();

    PackageManagerPendingFetchWatcher *w = pkgmgr->simulateOperation(QString("Install"), user1_name, user1_version, QString(""));
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_stringList_dataFetched.n_signals, 0);
    QCOMPARE(_string_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 1);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
  }

  void ignore_update_package_ok()
  {
    reset();

    PackageManagerPendingCallWatcher *w = pkgmgr->ignoreUpdatePackage(user1_name, user1_updateversion);
    connectSignals(w);
    waitCallFinished(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 0);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
    
    delete w;
    
    w = pkgmgr->refreshPackageCache("", "clearignored");
    connectSignals(w);
    wait(w);

    QCOMPARE(_error.name, QString());
    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 1);
    QCOMPARE(_operationStarted.operation, QString("Refresh"));
    QCOMPARE(_operationStarted.pkg, QString(""));
    QCOMPARE(_operationStarted.version, QString(""));
    QCOMPARE(_operationComplete.n_signals, 1);
    QCOMPARE(_operationComplete.operation, QString("Refresh"));
    QCOMPARE(_operationComplete.pkg, QString(""));
    QCOMPARE(_operationComplete.version, QString(""));
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    // QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
    QCOMPARE(_packageListUpdated.n_signals, 1);
  }

  void refresh_cache_ok()
  {
    reset();

    PackageManagerPendingCallWatcher *w = pkgmgr->refreshPackageCache("", "clearignored");
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 1);
    QCOMPARE(_operationStarted.operation, QString("Refresh"));
    QCOMPARE(_operationStarted.pkg, QString(""));
    QCOMPARE(_operationStarted.version, QString(""));
    QCOMPARE(_operationComplete.n_signals, 1);
    QCOMPARE(_operationComplete.operation, QString("Refresh"));
    QCOMPARE(_operationComplete.pkg, QString(""));
    QCOMPARE(_operationComplete.version, QString(""));
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    // QCOMPARE(_downloadProgress.size(), 0);
    QCOMPARE(_operationProgress.size(), 0);
    QCOMPARE(_packageListUpdated.n_signals, 0);
  }

  void install_ok()
  {
    reset();

    PackageManagerPendingCallWatcher *w = pkgmgr->install(user1_name);
    connectSignals(w);
    wait(w);

    verifyOperationProgress("Install", user1_name);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 1);
    QCOMPARE(_operationStarted.operation, QString("Install"));
    QCOMPARE(_operationComplete.n_signals, 1);
    QCOMPARE(_operationComplete.operation, QString("Install"));
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QVERIFY(_downloadProgress.size() >= 0);
    QVERIFY(_operationProgress.size() >= 2);

    _uninstall(user1_name);
  }

  void install_fail_data()
  {
    QTest::addColumn<QString>("package");
    QTest::addColumn<QString>("error");

    QTest::newRow("package not found")
      << QString("nonexistingpackage") << "PackageNotFound";
    // QTest::newRow("package not installed")
    //   << QString("hpm-user1") << "PackageNotInstalled";
  }

  void install_fail()
  {
    reset();

    QFETCH(QString, package);
    QFETCH(QString, error);

    PackageManagerPendingCallWatcher *w = pkgmgr->install(package);
    connectSignals(w);
    wait(w);

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    // QCOMPARE(_error.n_signals, 1);
    QCOMPARE(_operationStarted.n_signals, 0);
    QCOMPARE(_operationComplete.n_signals, 1);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QVERIFY(_downloadProgress.size() >= 0);
    QVERIFY(_operationProgress.size() >= 0);
  }

  void installFile_ok()
  {
    reset();

    PackageManagerPendingCallWatcher *w = pkgmgr->installFile(addDir(user1_filename));
    connectSignals(w);
    wait(w);

    verifyOperationProgress("InstallFile", addDir(user1_filename));

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 1);
    QCOMPARE(_operationComplete.n_signals, 1);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QVERIFY(_downloadProgress.size() >= 0);
    QVERIFY(_operationProgress.size() >= 2);

    _uninstall(user1_name);
  }

  void cancel_ok()
  {
    reset();

    system("stop-webserver.sh");
    system("start-webserver.sh 100");

    QString op = "Install";
    QString p = "hpm-user1";
    QString v = "0.1.0-1";
    PackageManagerPendingCallWatcher *w = pkgmgr->install(p, v);
    connectSignals(w);
    waitDownloadProgress(w);

    QVERIFY(_downloadProgress.size() >= 1);

    PackageManagerPendingCallWatcher *w2 = pkgmgr->cancelOperation(op, p, v);
    waitCallFinished(w2);

    wait(w);

    QCOMPARE(_operationComplete.result, fullError("OperationCanceled"));

    system("stop-webserver.sh");
    system("start-webserver.sh");
  }

  void install_long_ok()
  {
    reset();

    com::nokia::package_manager pm(PKGMGR_DBUS_SERVICE, PKGMGR_DBUS_PATH, QDBusConnection::systemBus(), NULL);
    pm.callTimeout = 10*1000;
    QDBusPendingReply<Package> reply = pm.fetch_current_operation_data();
    Package op = reply.value();
    QCOMPARE(op.p.size(), 0); // no operation

    QString pkg = "hpm-long-install"; // takes >= 40 s
    pkgmgr->callTimeout = 90*1000; // 90 s
    PackageManagerPendingCallWatcher *w = pkgmgr->install(pkg);
    connectSignals(w);

    QEventLoop l;
    pkgmgr->connect(pkgmgr.get(), SIGNAL(operationStarted(const QString &, const QString &, const QString &)),
		    &l, SLOT(quit()));
    wait(l, 10*1000);

    reply = pm.fetch_current_operation_data();
    op = reply.value();

    QCOMPARE(op.p["Operation"].toString(), QString("Install"));
    QCOMPARE(op.p["Name"].toString(), pkg);
    QCOMPARE(op.p["Version"].toString(), QString("0.1.0-1"));

    wait(w); // wait operation_complete

    verifyOperationProgress("Install", pkg);

    QCOMPARE(_error.name, QString());

    QCOMPARE(_list_dataFetched.n_signals, 0);
    QCOMPARE(_dataFetched.n_signals, 0);
    QCOMPARE(_error.n_signals, 0);
    QCOMPARE(_operationStarted.n_signals, 1);
    QCOMPARE(_operationComplete.n_signals, 1);
    QCOMPARE(_installDone.n_signals, 0);
    QCOMPARE(_uninstallDone.n_signals, 0);
    QVERIFY(_downloadProgress.size() >= 0);
    QVERIFY(_operationProgress.size() >= 2);

    _uninstall(pkg);
  }

protected Q_SLOTS:
  void list_dataFetched(const QList< QMap<QString, QVariant> >& list)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << list;
    }
    ++_list_dataFetched.n_signals;
    _list_dataFetched.list = list;
  }

  void stringList_dataFetched(const QStringList & list)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << list;
    }
    ++_stringList_dataFetched.n_signals;
    _stringList_dataFetched.list = list;
  }
  
  void string_dataFetched(const QString & str)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << str;
    }
    ++_string_dataFetched.n_signals;
    _string_dataFetched.str = str;
  }

  void dataFetched(const QMap<QString, QVariant>& package)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << package;
    }
    ++_dataFetched.n_signals;
    _dataFetched.package = package;
  }

  void dbusError(PackageManagerPendingCallWatcher *watcher)
  {
    ++_error.n_signals;
    _error.name = watcher->errorName();
    _error.message = watcher->errorMessage();
  }

  void operationStarted(const QString &operation, const QString &pkg, const QString &version)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << operation << pkg;
    }
    ++_operationStarted.n_signals;
    _operationStarted.operation = operation;
    _operationStarted.pkg = pkg;
    _operationStarted.version = version;
  }

  void operationComplete(const QString &operation, const QString &pkg, const QString &version, const QString &result, const bool need_reboot)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << operation << pkg << result;
    }
    ++_operationComplete.n_signals;
    _operationComplete.operation = operation;
    _operationComplete.pkg = pkg;
    _operationComplete.version = version;
    _operationComplete.result = result;
    _operationComplete.need_reboot = need_reboot;
  }

  void installDone()
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__;
    }
    ++_installDone.n_signals;
  }

  void uninstallDone()
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__;
    }
    ++_uninstallDone.n_signals;
  }

  void downloadProgress(const QString &operation, const QString &packagename, const QString &version, int already, int total)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << already << total;
    }
    DownloadProgress p;
    p.operation = operation;
    p.package = packagename;
    p.version = version;
    p.already = already;
    p.total = total;
    _downloadProgress.append(p);
  }

  void operationProgress(const QString &op, const QString &pkg, const QString &version, int already)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__ << op << pkg << already;
    }
    OperationProgress p;
    p.op = op;
    p.pkg = pkg;
    p.version = version;
    p.already = already;
    _operationProgress.append(p);
  }

  void packageListUpdated(const bool updates_changed)
  {
    if (DUMP_SIGNALS) {
      qDebug() << __func__;
    }
    ++_packageListUpdated.n_signals;
    _packageListUpdated.updates_changed = updates_changed;
  }

private:
  QString testrepo_dir;
  QString pool_dir;
  std::auto_ptr<PackageManager> pkgmgr;
  QString user1_name;
  QString user1_version;
  QString user1_updateversion;
  QString user1_filename;

  struct Signal
  {
    Signal() : n_signals() { }
    int n_signals;
  };

  struct ListDataFetched : Signal
  {
    QList< QMap<QString, QVariant> > list;
  } _list_dataFetched;

  struct StringListDataFetched : Signal
  {
    QStringList list;
  } _stringList_dataFetched;
  
  struct StringDataFetched : Signal
  {
    QString str;
  } _string_dataFetched;

  struct DataFetched : Signal
  {
    QMap<QString, QVariant> package;
  } _dataFetched;

  struct Error : Signal
  {
    QString name;
    QString message;
  } _error;

  struct OperationStarted : Signal
  {
    QString operation;
    QString pkg;
    QString version;
  } _operationStarted;
    
  struct OperationComplete : Signal
  {
    QString operation;
    QString pkg;
    QString version;
    QString result;
    bool need_reboot;
  } _operationComplete;

  struct InstallDone : Signal
  {
  } _installDone;

  struct UninstallDone : Signal
  {
  } _uninstallDone;

  struct DownloadProgress : Signal
  {
    DownloadProgress() : operation(""), package(""), already(), total() { }
    QString operation;
    QString package;
    QString version;
    int already;
    int total;
  };
  QList<DownloadProgress> _downloadProgress;
  
  struct OperationProgress : Signal
  {
    OperationProgress() : already() { }
    QString op;
    QString pkg;
    QString version;
    int already;
  };
  QList<OperationProgress> _operationProgress;

  struct UpgradesAvailable : Signal
  {
    bool updates_changed;
  } _packageListUpdated;
};

QTEST_MAIN(TestPkgmgrDbus)
// #include "test-libpkgmgr-dbus-qt.moc.cc"
