#!/usr/bin/python
#
# Copyright (C) 2009 - 2011 Nokia Corporation.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version
# 2 as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

import getopt
import sys
import os
import subprocess
import shlex
import time
import re
import copy


tools = {
            "massif" : ["--tool=massif"], 
            "memcheck" : ["--tool=memcheck"], 
            "callgrind" : ["--tool=callgrind"],
            "response" : None
        }


pkgmgrd_tmp_dir = os.path.join(os.getenv("TMP", "/var/tmp"), "pkgmgrd_nft/")

massif_file_prefix = "pkgmgrd_massif_"

tool_files = {"massif" : os.path.join(pkgmgrd_tmp_dir, massif_file_prefix + "%p"), 
              "memcheck" : os.path.join(pkgmgrd_tmp_dir, "pkgmgrd_memcheck.xml"),
              "response" : os.path.join(pkgmgrd_tmp_dir, "pkgmgrd_response")
             }

valgrind_opts = { "memcheck" : {"--show-below-main" : "no", 
                                "--leak-check" : "full", 
                                "--leak-resolution" : "high", 
                                "--num-callers" : "40", 
                                "--xml" : "yes", 
                                "--xml-file" : tool_files["memcheck"], 
                                "--child-silent-after-fork" : "yes", 
                                "-q" : None, 
                                "--log-file" : "memcheck_errors"
                                },  
                  "massif" : { "--heap" : "yes", 
                              #"--peak-inaccuracy" : "5.0", 
                               "--massif-out-file" : tool_files["massif"], 
                               "--max-snapshots" : "100",
                               "--stacks" : "yes",
                               "--threshold" : "10.0" # less than 10.0 % of totals are aggretated (less output)
                               }, 
                  "callgrind" : { }

                } # valgrind_opts is tool : options dicitonary. Dict is used so that user can override these in cmd line
                
dont_run_pkgmgrd = False
verbose = False
fakeroot = ""
scratchbox = False
do_cleanup = True
ms_print = False

#       10        20        30        40        50        60        70        80
#2345678901234567890123456789012345678901234567890123456789012345678901234567890
__doc__ = """
SYNOPSIS
    run_test.py [-t tool] [-r number] [-h] [-v "valgrind_opts"]
                [-o FILE] [-e] [-V] [-p "COMMAND"] [-P "COMMAND"]
                [-d] [-n] [-m] [-T number] [-M FILE] ["arguments"]
    
DESCRIPTION
    Runs pkgmgrd and then the command given as argument. After the command has
    exited the script stops the pkgmgrd. The script runs pkgmgrd with valgrind 
    tools (--tool) switch. If response times are measured the daemon is run
    without valgrind.   
    
OPTIONS
   -t, --tool=TOOL            MANDATORY. One of: """ + ", ".join(tools) + """
   -r, --runs=RUNS            Optional (default is 1). The number of runs to 
                              perform.
   -h, --help                 Prints this help
   -v, --valgrind-opts="OPTS" Options for valgrind in double quotes (overrides
                              defaults)
   -o, --output-file=FILE     Redirect the stdout of the command to the
                              specified file.
   -e, --redirect-stderr      Redirect also the stderr of the command to the
                              output-file. If no --output-file has been 
                              specified then this option has no effect.
   -V, --verbose              Prints out pkgmgrd messages and other debug info
                              to stdout and stderr. 
   -p, --pre-cmd="COMMAND"    Command to run before the test command will be
                              run, in double quotes. The command is run right
                              after starting pkgmgrd.
   -P, --post-cmd="COMMAND"   Command to run after the test command has been 
                              run, in double quotes. The command is run before
                              stopping pkgmgrd.  
   -d, --dont-run-pkgmgrd     Don't run pkgmgrd (works only in HW environment!). 
                              This means that DBUS will initiate the pkgmgrd on 
                              demand. It will affect response times. This should
                              be used only when measuring 'worst case scenarios'
   -n, --no-cleanup           Do not clean up generated test data from 
                              $TMP/pkgmgrd_nft directory. Useful if you want to
                              examine the data with other tools.
   -m, --ms_print             When tool=massif, run ms_print for every 
                              generated massif output file. Prints the contents
                              to stdout. 
   -T, --target-value         Target value in milliseconds for response time 
                              tool. 
   -M, --memory-output-file   Write memory usage stats to the specified file.

ARGUMENTS
   "command"                  Command to run in double quotes. Example:
                              "pkgmgr list-available"

"""

# TODO: Make this non-pkgmgrd dependant tool. Provide the measured tool as an option.

def check_tool(tool):
    if (tool in tools):
        return True
    else:
        return False



def parse_command_line():
    """
    Parses command line options and arguments. Exists if options or arguments are erroneous.
    Returns parsed variables as tuple.
    @return: tool: Tool to use, overriding_valgrind_opts: The overriding valgrind options,
    runs: The times to run, command: Command to run, output_file: File where stdout of the command is written, redirect_stderr: If specified, stderr is also written to stdout or output_file
    """
    global verbose, dont_run_pkgmgrd, do_cleanup, ms_print

    try:
        opts, args = getopt.getopt(sys.argv[1:], "ht:r:v:o:eVp:P:dnT:M:", ["help", "tool=", "runs=",
                "valgrind-opts=", "output-file=", "redirect-stderr", "verbose", "pre-cmd=", "post-cmd=",
                "dont-run-pkgmgrd", "no-cleanup", "target-value=", "memory-output-file="])
    except getopt.error, msg:
        print msg
        print "For help, use --help or -h"
        sys.exit(2)
    tool = None
    overriding_valgrind_opts = None
    output_file = None
    redirect_stderr = False
    runs = 1
    pre_cmd = None
    post_cmd = None
    tool_options = { }
    memory_output_file = None
    
    for o, a in opts:
        if o in ("-h", "--help"):
            print __doc__
            sys.exit(0)
        elif o in ("-t", "--tool"):
            tool = a
            if not check_tool(tool):
                print >> sys.stderr, "Sorry, no support for tool '%s'!" % ( tool )
                sys.exit(1)
        elif o in ("-r", "--runs"):
            try:
                runs = int(a)
            except ValueError:
                print >> sys.stderr, "Invalid parameter value given for 'runs'-parameter."
                sys.exit(2)
        elif o in ("-v", "--valgrind-opts"):
            overriding_valgrind_opts = dict([(partition[0], partition[2]) for 
                    partition in [item.partition("=") for item in a.split(" ")]])
        elif o in ("-o", "--output-file"):
            output_file = a
        elif o in ("-e", "--redirect-stderr"):
            redirect_stderr = True
        elif o in ("-V", "--verbose"):
            verbose = True
        elif o in ("-p", "--pre-cmd"):
            pre_cmd = a
        elif o in ("-P", "--post-cmd"):
            post_cmd = a
        elif o in ("-d", "--dont-run-pkgmgrd"):
            dont_run_pkgmgrd = True
        elif o in ("-n", "--no-cleanup"):
            do_cleanup = False
        elif o in ("-m", "--ms_print"):
            ms_print = True
        elif o in ("-T", "--target-value"):
            try:
                target_value = int(a)
            except ValueError:
                print >> sys.stderr, "Invalid parameter value given for 'target-value'-parameter."
                sys.exit(2)
                
            tool_options.update( {"target-value" : target_value } )
        elif o in ("-M", "--memory-output-file"):
            memory_output_file = a
    
    if len(args) != 1:
        print >> sys.stderr, "Invalid number of arguments. For help, use --help or -h."
        sys.exit(2)
    if not tool:
        print >> sys.stderr, "No tool was given. For help, use --help or -h"
        sys.exit(2)

    command = args[0]
    return tool, overriding_valgrind_opts, runs, command, pre_cmd, post_cmd, output_file, redirect_stderr, tool_options, memory_output_file


def run_tests(tool, overriding_valgrind_opts, runs, command, pre_cmd, post_cmd, output_file, redirect_stderr, tool_options, memory_output_file):
    global verbose
    
    returncode = 0
    if not dont_run_pkgmgrd:
#        if tool == "massif": 
#            overriding_valgrind_opts.update("--massif-out-file", )
        run_pkgmgrd(tool, overriding_valgrind_opts) # Run the Package manager daemon 
        time.sleep(3) # FIXME: Make a better solution to check whether the daemon has initialized itself and is running
    
    if pre_cmd:
        if verbose: print "Running pre-test command..."
        preretcode = run_subprocess(pre_cmd, "", wait=True, stdout_filename=output_file, redirect_stderr=redirect_stderr)
        if preretcode != 0: 
            print >> sys.stderr, "Pre-test command returned non-zero return code: %d" % (preretcode)
            return preretcode
        
    if verbose: print "Running test command..."
    try:
        if tool == "response":
            #file = open(tool_files[tool], "w")
            target_value = tool_options.get("target-value", 0) # test case will fail if option is not specified
            from datetime import datetime
            
            procname = "pkgmgrd"
            memsizecmd = "cat /proc/`pidof %s`/stat|cut -d \" \" -f 23|tr -d \"\n\" >> %s; echo -n \" \" >> %s" % (procname, memory_output_file, memory_output_file)
            if memory_output_file:
                os.system("rm -f %s" % memory_output_file)

            # Print initial test process memory usage
            if memory_output_file:
                os.system(memsizecmd)

            print "Total runs: %d" % runs
            total = 0
            for i in range(0, runs): # Run the command n times that the user has specified
                start = datetime.now()
                returncode = run_subprocess(command, "", wait=True, stdout_filename=output_file, redirect_stderr=redirect_stderr)
                if returncode != 0: 
                    print >> sys.stderr, "Test command returned non-zero return code: %d" % (returncode)
                    return returncode
                delta = (datetime.now() - start)
                elapsed = (delta.days * 24 * 60 * 60 * 1000) + (delta.seconds * 1000) + (delta.microseconds / 1000) # calculate difference in milliseconds
                total += elapsed
                print "Run %d duration: %dms" % (i + 1, elapsed)
                # Print test process memory usage into a file after every testrun
                if memory_output_file:
                    os.system(memsizecmd)

            if memory_output_file:
                os.system("echo \"\" >> %s" % memory_output_file)

            average = (total / runs)
            if runs > 1:
                print "Average duration: %dms" % average

            if average > target_value:
                message = "Target value of %dms was exceeded. Actual result was %dms." % (target_value, average)
                print >> sys.stderr, message
                returncode = 1

            #file.close()
        else:
            for i in range(0, runs): # Run the command n times that the user has specified
                returncode = run_subprocess(command, "", wait=True, stdout_filename=output_file, redirect_stderr=redirect_stderr)
                if returncode != 0: 
                    print >> sys.stderr, "Subprocess returned non-zero return code: %d" % (returncode)
                    return returncode

        if post_cmd:
            if verbose: print "Running post-test command..."
            postretcode = run_subprocess(post_cmd, "", wait=True, stdout_filename=output_file, redirect_stderr=redirect_stderr)
            if postretcode != 0: 
                print >> sys.stderr, "Post-test command returned non-zero return code: %d" % (postretcode)
                return postretcode
    except Exception, e:
        import traceback
        print >> sys.stderr, "Error occurred: %s" % (traceback.format_exc (e))
    finally:
        stop_pkgmgrd()
        return returncode
    
def clean_up(tool, initial=False):
    global verbose, do_cleanup
    if tool == "massif":
        files = search_massif_data_files()
        for file in files:
            if do_cleanup or initial:
                if verbose: print "Removing file %s" % (file)
                try: 
                    os.remove(file)
                except OSError, e:
                    pass # okay, it doesn't exist.
                
            else:
                #if the result files are kept, we must rename them so that they will not get processed again
                splitted_path = os.path.split(file)
                splitted_filename = splitted_path[1].split(massif_file_prefix)
                newpath = os.path.join(splitted_path[0], "pkgmgrd_processed_massif_%s" % (splitted_filename[1]))
                if verbose: print "Renaming file '%s' to '%s'" % (file, newpath) 
                os.rename(file, newpath)
    elif do_cleanup or initial:
        try: 
            os.remove(tool_files[tool])
        except OSError, e:
            pass # okay, it doesn't exist.



def main():
    global fakeroot, scratchbox, do_cleanup, pkgmgrd_tmp_dir
    tool, overriding_valgrind_opts, runs, command, pre_cmd, post_cmd, output_file, redirect_stderr, tool_options, memory_output_file = parse_command_line()
    
    if not os.path.exists(pkgmgrd_tmp_dir):
        os.mkdir(pkgmgrd_tmp_dir)
    
    clean_up(tool, initial=True) # clean up previous results for this tool so that they do not interfere  

    returncode = 0
    
    try:        
        
        if os.getenv("SBOX_PRELOAD"):
            scratchbox = True # We're in scratchbox environment
            os.putenv("LD_PRELOAD", os.getenv("SBOX_PRELOAD")) # Add LD_PRELOAD from environment to python env
            fakeroot = "fakeroot "
        
        returncode = run_tests(tool, overriding_valgrind_opts, runs, command, pre_cmd, post_cmd, output_file, redirect_stderr, tool_options, memory_output_file)
        
        if returncode == 0:        
            analyze_results(tool) 
    except Exception, e:
        import traceback
        print >> sys.stderr, "Error occurred: %s" % (traceback.format_exc (e))
    finally:
        clean_up(tool)

        if returncode is not None:
            return returncode
        else:
            print >> sys.stderr, "No return code from scripts. Returning with abnormal return code (1)."
            return 1 # error

    

def run_subprocess(path, args, wait = False,  stdout_filename = None, redirect_stderr = False):
    """
    Runs a command with subprocess.Popen. Returns the Popen object. 
    If wait is True then runs the command with subprocess.call
    """
    global verbose
    command = path
    stdout_file = sys.stdout
    stderr_file = sys.stderr
    
    if stdout_filename:
        stdout_file = open(stdout_filename, "w")
    
    if redirect_stderr:
        stderr_file = stdout_file
        
    if args and len(args) > 0: 
        command += " %s" % (" ".join(args))
    
    command = shlex.split(command)
    if verbose: print " ".join(command) #FIXME: Debug-code. Remove this when not needed anymore
    
    if wait:
        return subprocess.call(command, stdout=stdout_file, stderr=stderr_file, env=os.environ)
    else:
        return subprocess.Popen(command, stdout=stdout_file, stderr=stderr_file)

def run_pkgmgrd(tool, overriding_valgrind_opts):
    global verbose
    stdout_filename = None
    redirect_stderr = None
    if not verbose:
        stdout_filename='/dev/null'
        redirect_stderr = True
        
        
    if tool == "response":
        if verbose: print "Running pkgmgrd..." 
        run_subprocess(fakeroot + "pkgmgrd", None, stdout_filename = stdout_filename, redirect_stderr = redirect_stderr)
    else:
        if verbose: print "Running pkgmgrd with valgrind..."
        valgrind_args = generate_valgrind_arguments(tool, overriding_valgrind_opts, ["pkgmgrd"])
        run_subprocess(fakeroot + "valgrind", valgrind_args, stdout_filename=stdout_filename, redirect_stderr = redirect_stderr)

def stop_pkgmgrd():
    global verbose
    try:
        f = open("/var/lock/LCK..pkgmgrd.lock")
        pid = f.read()
        f.close()
        if verbose: print "Stopping pkgmgrd..."
        run_subprocess("kill", [pid], wait=True)
        while os.path.exists("/var/lock/LCK..pkgmgrd.lock"):
            time.sleep(0.1)
    except IOError:
        pass


def generate_valgrind_arguments(tool, overriding_valgrind_opts, args):
    """
    Generates the argument list for given tool and given overriding options
    """
    valgrind_arguments = tools[tool]
    valopts = valgrind_opts[tool]
    if overriding_valgrind_opts is not None:
        valopts.update(overriding_valgrind_opts)
        
    valopts_list = []
    for key,  value in valopts.iteritems():
        if not value or value == '': 
            valopts_list.append(key)
        else: 
            valopts_list.append("%s=%s" % (key,  value))
            
    valgrind_arguments.extend(valopts_list)
    valgrind_arguments.extend(args)
    
    return valgrind_arguments
    

class SnapshotItem:
    """
    Represents a snapshot item in massif output file. See valgrind manual for details.
    """
    item = 0
    time = 0
    mem_total_heap_B = 0
    mem_useful_heap_B = 0
    mem_heap_extra_B = 0
    mem_stacks_B = 0
    heap_tree = ""
   
    def __str__(self):
        return "(item=%d, time=%d, mem_total_heap_B=%d, mem_useful_heap_B=%d, mem_heap_extra_B=%d, mem_stacks_B=%d, heap_tree=%s)" % (self.item,  self.time, self.mem_total_heap_B, self.mem_useful_heap_B, self.mem_heap_extra_B, self.mem_stacks_B, self.heap_tree)
       

def search_massif_data_files():
    global verbose, pkgmgrd_tmp_dir
    data_files = []
    
    for path in os.listdir(pkgmgrd_tmp_dir):
        if path.startswith(massif_file_prefix):
            data_files.append(os.path.join(pkgmgrd_tmp_dir, path))
    
    return data_files
            
"""
Parses massif data and returns dictionary which contains file-specific lists of Snapshot items

"""
def parse_massif_data():
    global verbose, do_cleanup
    
    file_snapshots = {} 
    data_files = search_massif_data_files()
    for file in data_files:
        if verbose: print "Parsing massif output file %s..." % file        
        f = open(file)
        massif_data = f.read()
        f.close()
            
        massif_data = massif_data.replace("#-----------\n", "") # remove separators. We don't need them
        snapshot_re = re.compile("(snapshot=\d{0,3}\n)")
        massif_snapshots = snapshot_re.split(massif_data)  # split data by snapshot identifiers
        
        snapshots = []
        snapshot = SnapshotItem()
        item = -1 # The data before snapshot=0
        
        for massif_snapshot in massif_snapshots:
            if snapshot_re.match(massif_snapshot):
                
                if item != -1:
                    # Item is complete. Add it to list
                    snapshots.append(snapshot)
                    snapshot = SnapshotItem()
                
                item = int(massif_snapshot.split('=')[1])
                snapshot.item = item
            elif item != -1:
                lines = massif_snapshot.split("\n")
                snapshot.time = int(lines[0].split('=')[1])
                snapshot.mem_total_heap_B = int(lines[1].split('=')[1]) 
                snapshot.mem_heap_extra_B = int(lines[2].split('=')[1])
                snapshot.mem_useful_heap_B = snapshot.mem_total_heap_B - snapshot.mem_heap_extra_B
                snapshot.mem_stacks_B = int(lines[3].split('=')[1])
                snapshot.heap_tree = lines[4].split('=')[1]
                if (snapshot.heap_tree != "empty"):
                    snapshot.heap_tree += "\n" + "\n".join(lines[5:])
        
        snapshots.append(snapshot)
        file_snapshots.update({file : snapshots})

    return file_snapshots


def analyze_results(tool):
    if tool == "massif":
        peak_process_snapshot = None
        parsed_data = parse_massif_data()
        for path, snapshots in parsed_data.iteritems(): 
            peak_snapshot = SnapshotItem()
            avg_snapshot= SnapshotItem()
            totals_snapshot = SnapshotItem()
            last_snapshot = snapshots[len(snapshots) - 1]
            
            # analyze peak mem usage
            for snapshot in snapshots:
                if (snapshot.mem_useful_heap_B > peak_snapshot.mem_useful_heap_B):
                    peak_snapshot = snapshot
                
                if peak_snapshot and (not peak_process_snapshot or peak_snapshot.mem_useful_heap_B > peak_process_snapshot.mem_useful_heap_B):
                    peak_process_snapshot = copy.deepcopy(peak_snapshot)
                    peak_process_snapshot.item = path
                    
                totals_snapshot.mem_total_heap_B += snapshot.mem_total_heap_B
                totals_snapshot.mem_useful_heap_B += snapshot.mem_useful_heap_B
                totals_snapshot.mem_heap_extra_B += snapshot.mem_heap_extra_B
                totals_snapshot.mem_stacks_B += snapshot.mem_stacks_B
            
            avg_snapshot.mem_total_heap_B = totals_snapshot.mem_total_heap_B / len(snapshots)
            avg_snapshot.mem_useful_heap_B = totals_snapshot.mem_useful_heap_B / len(snapshots)
            avg_snapshot.mem_heap_extra_B = totals_snapshot.mem_heap_extra_B / len(snapshots)
            avg_snapshot.mem_stacks_B = totals_snapshot.mem_stacks_B / len(snapshots)
            
            print "Summary for file: %s" % path
            #TODO: HTML output to file?
            
            print "  Total amount of snapshots: %d" % (len(snapshots))
            print "  Peak snapshot index: %d. Peak heap total bytes: %d kB. " \
                  "Peak useful bytes: %d kB. Peak heap extra bytes: %d kB. " \
                  "Peak stacks: %d kB " % (peak_snapshot.item, peak_snapshot.mem_total_heap_B / 1024,  
                                                 peak_snapshot.mem_useful_heap_B / 1024,
                                                 peak_snapshot.mem_heap_extra_B / 1024,  
                                                 peak_snapshot.mem_stacks_B / 1024)
                                                    
            print "  Last heap total bytes: %d kB. Last heap useful bytes: %d kB. " \
                  "Last heap extra bytes: %d kB. " \
                  "Last stacks: %d kB" % (last_snapshot.mem_total_heap_B / 1024,  
                                          last_snapshot.mem_useful_heap_B / 1024,
                                          last_snapshot.mem_heap_extra_B / 1024,  
                                          last_snapshot.mem_stacks_B / 1024)
            print "  Average heap total bytes: %d kB. Average heap useful bytes: %d kB. " \
                  "Average heap extra bytes: %d kB. Average stacks: %d KB " % (avg_snapshot.mem_total_heap_B / 1024,
                                                    avg_snapshot.mem_useful_heap_B / 1024, 
                                                    avg_snapshot.mem_heap_extra_B / 1024,  
                                                    avg_snapshot.mem_stacks_B / 1024)
                  
            
            if ms_print:
                print "  ms_print output:" # begin/end "tags" if someone wants to strip this data. 
                sys.stdout.flush()
                run_subprocess("ms_print", [path], wait=True)
                print "  ms_print output ends."             
                  
    
        if peak_process_snapshot and len(parsed_data) > 1:
            print "Peak process data file was: %s" % peak_process_snapshot.item
            print "  Peak heap total bytes: %d kB. Peak heap useful bytes: %d kB. " \
                  "Peak heap extra bytes: %d kB. " \
                  "Peak stacks: %d kB " % (peak_process_snapshot.mem_total_heap_B / 1024,
                                           peak_process_snapshot.mem_useful_heap_B / 1024,  
                                           peak_process_snapshot.mem_heap_extra_B / 1024,  
                                           peak_process_snapshot.mem_stacks_B / 1024)
        
        
    elif tool == "response":
        pass #disabled the file usage for now
        #Only print the output.
        #file = open(tool_files[tool])
        #for line in file.readlines():
        #    print line,
        #file.close()

if __name__ == "__main__":
    sys.exit(main())
