/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include "netconnection.h"
#include <apt-pkg/configuration.h>
#include <apt-pkg/init.h>
#include <cassert>
#include <cctype>
#include <cstdio>
#include <cstdlib>
#include <dbus/dbus.h>
#include <dbus/dbus-glib-bindings.h>
#include <dbus/dbus-glib-lowlevel.h>
#include <termios.h>

struct termios oldt, newt;
static bool canonical_mode = false;

static void
canonical_mode_off ()
{
  tcgetattr (STDIN_FILENO, &oldt);
  newt = oldt;
  newt.c_lflag &= ~(ICANON /* | ECHO */ );
  tcsetattr (STDIN_FILENO, TCSANOW, &newt);
  canonical_mode = true;
}

static void
canonical_mode_on ()
{
  if (canonical_mode)
    {
      tcsetattr (STDIN_FILENO, TCSANOW, &oldt);
      canonical_mode = false;
    }
}

static int
getch ()
{
  canonical_mode_off ();
  int ch = getchar ();
  canonical_mode_on ();

  return ch;
}

static bool
dbus_connection_init (GMainContext *context)
{
  DBusConnection *conn;
  DBusError err;

  dbus_error_init (&err);
  conn = dbus_bus_get (DBUS_BUS_SYSTEM, &err);
  if (!conn)
    {
      fprintf (stderr, "dbus_bus_get() failed: %s: %s\n",
	       err.name, err.message);
      dbus_error_free (&err);
      return false;
    }

  dbus_connection_setup_with_g_main (conn, context);

  return true;
}

static gpointer
run_main_loop (gpointer data)
{
  g_main_loop_run (static_cast<GMainLoop *>(data));

  return NULL;
}

static void log_handler (const gchar *log_domain,
			 GLogLevelFlags log_level,
			 const gchar *message,
			 gpointer unused_data)
{
  printf ("%s\n", message);
}

int
main (int argc, char *argv[])
{
  printf ("Test NetConnection state machine\n");

  g_log_set_handler (NULL,
		     GLogLevelFlags (G_LOG_LEVEL_ERROR | G_LOG_LEVEL_CRITICAL
				     | G_LOG_LEVEL_WARNING |
				     G_LOG_LEVEL_MESSAGE | G_LOG_LEVEL_INFO |
				     G_LOG_LEVEL_DEBUG | G_LOG_FLAG_FATAL |
				     G_LOG_FLAG_RECURSION), log_handler,
		     NULL);

  g_debug ("g_debug");

  atexit (canonical_mode_on);

  g_type_init ();
  g_thread_init (NULL);
  dbus_g_thread_init ();

  // Create main loop as backend does
  GMainContext *context = g_main_context_new ();
  GMainLoop *loop = g_main_loop_new (context, FALSE);
  g_thread_create (run_main_loop, loop, TRUE, NULL);

  // GMainContext *context = g_main_loop_get_context (loop);
  assert (dbus_connection_init (context));

  // pkgInitConfig(*_config);
  _config->Set ("quiet", 0);
  _config->Set ("Debug::NetConnection", true);

  NetConnection conn (context);
  NetConnection::ConnectionStatus status;

  for (;;)
    {
      printf ("C - Connect (automatic)\n"
	      "c - Connect (manual)\n"
	      "d - Disconnect\n" "e - Exit\n" "Press [Ccde]: ");
      fflush (stdout);
      int c = getch ();
      puts ("");

      switch (c)
	{
	case 'C':
	  status = conn.connect (true);
	  break;

	case 'c':
	  status = conn.connect (false);
	  break;

	case 'd':
	  conn.disconnect ();
	  break;

	case 'E':
	case 'e':
	  return 0;

	default:
	  break;
	}
    }

  // puts ("Stopping...");
  // conn.stop ();
  // puts ("Stopped");

  return 0;
}
