#!/usr/bin/python
#
# Copyright (C) 2009 - 2011 Nokia Corporation.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version
# 2 as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

#
# Make .deb package for testing.

import commands
import errno
import sys
import getopt
import os
import string
import shutil

workdir = '.'

class Package:
    def __init__(self, package, version, preinst_delay=0, postinst_delay=0, prerm_delay=0, postrm_delay=0):
        self.fields = {'Architecture': 'i386',
                       'Section': 'user/desktop',
                       'Priority': 'extra',
                       'Installed-Size': '16',
                       'Maintainer': 'Maintainer Name <maintainer@package>'}
        self.package = package
        self.version = version
        self.preinst_delay = preinst_delay
        self.postinst_delay = postinst_delay
        self.prerm_delay = prerm_delay
        self.postrm_delay = postrm_delay
        self.files = []

    def add_field(self, name, value):
        self.fields[name] = value

    def _write_tag(self, f, name, value):
        print >>f, "%s: %s" % (name, value)

    def _write_field(self, f, name):
        if name in self.fields:
            print >>f, "%s: %s" % (name, self.fields[name])

    def get_long_description(self):
        return "%s\n Long description for %s" % (self.description, self.package)

    def make_control_file(self, filename):
        f = open(filename, 'w')
        self._write_tag(f, 'Package', self.package)
        self._write_tag(f, 'Version', self.version)

        # Write known tags.  Tags without localized postfix must
        # precede the postfixed tags.
        keys = ('Section', 'Priority', 'Architecture', 'Depends', 'Conflicts', 'Replaces', 'Installed-Size', 'Maintainer', 'Description', 'Maemo-Display-Name', 'Maemo-Required-Free-Space', 'Maemo-Upgrade-Description')
        all_keys = self.fields.keys()
        for k in keys:
            self._write_field(f, k)
            if k in all_keys:
                all_keys.remove(k)

        # Write extra tags
        for k in all_keys:
            self._write_field(f, k)
        f.close()

    def make_preinst_file(self, filename):
        f = open(filename, 'w')
        print >>f, """#!/bin/sh
set -e

case "$1" in
    install|upgrade)
        sleep %d
        ;;

    abort-upgrade)
        ;;

    *)
        echo "preinst called with unknown argument '$1'" >&2
        exit 1
        ;;
esac

exit 0""" % self.preinst_delay
        f.close()
        os.chmod(filename, 0755)

    def make_postinst_file(self, filename):
        f = open(filename, 'w')
        print >>f, """#!/bin/sh
set -e

case "$1" in
    configure)
        sleep %d
        ;;

    abort-upgrade|abort-remove|abort-deconfigure)
        ;;

    *)
        echo "postinst called with unknown argument '$1'" >&2
        exit 1
        ;;
esac

exit 0""" % self.postinst_delay
        f.close()
        os.chmod(filename, 0755)

    def make_prerm_file(self, filename):
        f = open(filename, 'w')
        print >>f, """#!/bin/sh
set -e

case "$1" in
    remove|upgrade|deconfigure)
        sleep %d
        ;;

    failed-upgrade)
        ;;

    *)
        echo "prerm called with unknown argument '$1'" >&2
        exit 1
        ;;
esac

exit 0""" % self.prerm_delay
        f.close()
        os.chmod(filename, 0755)

    def make_postrm_file(self, filename):
        f = open(filename, 'w')
        print >>f, """#!/bin/sh
set -e

case "$1" in
    purge|remove|upgrade|failed-upgrade|abort-install|abort-upgrade|disappear)
        sleep %d
        ;;

    *)
        echo "postrm called with unknown argument '$1'" >&2
        exit 1
        ;;
esac

exit 0""" % self.postrm_delay
        f.close()
        os.chmod(filename, 0755)

    def make(self):
        global workdir

        pkgdir = '%s/%s-%s' % (workdir, self.package, self.version)
        d = '%s/DEBIAN' % pkgdir
        try:
            os.makedirs(d)
        except OSError, e:
            if e.errno != errno.EEXIST:
                raise

        digsigsums = '%s/%s' % (d, 'digsigsums')
        if os.path.exists(digsigsums):
            os.remove(digsigsums)

        for f in self.files:
            pathname = '%s%s' % (pkgdir, f[1])
            try:
                os.makedirs(os.path.dirname(pathname))
            except OSError, e:
                if e.errno != errno.EEXIST:
                    raise

            shutil.copyfile(f[0], pathname)

        self.make_control_file('%s/control' % d)
        self.make_preinst_file('%s/preinst' % d)
        self.make_postinst_file('%s/postinst' % d)
        self.make_prerm_file('%s/prerm' % d)
        self.make_postrm_file('%s/postrm' % d)

        ret = commands.getstatusoutput('fakeroot dpkg-deb -b %s %s' % (pkgdir, workdir))
        if ret[0] != 0:
            print >>sys.stderr, ret[1]
            return 1

        return 0

def decode(s):
    return s.replace('\\n', '\n')

def print_usage(msg=None):
    if msg:
        print >>sys.stderr, msg
    print >>sys.stderr, """usage: %s [OPTION...] PACKAGE VERSION

Options:
  -h, --help               print this help message
  -d, --build-dir=DIR      make package in DIR
  -f, --file=FILE=PATH     add FILE as PATH in package
  -F, --field=FIELD=VALUE  add FIELD: VALUE to debian/control
  -i, --preinst-delay=T    set preint script delay to T seconds
  -I, --postinst-delay=T   set postisnt script delay to T seconds
""" % sys.argv[0]

def main(argv=None):
    global workdir

    if argv is None:
        argv = sys.argv
    try:
        opts, args = getopt.getopt(sys.argv[1:],
                                   'd:hf:F:i:I:r:R:',
                                   ['build-dir=', 'help', 'file=', 'field=', 'preinst-delay=', 'postinst-delay=', 'prerm-delay=', 'postrm-delay='])
    except getopt.error, msg:
        print_usage(msg)
        return 2

    if len(args) != 2:
        print len(args)
        print_usage()
        return 2

    p = Package(args[0], args[1])

    for o, a in opts:
        if o in ('-h', '--help'):
            print_usage()
            sys.exit(0)
        elif o in ('-d', '--build-dir'):
            workdir = a
        elif o in ('-i', '--preinst-delay'):
            p.preinst_delay = string.atoi(a)
        elif o in ('-I', '--postinst-delay'):
            p.postinst_delay = string.atoi(a)
        elif o in ('-r', '--prerm-delay'):
            p.prerm_delay = string.atoi(a)
        elif o in ('-R', '--postrm-delay'):
            p.postrm_delay = string.atoi(a)
        elif o in ('-f', '--file'):
            i = a.find('=')
            if a < 0:
                print_usage('Invalid file format')
                return 2
            name = a[0:i]
            path = a[i+1:]
            p.files.append([name, path])
        elif o in ('-F', '--field'):
            i = a.find('=')
            if a < 0:
                print_usage('Invalid field format')
                return 2
            key = a[0:i]
            value = a[i+1:]
            p.add_field(key, decode(value))

    p.make()

if __name__ == '__main__':
    sys.exit(main())

