#!/usr/bin/python
#
# Copyright (C) 2009 - 2011 Nokia Corporation.
# 
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License version
# 2 as published by the Free Software Foundation.
# 
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see http://www.gnu.org/licenses/.

#
# Simple web server to serve Debian repository files

import atexit
import getopt
import httplib
import string
import mimetypes
import os
import signal
import sys
import time
import itertools
from BaseHTTPServer import BaseHTTPRequestHandler, HTTPServer

# The bytes per second to send per request (not completely accurate due to implementation but should fit our purposes)
send_bps = None

PIDFILE = '/var/run/hpm-webserver.pid'
DOCUMENT_ROOT = '/var/lib/package-manager-testrepo'
PORT = 8080
LOGFILE = '/var/tmp/hpm-webserver.log'

class Daemon:
    """
    A generic daemon class.
    Usage: subclass the Daemon class and override the run() method

    Lifted from: http://www.jejik.com/articles/2007/02/a_simple_unix_linux_daemon_in_python/
    """
    def __init__(self, pidfile, stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
        self.stdin = stdin
        self.stdout = stdout
        self.stderr = stderr
        self.pidfile = pidfile

    def daemonize(self):
        """
        do the UNIX double-fork magic, see Stevens' "Advanced
        Programming in the UNIX Environment" for details (ISBN 0201563177)
        http://www.erlenstar.demon.co.uk/unix/faq_2.html#SEC16
        """
        try:
            pid = os.fork()
            if pid > 0:
                # exit first parent
                sys.exit(0)
        except OSError, e:
            sys.stderr.write("fork #1 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # decouple from parent environment
        os.chdir("/")
        os.setsid()
        os.umask(0)

        # do second fork
        try:
            pid = os.fork()
            if pid > 0:
                # exit from second parent
                sys.exit(0)
        except OSError, e:
            sys.stderr.write("fork #2 failed: %d (%s)\n" % (e.errno, e.strerror))
            sys.exit(1)

        # redirect standard file descriptors
        sys.stdout.flush()
        sys.stderr.flush()
        si = file(self.stdin, 'r')
        so = file(self.stdout, 'a+')
        se = file(self.stderr, 'a+', 0)
        os.dup2(si.fileno(), sys.stdin.fileno())
        os.dup2(so.fileno(), sys.stdout.fileno())
        os.dup2(se.fileno(), sys.stderr.fileno())

        # write pidfile
        atexit.register(self.delpid)
        pid = str(os.getpid())
        f = file(self.pidfile, 'w+')
        f.write("%s\n" % pid)
        f.flush()
#        os.fsync(f.fileno())
        f.close()

    def delpid(self):
        os.remove(self.pidfile)

    def start(self):
        """
        Start the daemon
        """
        # Check for a pidfile to see if the daemon already runs
        try:
            pf = file(self.pidfile,'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if pid:
            message = "pidfile %s already exist. Daemon already running?\n"
            sys.stderr.write(message % self.pidfile)
            sys.exit(1)

        # Start the daemon
        self.daemonize()
        self.run()

    def stop(self):
        """
        Stop the daemon
        """
        # Get the pid from the pidfile
        try:
            pf = file(self.pidfile,'r')
            pid = int(pf.read().strip())
            pf.close()
        except IOError:
            pid = None

        if not pid:
            message = "pidfile %s does not exist. Daemon not running?\n"
            sys.stderr.write(message % self.pidfile)
            return # not an error in a restart

        # Try killing the daemon process
        try:
            while 1:
                os.kill(pid, signal.SIGTERM)
                time.sleep(0.1)
        except OSError, err:
            err = str(err)
            if err.find("No such process") > 0:
                if os.path.exists(self.pidfile):
                    os.remove(self.pidfile)
            else:
                print str(err)
                sys.exit(1)

    def restart(self):
        """
        Restart the daemon
        """
        self.stop()
        self.start()

    def run(self):
        """
        You should override this method when you subclass Daemon. It will be called after the process has been
        daemonized by start() or restart().
        """

class Handler(BaseHTTPRequestHandler):

    def do_GET(self):
        global send_bps
        
        mimetype = 'text/plain'
        mime = {'.bz2': 'application/x-bz2',
                '.gz': 'application/x-gzip',
                '.tar.gz': 'application/x-tar-gz',
                '.deb': 'application/x-deb'}
        for k, v in mime.items():
            if self.path.endswith(k):
                mimetype = v
                break

        try:
            #fn = '%s%s%s' % (curdir, sep, self.path)
            fn = '%s%s' % (os.curdir, self.path)
            # print fn
            f = open(fn, 'r')
            s = f.read()
            self.send_response(200)
            self.send_header('Content-type', mimetype)
            self.send_header("Content-Length", len(s))
            self.end_headers()
            
            if send_bps is not None:
                slice = send_bps
                for i in range(0,  len(s),  send_bps):
                    if i + send_bps > len(s): slice = len(s) - i
                    self.wfile.write(s[i:i+slice])
                    time.sleep(1)
            else:
                self.wfile.write(s)
               
            self.wfile.flush()
                
            f.close()
            self.log_message('sent %s (%d bytes)', self.path, len(s))
        except IOError:
            self.send_error(404, 'File not found: %s' % self.path)

#    def do_QUIT(self):
#        """send 200 OK response, and set server.stop to True"""
#        print "Server QUIT, stopping..."
#        sys.stdout.flush()
#        self.send_response(200)
#        self.end_headers()
#        self.server.stop = True

class StoppableHTTPServer(HTTPServer):
    def serve_forever(self):
        """Handle one request at a time until stopped."""
        self.stop = False
        while not self.stop:
            self.handle_request()

class WebServer(Daemon):
    def __init__(self, document_root, port, pidfile, stdin='/dev/null', stdout='/dev/null', stderr='/dev/null'):
        Daemon.__init__(self, pidfile, stdin, stdout, stderr)
        self.document_root = document_root
        self.port = port
        self.keep_running = True
        self.server = None

    def sig_handler(self, signum, frame):
        print 'Received signal %d, stopping' % signum
        sys.stdout.flush()
        # conn = httplib.HTTPConnection("127.0.0.1:%d" % self.port)
        # conn.request("QUIT", "/")
        # conn.getresponse()
        self.server.stop = True
   
    def run(self):
        signal.signal(signal.SIGTERM, self.sig_handler)
        os.chdir(self.document_root)
        print 'Document root: %s' % self.document_root
        print 'Listening to port: %d' % self.port
        sys.stdout.flush()
        # mimetypes.init()
        self.server = StoppableHTTPServer(('', self.port), Handler)
        self.server.serve_forever()

def print_usage():
    print """usage: %s start|stop|restart""" % sys.argv[0]

def main():
    global send_bps
    verbose = False

    try:
        opts, args = getopt.getopt(sys.argv[1:], 'hvs:',
                                   ['help', 'verbose',  'send_bps:'])
    except getopt.GetoptError, err:
        print str(err)
        print_usage()
        sys.exit(2)

    for o, a in opts:
        if o in ('-h', '--help'):
            print_usage()
            sys.exit(0)
        elif o in ('-v', '--verbose'):
            verbose = True
        elif o in ('-s',  '--send_bps'):
            send_bps = int(a) 

    if verbose:
        stdout = stderr = LOGFILE
    else:
        stdout = stderr = '/dev/null'

    daemon = WebServer(DOCUMENT_ROOT, PORT, PIDFILE, '/dev/null', stdout, stderr)

    if len(args) == 1:
        cmd = args[0]
        if cmd == 'start':
            daemon.start()
        elif cmd == 'stop':
            daemon.stop()
        elif cmd == 'restart':
            daemon.restart()
        else:
            print "Unknown command %s" % cmd
            sys.exit(2)
        sys.exit(0)
    else:
        print_usage()
        sys.exit(2)

if __name__ == '__main__':
    main()
