/*
 * Copyright (C) 2009 - 2011 Nokia Corporation.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see http://www.gnu.org/licenses/.
 *
 */

#include <QMap>
#include <QDBusReply>
#include <QList>
#include <QFile>
#include <QSettings>

#include "updater.h"
#include "packagemanager.h"

Updater::Updater() : timed(NULL)
{
    timed = new TimedOperations();
    if ( !timed ) {
        //TODO: throw exception
        qDebug() << Q_FUNC_INFO << "Error: couldn't create timed!";
    }
}

Updater::~Updater()
{
    if ( timed ) {
        delete timed;
        timed = NULL;
    }
}

void Updater::reschedule(bool error, int min)
{
    qDebug() << Q_FUNC_INFO << "Error:" << error;

    Maemo::Timed::Event event;

    // Compile the event object
    event.setAttribute(TIMED_ATTRIBUTE_KEY_APP, qApp->applicationName());
    event.setTriggerIfMissedFlag();
    event.setKeepAliveFlag();   // Only this binary should be able to remove the event

    // Add the event action (call this binary with the appropriate parameter to launch pkgmgrd.refresh_cache() -call)
    Maemo::Timed::Event::Action &action = event.addAction();

    QString command = QString("%1 %2").arg(APP_BINARY, ARG_REFRESH_CACHE);

    qDebug() << Q_FUNC_INFO << "Command: " << command;
    action.whenTriggered(); // Trigger the event on TRIGGERED-state (?!)
    
    if ( getenv ("_SBOX_DIR") != NULL ) {
        action.runCommand(command, "myusername"); // To test the command on scratchbox use your own username
        qDebug() << Q_FUNC_INFO << "In sbox.";
    } else {
        action.runCommand(command);
        qDebug() << Q_FUNC_INFO << "In device.";
    }

    // Add the event recurrence (every 24 hours from now. In case of error every hour)
    // TODO: Should the recurrence be somehow configurable?
    Maemo::Timed::Event::Recurrence &recurrence = event.addRecurrence();

    int h = QDateTime::currentDateTime().time().hour();
    int m = QDateTime::currentDateTime().time().minute();
    qDebug() << Q_FUNC_INFO << "Time:" << h << ":" << m;

    if ( error || min > 0 ) {
        for ( int i = 0; i <= 23 ; ++i ) {
            recurrence.addHour(i);
        }
        m += min;
        m = m % 60;
    } else {
        recurrence.addHour(h);
    }

    recurrence.addMinute(m);
    recurrence.everyDayOfMonth();
    recurrence.everyDayOfWeek();
    recurrence.everyMonth();

    uint cookieId = 0;
    if ( timed->checkForEvents(qApp->applicationName(), &cookieId) ) {
        if ( !timed->replaceEventToTimed( event, cookieId ) ) {
            qDebug() << Q_FUNC_INFO << "Failed to replace the event into timed :(";
            exit( RETURN_VALUE_TIMED_ERROR );
        }
    } else {
        if ( ! timed->addEventToTimed( event ) ) {
            qDebug() << Q_FUNC_INFO << "Failed to add the event into timed :(";
            exit( RETURN_VALUE_TIMED_ERROR );
        }
    }
    
    exit( RETURN_VALUE_OK );
}

void Updater::operationStarted(const QString &operation, const QString &packagename, const QString &version)
{
    qDebug() << Q_FUNC_INFO << "Operation started:" << operation;
}

void Updater::dbusError( PackageManagerPendingCallWatcher* watcher )
{
    qDebug() << Q_FUNC_INFO << "D-Bus error!" << watcher->errorName();
    reschedule(true);
}

// When cache refresh is finished this slot is called.
void Updater::operationComplete(const QString &operation, const QString &packagename, const QString &version, const QString &result, const bool need_reboot)
{
    if ( operation.compare(QString("Refresh")) != 0 ) {
        qDebug() << Q_FUNC_INFO << "Operation not refresh.";
        return;
    }
    qDebug() << Q_FUNC_INFO << "Result:" << result;

    reschedule( result.compare(QString("")) != 0 && result.compare(ERROR_PARTIAL_SUCCESS) != 0 );
}

bool Updater::refreshCache()
{
    qDebug() << Q_FUNC_INFO << "Refreshing the pkgmgrd cache...";
    PackageManager* packageManager = new PackageManager();
    
    connect(packageManager, SIGNAL(operationStarted(const QString&,const QString&,const QString&)),
            this, SLOT(operationStarted(const QString&,const QString&,const QString&)) );

    connect(packageManager, SIGNAL(operationComplete(const QString&,const QString&,const QString&,const QString&,const bool)),
            this, SLOT(operationComplete(const QString&,const QString&,const QString&,const QString&,const bool)) );
    
    PackageManagerPendingCallWatcher* callWatcher = packageManager->refreshPackageCache(QString(""),
                                                                                        QString("automatic,clearignored"));
    
    connect(callWatcher, SIGNAL(dbusError(PackageManagerPendingCallWatcher*)),
            this, SLOT(dbusError(PackageManagerPendingCallWatcher*)) );
    
    if ( callWatcher->errorCode() != PackageManager::ErrorNone ) {
        qDebug() << Q_FUNC_INFO << "Error while updating the packagemanager cache.";
        return false;
    }
    
    return true;
}

void Updater::addEventIfNoneExists()
{
    qDebug() << Q_FUNC_INFO << "Checking for events...";

    uint cookieId = 0;
    if ( timed->checkForEvents(qApp->applicationName(), &cookieId) ) {
        qDebug() << Q_FUNC_INFO << "Our event already exists in the timed. No need to add it again :)";
        exit( RETURN_VALUE_OK );
    }
    
    int min = 0;
    if ( ! QFile::exists( CACHE_REFRESHED_FILE ) ) {
        min = 5; // Default value
        QSettings* config = new QSettings(CONFIG_FILE, QSettings::IniFormat);
        min = config->value("TimedTasker/FirstRefreshCache", min).toInt();
        delete config;
        qDebug() << Q_FUNC_INFO << "Min: " << min;
    }
    
    reschedule(false, min);
}

bool Updater::cancelEvent()
{
    qDebug() << Q_FUNC_INFO << "Checking for the event...";

    uint cookieId = 0;
    if ( ! timed->checkForEvents(qApp->applicationName(), &cookieId) ) {
        qDebug() << Q_FUNC_INFO << "The event can not be found";
        return false;
    }
    if ( cookieId == 0 ) {
        qDebug() << Q_FUNC_INFO << "The event id is zero";
        return false;
    }
    
    return timed->cancelEvent(cookieId);
}

