/*
 * This file is part of the package-manager.
 *
 * This file is derived from hildon-application-manager.
 *
 * Parts of this file are derived from apt.  Apt is copyright 1997,
 * 1998, 1999 Jason Gunthorpe and others.
 *
 * Copyright (C) 2005 - 2011 Nokia Corporation.  All Rights reserved.
 *
 * Contact: Marius Vollmer <marius.vollmer@nokia.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version
 * 2 as published by the Free Software Foundation.
 *
 * This program is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA
 * 02110-1301 USA
 *
 */

/* This is the process that runs as root and does all the work.

   It is started from a separate program (as opposed to being forked
   directly from the GUI process) since that allows us to use sudo for
   starting it.

   This process communicates with the GUI process via some named pipes
   that are created by that process.  You can't really use it from the
   command line.

   It will output stuff to stdin and stderr, which the GUI process is
   supposed to catch and put into its log.

   The program tries hard not to exit prematurely.  Once the
   connection between the GUI process and this process has been
   established, the apt-worker is supposed to stick around until that
   connection is broken, even if it has to fail every request send to
   it.  This allows the user to try and fix the system after something
   went wrong, although the options are limited, of course.  The best
   example is a corrupted /etc/apt/sources.list: even tho you can't do
   anything related to packages, you still need the apt-worker to
   correct /etc/apt/sources.list itself in the UI.
*/

#include <stdlib.h>
#include <stdio.h>
#include <unistd.h>
#include <assert.h>
#include <stdarg.h>
#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#include <sys/types.h>
#include <sys/fcntl.h>
#include <errno.h>
#include <dirent.h>
#include <signal.h>
#include <ftw.h>
#include <sys/reboot.h>
#include <time.h>

#include <algorithm>
#include <fstream>
#include <sstream>
#include <iterator>
#include <map>
#include <set>

#include <apt-pkg/init.h>
#include <apt-pkg/error.h>
#include <apt-pkg/pkgcache.h>
#include <apt-pkg/pkgcachegen.h>
#include <apt-pkg/sourcelist.h>
#include <apt-pkg/acquire.h>
#include <apt-pkg/acquire-item.h>
#include <apt-pkg/cachefile.h>
#include <apt-pkg/strutl.h>
#include <apt-pkg/sptr.h>
#include <apt-pkg/packagemanager.h>
#include <apt-pkg/deblistparser.h>
#include <apt-pkg/debversion.h>
#include <apt-pkg/dpkgpm.h>
#include <apt-pkg/debsystem.h>
#include <apt-pkg/debversion.h>
#include <apt-pkg/orderlist.h>
#include <apt-pkg/algorithms.h>
#include <apt-pkg/metaindex.h>
#include <apt-pkg/debmetaindex.h>
#include <apt-pkg/md5.h>
#include <apt-pkg/sha1.h>
#include <apt-pkg/sha256.h>

#include <glib.h>
#include <glib/gstdio.h>
#include <glib/glist.h>
#include <glib/gstring.h>
#include <glib/gstrfuncs.h>
#include <glib/gmem.h>
#include <glib/gfileutils.h>
#include <glib/gslist.h>
#include <glib/gkeyfile.h>
#include <glib-object.h>

#include <qmbattery.h>

#include "apt-worker.h"
#include "apt-worker-utils.h"
#include "apt-worker-cache.h"
#include "apt-worker-operation.h"
#include "apt-worker-markpolicy.h"
#include "confutils.h"
#include "package.h"
#include "error.h"
// #include "domain.h"
#include "catalogues.h"
#include "glib-util.h"
#include "proxy.h"
#include "format-metadata.h"
#include "problemresolver.h"
#include "algorithms.h"

// #include "update-notifier-conf.h"

using namespace std;

/* Table of contents.
 
   COMPILE-TIME CONFIGURATION
   
   RUN-TIME CONFIGURATION

   GENERAL UTILITIES

   COMMUNICATING WITH THE FRONTEND

   STARTUP AND COMMAND DISPATCHER

   CACHE HANDLING
*/


/** COMPILE-TIME CONFIGURATION
 */

/* Defining this to non-zero will also recognize packages in the
   "maemo" section as user packages.  There are still packages
   floating around that follow this old rule.
*/
#define ENABLE_OLD_MAEMO_SECTION_TEST 1

/* Requests up to this size are put into a stack allocated buffer.
 */
// #define FIXED_REQUEST_BUF_SIZE 4096

/* The location where we keep our lock.
 */
#define APT_WORKER_LOCK "/var/lib/package-manager/apt-worker-lock"

/* Temporary catalogues and temporary sources.list */
#define TEMP_CATALOGUE_CONF "/var/lib/package-manager/catalogues.temp"

#define CACHE_REFRESHED_FILE "/var/lib/package-manager/cacherefreshed"

/* APT CACHE ARCHIVES DIRECTORIES */
#define DEFAULT_DIR_CACHE_ARCHIVES "archives/"
#define ALT_DIR_CACHE_ARCHIVES ".apt-archive-cache/"

/* Files related to the 'check for updates' process */

/* Domain names associated with "OS" and "Nokia" updates */
#define OS_UPDATES_DOMAIN_NAME "nokia-system"
#define NOKIA_UPDATES_DOMAIN_NAME "nokia-certified"

/* absolute path to directory where dpkg keeps it's package information */

#define DPKG_INFO_PATH "/var/lib/dpkg/info/"

#define REBOOT_SCRIPT "/usr/lib/package-manager/reboot.sh"

#define IGNORED_PACKAGE_LIST "/var/lib/package-manager/ignored-packages"
#define AVAILABLE_PACKAGE_LIST "/var/lib/package-manager/available-packages"
#define RESTORE_PACKAGE_LIST_TMP "/home/user/.package-manager/installed-packages.restored"
#define RESTORE_PACKAGE_LIST "/var/lib/package-manager/installed-packages.restored"

#define LOW_BATTERY_LEVEL 50
#define CRITICAL_BATTERY_LEVEL 15

/* You know what this means.
 */
#define DEBUG
/*#define DEBUG_COMMANDS*/

extern int verbosity;

#ifdef DEBUG
#define DBG(fmt, ...) g_debug("%s(): " fmt, __func__, ## __VA_ARGS__)
#else
#define DBG(...)
#endif

/** RUN-TIME CONFIGURATION
 */

bool flag_enable_notifications = true;

bool flag_collect_dpkg_output = true;

/* Setting this to false will not use MMC to save the packages when
   downloading them.
*/
//bool flag_download_packages_to_mmc = false;

/* List of packages found with the 'system-update' flag (SSU packages)
 */
GArray *ssu_packages = NULL;

/* Flag to tell whether the ssu_packages GArray needs to be refreshed
 */
gboolean ssu_packages_needs_refresh = true;
static PkgTsStorage *ignored_update_pkgs = NULL;
static PkgTsStorage *available_update_pkgs = NULL;
static PkgTsStorage *restore_pkgs = NULL;
gboolean new_restore = false;

GList* pkgs_to_be_deleted = NULL;

extern GStaticRWLock cache_rw_lock_;

/** APT WORKER MULTI STATE MANAGEMENT
 */




/* Open FILENAME with FLAGS, or die.
 */
// static int
// must_open (char *filename, int flags)
// {
//   int fd = open (filename, flags);
//   if (fd < 0)
//     {
//       perror (filename);
//       exit (1);
//     }
//   return fd;
// }

// static void
// must_set_flags (int fd, int flags)
// {
//   if (fcntl (fd, F_SETFL, flags) < 0)
//     {
//       perror ("apt-worker fcntl");
//       exit (1);
//     }
// }

// static void
// block_for_read (int fd)
// {
//   fd_set set;
//   FD_ZERO (&set);
//   FD_SET (fd, &set);

//   if (select (fd+1, &set, NULL, NULL, NULL) < 0)
//     {
//       perror ("apt-worker select");
//       exit (1);
//     }
// }

// static int
// read_byte (int fd)
// {
//   unsigned char byte;
//   if (read (fd, &byte, 1) == 1)
//     return byte;
//   return -1;
// }

/* DRAIN_FD reads all bytes from FD that are available.
*/
// static void
// drain_fd (int fd)
// {
//   while (read_byte (fd) >= 0)
//     ;
// }



/** COMMUNICATING WITH THE FRONTEND.
 
   The communication with the frontend happens over four
   unidirectional fifos: requests are read from INPUT_FD and responses
   are sent back via OUTPUT_FD.  No new request is read until the
   response to the current one has been completely sent.

   The data read from INPUT_FD must follow the request format
   specified in <apt-worker-proto.h>.  The data written to OUTPUT_FD
   follows the response format specified there.

   The CANCEL_FD is polled periodically and when something is
   available to be read, the current operation is aborted.  There is
   currently no meaning defined for the actual bytes that are sent,
   the mere arrival of a byte triggers the abort.

   When using the libapt-pkg PackageManager, it is configured in such
   a way that it sends it "pmstatus:" message lines to STATUS_FD.
   Other asynchronous status reports are sent as spontaneous
   APTCMD_STATUS responses via OUTPUT_FD.  'Spontaneous' should mean
   that no request is required to receive APTCMD_STATUS responses.  In
   fact, APTCMD_STATUS requests are treated as an error by the
   apt-worker.

   Logging and debug output, and output from dpkg and the maintainer
   scripts appears normally on stdout and stderr of the apt-worker
   process.
*/

// int input_fd, output_fd, status_fd, cancel_fd;
int status_fd = -1;

/* MUST_READ and MUST_WRITE read and write blocks of raw bytes from
   INPUT_FD and to OUTPUT_FD.  If they return, they have succeeded and
   read or written the whole block.
*/

// void
// must_read (void *buf, size_t n)
// {
//   int r;

//   while (n > 0)
//     {
//       r = read (input_fd, buf, n);
//       if (r < 0)
// 	{
// 	  perror ("apt-worker read");
// 	  exit (1);
// 	}
//       else if (r == 0)
// 	{
// 	  DBG ("exiting");
// 	  exit (0);
// 	}
//       n -= r;
//       buf = ((char *)buf) + r;
//     }
// }

// static void
// must_write (void *buf, ssize_t n)
// {
//   if (n > 0 && write (output_fd, buf, n) != n)
//     {
//       perror ("apt-worker write");
//       exit (1);
//     }
// }

/* This function sends a response on OUTPUT_FD with the given CMD and
   SEQ.  It either succeeds or does not return.
*/
// void
// send_response_raw (int cmd, int seq, void *response, size_t len)
// {
//   apt_response_header res = { cmd, seq, len };
//   must_write (&res, sizeof (res));
//   must_write (response, len);
// }

/* Fabricate and send a APTCMD_STATUS response.  Parameters OP,
   ALREADY, and TOTAL are as specified in apt-worker-proto.h.

   A status response is only sent when there is enough change since
   the last time.  The following counts as 'enough': ALREADY has
   decreased, it has increased by more than MIN_CHANGE, it is equal to
   -1, LAST_TOTAL has changed, or OP has changed.
*/


/** STARTUP AND COMMAND DISPATCHER.
 */

/* Since the apt-worker only works on a single command at a time, we
   use two global encoder and decoder engines that manage the
   parameters of the request and the result values of the response.

   Handlers of specific commands will read the parameters from REQUEST
   and put the results into RESPONSE.  The command dispatcher will
   prepare REQUEST before calling the command handler and ship out
   RESPONSE after it returned.
*/
// apt_proto_decoder request;
// apt_proto_encoder response;


/** MAPPING FUNCTION TO FILTER ERROR DETAILS
    IN CATALOGUES CONFIGURATION FILE
*/
// static xexp *map_catalogue_error_details (xexp *x);


// void
// handle_request ()
// {
//   apt_request_header req;
//   char stack_reqbuf[FIXED_REQUEST_BUF_SIZE];
//   char *reqbuf;
//   AptWorkerCache * awc = 0;
//   time_t last_modified = -1;

//   must_read (&req, sizeof (req));

// #ifdef DEBUG_COMMANDS
//   DBG ("got req %s/%d/%d", cmd_names[req.cmd], req.seq, req.len);
// #endif

//   reqbuf = alloc_buf (req.len, stack_reqbuf, FIXED_REQUEST_BUF_SIZE);
//   must_read (reqbuf, req.len);

//   drain_fd (cancel_fd);

//   request.reset (reqbuf, req.len);
//   response.reset ();

//   awc = AptWorkerCache::GetCurrent ();
//   awc->init_cache_after_request = false; // let's reset it now

//   /* Re-read domains conf file if modified */
//   last_modified = file_last_modified (PACKAGE_DOMAINS);
//   if (last_modified != domains_last_modified)
//     read_domain_conf ();

//   switch (req.cmd)
//     {

//     case APTCMD_NOOP:
//       // Nothing to do.
//       break;

//     case APTCMD_GET_PACKAGE_LIST:
//       cmd_get_package_list ();
//       break;

//     case APTCMD_GET_PACKAGE_INFO:
//       cmd_get_package_info ();
//       break;

//     case APTCMD_GET_PACKAGE_DETAILS:
//       cmd_get_package_details ();
//       break;

//     case APTCMD_CHECK_UPDATES:
//       cmd_check_updates ();
//       break;

//     case APTCMD_GET_CATALOGUES:
//       cmd_get_catalogues ();
//       break;

//     case APTCMD_SET_CATALOGUES:
//       cmd_set_catalogues ();
//       break;

//     case APTCMD_ADD_TEMP_CATALOGUES:
//       cmd_add_temp_catalogues ();
//       break;

//     case APTCMD_RM_TEMP_CATALOGUES:
//       cmd_rm_temp_catalogues ();
//       break;

//     case APTCMD_INSTALL_CHECK:
//       cmd_install_check ();
//       break;

//     case APTCMD_DOWNLOAD_PACKAGE:
//       cmd_download_package ();
//       break;

//     case APTCMD_INSTALL_PACKAGE:
//       cmd_install_package ();
//       break;

//     case APTCMD_REMOVE_CHECK:
//       cmd_remove_check ();
//       break;

//     case APTCMD_REMOVE_PACKAGE:
//       cmd_remove_package ();
//       break;

//     case APTCMD_CLEAN:
//       cmd_clean ();
//       break;

//     case APTCMD_GET_FILE_DETAILS:
//       cmd_get_file_details ();
//       break;

//     case APTCMD_INSTALL_FILE:
//       cmd_install_file ();
//       break;

//     case APTCMD_SAVE_BACKUP_DATA:
//       cmd_save_backup_data ();
//       break;

//     case APTCMD_GET_SYSTEM_UPDATE_PACKAGES:
//       cmd_get_system_update_packages ();
//       break;

//     case APTCMD_REBOOT:
//       cmd_reboot ();
//       break;

//     case APTCMD_SET_OPTIONS:
//       cmd_set_options ();
//       break;

//     case APTCMD_SET_ENV:
//       cmd_set_env ();
//       break;

//     case APTCMD_THIRD_PARTY_POLICY_CHECK:
//       cmd_third_party_policy_check ();
//       break;

//     default:
//       log_stderr ("unrecognized request: %d", req.cmd);
//       break;
//     }

//   _error->DumpErrors ();

//   send_response_raw (req.cmd, req.seq,
// 		     response.get_buf (), response.get_len ());

// #ifdef DEBUG_COMMANDS
//   DBG ("sent resp %s/%d/%d",
//        cmd_names[req.cmd], req.seq, response.get_len ());
// #endif

//   free_buf (reqbuf, stack_reqbuf);

//   if (awc->init_cache_after_request)
//     {
//       cache_init (false);
//       _error->DumpErrors ();
//     }
// }

#ifdef HAVE_APT_TRUST_HOOK
// static int index_trust_level_for_package (pkgIndexFile *index,
// 					  const pkgCache::VerIterator &ver);
#endif

extern const char *lc_messages;

// static void
// usage ()
// {
//   fprintf (stderr, "Usage: apt-worker check-for-updates [http_proxy]\n");
//   fprintf (stderr, "       apt-worker rescue [package] [archives]\n");
//   exit (1);
// }

/* Try to get the lock specified by FILE.  If this fails because
   someone else has the lock, a string is returned with the content of
   the file.  Otherwise MY_CONTENT is written to the file and NULL is
   returned.

   Other kinds of failures (out of space, insufficient permissions,
   etc) will terminate this process.

   The lock will be released when this process exits.
*/

// static char *
// try_lock (const char *file, const char *my_content)
// {
//   int lock_fd = open (file, O_RDWR | O_CREAT, 0640);
//   if (lock_fd < 0)
//     {
//       log_stderr ("Can't open %s: %m", file);
//       exit (1);
//     }

//   SetCloseExec (lock_fd, true);

//   struct flock fl;
//   fl.l_type = F_WRLCK;
//   fl.l_whence = SEEK_SET;
//   fl.l_start = 0;
//   fl.l_len = 0;

//   if (fcntl (lock_fd, F_SETLK, &fl) == -1)
//     {
//       char buf[256];
//       int n;
      
//       if (errno == ENOLCK)
// 	{
// 	  log_stderr ("locking not supported on %s: %m", file);
// 	  exit (1);
// 	}
     
//       /* We didn't get the lock.
//        */

//       n = read (lock_fd, buf, 255);
//       if (n < 0)
// 	{
// 	  log_stderr ("can't read lock %s: %m", file);
// 	  exit (1);
// 	}

//       buf[n] = '\0';
//       return g_strdup (buf);
//     }

//   /* We have the lock.
//    */

//   if (ftruncate (lock_fd, 0) < 0)
//     {
//       log_stderr ("can't truncate lock %s: %m", file);
//       exit (1);
//     }
    
//   int n = strlen (my_content);
//   if (write (lock_fd, my_content, n) != n)
//     {
//       log_stderr ("can't write lock %s: %m", file);
//       exit (1);
//     }
     
//   return NULL;
// }

// static void
// get_apt_worker_lock (bool weak)
// {
//   char *mine = g_strdup_printf ("%c %d\n", weak? 'w' : 's', getpid ());
//   char *his = NULL;
//   int termination_attempts = 0;
//   int lock_attempts = 0;

//   while (true)
//     {
//       g_free (his);
//       his = try_lock (APT_WORKER_LOCK, mine);
      
//       if (his)
// 	{
// 	  char his_type;
// 	  int his_pid;
	  
// 	  if (sscanf (his, "%c %d", &his_type, &his_pid) != 2)
// 	    {
// 	      log_stderr ("can't parse lock.");
// 	      exit (1);
// 	    }

// 	  if (weak || his_type != 'w')
// 	    {
// 	      if (lock_attempts < 5)
// 	        {
// 	          lock_attempts++;
// 	          sleep (1);
// 	          continue;
// 	        }
// 	      else
// 	        {
// 	          log_stderr ("too weak to get lock from %d.", his_pid);
// 	          exit (1);
// 	        }
// 	    }
// 	  else if (termination_attempts < 5)
// 	    {
// 	      termination_attempts += 1;
// 	      log_stderr ("terminating %d to get lock.", his_pid);
// 	      kill (his_pid, SIGTERM);
// 	      sleep (1);
// 	      continue;
// 	    }
// 	  else
// 	    {
// 	      /* The big hammer.
// 	       */
// 	      log_stderr ("killing %d to get lock.", his_pid);
// 	      kill (his_pid, SIGKILL);
// 	      unlink (APT_WORKER_LOCK);
// 	      sleep (1);
// 	      continue;
// 	    }
// 	}
//       else
// 	{
// 	  /* It's ours.
// 	   */
// 	  g_free (mine);
// 	  return;
// 	}
//     }
// }

/* MMC default mountpoints */
#define INTERNAL_MMC_MOUNTPOINT  "/home/user/MyDocs"
#define REMOVABLE_MMC_MOUNTPOINT "/media/mmc1"

static int
must_open (const char *filename, int flags)
{
	int fd = open (filename, flags);
	if (fd < 0) {
		DBG ("open failed: %s: %d: %s", filename, errno, strerror(errno));
		return -1;
	}
	return fd;
}

int
must_open_nonblock (const char *filename, int flags)
{
	int fd = must_open (filename, flags | O_NONBLOCK);
	if (fd >= 0) {
		if (fcntl (fd, F_SETFL, flags) < 0) {
			DBG ("fcntl failed");
		}
	}
	return fd;
}

bool
must_mkfifo (const char *filename, int mode)
{
	if (unlink (filename) < 0 && errno != ENOENT) {
		DBG ("unlink failed: %s", filename);
	}

	if (mkfifo (filename, mode) < 0) {
		DBG ("mkfifo failed: %s", filename);
		return false;
	}
	return true;
}
static void
__setup_dpkg_status_ind()
{
	must_mkfifo ("/tmp/apt-worker.status", 0600);
	int read_status_fd = must_open_nonblock ("/tmp/apt-worker.status",
					O_RDONLY);
	DBG("read_status_fd: %d", read_status_fd);
	status_fd = must_open_nonblock ("/tmp/apt-worker.status", O_WRONLY);
	DBG("status_fd: %d", status_fd);

	status_ind_setup_from_fd(read_status_fd);
}
extern DownloadProgressFunc download_progress_func;
extern OperationStartedFunc operation_started_func;
extern OperationProgressFunc operation_progress_func;
extern OperationCompleteFunc operation_complete_func;
extern PackageListUpdatedFunc package_list_updated_func;
extern PackageIgnoredFunc package_ignored_func;
extern gpointer signal_data;

void
apt_worker_init (DownloadProgressFunc download_progress_fn,
        OperationStartedFunc operation_started_fn,
		OperationProgressFunc operation_progress_fn,
		OperationCompleteFunc operation_complete_fn,
		PackageListUpdatedFunc package_list_updated_fn,
		PackageIgnoredFunc package_ignored_fn,
		gpointer data)
{
  lc_messages = getenv ("LC_MESSAGES");
  DBG ("LC_MESSAGES %s", lc_messages);

  DBG ("OSSO_PRODUCT_HARDWARE %s", getenv ("OSSO_PRODUCT_HARDWARE"));

  download_progress_func = download_progress_fn;
  operation_started_func = operation_started_fn;
  operation_progress_func = operation_progress_fn;
  operation_complete_func = operation_complete_fn;
  package_list_updated_func = package_list_updated_fn;
  package_ignored_func = package_ignored_fn;
  signal_data = data;
  load_system_settings ();
  read_domain_conf ();

  AptWorkerCache::Initialize ();
  if (getenv("https_proxy") == NULL) {
	  string https_proxy = _config->Find("Acquire::https::Proxy");
	  if (!https_proxy.empty()) {
		  DBG ("Setting https_proxy: '%s'", https_proxy.c_str());
		  setenv("https_proxy", https_proxy.c_str(), 1);
	  }
  } else {
	  DBG ("https_proxy: '%s'", getenv("https_proxy"));
  }

  unlink (_config->FindFile ("Dir::Cache::pkgcache-read").c_str ());
  unlink (_config->FindFile ("Dir::State::status-read").c_str ());
  unlink (_config->FindFile ("Dir::Aegis::Restok::conf-read").c_str ());

  cache_init (false);

#ifdef HAVE_APT_TRUST_HOOK
//   apt_set_index_trust_level_for_package_hook (index_trust_level_for_package);
#endif

  clean_temp_catalogues ();

  // initialize the MMC mount points with defaults
  setenv ("INTERNAL_MMC_MOUNTPOINT", INTERNAL_MMC_MOUNTPOINT, 1);
  setenv ("REMOVABLE_MMC_MOUNTPOINT", REMOVABLE_MMC_MOUNTPOINT, 1);

  __setup_dpkg_status_ind();

  g_static_rw_lock_init (&cache_rw_lock_);
}

void
apt_worker_free ()
{
  if ( available_update_pkgs ) {
    delete available_update_pkgs;
    available_update_pkgs = 0;
  }
  if ( ignored_update_pkgs ) {
    delete ignored_update_pkgs;
    ignored_update_pkgs = 0;
  }

  free_dpkg_status_ind();
  cache_close();
  cache_free ();
  domain_conf_free ();
  free_system_settings ();
  delete AptWorkerCache::GetCurrent();
}

// void
// set_options (const char *options)
// {
//   if (strchr (options, 'B'))
//     flag_break_locks = true;

//   if (strchr (options, 'D'))
//     flag_allow_wrong_domains = true;

//   if (strchr (options, 'M'))
//     flag_download_packages_to_mmc = true;

//   if (strchr (options, 'A'))
//     flag_use_apt_algorithms = true;
// }

// void
// cmd_set_options ()
// {
//   const char *options = request.decode_string_in_place ();
//   set_options (options);
// }

// void
// cmd_set_env ()
// {
//   const char *http_proxy = request.decode_string_in_place ();
//   const char *https_proxy = request.decode_string_in_place ();
//   const char *internal_mmc = request.decode_string_in_place ();
//   const char *removable_mmc = request.decode_string_in_place ();

//   if (http_proxy)
//     {
//       setenv ("http_proxy", http_proxy, 1);
//       DBG ("http_proxy: %s", http_proxy);
//     }
//   else
//     {
//       // clear HTTP_PROXY from environment otherwise
//       unsetenv ("http_proxy");
//     }

//   if (https_proxy)
//     {
//       setenv ("https_proxy", https_proxy, 1);
//       DBG ("https_proxy: %s", https_proxy);
//     }
//   else
//     {
//       // clear HTTPS_PROXY from environment otherwise
//       unsetenv ("https_proxy");
//     }

//   if (internal_mmc)
//     {
//       setenv ("INTERNAL_MMC_MOUNTPOINT", internal_mmc, 1);
//       DBG ("INTERNAL_MMC_MOUNTPOINT: %s", internal_mmc);
//     }

//   if (removable_mmc)
//     {
//       setenv ("REMOVABLE_MMC_MOUNTPOINT", removable_mmc, 1);
//       DBG ("REMOVABLE_MMC_MOUNTPOINT: %s", removable_mmc);
//     }
// }

// int
// main (int argc, char **argv)
// {
//   if (argc == 1)
//     usage ();

//   argv += 1;
//   argc -= 1;

//   if (!strcmp (argv[0], "backend"))
//     {
//       const char *options;

//       if (argc != 6)
// 	{
// 	  log_stderr ("wrong invocation");
// 	  exit (1);
// 	}

//       DBG ("starting up");

//       char *input_pipe = is_fifo (argv[1]);
//       char *output_pipe = is_fifo (argv[2]);
//       char *status_pipe = is_fifo (argv[3]);
//       char *cancel_pipe = is_fifo (argv[4]);

//       if (!(input_pipe && output_pipe && status_pipe && cancel_pipe))
// 	{
// 	  g_free (input_pipe);
// 	  g_free (output_pipe);
// 	  g_free (status_pipe);
// 	  g_free (cancel_pipe);

// 	  log_stderr ("wrong fifo pipes specified");
// 	  exit (1);
// 	}

//       input_fd = must_open (input_pipe, O_RDONLY | O_NONBLOCK);
//       cancel_fd = must_open (cancel_pipe, O_RDONLY | O_NONBLOCK);
//       output_fd = must_open (output_pipe, O_WRONLY);
//       status_fd = must_open (status_pipe, O_WRONLY);

//       g_free (input_pipe);
//       g_free (output_pipe);
//       g_free (status_pipe);
//       g_free (cancel_pipe);

//       /* This tells the frontend that the fifos are open.
//        */
//       send_status (op_general, 0, 0, -1);

//       /* This blocks until the frontend has opened our input fifo for
// 	 writing.
//       */
//       block_for_read (input_fd);

//       /* Reset the O_NONBLOCK flag for the input_fd since we want to block
// 	 until a new request arrives.  The cancel_fd remains in
// 	 non-blocking mode since we just poll it periodically.
//       */
//       must_set_flags (input_fd, O_RDONLY);

//       options = argv[5];

//       DBG ("starting with pid %d, in %d, out %d, stat %d, cancel %d, options %s",
// 	   getpid (), input_fd, output_fd, status_fd, cancel_fd,
// 	   options);

//       set_options (options);

//       /* Don't let our heavy lifting starve the UI.
//        */
//       errno = 0;
//       if (nice (20) == -1 && errno != 0)
// 	log_stderr ("nice: %m");

//       get_apt_worker_lock (false);
//       misc_init ();

//       while (true)
// 	handle_request ();

//       return 0;
//     }
//   else if (!strcmp (argv[0], "check-for-updates"))
//     {
//       get_apt_worker_lock (true);
//       misc_init ();
//       return cmdline_check_updates (argv);
//     }
//   else if (!strcmp (argv[0], "rescue"))
//     {
//       return cmdline_rescue (argv);
//     }
//   else if (!strcmp (argv[0], "sleep"))
//     {
//       get_apt_worker_lock (argv[1] == NULL);
//       while (true)
// 	{
// 	  fprintf (stderr, "sleeping...\n");
// 	  sleep (5);
// 	}
//     }
//   else
//     usage ();
// }


/* We only report real progress information when reconstructing the
   cache and during downloads.  Only downloads can be cancelled.

   The following two classes allow us to hook into libapt-pkgs
   progress reporting mechanism.  Instances of UPDATE_PROGESS are used
   for cache related activities, and instances of DOWNLOAD_STATUS are
   used when 'acquiring' things.
*/

class DownloadStatus : public pkgAcquireStatus
{
  virtual bool
  MediaChange (string Media, string Drive)
  {
    return false;
  }

  virtual void Fetch(pkgAcquire::ItemDesc &Itm)
  {
    DBG("DownloadStatus::Fetch(URI=%s, Description=%s, ShortDesc=%s)",
	Itm.URI.c_str(),
	Itm.Description.c_str(),
	Itm.ShortDesc.c_str());
  }

  virtual bool
  Pulse (pkgAcquire *Owner)
  {
    pkgAcquireStatus::Pulse (Owner);

    // DBG("Pulse: CurrentItems: %lu, TotalItems %lu", CurrentItems, TotalItems);
    if (TotalBytes >= 2)
      send_download_progress_signal ((int)CurrentBytes, (int)TotalBytes, 1000);

    OperationData *op = cache_get_current_operation();
    return op != NULL ? !op->cancel_flag : true;
  }
};

bool
check_battery_level(GError **error, int batt_level)
{
    if ( in_scratchbox() ) {
        DBG("In scratchbox. So not checking the battery level.");
        return false;
    }

    using namespace MeeGo;
    DBG("");
    QmBattery* batt = new QmBattery();
    
    int pct = batt->getRemainingCapacityPct();
    QmBattery::ChargingState cs = batt->getChargingState();
    QmBattery::ChargerType ct = batt->getChargerType();
    
/* Charger types
    Unknown = -1,        
    None,                
    Wall,                
    USB_500mA,           
    USB_100mA  
*/

    DBG("Battery level: %d%%", pct);
    DBG("Charging: %s", (cs == QmBattery::StateNotCharging?"no":(cs == QmBattery::StateChargingFailed?"failed":"yes")) );
    DBG("Charger type: %s", ct == QmBattery::USB_100mA?"USB 100mA":(ct == QmBattery::None?"None":(ct == QmBattery::Wall?"Wall":(ct == QmBattery::USB_500mA?"USB 500mA":"Unknown"))) );

    delete batt;
    if ( pct <= batt_level ) {
        if ( cs != QmBattery::StateCharging ) {
            g_set_error(error, PKGMGRD_ERROR, ERROR_LOW_BATTERY_LEVEL,
                        "Battery level is too low: %d", pct);
            return true;
        }
        if ( ct != QmBattery::Wall && ct != QmBattery::USB_500mA ) {
            g_set_error(error, PKGMGRD_ERROR, ERROR_LOW_BATTERY_LEVEL, "Charger incompatible");
            return true;
        }
    }
    
    return false;
}

/* Initialize libapt-pkg if this has not been done already and
   (re-)create PACKAGE_CACHE.  If the cache can not be created,
   PACKAGE_CACHE is set to NULL and an appropriate message is output.
   */
void
apt_worker_reset_state ()
{
	AptWorkerCache * awc = 0;
	DBG("Reseting apt worker state");

	if(!ensure_cache(false)) {
		return;
	}
	awc = AptWorkerCache::GetCurrent ();
	if (awc->init_cache_after_request) {
		cache_init(false);
	}
	cache_reset();
}

bool is_in_list(GList *list, PackageVersion* pv)
{
    if ( list == NULL || pv == NULL )
        return false;
    GList *i = list;
    PackageVersion *i_pv = NULL;
    while ( i ) {
        i_pv = (PackageVersion *)i->data;
        if ( strcmp( i_pv->pkg, pv->pkg ) == 0 && strcmp( i_pv->version, pv->version ) == 0 )
            return true;
        i = g_list_next( i );
    }
    return false;
}

bool is_in_list(GList *list, const gchar* pkg, const gchar* version)
{
    if ( list == NULL || pkg == NULL )
        return false;
    PackageVersion *pv = package_version_new ();
    pv->pkg = g_strdup(pkg);
    pv->version = g_strdup(version);

    bool result = is_in_list(list, pv);
    g_free(pv->pkg);
    g_free(pv->version);
    g_free(pv);
    return result;
}

void free_list(GList *list)
{
    if ( list == NULL )
        return;
    GList *i = g_list_first(list);
    while ( i ) {
        PackageVersion *p = (PackageVersion*)i->data;
        package_version_free(p);
        i = g_list_next(i);
    }
    g_list_free(list);
}

void
mark_related (const pkgCache::VerIterator &ver)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  const pkgCache::PkgIterator &pkg = ver.ParentPkg();

  if (awc->cache->extra_info[pkg->ID].related)
    return;

  DBG ("related: %s", pkg.Name());
  awc->cache->extra_info[pkg->ID].related = true;

  pkgDepCache &cache = *awc->cache;

  if (pkg.State() == pkgCache::PkgIterator::NeedsUnpack)
    cache.SetReInstall (pkg, true);

  /* When there are some packages that might need configuring or
     unpacking, we also mark all dependencies of this package as
     related so that we try to unpack/configure them as well.
  */
  if (cache.BadCount () > 0)
    {
      for (pkgCache::DepIterator D = ver.DependsList(); D.end() == false;
	   D++)
	{
	  const pkgCache::PkgIterator &dep_pkg = D.TargetPkg ();
	  const pkgCache::VerIterator &dep_ver = dep_pkg.CurrentVer ();
	  if (!dep_ver.end())
	    mark_related (dep_ver);
	}
    }
}

// static bool
// any_newly_or_related_broken ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   if (awc->cache == NULL)
//     return false;

//   pkgDepCache &cache = *(awc->cache);
//   for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
//     {
//       if (cache[pkg].InstBroken() &&
// 	  (!cache[pkg].NowBroken() || is_related (pkg)))
// 	return true;
//     }
//   return false;
// }


/* Try to fix packages that have been broken by undoing soft changes.

   This is not really complicated, the code only looks impenetrable
   because of libapt-pkg's data structures.

   For each package that is broken for the planned operation, we try
   to fix it by undoing the removal of softly removed packages that it
   depends on.  We do this in a loop since a package that has been put
   back might be broken itself and be in need of fixing.
*/
void
fix_soft_packages ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  if (awc->cache == NULL)
    return;

  pkgDepCache &cache = *(awc->cache);

  bool something_changed;

  do 
    {
      DBG ("FIX");

      something_changed = false;
      for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
	{
	  if (cache[pkg].InstBroken())
	    {
	      pkgCache::DepIterator Dep =
		cache[pkg].InstVerIter(cache).DependsList();
	      for (; Dep.end() != true;)
		{
		  // Grok or groups
		  pkgCache::DepIterator Start = Dep;
		  bool Result = true;
		  for (bool LastOR = true;
		       Dep.end() == false && LastOR == true;
		       Dep++)
		    {
		      LastOR = ((Dep->CompareOp & pkgCache::Dep::Or)
				== pkgCache::Dep::Or);
		    
		      if ((cache[Dep] & pkgDepCache::DepInstall)
			  == pkgDepCache::DepInstall)
			Result = false;
		    }
		
		  // Dep is satisfied okay.
		  if (Result == false)
		    continue;

		  // Try to fix it by putting back the first softly
		  // removed target

		  for (bool LastOR = true;
		       Start.end() == false && LastOR == true;
		       Start++)
		    {
		      LastOR = ((Start->CompareOp & pkgCache::Dep::Or)
				== pkgCache::Dep::Or);
		    
		      pkgCache::PkgIterator Pkg = Start.TargetPkg ();

		      if (((cache[Start] & pkgDepCache::DepInstall)
			   != pkgDepCache::DepInstall)
			  && !Pkg.end()
			  && cache[Pkg].Delete()
              && awc->cache->extra_info[Pkg->ID].soft)
			{
			  DBG ("= %s", Pkg.Name());
			  cache_reset_package (Pkg);
			  something_changed = true;
			  break;
			}
		    }
		}
	    }
	}
    } while (something_changed);
}


#if 0
/* Determine whether PKG is a critical dependency of other packages
   thata re going to be installed.
 */
static bool
package_is_needed (pkgCache::PkgIterator &pkg)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  pkgCache::DepIterator Dep = pkg.RevDependsList();
  for (; Dep.end() != true; Dep++)
    {
      if (Dep->Type == pkgCache::Dep::PreDepends
	  || Dep->Type == pkgCache::Dep::Depends)
	{
	  pkgCache::PkgIterator other_pkg = Dep.ParentPkg();
	  pkgCache::VerIterator other_ver = Dep.ParentVer();
	  pkgCache::VerIterator inst_ver = cache[other_pkg].InstVerIter(cache);

	  if (other_ver == inst_ver)
	    return true;
	}
    }
  return false;
}
#endif

static void
print_error (GError *error)
{
  if (!error)
    return;

  DBG ("Error: %s: %s\n",
       pkgmgr_error_to_dbus_error ((PkgmgrdError) error->code),
       error->message);
}

  /* Try to satisfy dependencies.  We can't use MarkInstall with
     AutoInst == true since we don't like how it handles conflicts,
     and we have our own way of uninstalling packages.

     The code below is lifted from pkgDepCache::MarkInstall.  Sorry
     for introducing this mess here.
  */

static void check_inst_broken (pkgCacheFile &Cache, const pkgCache::VerIterator &Ver, int default_error, GError **error)
{
  bool Now = false;
  ostringstream out;
  bool downgrade = false;

  out << "The following packages have unmet dependencies:" << endl;

  for (pkgCache::PkgIterator I = Cache->PkgBegin(); !I.end(); I++)
    {
      if (!Cache[I].InstBroken())
	continue;

      // Print out each package and the failed dependencies
      out << "  " << I.Name() << ":";
      unsigned Indent = strlen(I.Name()) + 3;
      bool First = true;

      for (pkgCache::DepIterator D = Ver.DependsList(); D.end() == false;)
	{
	  // Compute a single dependency element (glob or)
	  pkgCache::DepIterator Start;
	  pkgCache::DepIterator End;
	  D.GlobOr(Start,End); // advances D

	  if (Cache->IsImportantDep(End) == false)
	    continue;

	  if (Now == true)
	    {
	    if ((Cache[End] & pkgDepCache::DepGNow) == pkgDepCache::DepGNow)
	       continue;
	    }
	  else
	    {
	      if ((Cache[End] & pkgDepCache::DepGInstall) == pkgDepCache::DepGInstall)
		continue;
	    }

	  bool FirstOr = true;
	  while (1)
	    {
	      if (First == false)
		for (unsigned J = 0; J != Indent; J++)
		  out << ' ';
	      First = false;

	      if (FirstOr == false)
		{
		  for (unsigned J = 0; J != strlen(End.DepType()) + 3; J++)
		    out << ' ';
		}
	      else
		out << ' ' << End.DepType() << ": ";
	      FirstOr = false;

	      out << Start.TargetPkg().Name();

	      // Show a quick summary of the version requirements
	      if (Start.TargetVer() != 0)
		out << " (" << Start.CompType() << " " << Start.TargetVer() << ")";

	      /* Show a summary of the target package if possible. In
		 the case of virtual packages we show nothing */
	      pkgCache::PkgIterator Targ = Start.TargetPkg();
	      if (Targ->ProvidesList == 0)
		{
		  out << ' ';
		  pkgCache::VerIterator Ver = Cache[Targ].InstVerIter(Cache);
		  if (Now == true)
		    Ver = Targ.CurrentVer();

		  if (Ver.end() == false)
		    {
		      if (Now == true)
			out << "but " << Ver.VerStr() << " is installed";
		      else
			out << "but " << Ver.VerStr() << " is to be installed";

		      // if (debVS.CmpVersion(Start.TargetVer(), Ver.VerStr()) < 0)
		      if (!Targ.CurrentVer().end() && debVS.CmpVersion(Start.TargetVer(), Ver.VerStr()) < 0)
			downgrade = true;
		    }
		  else
		    {
		      if (Cache[Targ].CandidateVerIter(Cache).end() == true)
			{
			  if (Targ->ProvidesList == 0)
			    out << "but it is not installable";
			  else
			    out << "but it is a virtual package";
			}
		      else
			out << (Now?"but it is not installed":"but it is not going to be installed");
		    }
		}

	      if (Start != End)
		out << " or";
	      out << endl;

	      if (Start == End)
		break;
	      Start++;
	    }
	}
    }

  if (downgrade)
  {
    g_set_error (error, PKGMGRD_ERROR, ERROR_NEED_DOWNGRADE,
		 "Package '%s' would result to package being dowgraded: %s",
		 Ver.ParentPkg ().Name (), out.str ().c_str ());
  }
  else
  {
    g_set_error (error, PKGMGRD_ERROR, default_error,
		 "Package '%s' has unmet dependencies: %s",
		 Ver.ParentPkg ().Name (), out.str ().c_str ());
  }

  print_error (*error);
}

static void
print_marked ()
{
  bool details = _config->FindB ("Debug::print_marked");
  if (details)
    DBG ("Marked packages:");

  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  if (cache.BrokenCount() > 0)
    fprintf (stderr, "[ Some installed packages are broken! ]\n");

  int n_new_install = 0, n_upgrade = 0, n_downgrade = 0, n_delete = 0,
    n_inst_broken = 0;

  for (pkgCache::PkgIterator pkg = cache.PkgBegin();
       pkg.end() != true;
       pkg++)
    {
      if (cache[pkg].NewInstall())
	{
          n_new_install++;
	  if (details)
	    DBG ("  NewInstall: %s (%s)",
		 pkg.Name(),
		 cache[pkg].CandidateVerIter(cache).VerStr());
	}
      else if (cache[pkg].Upgrade())
	{
          n_upgrade++;
	  if (details)
	    DBG ("  Upgrade:    %s (%s)",
		 pkg.Name(),
		 cache[pkg].CandidateVerIter(cache).VerStr());
	}
      else if (cache[pkg].Downgrade())
	{
          n_downgrade++;
	  if (details)
	    DBG ("  Downgrade:    %s (%s)",
		 pkg.Name(),
		 cache[pkg].CandidateVerIter(cache).VerStr());
	}
      else if (cache[pkg].Delete())
	{
          n_delete++;
	  if (details)
	    DBG ("  Delete:     %s (%s)",
		 pkg.Name(),
		 pkg.CurrentVer().VerStr());
	}

      if (cache[pkg].InstBroken())
	{
          n_inst_broken++;
	  if (details)
	    DBG ("  InstBroken: %s",
		 pkg.Name());
	}
    }

  DBG ("NewInstalls: %d, Upgrades: %d, Downgrades: %d, Deletes: %d, InstBrokens: %d\n",
       n_new_install, n_upgrade, n_downgrade, n_delete, n_inst_broken);
}

static void
add_dep (GList *&deps, const pkgCache::VerIterator &ver)
{
  if (!is_user_package (ver))
    return;

  PackageVersion *pv = package_version_new ();
  pkgCache::PkgIterator const& pkg = ver.ParentPkg();
  pv->pkg = g_strdup (pkg.Name());
  pv->version = g_strdup (ver.VerStr());
  deps = g_list_append (deps, pv);
}

static void
current_operation_add_data (const pkgCache::VerIterator &cand)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);
  pkgCache::PkgIterator const& pkg = cand.ParentPkg();

  cache_operation_lock ();
  OperationData *op = cache_get_current_operation ();
  if (!op)
    {
      cache_operation_unlock ();
      return;
    }

  if (!op->version || !strcmp (op->version, ""))
    op->version = g_strdup (cand.VerStr());

  if (op)
    {
      for (pkgCache::PkgIterator p = cache.PkgBegin(); !p.end(); p++)
	{
	  if (pkg == p)
	    continue;
	  pkgDepCache::StateCache &s = cache[p];
	  if (s.Keep())
	    continue;

	  if (s.NewInstall())
	    add_dep (op->installed_user_deps, s.CandidateVerIter(cache));
	  else if (s.Upgrade())
	    add_dep (op->updated_user_deps, s.CandidateVerIter(cache));
	  else if (s.Delete())
	    add_dep (op->removed_user_deps, p.CurrentVer());
	}
    }
  cache_operation_unlock ();
}

static bool mark_autoremoved (bool protect_user_packages);

static bool
mark_for_apt_install (myCacheFile &cacheFile, const pkgCache::VerIterator &cand, bool only_user, pkgCache::VerIterator *sys_ver, GError **error)
{
  MarkPolicyDepCache &cache = cacheFile.dep_cache ();
  myMarkPolicy mark_policy;
  cache.SetMarkPolicy (&mark_policy);
  if (only_user)
    mark_policy.setMode (myMarkPolicy::UserInstall);

  pkgCache::PkgIterator const& pkg = cand.ParentPkg();
  pkgDepCache::StateCache &State = cache[pkg];
  bool ok;

  cache.SetCandidateVersion(cand);

  ProblemResolver Fix(&cache);
  Fix.Clear(pkg);
  Fix.Protect(pkg);

  ok = cache.MarkInstall1(pkg, false, 0, true, false, only_user);
  mark_policy.error (error);
  if (*error)
    {
      if (sys_ver)
	*sys_ver = mark_policy.illegalSysVer ();
      return false;
    }

  if (!State.Install())
    {
      if (pkg->CurrentVer && is_downloadable (pkg.CurrentVer()))
	cache.SetReInstall(pkg, true);
    }

  if (only_user)
    {
      if (State.InstBroken() || State.InstPolicyBroken())
	{
	  ok = cache.MarkInstall1(pkg, true, 0, true, false, only_user);
	  mark_policy.error (error);
	  if (*error)
	    {
	      if (sys_ver)
		*sys_ver = mark_policy.illegalSysVer ();
	      return false;
	    }
	}
    }
  else
    {
      cache.MarkInstall1 (pkg, true, 0, true, false, only_user);

      // Protect deps marked for install to increase problem resolver
      // score.
      ProtectInstallDeps (pkg, Fix);

      MarkDepsWithBetterCandidates (pkg, Fix);

      // Mark deps for install whose versions may be less than the
      // candidate version.  This allows for installing nongreatest
      // versions or downgrading installed versions.
      if (State.InstBroken() || State.InstPolicyBroken())
	ForceDeps (pkg, Fix);
    }

  // if (only_user && !check_no_marked_system_pkgs (cache, pkg, sys_ver, error))
  //   return false;

  Fix.InstallProtect();
  if (only_user)
    {
      // When installing 3rd party packages, we have to fix what we
      // break because the resolver might otherwise decide to remove
      // the OS.

      if (!Fix.Resolve(true, true))
	_error->Discard();

      mark_policy.error (error);
      if (*error)
	{
	  *sys_ver = mark_policy.illegalSysVer();
	  return false;
	}
    }
  else
    {
      // When installing a OS update, we remove what we break.

      if (!Fix.Resolve(false))
	_error->Discard();

      // if (*error)
      // 	return false;
    }

  /* If something went wrong already, we don't remove unneeded
     packages since that might interfere with the error reporting.
  */
  if (cache.BrokenCount() == 0)
    {
      /* Autoremoval is not actually done yet during upgrades since
         any failure to remove one of the unneeded packages will make
         the whole operation fail.  This is not acceptable, especially
         not for OS updates.

         See NB#265965.
       */
      
      // mark_autoremoved (true);
    }

  print_marked ();

  if (cache.BrokenCount() != 0)
    {
      check_inst_broken (cacheFile, cand, ERROR_MISSING_DEPENDENCIES, error);
      // ostringstream buf;
      // print_inst_broken (cacheFile, cand, buf);
      // g_set_error (error, PKGMGRD_ERROR, ERROR_MISSING_DEPENDENCIES,
      // 		  "Package '%s' has unmet dependencies: %s",
      // 		   pkg.Name (), buf.str ().c_str ());
      return false;
    }

  if (cache_get_current_operation())
    current_operation_add_data (cand);

  // if (only_user && !check_only_user_deps_marked (cand, error))
  //   return false;

  // Might be marked for reinstall, in which case State.Install() ==
  // false
  // assert (State.Install());

  return true;
}

/* Mark a package for installation, using a 'no-surprises' approach
   suitable for the Package Manager.

   Concretely, installing a package will never automatically remove
   other packages.  Thus, we undo the removals scheduled by
   MarkInstall.  Doing this will break the original package, but that
   is what we want.
*/

static gboolean
mark_for_install_1 (pkgCache::PkgIterator &pkg,
		    pkgCache::VerIterator ver,
		gboolean only_user,
		gboolean only_os_update,
		gboolean upgrade,
		gboolean dependency,
		const gchar *orig_pkg,
		pkgCache::VerIterator *sys_ver,
		GError **error)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  pkgCache::VerIterator cur = pkg.CurrentVer ();
//   DBG("%s: bad count: %d",pkg.Name(), cache.BadCount());
//   DBG("%s: now state: %d",pkg.Name(), pkg.State());
  bool failed_install = awc->cache->failed_installs.contains (pkg.Name ());
  if (!dependency) {
	  if (upgrade
	      && cur.end()
	      && !failed_install)
	  {
		  DBG("Package is not installed!");
		  g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_INSTALLED,
			  "Package '%s' is not installed", pkg.Name());
		  return FALSE;
	  } else if (!upgrade && !cur.end () && !is_broken (pkg)) {
		  DBG("Package is already installed!");
		  g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_ALREADY_INSTALLED,
			  "Package '%s' is already installed", pkg.Name());
		  return FALSE;
	  }
  }
  pkgCache::VerIterator candidate;
  if (!ver.end ())
    candidate = ver;
  else
    candidate = cache[pkg].CandidateVerIter(cache);

  if (candidate.end()) {
	  DBG("No candidate for package '%s'", pkg.Name());
	  g_set_error(error,PKGMGRD_ERROR, ERROR_NO_CANDIDATE, "No candidate for package '%s'", pkg.Name());
	  return FALSE;
  }
  if (!dependency && upgrade && !failed_install)
    {
      if (debVS.CmpVersion(cur.VerStr(), candidate.VerStr()) >= 0
	  && !is_broken (pkg))
	{
	  DBG("Package '%s' already newest version '%s'", pkg.Name(), cur.VerStr());
	  g_set_error(error,PKGMGRD_ERROR, ERROR_PACKAGE_ALREADY_NEWEST, "Package '%s' already newest version '%s'", pkg.Name(), cur.VerStr());
	  return FALSE;
	}
    }
  if (!dependency && cache_get_current_operation())
	  cache_operation_set_version (candidate.VerStr());

  if ( only_os_update ) {
    if ( !is_os_meta_package (candidate) ) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE,
		   "Package '%s' version '%s' is not system metapackage",
		   pkg.Name (),
		   candidate.VerStr ());
      return FALSE;
    }
  } else { // !only_os_update
    if (!is_user_package (candidate))
      {
	g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_USER,
		     "Package '%s' is not user package", pkg.Name());
	return FALSE;
      }

    if ( is_os_meta_package(candidate) ) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_IS_SYSTEM_PACKAGE,
		   "Package '%s' version '%s' is system package",
		   pkg.Name (),
		   candidate.VerStr ());
      return FALSE;
    }
    if ( !cur.end() && is_os_meta_package(cur) ) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_IS_SYSTEM_PACKAGE,
		   "Package '%s' version '%s' is system package",
		   pkg.Name (),
		   cur.VerStr ());
      return FALSE;
    }
  }

  mark_related (candidate);

  return mark_for_apt_install (*awc->cache, candidate, !only_os_update, sys_ver, error);
}

static gboolean
mark_for_install (pkgCache::PkgIterator &pkg,
		  pkgCache::VerIterator ver,
		  gboolean only_user,
		  gboolean only_os_update,
		  gboolean upgrade,
		  pkgCache::VerIterator *sys_ver,
		  GError **error)
{
  DBG ("INSTALL %s (%s)", pkg.Name(), !ver.end () ? ver.VerStr () : "");
  gboolean res = TRUE;

  res = mark_for_install_1 (pkg, ver, only_user, only_os_update, upgrade, FALSE, pkg.Name(), sys_ver, error);

  if (!res)
    DBG("res: %d Error: %s", res, pkgmgr_error_to_dbus_error (PkgmgrdError ((*error)->code)));

  return res;
}

struct PkgVer {
  PkgVer (const char *pkg, const char *ver) : pkg(pkg), ver(ver) { }

  string pkg;
  string ver;

  bool operator<(PkgVer const& x) const
  {
    return pkg < x.pkg
      ? true
      : (pkg == x.pkg
	 ? ver < x.ver
	 : false);
  }

  bool operator==(PkgVer const& x) const { return pkg == x.pkg && ver == x.ver; }
};

/*!
  Check if system metapackage (SMP) is marked for operation.  \a error
  is set, if it is marked for other than upgrade.

  \return true, if SMP is marked.

  \post (*error)->code == ERROR_PACKAGE_NOT_COMPATIBLE, if SMP is
  marked for other than upgrade.
 */
static bool
check_marked_os_meta_package_upgrade (pkgCache::VerIterator &sys_ver, GError **error)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  string smp = os_meta_package_name ();
  if (smp.empty ())
    return false;

  pkgCache::PkgIterator p = cache.FindPkg (smp.c_str ());
  if (p.end ())
    return false;

  pkgCache::VerIterator cur = p.CurrentVer ();
  if (cur.end ())
    return false;

  pkgDepCache::StateCache &s = cache[p];

  if (s.Install ())
    {
      pkgCache::VerIterator cand = cache[p].CandidateVerIter (cache);

      if (cand.CompareVer (cur) <= 0)
	{
	  g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE,
		       "Package would downgrade system metapackage '%s' from version '%s' to '%s'",
		       p.Name (),
		       cur.VerStr (),
		       cand.VerStr ());
	  return true;
	}

      sys_ver = cand;
      return true;
    }
  else if (s.Delete ())
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE,
		   "Package would remove system metapackage '%s' version '%s'",
		   p.Name (),
		   cur.VerStr ());
      return true;
    }

  return false;
}

static bool
check_marked_system_update(pkgCache::VerIterator &sys_ver)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *(awc->cache);
    
    for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++) {
        pkgCache::VerIterator cur = p.CurrentVer();
        if ( cache[p].Mode == pkgDepCache::ModeInstall ) {
            pkgCache::VerIterator ver = cache[p].CandidateVerIter(cache);
            if ( is_os_meta_package(ver) ) {
                sys_ver = ver;
                return true;
            } else if ( !cur.end() && is_os_meta_package(cur) ) {
                sys_ver = cur;
                return true;
            }
        }
        if ( cache[p].Mode == pkgDepCache::ModeDelete && is_os_meta_package(cur) ) {
            sys_ver = cur;
            return true;
        }
    }

    return false;
}

static bool
check_if_user_packages_marked_for_uninstall (const pkgCache::VerIterator &cand, GError **error = NULL)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  for (pkgCache::PkgIterator pkg = cache.PkgBegin (); !pkg.end (); pkg++)
    {
      pkgDepCache::StateCache &s = cache[pkg];
      if (!s.Delete ())
	continue;
      
      if ( is_in_list(pkgs_to_be_deleted, pkg.Name(), "") ) {
        continue;
      }

      pkgCache::VerIterator const& cur = pkg.CurrentVer ();
      if (!is_user_package (cur))
	continue;
      if (is_os_meta_package (cur))
	continue;

      // If cand conflicts or breaks pkg, no error
      // See SWP#PkMgr-557
      bool is_dep = false;
      bool conflicts_or_breaks = false;

      for (pkgCache::DepIterator dep = cand.DependsList (); !dep.end (); dep++)
	{
	  pkgCache::Dependency const& d = *dep;
	  pkgCache::PkgIterator const& target = dep.TargetPkg ();
	  if (target.end ())
	    continue;

	  if (dep.TargetPkg ()->ID != pkg->ID)
	      continue;

	  is_dep = true;

	  if (!(d.Type == pkgCache::Dep::Conflicts
		|| d.Type == pkgCache::Dep::DpkgBreaks))
	    {
	      DBG ("%s marked for deletion, without %s conflicting/breaking it",
		   target.Name (), cand.ParentPkg ().Name ());
	    }
	  else
	    {
	      DBG ("%s marked for deletion, but ok since %s conflicts/breaks it",
		   target.Name (), cand.ParentPkg ().Name ());
	      conflicts_or_breaks = true;
	      
	      PackageVersion *pv = package_version_new ();
              pv->pkg = g_strdup( target.Name() );
              pv->version = g_strdup( "" );
              pkgs_to_be_deleted = g_list_append( pkgs_to_be_deleted, pv);
	    }
	  break;
	}

      if (is_dep && conflicts_or_breaks)
	continue;

      if (error)
	g_set_error (error, PKGMGRD_ERROR, ERROR_DEPENDENCIES_PREVENT_OPERATION,
		     "User package '%s' would be removed", pkg.Name ());

      return true;
    }

  return false;
}

static bool
check_if_packages_marked_for_downgrade (GError **error = NULL)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  for (pkgCache::PkgIterator pkg = cache.PkgBegin (); !pkg.end (); pkg++)
    {
      pkgDepCache::StateCache &s = cache[pkg];
      if (s.Downgrade ())
	{
	  if (error)
	    g_set_error (error, PKGMGRD_ERROR, ERROR_DEPENDENCIES_PREVENT_OPERATION,
			 "Package '%s' would be downgraded", pkg.Name ());

	  return true;
	}
    }

  return false;
}

/* Mark the named package for installation.  This function also
   handles magic packages like "magic:sys".
*/

bool
mark_named_package_for_install (const char *package, const char *version, const char *domain, 
                                gboolean only_user, gboolean only_os_update, gboolean upgrade, GError **error)
{
	if (version && strlen(version) == 0)
		version = NULL;

	AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
        pkgDepCache &cache = *(awc->cache);
        pkgCache::PkgIterator pkg = cache.FindPkg (package);
        if (!pkg.end()) {
          pkgCache::VerIterator ver;
          if (version != NULL) {
            if (!(awc->cache->find_version (pkg, version, domain, ver, error)
                  && awc->cache->check_compatible_domains (ver, error))) {
              return false;
            }
	    if (!is_downloadable (ver))
	      {
		g_set_error (error, PKGMGRD_ERROR, ERROR_NO_CANDIDATE,
			     "No candidate for package '%s' version '%s'",
			     package, version);
		return false;
	      }
          }

          if (ver.end ())
            ver = cache[pkg].CandidateVerIter (cache);

          pkgCache::VerIterator sys_ver;
          gboolean res = mark_for_install (pkg, ver, only_user, only_os_update, upgrade, &sys_ver, error);

          if ( res && only_user && !only_os_update ) {
            if ( check_marked_os_meta_package_upgrade (sys_ver, error)) {
              if ( !*error ) {
                g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE,
                             "Package requires system upgrade '%s' version '%s'",
                             sys_ver.ParentPkg ().Name (),
                             sys_ver.VerStr ());
              }
              
              res = FALSE;
            } else if ( check_marked_system_update(sys_ver) ) {
              
              if ( !*error ) {
                g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_IS_SYSTEM_PACKAGE,
                             "Package '%s' version '%s' is system package",
                             sys_ver.ParentPkg ().Name (),
                             sys_ver.VerStr ());
              }
              
              res = FALSE;
            } else if (check_if_user_packages_marked_for_uninstall (ver, error)) {
              res = FALSE;
            } else if (check_if_packages_marked_for_downgrade (error)) {
              res = FALSE;
            }
          }
          
          DBG ("%d: %s", res, !*error ? "" : pkgmgr_error_to_dbus_error (PkgmgrdError ((*error)->code)));
          
          return res == TRUE;
        } else {
          g_set_error(error,PKGMGRD_ERROR, ERROR_NO_CANDIDATE, "No candidate for package '%s'", package);
          return false;
        }
}

class MyRootSetFunc : public pkgDepCache::DefaultRootSetFunc
{
  virtual bool InRootSet (const pkgCache::PkgIterator &pkg)
  {
    return pkg.CurrentVer() != 0 && is_user_package (pkg.CurrentVer());
  }
};

bool protect_user_packages_from_autoremoval = false;

static bool
mark_autoremoved (bool protect_user_packages)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgCacheFile &Cache = *awc->cache;

  if (protect_user_packages_from_autoremoval
      || protect_user_packages)
    {
      MyRootSetFunc rootset;
      Cache->MarkAndSweep (rootset);
    }
  else
    Cache->MarkAndSweep ();

  bool purgePkgs = _config->FindB("APT::Get::Purge", false);

  // look over the cache to see what can be removed
  for (pkgCache::PkgIterator Pkg = Cache->PkgBegin(); ! Pkg.end(); ++Pkg)
    {
      if (Cache[Pkg].Garbage)
	{
          if(Pkg.CurrentVer() != 0 &&
             Pkg->CurrentState != pkgCache::State::ConfigFiles)
            {
              g_debug ("autoremoving %s", Pkg.Name());
              Cache->MarkDelete(Pkg, purgePkgs);
            }
          else
            Cache->MarkKeep(Pkg, false, false);
        }
    }

  return true;
}

static bool
mark_for_apt_remove (const pkgCache::PkgIterator &pkg, bool only_user, GError **error)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  MarkPolicyDepCache &Cache = awc->cache->dep_cache ();
  myMarkPolicy mark_policy;
  Cache.SetMarkPolicy (&mark_policy);
  if (only_user)
    mark_policy.setMode (myMarkPolicy::UserInstall);

  pkgCache::VerIterator cur = pkg.CurrentVer ();

  mark_related (cur);

  pkgProblemResolver Fix(&Cache);
  Fix.Clear(pkg);
  Fix.Protect(pkg);
  Fix.Remove(pkg);

  Cache.MarkDelete1(pkg, _config->FindB("APT::Get::Purge", false));
  mark_policy.error (error);
  if (*error)
    return false;

  if (Cache.BrokenCount() != 0)
    {
      check_inst_broken (*awc->cache, cur, ERROR_DEPENDENCIES_PREVENT_UNINSTALL, error);
      return false;
    }

  Fix.InstallProtect();
  if (!Fix.Resolve(false))
    _error->Discard();

  mark_policy.error (error);
  if (*error)
    return false;

  mark_autoremoved (false);
  print_marked ();

  current_operation_add_data (cur);

  return true;
}

static gboolean
mark_for_remove_1 (pkgCache::PkgIterator &pkg, gboolean only_user, GError **error)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  DBG ("%s - %s", __PRETTY_FUNCTION__, pkg.Name());

  if (is_related (pkg)) {
    return TRUE;
  } else {
    awc->cache->extra_info[pkg->ID].related = true;
  }

  pkgCache::VerIterator cur = pkg.CurrentVer ();
  if (cur.end ()) {
	  DBG("Package is not installed!");
	  g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_INSTALLED,
		  "Package '%s' is not installed", pkg.Name());
	  return FALSE;
  }

  if (only_user && !is_user_package(cur)) {
	  DBG("Refusing to remove non-user package!");
	  g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_USER,
		  "Package '%s' is not user package", pkg.Name());
	  return FALSE;
  }

  if (only_user && is_os_meta_package(cur))
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE,
		   "Package '%s' is system metapackage and cannot be removed",
		   pkg.Name());
      return FALSE;
    }

  return mark_for_apt_remove (pkg, only_user, error);
}

static gboolean
mark_for_remove (pkgCache::PkgIterator &pkg, gboolean only_user, GError **error)
{
  DBG ("REMOVE %s", pkg.Name());
  gboolean res = TRUE;
  res = mark_for_remove_1 (pkg, only_user, error);
  return res;
}


/** COMMAND HANDLERS
 */

/* APTCMD_GET_PACKAGE_LIST 

   The get_package_list command can do some filtering and we have a
   few utility functions for implementing the necessary checks.  The
   check generally take cache iterators to identify a package or a
   version.
 */

static gchar*
get_package_repository (pkgCache::VerIterator Version);

// static bool
// name_matches_pattern (pkgCache::PkgIterator &pkg,
// 		      const char *pattern)
// {
//   bool match = false;
//   char **words = g_strsplit (pattern, " ", 0);
//   int i;

//   if (words == NULL)
//     return false;

//   for (i = 0; words[i] != NULL; i++)
//     if (strcasestr (pkg.Name(), words[i]))
//       match = true;
//     else
//       {
//         match = false;
//         break;
//       }

//   g_strfreev (words);
//   return match;
// }

// static bool
// description_matches_pattern (pkgCache::VerIterator &ver,
// 			     const char *pattern)
// {
//   bool match = false;
//   char **words = g_strsplit (pattern, " ", 0);
//   package_record rec (ver);
//   const char *desc = rec.P.LongDesc().c_str();
//   int i;

//   if (words == NULL)
//     return false;

//   for (i = 0; words[i] != NULL; i++)
//     if (strcasestr (desc, words[i]))  // XXX - UTF8?
//       match = true;
//     else
//       {
//         match = false;
//         break;
//       }

//   g_strfreev (words);
//   return match;
// }

static string
get_description (gboolean update_desc,
		 pkgCache::PkgIterator &pkg,
		 package_record &rec)
{
  string res;

  if (update_desc == 1 && !pkg.CurrentVer().end())
    res = rec.get_localized_multiline_string ("Maemo-Upgrade-Description");

  if (res.empty())
    {
      /* XXX - support apt's own method of localizing descriptions as
	       well.
      */
      res = rec.get_localized_multiline_string ("Description");
    }

  return res;
}

// static string
// get_short_description (int summary_kind,
// 		       pkgCache::PkgIterator &pkg,
// 		       package_record &rec)
// {
//   string res = get_description (summary_kind, pkg, rec);
//   string::size_type pos = res.find('\n');
//   if (pos != string::npos)
//     return string (res,0,pos);
//   return res;
// }

static string
get_long_description (gboolean update_desc,
		      pkgCache::PkgIterator &pkg,
		      package_record &rec)
{
  return get_description (update_desc, pkg, rec);
}

static char *
get_icon (package_record &rec)
{
  return rec.get ("Maemo-Icon-26");
}


static string
get_display_name (package_record &rec)
{
  return rec.get_localized_string ("Maemo-Display-Name");
}

static int64_t
get_required_free_space (package_record &rec)
{
  return 1024 * (int64_t) rec.get_int ("Maemo-Required-Free-Space", 0);
}

static int64_t
get_installed_size_bytes (const pkgCache::VerIterator &ver)
{
  return ver->InstalledSize;
}

static int64_t
get_installed_size_delta_bytes (const pkgCache::VerIterator &cand, const pkgCache::VerIterator &cur = pkgCache::VerIterator())
{
  int64_t dsize = 0;
  if (!cand.end())
    dsize += get_installed_size_bytes (cand);

  if (!cur.end())
    dsize -= get_installed_size_bytes (cur);

  return dsize;
}

static void
pkg_insert_version_info (Package *pkgli,
			pkgCache::VerIterator &ver,
			gboolean update_desc,
			bool need_reboot = false)
{
	g_return_if_fail(!ver.end());
	package_record rec (ver);

	pkgli->version = g_strdup(ver.VerStr ());
	pkgli->installed_size = ver->InstalledSize;
	pkgli->download_size = ver->Size;
	pkgli->category = g_strdup(user_section_to_category(ver.Section ()));
	string pretty = get_display_name (rec);
	pkgli->display_name = g_strdup(pretty.c_str());
	pkgCache::PkgIterator pkg = ver.ParentPkg();
	pkgli->description =
		g_strdup(get_long_description (update_desc, pkg, rec).c_str());
	pkgli->icon_data = get_icon (rec);
	pkgli->repository = get_package_repository (ver);

	if ( is_os_meta_package(ver) )
          pkgli->flag_system_update = TRUE;
        
        if (need_reboot)
          pkgli->flag_reboot = TRUE;
        else {
          package_record rec (ver);
          int flags = get_flags (rec);
          if (flags & pkgflag_reboot )
            pkgli->flag_reboot = TRUE;
	}
}

static bool
check_required_os_upgrade (pkgCache::VerIterator const& ver,
			   pkgCache::VerIterator &sys_ver);

static void
pkg_insert_os_meta_package_details (Package *p, const pkgCache::VerIterator &ver)
{
  pkgCache::VerIterator sys_ver;

  if (check_required_os_upgrade (ver, sys_ver)
      && !sys_ver.end ())
    {
      pkgCache::PkgIterator sys_pkg = sys_ver.ParentPkg ();
      p->sys_upgrade_name = g_strdup (sys_pkg.Name ());
      p->sys_upgrade_version = g_strdup (sys_ver.VerStr ());
    }
}

static void
_insert_str (GHashTable *map, const char *key, gchar *val)
{
  g_hash_table_insert_string_value (map, key, val);
  g_free (val);
}

static const char *
_find_string (const char *strings[], const char *s)
{
  for (int i = 0; strings[i] != 0; ++i)
    {
      if (!strcmp (strings[i], s))
	return strings[i];
    }

  return 0;
}

static void
pkg_map_insert_extra_fields(GHashTable *map,
		       pkgCache::VerIterator &ver,
                       bool only_display_name = false,
                       bool but_also_include_descriptions = false)
{
  // Insert additional fields not listed in skip_fields array.

  static const char *skip_fields[] = {
    "Description",
    "Installed-Size",
    "Maemo-Display-Name",
    "Maemo-Icon-26",
    "Package",
    "Size",
    "Version",
    0
  };
  package_record rec (ver);
  for (FieldIterator i (rec.section); !i.end (); ++i)
    {
      string const& key = i.key ();

      if (!_find_string (skip_fields, key.c_str ()))
	{
	  string const &value = i.value ();
	  
	  if ((!only_display_name || but_also_include_descriptions)
	      && (g_str_has_prefix (key.c_str (), "Description")
	      || g_str_has_prefix (key.c_str (), "Maemo-Upgrade-Description")))
	    {
	      gchar *f = format_multiline_field (value.c_str ());
	      g_hash_table_insert_string_value (map,
						key.c_str (),
						f);
	      g_free (f);
	    }
	  else
	    {
	      if ( only_display_name ) {
                if ( g_str_has_prefix (key.c_str (), "Maemo-Display-Name") ) {
	          g_hash_table_insert_string_value (map,
						    key.c_str (),
						    value.c_str ());
		}
	      } else {
                g_hash_table_insert_string_value (map,
                                                  key.c_str (),
						  value.c_str ());
              }
	    }
	}
    }
}

static gboolean
pkg_insert_details(Package *pkgli,
		pkgCache::PkgIterator pkg,
		pkgCache::VerIterator ver,
		const char* failed_version,
		gboolean update_desc,
		GError **error,
                bool need_reboot = false,
                bool check_os_details = true,
                bool include_descriptions = false)
{
	gboolean res = TRUE;

	// Name
	pkgli->name = g_strdup(pkg.Name ());

	// Broken.
	pkgli->is_broken = is_broken (pkg);

	if (update_desc && check_os_details)
		pkg_insert_os_meta_package_details (pkgli, ver);

        if ( !ver.end() ) {
            pkg_insert_version_info (pkgli, ver, update_desc, need_reboot);

            pkgli->extra_fields = g_hash_table_new (g_str_hash, g_str_equal);
            pkg_map_insert_extra_fields(pkgli->extra_fields, ver, true, include_descriptions);
        } else if ( failed_version ) {
            pkgli->version = g_strdup(failed_version);
        }

// 	response.encode_string (rec.P.Maintainer().c_str());
// 	response.encode_string 
// 		(get_long_description (summary_kind, pkg, rec).c_str());
// 	encode_dependencies (ver);
// 	encode_package_repository (ver, summary_kind);
// 	if (summary_kind == 1)
// 		encode_install_summary (package);
// 	else if (summary_kind == 2)
// 		encode_remove_summary (pkg);
// 	else
// 		response.encode_int (sumtype_end);
	return res;
}
// static void
// encode_empty_version_info (bool include_size, Package *pkgli)
// {
// //   g_free(pkgli->version);
//   pkgli->version = g_strdup("");
//   if (include_size)
//     pkgli->installed_size = 0;
// //   g_free(pkgli->section);
//   pkgli->section = g_strdup("");
// //   g_free(pkgli->display_name);
//   pkgli->display_name = g_strdup("");
// //   g_free(pkgli->description);
//   pkgli->description = g_strdup("");
// //   g_free(pkgli->icon_data);
//   pkgli->icon_data = g_strdup ("");
// }

static void
pkg_map_insert_fields (GHashTable *map,
		       pkgCache::VerIterator &ver,
		       gboolean update_desc,
		       bool need_reboot=false)
{
  g_return_if_fail (!ver.end ());

  package_record rec (ver);
  pkgCache::PkgIterator pkg = ver.ParentPkg();

  g_hash_table_insert_string_value (map, "Version", ver.VerStr ());
  g_hash_table_insert_int64_value (map, "InstalledSize", ver->InstalledSize);
  g_hash_table_insert_int64_value (map, "Size", ver->Size);
  g_hash_table_insert_string_value (map, "Category",
				    user_section_to_category (ver.Section ()));
  g_hash_table_insert_string_value (map, "DisplayName",
				    get_display_name (rec).c_str ());
  g_hash_table_insert_string_value (map, "Description",
				    get_long_description (update_desc, pkg, rec).c_str ());
  _insert_str (map, "IconData", get_icon (rec));
  _insert_str (map, "Repository", get_package_repository (ver));

  if ( is_os_meta_package(ver) )
    g_hash_table_insert_boolean_value (map, "FlagSystemUpdate", TRUE);
  
  if (need_reboot)
    g_hash_table_insert_boolean_value (map, "FlagReboot", TRUE);
  else {
    package_record rec (ver);
    int flags = get_flags (rec);
    if (flags & pkgflag_reboot )
      g_hash_table_insert_boolean_value (map, "FlagReboot", TRUE);
  }

  pkg_map_insert_extra_fields(map, ver);
}

/*!
  Check if installation or upgrade of \a ver would succeeed without
  installation of additional non-user packages.

  \param[out] sys_ver System metapackage version that would enable
  installation of \a ver, if applicable.

  \return true, if additional non-user packages are required
  \return false, if no new non-user packages are required
*/
static bool
check_required_os_upgrade (pkgCache::VerIterator const& ver,
			       pkgCache::VerIterator &sys_ver)
{
  pkgCache::PkgIterator pkg = ver.ParentPkg ();
  pkgCache::VerIterator cur = pkg.CurrentVer ();
  bool upgrade = !cur.end () ? debVS.CmpVersion (ver.VerStr (), cur.VerStr ()) : false;

  GError *err = NULL;
  mark_for_install (pkg, ver, TRUE, FALSE, upgrade, &sys_ver, &err);
  bool is_required = !sys_ver.end ();

  cache_reset_packages ();

  return is_required;
}

static void
pkg_map_insert_failed_install_data (GHashTable *p, const char *package, const char *version)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    if (!awc)
      return;

    if ( awc->cache->failed_installs.get_data (package, version, p))
    {
        DBG ("install failure found for %s (%s)", package, version);
    }
}

static void
pkg_map_insert_failed_update_data (GHashTable *p, const char *package, const char *version)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    if (!awc)
        return;

    if( awc->cache->failed_upgrades.get_data (package, version, p) )
    {
        DBG ("update failure found for %s (%s)", package, version);
    }
}

static GValue *package_value_list_to_gvalue (GList *l);

static void
pkg_map_insert_deps (GHashTable *map,
		     PkgType install_status,
		     pkgCache::PkgIterator pkg,
		     pkgCache::VerIterator ver)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

    pkgDepCache &cache = *(awc->cache);
    
    GError* error = NULL;

    gboolean upgrade = install_status == PKG_TYPE_UPDATE;
    gboolean is_omp = is_os_meta_package (ver);
    pkgCache::VerIterator sys_ver;
    bool ok = mark_for_install(pkg, ver, TRUE, is_omp, upgrade, &sys_ver, &error);

    if (ok || !sys_ver.end ()) {
        GList* install_deps = NULL;
        GList* upgrade_deps = NULL;
        GList* remove_pkgs = NULL;

        gint64 download_size = 0;
        gint64 install_size = 0;

        for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++) {
//            pkgDepCache::StateCache &state = cache[p];
            if ( pkg == p ) {
                continue;
            }
            if (cache[p].Mode == pkgDepCache::ModeInstall) {
                pkgCache::VerIterator ver = cache[p].CandidateVerIter(cache);
                pkgCache::VerIterator cur = p.CurrentVer();

                if ( cur.end() ) {
		    PackageVersion *pv = package_version_new ();
		    pv->pkg = g_strdup (p.Name ());
		    pv->version = g_strdup (ver.VerStr ());
		    install_deps = g_list_append (install_deps, pv);
                } else {
		    PackageVersion *pv = package_version_new ();
		    pv->pkg = g_strdup (p.Name ());
		    pv->version = g_strdup (ver.VerStr ());
		    upgrade_deps = g_list_append (upgrade_deps, pv);
                }

		if (!cur.end())
		    install_size += get_installed_size_delta_bytes (ver, cur);
		else
		    install_size += get_installed_size_bytes (ver);
            } else if (cache[p].Mode == pkgDepCache::ModeDelete) {
                pkgCache::VerIterator ver = p.CurrentVer();

		{
		    PackageVersion *pv = package_version_new ();
		    pv->pkg = g_strdup (p.Name ());
		    pv->version = g_strdup (ver.VerStr ());
		    remove_pkgs = g_list_append (remove_pkgs, pv);
		}

		install_size -= get_installed_size_bytes (ver);
            }
        }

        download_size = (gint64) cache.DebSize ();
//        gint64 install_sz = (gint64) cache.UsrSize ();
//        g_print("%s Download size: %lld Install size: %lld Install sz: %lld\n", __PRETTY_FUNCTION__, download_size, install_size, install_sz);
        
        if ( install_deps ) {
	    GValue *val = package_value_list_to_gvalue (install_deps);
	    g_hash_table_insert (map, g_strdup ("InstallDeps"), val);
        }
        if ( upgrade_deps ) {
	    GValue *val = package_value_list_to_gvalue (upgrade_deps);
	    g_hash_table_insert (map, g_strdup ("UpgradeDeps"), val);
        }
        if ( remove_pkgs ) {
	    GValue *val = package_value_list_to_gvalue (remove_pkgs);
	    g_hash_table_insert (map, g_strdup ("RemovePkgs"), val);
            // g_hash_table_insert_string_value (map, "RemovePkgs", remove_pkgs);
        }

        g_hash_table_insert_int64_value (map, "Size", ver->Size);
        g_hash_table_insert_int64_value (map, "DownloadSize", download_size);
        g_hash_table_insert_int64_value (map, "TotalInstalledSize", install_size);

	if (!sys_ver.end ())
	{
	    pkgCache::PkgIterator sys_pkg = sys_ver.ParentPkg ();
	    g_hash_table_insert_string_value (map, "DependsSystemUpgradeName", sys_pkg.Name ());
	    g_hash_table_insert_string_value (map, "DependsSystemUpgradeVersion", sys_ver.VerStr ());
	}

	if (error)
	    g_error_free (error);

//        g_print("%s Download size: %lld Install size: %lld\n", __PRETTY_FUNCTION__, download_size, install_size);
    } else {
//        g_print("%s Packages NOT marked.\n", __PRETTY_FUNCTION__);
        // TODO: Now what?
    }
    
    cache_reset_packages ();
}

static void
pkg_map_insert_details (GHashTable *map,
                        pkgCache::PkgIterator pkg,
                        pkgCache::VerIterator ver,
                        gboolean update_desc,
                        PkgType type,
                        bool need_reboot=false)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

  g_hash_table_insert_string_value (map, "Name", pkg.Name ());

  gboolean is_broken = ::is_broken (pkg);

  g_hash_table_insert_boolean_value (map, "IsBroken", is_broken);

  g_hash_table_insert_string_value (map, "Origin", awc->cache->get_domain_name(ver));

  if( type & PKG_TYPE_UPDATE ) {
    DBG ("checking for update exception for package %s %s", pkg.Name(), ver.VerStr() );
    pkg_map_insert_failed_update_data (map, pkg.Name (), ver.VerStr ());
  }
  else{
    DBG ("checking for installation exception for package %s %s", pkg.Name(), ver.VerStr() );
    pkg_map_insert_failed_install_data (map, pkg.Name (), ver.VerStr ());
  }

  // pkg_map_insert_system_metapackage_details (map, ver);

  pkg_map_insert_fields (map, ver, update_desc, need_reboot);
}

static void
ssu_packages_free ()
{
  if (ssu_packages != NULL)
    {
      g_array_free (ssu_packages, TRUE);
      ssu_packages = NULL;
    }
}

static void
ssu_packages_set (GSList *ssu_list)
{
  ssu_packages_free ();

  if (ssu_list != NULL)
    {
      ssu_packages = g_array_sized_new (TRUE, FALSE, sizeof (gchar*),
                                        g_slist_length (ssu_list));
      for (GSList *item = g_slist_reverse (ssu_list);
           item; item = g_slist_next (item))
        {
          /* Do not strdup strings as they will be used in the GArray */
          ssu_packages = g_array_append_val (ssu_packages, item->data);
        }
    }
}

static gint
package_name_cmp (gconstpointer a, gconstpointer b)
{
	const Package *pkga = static_cast<const Package*>(a);
	const Package *pkgb = static_cast<const Package*>(b);
	return g_strcmp0(pkga->name, pkgb->name);
}

static gboolean
category_matches(pkgCache::VerIterator &ver, const gchar *category)
{
	if (!category) {
		return TRUE;
	}
	const gchar *cat = user_section_to_category(ver.Section ());
// 	DBG("Comparing '%s' and '%s'",category, cat);
	return g_strcmp0(category, cat) == 0;
}

static bool
is_reboot_needed_package_marked_for_install()
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *(awc->cache);

    int flags = 0;
    for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++) {
        if (cache[p].Mode == pkgDepCache::ModeInstall) {
            pkgCache::VerIterator ver = cache[p].CandidateVerIter(cache);
            package_record rec (ver);

            flags = get_flags(rec);
            if ( need_reboot (flags) ) {
                return true;
            }
        }
    }

    return false;
}

/*!
 Return true, if \a ver or any of its dependency package would require a reboot.
*/
static bool
require_reboot (pkgCache::VerIterator ver)
{
	AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
	if (awc->cache == NULL)
		return false;

	GError *error = NULL;

	pkgCache::PkgIterator pkg = ver.ParentPkg ();
	bool b = mark_for_install (pkg,
				   ver,
				   TRUE,
				   TRUE,
				   FALSE,
				   0,
				   &error);
	if ( !b ) {
	    g_error_free(error);
	    error = NULL;
	    b = mark_for_install (pkg,
				   ver,
				   TRUE,
				   TRUE,
				   TRUE,
				   0,
				   &error);
        if ( !b ) {
    	    g_error_free(error);
    	}
	}
	if ( b ) {
		b = is_reboot_needed_package_marked_for_install();
	}

	cache_reset_packages();

	return b;
}

static gboolean
__filter_package_list (pkgCache::PkgIterator pkg,
		pkgCache::VerIterator installed,
		pkgCache::VerIterator candidate,
		PkgType type,
		const gchar *category,
		gboolean only_user,
		pkgCache::VerIterator *req_sys_pkg)
{
	// skip non user packages if requested.  Both the installed and
	// candidate versions must be non-user packages for a package to
	// be skipped completely.
	//
	if ( only_user ) {
	    if ( (installed.end () || !is_user_package (installed))
	       && (candidate.end () || !is_user_package (candidate))) {
//   	      DBG("pkg: '%s': not user package", pkg.Name ());
	        return TRUE;
		}
		// skip system update meta-packages that are not installed
	    if (installed.end () && !candidate.end ()) {
              if (is_os_meta_package (candidate))
                return TRUE;
	    }
        // Also skip non user update packages
        if ( type & PKG_TYPE_UPDATE && !candidate.end() && !is_user_package(candidate) ) {
            DBG("pkg: '%s': update candidate not user", pkg.Name ());
            return TRUE;
        }
	}

	// skip packages that are not installed and not available
	if (installed.end () && candidate.end ()) {
		// DBG("pkg: '%s': not installed, not available", pkg.Name ());
		return TRUE;
	}

    if ( type & PKG_TYPE_UPDATE ) {
        if (only_user && !installed.end() && !candidate.end () && is_os_meta_package(installed) != is_os_meta_package(candidate) ) {
		    return TRUE;
	    }
	    if (installed.end ()) {
	        return TRUE;
	    }
	    if ( candidate.end() ) {
		    DBG("pkg: '%s': wanted with status 'update' but not available", pkg.Name ());
		    return TRUE;
	    }
	    int c = debVS.CmpVersion(candidate.VerStr(), installed.VerStr());
	    if (c < 0 || (c == 0 && !is_broken(pkg))) {
            DBG("'%s': %s not update to %s", pkg.Name(), candidate.VerStr(), installed.VerStr());
            return TRUE;
	    }
	} else if ( type & PKG_TYPE_NOT_INSTALLED ) {
    	// skip non-available packages if requested
	    if ( candidate.end () || !installed.end()) {
//	    	DBG("pkg: '%s': wanted with status 'not installed' but not available or is installed"
//	    		, pkg.Name ());
		    return TRUE;
	    }
	} else if ( type & PKG_TYPE_INSTALLED ) {
	    if ( installed.end ()) {
//    		DBG("pkg: '%s': wanted with status 'installed' but not installed"
//    			, pkg.Name ());
		    return TRUE;
	    }
	}

    OperationData* cur = cache_get_current_operation();
    if ( cur && cur->type == OPERATION_UNINSTALL && strcmp(pkg.Name(), cur->pkg) == 0 && get_operation_started_signal_sent() ) {
        DBG("pkg: '%s': is being uninstalled", pkg.Name() );
        return TRUE;
    }

//       if(!installed.end()) {
// 	      DBG("pkg: '%s' version %s installed", pkg.Name (), installed.VerStr());
//       }
//       if(!candidate.end()) {
// 	      DBG("pkg: '%s' version %s available", pkg.Name (), candidate.VerStr());
//       }
	if ((type & (PKG_TYPE_NOT_INSTALLED | PKG_TYPE_UPDATE))
		&& !category_matches(candidate, category)) {
		return TRUE;
	} else if (type & PKG_TYPE_INSTALLED
		&& !category_matches(installed, category)) {
		return TRUE;
	}
	return FALSE;
}

void
ensure_ignored_update_pkg_list()
{
    if ( ignored_update_pkgs == NULL ) {
        ignored_update_pkgs = new PkgTsStorage();
        ignored_update_pkgs->readFromFile( IGNORED_PACKAGE_LIST );
    }
}

void
ensure_available_update_pkg_list()
{
    if ( available_update_pkgs == NULL ) {
        available_update_pkgs = new PkgTsStorage();
        available_update_pkgs->readFromFile( AVAILABLE_PACKAGE_LIST );
    }
}

void
ensure_restore_pkg_list()
{
    if ( restore_pkgs == NULL ) {
        restore_pkgs = new PkgTsStorage();
    }

    string tmp_list(RESTORE_PACKAGE_LIST_TMP);
    if (!g_file_test("/home/user", G_FILE_TEST_IS_DIR) && getenv("HOME"))
    {
	string home(getenv("HOME"));
	tmp_list.replace(0, sizeof("/home/user")-1, home);
    }

    if ( restore_pkgs->readFromFile( tmp_list.c_str() ) ) {
        restore_pkgs->saveToFile( RESTORE_PACKAGE_LIST );
        new_restore = true;
        g_remove( tmp_list.c_str() );
    } else {
        new_restore = false;
        restore_pkgs->readFromFile( RESTORE_PACKAGE_LIST );
    }
}

void
cmd_get_package_list(const gchar *category,
                     PkgType type,
                     const gchar *pattern,
                     GList **packages,
                     PkgSort sort,
                     GError **error)
{
    GSList *ssu_pkgs_found = NULL;
    gboolean only_user = TRUE;

    if (!ensure_cache (true)) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
      return;
    }

    cache_reset_packages ();
    current_operation = 0;

    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

    pkgDepCache &cache = *(awc->cache);
    myCacheFile* c = awc->cache;

    for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++) {
        /* Get installed and candidate iterators for current package */
        pkgCache::VerIterator installed = pkg.CurrentVer ();
        pkgCache::VerIterator candidate = cache[pkg].CandidateVerIter(cache);

        // Look for the SSU package if needed
        if (ssu_packages_needs_refresh && (!installed.end () || !candidate.end ())) {
            pkgCache::VerIterator viter = !candidate.end ()
                ? candidate
                : installed;
            if (is_os_meta_package (viter)) {
                /* Add it to the local GSList */
                ssu_pkgs_found = g_slist_prepend (ssu_pkgs_found,
                                                  g_strdup (pkg.Name ()));
            }
        }

        const char *failed_install_version = NULL;
        const char *failed_upgrade_version = NULL;

        if (type == PKG_TYPE_INSTALLED || type == PKG_TYPE_UPDATE) {
            const char *v = awc->cache->failed_installs.failed_version(pkg.Name());
            if (v && (installed.end () || !strcmp (installed.VerStr (), v))) {
                awc->cache->find_version(pkg, v, NULL, installed);
                failed_install_version = v;
                DBG ("installation exception found for package %s (%s)", pkg.Name(), failed_install_version);
            }
        }

        if (!installed.end () && !candidate.end () && type == PKG_TYPE_UPDATE) {
            const char *v = awc->cache->failed_upgrades.failed_version(pkg.Name());
            if (v && !strcmp (candidate.VerStr (), v)) {
                awc->cache->find_version (pkg, v, NULL, candidate);
                failed_upgrade_version = v;
                DBG ("update exception found for package %s (%s)", pkg.Name(), failed_upgrade_version);
            }
        }

        pkgCache::VerIterator sys_ver;
        if ( ( type != PKG_TYPE_INSTALLED || !failed_install_version) && __filter_package_list(pkg, installed, candidate, type, category, only_user, &sys_ver)) {
            continue;
        }
      // skip packages that don't match the pattern if requested
      //
//       if (pattern
// 	  && !(name_matches_pattern (pkg, pattern)
// 	       || (!installed.end ()
// 		   && description_matches_pattern (installed, pattern))
// 	       || (!candidate.end ()
// 		       && description_matches_pattern (candidate, pattern)))) {
// // 	      DBG("pkg: '%s': does not match pattern '%s'", pkg.Name (), pattern);
// 	continue;
//       }

        Package *pkgli = package_new();
        pkgli->error_name = NULL;
        pkgli->error_message = NULL;
        pkgli->type = type;

        bool ok = false;
        if ( type & PKG_TYPE_INSTALLED ) {
            pkgli->installed_timestamp = c->package_timestamp(pkg.Name(), &ok);
            if ( ! ok ) {
	      // DBG("No install timestamp for installed package '%s'", pkg.Name());
            } else {
                ok = false;
                const gchar* fn = c->package_filename(pkg.Name(), &ok);
                if ( ok && fn && strcmp(fn, "") != 0 ) {
                    pkgli->filename = g_strdup(fn);
                    pkgli->flag_local_file = TRUE;
                }
            }
        }

        pkgCache::VerIterator wanted_version;
        gboolean update_desc = FALSE;
        if (type & (PKG_TYPE_NOT_INSTALLED | PKG_TYPE_UPDATE)) {
            wanted_version = candidate;
            update_desc = TRUE;
            if ( type & PKG_TYPE_UPDATE ) {
                ok = false;
                ensure_ignored_update_pkg_list();
                ignored_update_pkgs->getPackageTimestamp(pkg.Name(), wanted_version.VerStr(), &ok);
                if ( ok ) { // Package is in the ignored list
                    pkgli->flag_ignored = TRUE;
                }
                ok = false;
                ensure_available_update_pkg_list();
                time_t ts = available_update_pkgs->getPackageTimestamp(pkg.Name(), wanted_version.VerStr(), &ok);
                if ( ok ) {
                    pkgli->available_timestamp = ts;
                }
            }
        } else { /* status & PKG_TYPE_INSTALLED */
            wanted_version = installed;
        }

        if ( !(type & PKG_TYPE_UPDATE) && failed_install_version && (wanted_version.end() || !strcmp (wanted_version.VerStr (), failed_install_version))) {
            const char *en, *em, *fn, *dm;
            awc->cache->failed_installs.get_data(pkg.Name(), failed_install_version, en, em, fn, dm);
            pkgli->error_name = g_strdup(en);
            pkgli->error_message = g_strdup(em);
            DBG ("setting install exception %s for package %s", pkgli->error_name, pkg.Name() );
            if ( fn && strcmp(fn, "") != 0 ) {
                pkgli->filename = g_strdup(fn);
                pkgli->flag_local_file = TRUE;
            }
        }
        else if (failed_upgrade_version && !strcmp (wanted_version.VerStr (), failed_upgrade_version)) {
            const char *en, *em, *fn, *dm;
            awc->cache->failed_upgrades.get_data(pkg.Name(), failed_upgrade_version, en, em, fn, dm);
            pkgli->error_name = g_strdup(en);
            pkgli->error_message = g_strdup(em);
            DBG ("setting update exception %s for package %s", pkgli->error_name, pkg.Name() );
        }
        
        if (!pkg_insert_details(pkgli, pkg, wanted_version, failed_install_version, update_desc, error,
                                false, false, is_os_meta_package (wanted_version))) {
            DBG("Internal error");
            package_free (pkgli);
            return;
        }
      // Available version
      //
      // We only offer an available version if the package is not
      // installed at all, or if the available version is newer than
      // the installed one, or if the installed version is broken.

//       if (!candidate.end ()
// 	  && (installed.end ()
// 	      || installed.CompareVer (candidate) < 0
// 	      || broken))
// 	      encode_version_info (1, candidate, false, pkgli);
//       else
// 	      encode_empty_version_info (false, pkgli);

        switch(sort) {
            case PKG_SORT_NAME:
            case PKG_SORT_RELEASE_TIME:
            case PKG_SORT_INSTALL_TIME:
            *packages = g_list_insert_sorted(*packages, pkgli, package_name_cmp);
        }
        pkgli = NULL;
    }

    /* Update the global GArray, if needed */
    if (ssu_packages_needs_refresh) {
        ssu_packages_set (ssu_pkgs_found);

        if (ssu_pkgs_found != NULL) {
            /* Free local GSList */
            g_slist_free (ssu_pkgs_found);
            ssu_pkgs_found = NULL;
        }

        /* Update global flag */
        ssu_packages_needs_refresh = false;
    }

//   if (show_magic_sys)
//     {
      // Append the "magic:sys" package that represents all system
      // packages.  This artificial package is identified by its name
      // and handled specially by MARK_NAMED_PACKAGE_FOR_INSTALL, etc.

      // Name
//       response.encode_string ("magic:sys");

      // Broken?  XXX - give real information here
//       response.encode_int (FALSE);

      // Installed version
//       response.encode_string ("");
//       response.encode_int64 (1000);
//       response.encode_string ("user/system");
//       response.encode_string ("Operating System");
//       response.encode_string ("All system packages");
//       response.encode_string (NULL);

      // Available version
//       response.encode_string ("");
//       response.encode_string ("user/system");
//       response.encode_string ("Operating System");
//       response.encode_string ("Updates to all system packages");
//       response.encode_string (NULL);
//     }
}

// void
// cmd_get_system_update_packages ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

//   if (!ensure_cache (true))
//     return;

//   pkgDepCache &cache = *(awc->cache);

//   for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
//     {
//       pkgCache::VerIterator installed = pkg.CurrentVer ();
//       pkgCache::VerIterator candidate = cache[pkg].CandidateVerIter(cache);

//       if (installed.end () || candidate.end ())
// 	continue;

//       package_record rec (candidate);
//       int flags = get_flags (rec);
//       if (flags & pkgflag_system_update)
// 	response.encode_string (pkg.Name ());
//     }

//   response.encode_string (NULL);
// }

/* APTCMD_GET_PACKAGE_INFO

   This command performs a simulated install and removal of the
   specified package to gather the requested information.
 */

// static int
// installable_status_1 (pkgCache::PkgIterator &pkg)
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &Cache = *(awc->cache);
//   pkgCache::VerIterator Ver = Cache[pkg].InstVerIter(Cache);

//   bool some_missing = false, some_conflicting = false;

//   if (Ver.end() == true)
//     return status_uninstallable;
      
//   for (pkgCache::DepIterator D = Ver.DependsList(); D.end() == false;)
//     {
//       // Compute a single dependency element (glob or)
//       pkgCache::DepIterator Start;
//       pkgCache::DepIterator End;
//       D.GlobOr(Start,End); // advances D

//       if ((Cache[End] & pkgDepCache::DepGInstall)
// 	  == pkgDepCache::DepGInstall)
// 	continue;

//       if (Start->Type == pkgCache::Dep::PreDepends ||
// 	  Start->Type == pkgCache::Dep::Depends)
// 	some_missing = true;
//       else if (Start->Type == pkgCache::Dep::Conflicts)
// 	some_conflicting = true;
//     }

//   if (some_missing)
//     return status_missing;
//   if (some_conflicting)
//     return status_conflicting;

//   return status_uninstallable;
// }

static PkgInstallStatus
combine_status (PkgInstallStatus s1, PkgInstallStatus s2)
{
  return max (s1, s2);
}

// static int
// installable_status ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &cache = *(awc->cache);
//   int installable_status = status_uninstallable;

//   for (pkgCache::PkgIterator pkg = cache.PkgBegin();
//        pkg.end() != true;
//        pkg++)
//     {
//       /* If a non-related package gets newly broken, we report this as
// 	 a conflict.  If a related package is broken, we take a closer
// 	 look.
//       */
//       if (cache[pkg].InstBroken())
// 	{
// 	  if (is_related (pkg))
// 	    {
// 	      installable_status =
// 		combine_status (installable_status_1 (pkg),
// 				installable_status);
// 	    }
// 	  else if (!cache[pkg].NowBroken())
// 	    {
// 	      installable_status =
// 		combine_status (status_conflicting,
// 				installable_status);
// 	    }
// 	}
//     }

//   return installable_status;
// }

// static int
// removable_status ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &cache = *(awc->cache);
//   for (pkgCache::PkgIterator pkg = cache.PkgBegin();
//        pkg.end() != true;
//        pkg++)
//     {
//       if (cache[pkg].InstBroken())
// 	return status_needed;
//     }

//   return status_uninstallable;
// }

// void
// cmd_get_package_info ()
// {
//   const char *package = request.decode_string_in_place ();
//   bool only_installable_info = request.decode_int ();

//   apt_proto_package_info info;

//   info.installable_status = status_unknown;
//   info.download_size = 0;
//   info.install_user_size_delta = 0;
//   info.required_free_space = 0;
//   info.install_flags = 0;
//   info.removable_status = status_unknown;
//   info.remove_user_size_delta = 0;

//   if (ensure_cache (true))
//     {
//       AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//       pkgDepCache &cache = *(awc->cache);
//       pkgCache::PkgIterator pkg = cache.FindPkg (package);

//       // simulate install

//       mark_named_package_for_install (package);
//       if (any_newly_or_related_broken ())
// 	info.installable_status = installable_status ();
//       else
// 	info.installable_status = status_installable;
//       info.download_size = (int64_t) cache.DebSize ();
//       info.install_user_size_delta = (int64_t) cache.UsrSize ();

//       for (pkgCache::PkgIterator pkg = cache.PkgBegin();
// 	   pkg.end() != true;
// 	   pkg++)
// 	{
// 	  if (is_related (pkg)
// 	      && (cache[pkg].Upgrade()
// 		  || pkg.State() != pkgCache::PkgIterator::NeedsNothing))
// 	    {
// 	      pkgCache::VerIterator ver = cache[pkg].CandidateVerIter(cache);
// 	      package_record rec (ver);
// 	      info.install_flags |= get_flags (rec);
// 	      info.required_free_space += get_required_free_space (rec);
// 	    }
// 	}

//       if (!only_installable_info)
// 	{
// 	  // simulate remove

// 	  if (!strcmp (package, "magic:sys"))
// 	    {
// 	      info.removable_status = status_system_update_unremovable;
// 	    }
// 	  else
// 	    {
// 	      if (!pkg.end())
// 		mark_for_remove (pkg);

// 	      for (pkgCache::PkgIterator pkg = cache.PkgBegin();
// 		   pkg.end() != true;
// 		   pkg++)
// 		{
// 		  if (cache[pkg].Delete())
// 		    {
// 		      pkgCache::VerIterator ver = pkg.CurrentVer ();
// 		      package_record rec (ver);
// 		      int flags = get_flags (rec);
// 		      if (flags & pkgflag_system_update)
// 			{
// 			  info.removable_status =
// 			    status_system_update_unremovable;
// 			  break;
// 			}
// 		    }
// 		}

// 	      if (info.removable_status == status_unknown)
// 		{
// 		  if (any_newly_or_related_broken ())
// 		    info.removable_status = removable_status ();
// 		  else
// 		    info.removable_status = status_installable;
// 		}
// 	      info.remove_user_size_delta = (int64_t) cache.UsrSize ();
// 	    }
// 	}
//     }

//   response.encode_mem (&info, sizeof (apt_proto_package_info));
// }

/* APTCMD_GET_PACKAGE_DETAILS
   
   Like APTCMD_GET_PACKAGE_INFO, this command performs a simulated
   install or removal (as requested), but it gathers a lot more
   information about the package and what is happening.
*/

// static void
// append_display_name (GString *str, const pkgCache::PkgIterator &pkg)
// {
//   pkgCache::VerIterator ver = pkg.CurrentVer();
//   if (!ver.end())
//     {
//       package_record rec (ver);
//       string display_name = get_display_name (rec);
//       if (!display_name.empty())
// 	{
// 	  g_string_append (str, display_name.c_str());
// 	  return;
// 	}
//     }
  
//   g_string_append (str, pkg.Name());
// }

// void
// encode_dependencies (pkgCache::VerIterator &ver)
// {
//   for (pkgCache::DepIterator dep = ver.DependsList(); !dep.end(); )
//     {
//       GString *str;
//       apt_proto_deptype type;
//       pkgCache::DepIterator start;
//       pkgCache::DepIterator end;
//       dep.GlobOr(start, end);

//       if (start->Type == pkgCache::Dep::PreDepends ||
// 	  start->Type == pkgCache::Dep::Depends)
// 	type = deptype_depends;
//       else if (start->Type == pkgCache::Dep::Conflicts)
// 	type = deptype_conflicts;
//       else 
// 	continue;

//       str = g_string_new ("");
//       while (1)
// 	{
// 	  g_string_append_printf (str, " %s", start.TargetPkg().Name());
// 	  if (start.TargetVer() != 0)
// 	    g_string_append_printf (str, " (%s %s)",
// 				    start.CompType(), start.TargetVer());
	  
// 	  if (start == end)
// 	    break;
// 	  g_string_append_printf (str, " |");
// 	  start++;
// 	}

//       response.encode_int (type);
//       response.encode_string (str->str);
//       g_string_free (str, 1);
//     }

//   response.encode_int (deptype_end);
// }

// void
// encode_broken (pkgCache::PkgIterator &pkg,
// 	       const char *want)
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &Cache = *(awc->cache);
//   pkgCache::VerIterator Ver = Cache[pkg].InstVerIter(Cache);
      
//   if (Ver.end() == true)
//     return;
      
//   for (pkgCache::DepIterator D = Ver.DependsList(); D.end() == false;)
//     {
//       GString *str;
//       apt_proto_sumtype type;

//       // Compute a single dependency element (glob or)
//       pkgCache::DepIterator Start;
//       pkgCache::DepIterator End;
//       D.GlobOr(Start,End); // advances D

//       if ((Cache[End] & pkgDepCache::DepGInstall)
// 	  == pkgDepCache::DepGInstall)
// 	continue;

//       if (Start->Type == pkgCache::Dep::PreDepends ||
// 	  Start->Type == pkgCache::Dep::Depends)
// 	type = sumtype_missing;
//       else if (Start->Type == pkgCache::Dep::Conflicts)
// 	type = sumtype_conflicting;
//       else
// 	continue;

//       str = g_string_new ("");
//       while (1)
// 	{
// 	  /* Show a summary of the target package if possible. In the case
// 	     of virtual packages we show nothing 
// 	  */
// 	  pkgCache::PkgIterator target = Start.TargetPkg ();

// 	  /* Never blame conflicts on the package that we want to
// 	     install.
// 	  */
// 	  if (strcmp (target.Name(), want) == 0
// 	      && Start->Type == pkgCache::Dep::Conflicts)
// 	    append_display_name (str, pkg);
// 	  else
// 	    {
// 	      append_display_name (str, target);
// 	      if (Start.TargetVer() != 0)
// 		g_string_append_printf (str, " (%s %s)",
// 					Start.CompType(), Start.TargetVer());
// 	    }

// 	  if (Start != End)
// 	    g_string_append_printf (str, " | ");
	  
// 	  if (Start == End)
// 	    break;
// 	  Start++;
// 	}

//       response.encode_int (type);
//       response.encode_string (str->str);
//       g_string_free (str, 1);
//     }
// }

static gchar*
chop_uri (gchar *uri)
{
  while (uri[0] && uri[strlen (uri) - 1] == '/')
    uri[strlen (uri) - 1] = '\0';

  return uri;
}

static gchar*
get_package_repository (pkgCache::VerIterator Version)
{
  // XXX - I don't know how to assure that the "selected" VerFile in the
  //       VerFileItertor is the installable VerFile, but always seems
  //       to be the first one in the iterator.

  pkgSourceList List;
  pkgIndexFile *Index;

  if (List.ReadMainList ())
    {
      if (Version.end () == false)
        {
          pkgCache::VerFileIterator Vf = Version.FileList ();
          if (Vf.end () == false)
            {
              if (List.FindIndex (Vf.File (), Index))
                {
                  char *archive_info, *archive_uri, *dist, *comp;

                  archive_uri = archive_info = comp = dist = NULL;

                  // extract the uri
                  archive_uri =
                    chop_uri (g_strdup (Index->ArchiveURI ("").c_str ()));

                  archive_info =
                    g_strdup (Index->ArchiveInfo (Version).c_str ());

                  // extract the dist
                  char **info = g_strsplit_set (archive_info, " ", -1);
                  if (info != NULL)
                    {
                      if (info[1] != NULL)
                        dist = g_strdup (info[1]);
                      g_strfreev (info);
                    }

                  // the component
                  comp = g_strdup (Vf.File ().Component ());

                  // nasty hack for some repositories which join the
                  // component to the distribution (dist/comp)
                  if (comp != NULL && comp[0] != '\0')
                    {
                      char *p = g_strstr_len (dist, -1, comp);
                      if (p != NULL && p[0] != '\0' && p - dist - 1 > 0)
                        {
                          char* tmp = g_strndup (dist, p - dist - 1);
                          g_free (dist);
                          dist = tmp;
                        }
                    }

                  char* catalogue = find_catalogue_by_info (archive_uri,
                                                            dist, comp, FALSE);

                  g_free (archive_uri);
                  g_free (dist);
                  g_free (comp);
                  if (catalogue != NULL) {
			  g_free(archive_info);
			  return catalogue;
		  } else {
			  return archive_info;
		  }
                }
            }
        }
    }
  return g_strdup("local-file");
}

// void
// encode_package_and_version (pkgCache::VerIterator ver)
// {
//   package_record rec (ver);
//   GString *str = g_string_new ("");
//   string pretty = get_display_name (rec);
//   if (!pretty.empty())
//     g_string_append (str, pretty.c_str());
//   else
//     g_string_append (str, ver.ParentPkg().Name());
//   g_string_append_printf (str, " (%s)", ver.VerStr());
//   response.encode_string (str->str);
//   g_string_free (str, 1);
// }

// void
// encode_install_summary (const char *want)
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &cache = *(awc->cache);

//   if (cache.BrokenCount() > 0)
//     fprintf (stderr, "[ Some installed packages are broken! ]\n");

//   mark_named_package_for_install (want);

//   for (pkgCache::PkgIterator pkg = cache.PkgBegin();
//        pkg.end() != true;
//        pkg++)
//     {
//       if (cache[pkg].NewInstall())
// 	{
// 	  response.encode_int (sumtype_installing);
// 	  encode_package_and_version (cache[pkg].CandidateVerIter(cache));
// 	}
//       else if (cache[pkg].Upgrade())
// 	{
// 	  response.encode_int (sumtype_upgrading);
// 	  encode_package_and_version (cache[pkg].CandidateVerIter(cache));
// 	}
//       else if (cache[pkg].Delete())
// 	{
// 	  response.encode_int (sumtype_removing);
// 	  encode_package_and_version (pkg.CurrentVer());
// 	}

//       if (cache[pkg].InstBroken())
// 	{
// 	  if (is_related (pkg))
// 	    encode_broken (pkg, want);
// 	  else if (!cache[pkg].NowBroken())
// 	    {
// 	      response.encode_int (sumtype_conflicting);
// 	      encode_package_and_version (pkg.CurrentVer());
// 	    }
// 	}
//     }

//   response.encode_int (sumtype_end);
// }

// void
// encode_remove_summary (pkgCache::PkgIterator &want)
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &cache = *(awc->cache);

//   if (cache.BrokenCount() > 0)
//     log_stderr ("[ Some installed packages are broken! ]\n");

//   mark_for_remove (want);

//   for (pkgCache::PkgIterator pkg = cache.PkgBegin();
//        pkg.end() != true;
//        pkg++)
//     {
//       if (cache[pkg].Delete())
// 	{
// 	  response.encode_int (sumtype_removing);
// 	  encode_package_and_version (pkg.CurrentVer());
// 	}

//       if (cache[pkg].InstBroken() && !cache[pkg].NowBroken())
// 	{
// 	  response.encode_int (sumtype_needed_by);
// 	  encode_package_and_version (pkg.CurrentVer());
// 	}
//     }

//   response.encode_int (sumtype_end);
// }

static PkgType
get_package_version_status(pkgCache::PkgIterator pkg, pkgCache::VerIterator ver, bool only_installed = false)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    const char* failed_install_version = awc->cache->failed_installs.failed_version(pkg.Name());
    bool broken = false;
    const char* installed_version = 0;
    pkgCache::VerIterator installed = pkg.CurrentVer ();

    if ( !installed.end() ) {
        installed_version = installed.VerStr();
        broken = is_broken( pkg );
    } else {
        installed_version = failed_install_version;
    }

    if( installed_version ) {
        DBG("Package %s has installed version %s, is broken: %s", pkg.Name(), installed_version, broken ? "true" : "false" );
        int c = debVS.CmpVersion( ver.VerStr(), installed_version );
        if (c == 0) {
            // TODO: Should also check if the installation has been broken by update operation
            if( !only_installed && broken ) {
                return PKG_TYPE_UPDATE;
            } else {
                return PKG_TYPE_INSTALLED;
            }
        } else if (c > 0) {
            return PKG_TYPE_UPDATE;
        } else {
            return PKG_TYPE_OLDER_THAN_INSTALLED;
        }
    }
    return PKG_TYPE_NOT_INSTALLED;
}

GHashTable *
cmd_get_package_details_map (const char *package,
			     const char *version,
                             const char *domain,
			     gboolean installed,
			     GError **error)
{
//   if (!strcmp (package, "magic:sys"))
//     {
//       response.encode_string ("");      // maintainer
//       response.encode_string
//         ("This is an artificial package that represents all\n"
//          "system packages that are installed on your device.");
//       response.encode_int (deptype_end);  // dependencies
//       response.encode_string (NULL);      // repository
//       if (summary_kind == 1)
//         encode_install_summary (package);
//       else
//         response.encode_int (sumtype_end);
//     }
//   else
//     {

    cache_reset_packages ();
    current_operation = 0;

    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *awc->cache;
    myCacheFile* c = awc->cache;
    pkgCache::PkgIterator pkg;
    pkgCache::VerIterator ver;
    GHashTable *map = NULL;
    bool failed_install = false;
    const char *failed_install_version = NULL;

    if (installed) {
        pkg = cache.FindPkg (package);
        const char *v = awc->cache->failed_installs.failed_version(package);
        if ( v ) {
            pkgCache::VerIterator tmp;
            awc->cache->find_version(package, v, NULL, tmp);
            failed_install_version = v;
            failed_install = true;
        }
        
        if (pkg.end () ) {
            if ( !failed_install ) {
                g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
                       "Cannot find package: %s", package);
            }
        } else {
            ver = pkg.CurrentVer ();
            if (ver.end ()) {
                if (awc->cache->failed_installs.contains (package)) {
                     failed_install = true;
                     awc->cache->find_version (pkg, 
                                               awc->cache->failed_installs.failed_version (package),
                                               awc->cache->failed_installs.failed_domain (package),
                                               ver);
                }

                if (ver.end () && !failed_install ) {
                    g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_INSTALLED, "Package '%s' is not installed", package);
                }

            }
        }
    } else {
        awc->cache->find_version (package, version, domain, pkg, ver, error);
    }

    if ( !installed ) {
        pkgCache::VerIterator cur = pkg.CurrentVer();
        if ( !cur.end() && !ver.end() && is_os_meta_package(cur) != is_os_meta_package(ver) ) {
            g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
                        "Cannot find package: %s", package);
        }
    }

    if (*error)
        return NULL;

    map = g_hash_table_new (g_str_hash, g_str_equal);

    PkgType t;
    if (installed /*&& !failed_install*/) {
        t = PKG_TYPE_INSTALLED;
    } else {
        t = get_package_version_status (pkg, ver);
    }

    g_hash_table_insert_string_value (map, "Type",
				      package_type_to_string (t));

    bool ok = false;
    bool need_reboot = false;
    if ( installed ) {
        g_hash_table_insert_int64_value (map, "InstalledTimestamp", c->package_timestamp(package, &ok));
        if ( !ok ) {
            DBG("Internal error! Timestamp can not be found for a package!");
        } else {
            ok = false;
            const gchar* fn = c->package_filename(package, &ok);
            if ( ok && fn && strcmp(fn, "") != 0 ) {
                g_hash_table_insert_string_value(map, "Filename", fn);
                g_hash_table_insert_boolean_value(map, "FlagLocalFile", TRUE);
            }
        }
    }
    if ( !ver.end() && is_os_meta_package(ver) ) {
        need_reboot = require_reboot(ver);
    }
    if ( t & PKG_TYPE_UPDATE ) {
        ok = false;
        ensure_ignored_update_pkg_list();
        ignored_update_pkgs->getPackageTimestamp(pkg.Name(), ver.VerStr(), &ok);
        if ( ok ) { // Package is in the ignored list
            g_hash_table_insert_boolean_value(map, "FlagIgnored", TRUE);
        }
        ok = false;
        ensure_available_update_pkg_list();
        time_t ts = available_update_pkgs->getPackageTimestamp(pkg.Name(), ver.VerStr(), &ok);
        if ( ok ) {
            g_hash_table_insert_int64_value (map, "AvailableTimestamp", ts);
        }
    }

    if ( !pkg.end() && !ver.end() ) {
        // Use data from cache
        pkg_map_insert_deps(map, t, pkg, ver);
        pkg_map_insert_details (map, pkg, ver, FALSE, t, need_reboot);
    } else if ( failed_install ) {
        // Only use what we know + possible exception
        g_hash_table_insert_string_value(map, "Name", package);
        g_hash_table_insert_string_value(map, "Version", failed_install_version);
        if( t & PKG_TYPE_UPDATE ){
            pkg_map_insert_failed_update_data (map, package, failed_install_version);
        } else {
            pkg_map_insert_failed_install_data (map, package, failed_install_version);
        }
    }

    return map;
}

/*
static void
print_metadata (pkgCache::PkgIterator pkg, gpointer user_data)
{
  printf ("name: %s\n", pkg.Name());
}
*/

// static void
// print (pkgCache::DepIterator dep)
// {
//   printf ("%s: ", __func__);

//   for (; !dep.end (); dep++)
//     {
//       printf (" '%s' ", dep.TargetPkg ().Name ());
//     }

//   printf ("\n");
// }

static gboolean
walk_dependencies (pkgCache::PkgIterator pkg,
		   int depth,
		   void (*callback) (pkgCache::PkgIterator, gpointer),
		   gpointer user_data)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);
  gboolean result = TRUE;

  if (depth > 100)
    return FALSE;

  pkgCache::VerIterator ver = cache[pkg].InstVerIter (cache);
  if (ver.end ())
    ver = cache[pkg].CandidateVerIter (cache);

  pkgCache::DepIterator Dep = ver.DependsList ();
  // print (Dep);

  if (!is_user_package (ver))
    return FALSE;
  
  callback (pkg, user_data);

  while (!Dep.end ())
    {
      // printf ("%s\n", Dep.TargetPkg ().Name ());

      // Grok or groups
      pkgCache::DepIterator Start = Dep;
      bool Result = true;
      unsigned Ors = 0;
      for (bool LastOR = true;
	   !Dep.end () && LastOR;
	   Dep++, Ors++)
	{
	  LastOR = (Dep->CompareOp & pkgCache::Dep::Or) == pkgCache::Dep::Or;

	  // if ((cache[Dep] & pkgDepCache::DepInstall) == pkgDepCache::DepInstall)
	  //   Result = false;
	}

      // Dep is satisfied okay.
      if (!Result)
	continue;

      /* Check if this dep should be consider for install. If it is a
         user defined important dep and we are installed a new package
         then it will be installed. Otherwise we only worry about
         critical deps */
      if (!cache.IsImportantDep (Start))
	continue;

      if (pkg->CurrentVer != 0 && !Start.IsCritical ())
	continue;

      /* If we are in an or group locate the first or that can
         succeed. We have already cached this.. */
      for (; Ors > 1
	     && (cache[Start] & pkgDepCache::DepCVer) != pkgDepCache::DepCVer;
	   Ors--)
	Start++;

      /* This bit is for processing the possibilty of an install/upgrade
         fixing the problem */
      SPtrArray<pkgCache::Version *> List = Start.AllTargets ();
      if ((cache[Start] & pkgDepCache::DepCVer) == pkgDepCache::DepCVer)
	{
	  // Right, find the best version to install..
	  pkgCache::Version **Cur = List;
	  pkgCache::PkgIterator P = Start.TargetPkg ();
	  pkgCache::PkgIterator InstPkg (cache, 0);

	  // See if there are direct matches (at the start of the list)
	  for (; *Cur != 0 && (*Cur)->ParentPkg == P.Index (); Cur++)
	    {
	      pkgCache &pkgcache = cache.GetCache ();
	      pkgCache::PkgIterator Pkg (pkgcache,
					 pkgcache.PkgP + (*Cur)->ParentPkg);
	      if (cache[Pkg].CandidateVer != *Cur)
		continue;
	      InstPkg = Pkg;

	      break;
	    }

	  // Select the highest priority providing package
	  if (InstPkg.end ())
	    {
	      pkgPrioSortList (cache, Cur);
	      for (; *Cur != 0; Cur++)
		{
		  pkgCache &pkgcache = cache.GetCache ();
		  pkgCache::PkgIterator
		    Pkg (pkgcache, pkgcache.PkgP + (*Cur)->ParentPkg);
		  if (cache[Pkg].CandidateVer != *Cur)
		    continue;
		  InstPkg = Pkg;

		  break;
		}
	    }

	  if (!InstPkg.end ())
	    {
	      gboolean res = walk_dependencies (InstPkg, depth + 1,
						callback, user_data);
	      if (!res)
		{
		  result = FALSE;

		  break;
		}
	    }

	  continue;
	}
    }

  return result;
}

static GHashTable *
make_package_map (pkgCache::PkgIterator pkg)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);

  pkgCache::VerIterator ver = pkg.CurrentVer ();
  PkgType t;
  if (!ver.end ())
    t = PKG_TYPE_INSTALLED;
  else
    {
      ver = cache[pkg].CandidateVerIter (cache);
      t = get_package_version_status (pkg, ver);
    }

  g_hash_table_insert_string_value (map, "Type",
				    package_type_to_string (t));

  pkg_map_insert_details (map, pkg, ver, FALSE, t);

  return map;
}

static void
add_package_metadata (pkgCache::PkgIterator pkg, gpointer user_data)
{
  GList **P = static_cast<GList **> (user_data);

  *P = g_list_prepend (*P, make_package_map (pkg));
}

static gboolean
get_package_metadata (pkgCache::PkgIterator pkg,
		      void (*callback) (pkgCache::PkgIterator, gpointer),
		      gpointer user_data)
{
  // AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  // pkgDepCache &cache = *(awc->cache);
  // cache.MarkInstall (pkg, false);

  return walk_dependencies (pkg, 0, callback, user_data);
}

GList *
cmd_get_package_dep_details (const char *package,
			     const char *version,
			     GError **error)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *awc->cache;
  pkgCache::PkgIterator pkg;
  pkgCache::VerIterator ver;
  GList *P = NULL;

  DBG("");

  pkg = cache.FindPkg (package);
  if (pkg.end ())
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
		   "Cannot find package: %s", package);
    }
  else
    {
      get_package_metadata (pkg, add_package_metadata, &P);
      P = g_list_reverse (P);
    }

  return P;
  // else if (!version)
  //   {
  //     ver = pkg.CurrentVer ();
  //     if (ver.end ())
  // 	{
  // 	  ver = cache[pkg].CandidateVerIter (cache);
  // 	  if (ver.end ())
  // 	    {
  // 	      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_INSTALLED,
  // 			   "Package '%s' is not installed or has no installation candidate", package);
  // 	    }
  // 	}
  //   }
  // else if (!find_package_version (awc->cache, pkg, ver, package, version))
  //   {
  //     g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
  // 		   "Cannot find package (version): %s (%s)", package, version);
  //   }

  // if (!ver.end ())
  //   {
  //     walk_dependencies (pkg, print_dep, NULL);
  //   }
}

bool
installed_only(GList *pkgs)
{
    DBG("");

    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    myCacheFile* c = awc->cache;

    GList* i = g_list_first(pkgs);
    package_info* p = NULL;
    pkgCache::PkgIterator pkg;
    pkgCache::VerIterator ver;
    Package* pkgli = NULL;
    while ( i ) {
        p = (package_info *) i->data;

        if ( p->filename && strcmp(p->filename, "") != 0 ) {
            pkgli = NULL;
            GError * error = NULL;
            pkgli = cmd_get_file_details (p->filename, &error);
            if ( error ) {
                g_error_free( error );
            }
            if ( pkgli ) {
                if ( pkgli->type != PKG_TYPE_INSTALLED ) {
                    return false;
                }
                package_free(pkgli);
            }
        } else if ( c->find_version(p->package, p->version, NULL, pkg, ver) ) {
            pkgCache::VerIterator installed = pkg.CurrentVer ();
            if ( installed.end() ) {
                return false;
            }
            if ( debVS.CmpVersion(installed.VerStr(), p->version) != 0 ) {
                return false;
            }
        }
        i = g_list_next(i);
    }

    DBG("out (true)");
    return true;
}

gchar*
cmd_get_restore_status(GError **error)
{
    ensure_restore_pkg_list();
    
    GList* pkgs = restore_pkgs->sortedList();
    if ( pkgs == NULL || installed_only(pkgs) ) {
        new_restore = true;
        return g_strdup("NoRestorableContent");
    }
    g_list_free( pkgs );

    if ( new_restore ) {
        return g_strdup("NewRestorePerformed");
    } else {
        return g_strdup("RestorableContentAvailable");
    }
}

gboolean
cmd_get_restorable_packages(GList** packages, GError **error)
{
    ensure_restore_pkg_list();
    GList* pkgs = restore_pkgs->sortedList();
    if ( pkgs == NULL ) {
        DBG("%s out (empty list)", __PRETTY_FUNCTION__);
        return TRUE;
    }

    if (!ensure_cache (true)) {
        g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
        return FALSE;
    }
    
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *(awc->cache);
    myCacheFile* c = awc->cache;

    GList* i = g_list_first( pkgs );
    package_info* p = NULL;
    
    while ( i ) {
        pkgCache::PkgIterator pkg;
        pkgCache::VerIterator ver;
        p = (package_info *) i->data;

        Package *pkgli = NULL;

        bool ok = false;
        const gchar* fn = restore_pkgs->getPackageFilename(p->package, &ok);
        if ( ok && fn && strcmp(fn, "") != 0 ) {
            if ( g_file_test(fn, G_FILE_TEST_EXISTS) ) {
                GError* e = NULL;
                pkgli = cmd_get_file_details (fn, &e);
                if ( e ) {
                    g_error_free(e);
                    e = NULL;
                }
                if ( pkgli->type & PKG_TYPE_INSTALLED ) {
                    package_free (pkgli);
                    i = g_list_next(i);
                    continue;
                }
            }
            if ( !pkgli ) {
                pkgli = package_new();
                pkgli->name = g_strdup(p->package);
                pkgli->version = g_strdup(p->version);
                pkgli->type = PKG_TYPE_NOT_AVAILABLE;
                pkgli->flag_local_file = TRUE;
            }
            pkgli->filename = g_strdup(fn);
        } else if (c->find_version (p->package, p->version, NULL, pkg, ver)) {
            /* Get installed and candidate iterators for current package */
            pkgCache::VerIterator installed = pkg.CurrentVer ();
            pkgCache::VerIterator candidate = cache[pkg].CandidateVerIter(cache);

/*
            PKG_TYPE_UNKNOWN = 0, // < unknown
    	    PKG_TYPE_INSTALLED = 1, // < installed
    	    PKG_TYPE_UPDATE    = 2, // < installed and new version available
    	    PKG_TYPE_NOT_INSTALLED = 4, // < available but not installed
    	    PKG_TYPE_OLDER_THAN_INSTALLED = 8, // < older than installed version
*/
            PkgType type;
            // TODO: There is a possibility that the version of the package is older than installed version.
            if ( !installed.end() ) {
                if ( debVS.CmpVersion(installed.VerStr(), ver.VerStr()) == 0 ) {
                    //type = PKG_TYPE_INSTALLED;
                    i = g_list_next(i);
                    continue;
                } else if ( !candidate.end() ) {
                    if (debVS.CmpVersion(candidate.VerStr(), ver.VerStr()) == 0 ) {
                        type = PKG_TYPE_UPDATE;
                    } else {
                        type = PKG_TYPE_OLDER_THAN_INSTALLED; // TODO: ??
                    }
                } else {
                    type = PKG_TYPE_OLDER_THAN_INSTALLED; // TODO: ??
                }
            } else { 
                type = PKG_TYPE_NOT_INSTALLED;
            }
            
            pkgli = package_new();
            pkgli->type = type;

            gboolean update_desc = FALSE;
            if (pkgli->type & (PKG_TYPE_NOT_INSTALLED | PKG_TYPE_UPDATE)) {
                update_desc = TRUE;
                if ( pkgli->type & PKG_TYPE_UPDATE ) {
                    ok = false;
                    ensure_available_update_pkg_list();
                    time_t ts = available_update_pkgs->getPackageTimestamp(pkg.Name(), ver.VerStr(), &ok);
                    if ( ok ) {
                        pkgli->available_timestamp = ts;
                    }
                }
            }

    // TODO: ??
    /*      if (failed_install_version) {
	          const char *en, *em, *fn;
	          awc->cache->failed_installs.get_data(pkg.Name(), failed_install_version, en, em, fn);
	          pkgli->error_name = g_strdup(en);
	          pkgli->error_message = g_strdup(em);
          } */
          
            bool need_reboot = false;
            if ( is_os_meta_package(ver) ) {
                need_reboot = require_reboot(ver);
            }
				        
            if (!pkg_insert_details(pkgli, pkg, ver, NULL, update_desc, error, need_reboot)) {
                DBG("Internal error");
                package_free (pkgli);
                i = g_list_next(i);
                continue;
            }

        } else {
            pkgli = package_new();
            pkgli->name = g_strdup(p->package);
            pkgli->version = g_strdup(p->version);
            pkgli->type = PKG_TYPE_NOT_AVAILABLE;
        }
        
        *packages = g_list_append(*packages, pkgli);
        pkgli = NULL;
        
        i = g_list_next(i);
    }

    return TRUE;
}

/* APTCMD_THIRD_PARTY_POLICY_CHECK
*/

/* APTCMD_CHECK_UPDATES
*/

static bool
filter_with_domain (metaIndex *m, const char *domain)
{
  if (domain == NULL || *domain == 0)
    return true;

  domain_t d = get_domain (m);
  return strcmp (domain_get_name (d), domain) == 0;
}

static bool
download_lists (xexp *catalogues_for_report,
		bool with_status, ResultCode *result,
                const char *domain = NULL)
{
  *result = rescode_failure;

  // Get the source list
  pkgSourceList List;
  if (List.ReadMainList () == false)
    return false;

  // Lock the list directory
  FileFd Lock;
  if (_config->FindB("Debug::NoLocking",false) == false)
    {
      Lock.Fd (ForceLock (_config->FindDir("Dir::State::Lists") + "lock"));
      if (_error->PendingError () == true)
	{
	  _error->Error ("Unable to lock the list directory");
	  return false;
	}
    }

  // Create the download object
  DownloadStatus Stat;
  pkgAcquire Fetcher (with_status ? &Stat : NULL);

  // Populate it with the source selection
  int n_domains = 0;
  set_sources_for_get_domain (&List);
  for (pkgSourceList::const_iterator I = List.begin(); I != List.end(); I++)
    {
      metaIndex *m = (*I);
      if (filter_with_domain (m, domain))
        {
	  if (domain)
	    DBG ("Updating domain %s", domain);

	  ++n_domains;
          if (m->GetIndexes(&Fetcher,false) == false)
            {
              set_sources_for_get_domain (NULL);
              return false;
            }
        }
    }
  set_sources_for_get_domain (NULL);

  if (n_domains == 0)
    {
      DBG ("Updating all domains");
      // Update all sources
      for (pkgSourceList::const_iterator I = List.begin(); I != List.end(); I++)
	{
	  metaIndex *m = (*I);
	  if (m->GetIndexes(&Fetcher,false) == false)
	    return false;
	}
    }

  // Run it
  if (Fetcher.Run() != pkgAcquire::Continue) {
    return false;
  }

  bool some_failed = false;
  bool some_succeeded = false;
  for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin();
       I != Fetcher.ItemsEnd(); I++)
    {
      if ((*I)->Status == pkgAcquire::Item::StatDone)
        {
          /* Only count 'complete' files as a success since we are
             looking for signs of successful download activity here.

             Apt tries to download all kinds of non-existing files,
             and they show up here as successes, which would make us
             think that the repositories work at least partially,
             which might not be true.
           */
          if ((*I)->Complete)
            some_succeeded = true;
          continue;
        }

      (*I)->Finished();

      GList *cat_glist = find_catalogues_for_item_desc (catalogues_for_report,
                                                        (*I)->DescURI());

      for (GList *iter = cat_glist; iter; iter = g_list_next (iter))
	{
	  if (iter->data != NULL)
	    {
	      xexp *cat = (xexp *) (iter->data);

	      xexp *errors = xexp_aref (cat, "errors");
	      if (errors == NULL)
		{
		  errors = xexp_list_new ("errors");
		  xexp_append_1 (cat, errors);
		}

	      xexp *error = xexp_list_new ("error");
	      xexp_aset_text (error, "uri", (*I)->DescURI().c_str());
	      xexp_aset_text (error, "msg", (*I)->ErrorText.c_str());

	      xexp_append_1 (errors, error);
	    }
	}

      if (cat_glist != NULL)
	g_list_free (cat_glist);

      _error->Error ("Failed to fetch %s  %s", (*I)->DescURI().c_str(),
                     (*I)->ErrorText.c_str());
      some_failed = true;
    }

  // Clean out any old list files, but only of this was a full refresh
  // and nothing has failed.
  if ((domain == NULL || *domain == '\0')
      && !some_failed
      && _config->FindB("APT::Get::List-Cleanup",true) == true)
    {
      Fetcher.Clean (_config->FindDir("Dir::State::lists"));
      Fetcher.Clean (_config->FindDir("Dir::State::lists") + "partial/");
    }

  if (some_failed && some_succeeded)
    *result = rescode_partial_success;
  else if (!some_failed)
    *result = rescode_success;
  
  // Run the success scripts if something was downloaded
  if(*result == rescode_success || *result == rescode_partial_success)
    RunScripts("APT::Update::Post-Invoke-Success");
  
  // Run the other scripts
  RunScripts("APT::Update::Post-Invoke");
  
  return true;
}

/* Duplicate the directory hierarchy at OLD to NEW by creating hard
   links for all the regular files.
*/

static int duplink_base;
static const char *duplink_new;

int
duplink_callback (const char *name, const struct stat *, int m, struct FTW *f)
{
  char *new_name = g_strdup_printf ("%s/%s",
				    duplink_new, name + duplink_base);

  // fprintf (stderr, "%s -> %s\n", new_name, name);

  if (m == FTW_D)
    {
      if (mkdir (new_name, 0777))
	{
	  perror (new_name);
	  g_free (new_name);
	  return -1;
	}
    }
  else if (m == FTW_F)
    {
      if (link (name, new_name))
	{
	  perror (new_name);
	  g_free (new_name);
	  return -1;
	}
    }

  g_free (new_name);
  return 0;
}

int
duplink_file_tree (const char *old_tree, const char *new_tree)
{
  duplink_base = strlen (old_tree);
  duplink_new = new_tree;
  return nftw (old_tree, duplink_callback, 10, 0);
}

/* Unlink a directory hirarchy.
 */

int
unlink_callback (const char *name, const struct stat *, int m, struct FTW *f)
{
  // fprintf (stderr, "- %s\n", name);

  if (m == FTW_DP)
    {
      if (rmdir (name))
	perror (name);
    }
  else if (m == FTW_F)
    {
      if (unlink (name))
	perror (name);
    }

  return 0;
}

int
unlink_file_tree (const char *tree)
{
  return nftw (tree, unlink_callback, 10, FTW_DEPTH);
}

ResultCode
update_package_cache (xexp *catalogues_for_report,
		      bool with_status,
                      const char *domain)
{
  /* XXX - We do the downloading in a 'transaction'.  If we get
           interrupted half-way through, all the old files are kept in
           place.  Libapt-pkg is careful not to leave partial files on
           disk, but when more than one file needs to be downloaded
           for a repository, we can still end up with an inconsistent
           state.  Worse, the cleanup will remove the "Packages" files
           when the corresponding "Release" has not been downloaded
           yet.

     Libapt-pkg should take care of this itself, but I am too much of
     a chicken to make that change now.
  */

  ResultCode result = rescode_failure;

  string lists_val = _config->Find("Dir::State::Lists");
  string lists_dir = _config->FindDir("Dir::State::Lists");

  if (lists_dir.length() > 0 && lists_dir[lists_dir.length()-1] == '/')
    lists_dir.erase(lists_dir.length()-1, 1);

  string lists_dir_new = lists_dir + ".new";
  string lists_dir_old = lists_dir + ".old";

  unlink_file_tree (lists_dir_new.c_str());
  duplink_file_tree (lists_dir.c_str(), lists_dir_new.c_str());
  _config->Set ("Dir::State::Lists", lists_dir_new);

  if (download_lists (catalogues_for_report, 
		      with_status, &result, domain))
    {
      cache_rw_lock ();

      /* complete transaction */
      unlink_file_tree (lists_dir_old.c_str());
      rename (lists_dir.c_str(), lists_dir_old.c_str());
      rename (lists_dir_new.c_str(), lists_dir.c_str());
      unlink_file_tree (lists_dir_old.c_str());
      _config->Set ("Dir::State::Lists", lists_val);

      cache_rw_unlock ();

      cache_init (with_status);
    }
  else
    {
      /* cleanup */
      _config->Set ("Dir::State::Lists", lists_val);
      unlink_file_tree (lists_dir_new.c_str());
    }

  return result;
}

// static gboolean
// add_temp_sources_list (xexp *tempcat)
// {
//   gboolean success;

//   success = (xexp_write_file (TEMP_CATALOGUE_CONF, tempcat) &&
//              write_sources_list (TEMP_APT_SOURCE_LIST, tempcat));

//   return success;
// }

bool
is_update_package_ignored(pkgCache::PkgIterator pkg, pkgCache::VerIterator candidate);

std::map<string, string> make_available_updates_map ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);
  std::map<string, string> U;

  for (pkgCache::PkgIterator pkg = cache.PkgBegin (); !pkg.end (); pkg++)
    {
      pkgCache::VerIterator cur = pkg.CurrentVer ();
      if (cur.end ())
	continue;

      pkgCache::VerIterator cand = cache[pkg].CandidateVerIter (cache);
      if (cand.end ())
	continue;

      if (debVS.CmpVersion (cand.VerStr (), cur.VerStr ()) <= 0)
	continue;

      if (!is_user_package (cand))
	continue;

      if (is_update_package_ignored (pkg, cand))
	continue;

      const char *type;
      if (is_os_meta_package (cur))
	{
	  if (!is_os_meta_package (cand))
	    continue;

	  type = "os";
	}
      else
	type = "other";

      U[pkg.Name ()] = cand.VerStr ();
      // U.insert (package_version_type (pkg.Name (), cand.VerStr (), type));
    }

  return U;
}

static void dout (std::map<string, string> const& s)
{
  std::ostringstream out;
  out << '[';

  for (std::map<string, string>::const_iterator i = s.begin (); i != s.end (); ++i)
    {
      out << i->first.c_str () << '=' << i->second.c_str () << ' ';
    }

  out << ']' << endl;

  DBG ("%s", out.str ().c_str ());
}

// static bool is_new_update_available ()
// {
//   read_available_updates_file ();
//   set<package_version_type> U1 = available_updates ();

//   using std::set_difference;
//   using std::inserter;

//   for (set<package_version_type>::const_iterator i = U1.begin (); i != U1.end (); ++i)
//     {
//       if (
//     }
//   set<package_version_type> D;
//   set_difference (U.begin (), U.end (),
// 		  U1.begin (), U1.end (),
// 		  inserter (D, D.begin ()));

  
// }

#define NOTIF_TOOL "/usr/lib/package-manager/pm-notification-tool.sh"

/*
static bool is_os_meta_package_ver (const char *package, const char *version, const char *domain)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgCache::PkgIterator pkg;
  pkgCache::VerIterator ver;

  if (awc->cache->find_version (package, version, domain, pkg, ver))
    return is_os_meta_package (ver);
  else
    return false;
}
*/
/*
static string get_update_description (const char *package, const char *version, const char *domain)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgCache::PkgIterator pkg;
  pkgCache::VerIterator ver;

  if (awc->cache->find_version (package, version, domain, pkg, ver))
    {
      package_record rec (ver);
      return get_description (TRUE, pkg, rec);
    }
  else
    return string ();
}
*/

/*!
  Create update notification if there are new updates.  Otherwise
  update the notification based on current state if it exists.
 */
static void update_upgrade_notification (bool only_update = true)
{
  if (!flag_enable_notifications)
    return;

  map<string, string> U0;
  read_available_updates_file (U0);

  map<string, string> U (make_available_updates_map ());
  typedef map<string, string>::const_iterator const_iterator;
  //int n = 0, n_sys = 0;
  //string desc;
  bool new_updates = false;
  //string sys_name, sys_ver;

  dout (U0);
  dout (U);

  for (const_iterator i = U.begin (); i != U.end (); ++i)
    {
      //bool sys = is_os_meta_package_ver (i->first.c_str (), i->second.c_str (), NULL);
      const_iterator j = U0.find (i->first);
      if (j == U0.end ()
	  || debVS.CmpVersion (i->second, j->second) > 0)
	{
	  new_updates = true;
	}

      /*if (sys)
	{
	  sys_name = i->first;
	  sys_ver = i->second;
	  desc = get_update_description (i->first.c_str (), i->second.c_str (), NULL);
	  // Cut description to first newline character
	  string::size_type n = desc.find ('\n');
	  if (n != string::npos)
	    desc.resize(n);
	  ++n_sys;
	}
      else
	++n;*/
    }
/*
    if( new_updates ){
        DBG ("Refreshing notifications (new updates available)");
        system("dbus-send --session --print-reply --type=\"method_call\" --dest=com.nokia.package_manager_notifications /com/nokia/package_manager_notifications com.nokia.package_manager_notifications.refresh boolean:true");
    }
    else{
        DBG ("Refreshing notifications");
        system("dbus-send --session --print-reply --type=\"method_call\" --dest=com.nokia.package_manager_notifications /com/nokia/package_manager_notifications com.nokia.package_manager_notifications.refresh boolean:false");
    }
*/
    std::ostringstream cmd;
    cmd << NOTIF_TOOL;

    if( new_updates ){
        DBG ("Refreshing notifications (new updates available)");
        cmd << " dbus-send --session --print-reply --type=\"method_call\" --dest=com.nokia.package_manager_notifications /com/nokia/package_manager_notifications com.nokia.package_manager_notifications.refresh boolean:true";
    }
    else{
        DBG ("Refreshing notifications");
        cmd << " dbus-send --session --print-reply --type=\"method_call\" --dest=com.nokia.package_manager_notifications /com/nokia/package_manager_notifications com.nokia.package_manager_notifications.refresh boolean:false";
    }
    DBG ("%s", cmd.str().c_str ());
    system (cmd.str().c_str ());
/*
  std::ostringstream cmd;
  cmd << NOTIF_TOOL;

  if (n_sys > 0)
    cmd << " -U " << sys_name << '=' << sys_ver;
  if (n > 0)
    {
      cmd << " -u " << n;
      if (only_update)
	cmd << " -o";
    }

  if (new_updates)
    cmd << " -c";

  if (!desc.empty ())
    {
      gchar *t = escape_for_shell (desc.c_str ());
      // cmd << " -t " << '\"' << desc.c_str () << '\"';
      cmd << " -t " << t;
      g_free (t);
    }

  DBG ("%s", cmd.str ().c_str ());
  system (cmd.str ().c_str ());
*/
  write_available_updates_file ();
}

typedef struct PackageVersions {
	gchar *pkg;
	gchar *current_ver;
	gchar *candidate_ver;
} PackageVersions;

PackageVersions*
find_and_remove_packageversions(const PackageVersions* pkg, GList** list)
{
    GList* i = g_list_first(*list);
    while ( i != NULL ) {
        PackageVersions *p = static_cast<PackageVersions*>(i->data);
        if ( strcmp(p->pkg, pkg->pkg) == 0 ) {
            *list = g_list_remove(*list, p);
            return p;
        }
        i = g_list_next(i);
    }
    return NULL;
}

void free_packageversions(PackageVersions* pkg)
{
    if ( pkg == NULL ) {
        return;
    }
    g_free( pkg->pkg );
    g_free( pkg->current_ver );
    g_free( pkg->candidate_ver );
    g_free( pkg );
}

void clear_packageversions_list(GList* list)
{
    if ( list == NULL ) {
        return;
    }
    
    for ( GList* i = g_list_first(list); i != NULL; i = g_list_next(i) ) {
        PackageVersions *data = static_cast<PackageVersions*>(i->data);
        free_packageversions( data );
        data = 0;
    }
    g_list_free(list);
}

// NOTE: This takes the ownership of the lists!
bool lists_differ(GList* list_old, GList* list_new, bool *update_changed, bool *new_updates_available = 0)
{
    ensure_available_update_pkg_list();

    bool package_changed = false;
    bool updates_changed = false;
    *update_changed = false;
    GList* i = g_list_first(list_new);

    while ( i ) {
        const PackageVersions *pkg_new = static_cast<const PackageVersions*>(i->data);
        PackageVersions *pkg_old = find_and_remove_packageversions(pkg_new, &list_old);

        if ( pkg_old ) {
            if ( strcmp(pkg_new->current_ver, pkg_old->current_ver) != 0 ) { // This should never be true!
                package_changed = true;
            }
            if ( strcmp(pkg_new->candidate_ver, pkg_old->candidate_ver) != 0 ) { // So there's a difference between candidate versions
                if ( debVS.CmpVersion(pkg_new->current_ver, pkg_new->candidate_ver) >= 0 ) { // True, if current is same or newer...
                    available_update_pkgs->removePackage(pkg_old->pkg, pkg_old->candidate_ver);
                    updates_changed = true;
                } else {
                  // DBG ("%s: new version: %s > %s",
                  // pkg_new->pkg, pkg_new->candidate_ver, pkg_new->current_ver);
                    if ( strcmp(pkg_new->current_ver, "") == 0 ) {
                        package_changed = true;
                    } else {
                        updates_changed = true;
                        available_update_pkgs->addPackage(pkg_new->pkg, pkg_new->candidate_ver);
			if (new_updates_available)
			    *new_updates_available = true;
                    }
                }
            }
        } else { // This means that there's a new package available.
          // DBG ("new updates available");
            package_changed = true;
            if (new_updates_available)
                *new_updates_available = true;
//            available_update_pkgs->addPackage(pkg_new->pkg, pkg_new->candidate_ver); // Do not add a timestamp for a new package, only for a new update.
        }

        if ( debVS.CmpVersion(pkg_new->current_ver, pkg_new->candidate_ver) < 0 ) {
            if ( strcmp(pkg_new->current_ver, "") != 0 && strcmp(pkg_new->candidate_ver, "") != 0 ) {
                bool ok = false;
                available_update_pkgs->getPackageTimestamp(pkg_new->pkg, pkg_new->candidate_ver, &ok);
                if ( ! ok ) {
                    updates_changed = true;
                    available_update_pkgs->addPackage(pkg_new->pkg, pkg_new->candidate_ver);
                }
            }
        }

        free_packageversions( pkg_old );
        i = g_list_next(i);
    }
    
    int cnt = g_list_length(list_old);
    if ( cnt > 0 ) { // This means that some packages has been removed. (The cnt tells how many).
        package_changed = true;
        GList* i = g_list_first(list_old);
        while ( i ) {
            const PackageVersions *pkg = static_cast<const PackageVersions*>(i->data);
            available_update_pkgs->removePackage(pkg->pkg, pkg->candidate_ver); // Try to remove the timestamp. Don't care if it can't be found.
            i = g_list_next(i);
        }
    }
    available_update_pkgs->saveToFile( AVAILABLE_PACKAGE_LIST );

    clear_packageversions_list(list_old);
    clear_packageversions_list(list_new);
    
    *update_changed = updates_changed;
    return ( package_changed or updates_changed );
}

static pkgCache::VerIterator
failed_install_version (const pkgCache::PkgIterator &pkg)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgCache::VerIterator ver;
  awc->cache->find_failed_install_version (pkg, ver, NULL);
  return ver;
}

GList*
get_update_package_list()
{
    if ( ! ensure_cache(false) ) {
        return NULL;
    }

    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *(awc->cache);
    GList *result = NULL;
    
    for (pkgCache::PkgIterator p = cache.PkgBegin(); !p.end (); p++) {
        const char* name = p.Name();
        pkgCache::VerIterator v0 = p.CurrentVer();
	if (v0.end())
	    v0 = failed_install_version(p);

        pkgCache::VerIterator v1 = cache[p].CandidateVerIter( cache );
        
        const char* ver0 = "";
        if (! v0.end() ) {
            ver0 = v0.VerStr();
        }
        const char* ver1 = "";
        if (! v1.end() ) {
            ver1 = v1.VerStr();
        }

        PackageVersions* pkg = new PackageVersions();
        pkg->pkg = strdup(name);
        pkg->current_ver = strdup(ver0);
        pkg->candidate_ver = strdup(ver1);
        result = g_list_append(result, pkg);
    }
    
    return result;
}

/*!
  Clear failed upgrades with a newer available candidate version.
*/
static bool
clear_old_failed_upgrades ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);
  bool changes = false;

  for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++)
    {
      pkgCache::VerIterator v0 = p.CurrentVer ();
      if (v0.end ())
	continue;

      pkgCache::VerIterator const& v1 = cache[p].CandidateVerIter(cache);
      if (v1.end ())
	continue;

      if (v1.CompareVer (v0) <= 0)
	continue;

      const char *v = awc->cache->failed_upgrades.failed_version (p.Name ());
      if (v && debVS.CmpVersion (v, v1.VerStr ()) < 0)
	{
	  awc->cache->failed_upgrades.clear (p.Name ());
	  changes = true;
	}
    }

  if (changes)
    awc->cache->failed_upgrades.write ();

  return changes;
}

#define AVAILABLE_UPDATES_FILE "/var/lib/package-manager/available-updates"

ResultCode
cmd_check_updates (gboolean with_status, const char *domain, bool clear_ignored)
{
  using std::map;

  GList* old_update_list = get_update_package_list();
  bool ignored_changed = false;

  if ( clear_ignored )
    {
      ensure_ignored_update_pkg_list();
      ignored_changed = ignored_update_pkgs->clearStorage();
      ignored_update_pkgs->saveToFile( IGNORED_PACKAGE_LIST );
    }

  xexp *catalogues = read_catalogues ();

  reset_catalogue_errors (catalogues);

  /* Update sources.list file before refreshing */
  update_sources_list (catalogues);

  ResultCode result_code = update_package_cache (catalogues, with_status, domain);

  if ((result_code == rescode_success)
      || (result_code == rescode_partial_success))
    {
      GList* new_update_list = get_update_package_list();
      /* Some packages with the 'system-update' could
         appear / disappear under this conditions */
      ssu_packages_needs_refresh = true;

      bool changed = false;
      bool new_updates_available = false;
      bool failed_changed = clear_old_failed_upgrades ();

      if ( lists_differ(old_update_list, new_update_list, &changed, &new_updates_available ) || ignored_changed || failed_changed ) {
        send_package_list_updated_signal (changed | ignored_changed | failed_changed);
      }

      update_upgrade_notification (false);
    } else {
      clear_packageversions_list(old_update_list);
    }
  
  /* Save potential errors to disk */
  save_failed_catalogues (catalogues);

//   response.encode_xexp (catalogues);
//   response.encode_int (result_code);

  if (catalogues) {
    xexp_free (catalogues);
  }

  return result_code;
}

// int
// cmdline_check_updates (char **argv)
// {
//   AptWorkerCache * awc = 0;
//   int result_code = -1;

//   awc = AptWorkerCache::GetCurrent ();
//   awc->init_cache_after_request = false;

//   if (awc->cache == NULL)
//     return 2;

//   if (argv[1])
//     {
//       DBG ("http_proxy: %s", argv[1]);
//       setenv ("http_proxy", argv[1], 1);
//     }

//   response.reset ();
//   request.reset (NULL, 0);
//   result_code = cmd_check_updates (false);

//   _error->DumpErrors ();

//   if (result_code == rescode_success
//       || result_code == rescode_partial_success)
//     return 0;
//   else
//     return 1;
// }

/* APTCMD_GET_CATALOGUES
 *
 * We also return the non-comment lines from all sources.list files in
 * order to let the user at least know that there are sources in use
 * that are not controlled by us.  The code for this is copied from
 * apt-pkg.
 */

// void
// append_system_sources (xexp *catalogues, string File)
// {
//    // Open the stream for reading
//    ifstream F(File.c_str(),ios::in /*| ios::nocreate*/);
//    if (!F != 0)
//      return;
   
//    // CNC:2003-12-10 - 300 is too short.
//    char Buffer[1024];

//    int CurLine = 0;
//    while (F.eof() == false)
//      {
//        F.getline(Buffer,sizeof(Buffer));
//        CurLine++;
//        _strtabexpand(Buffer,sizeof(Buffer));
//        if (F.fail() && !F.eof())
// 	 return;

//        char *I;
//        // CNC:2003-02-20 - Do not break if '#' is inside [].
//        for (I = Buffer; *I != 0 && *I != '#'; I++)
//          if (*I == '[')
// 	   for (I++; *I != 0 && *I != ']'; I++);
//        *I = 0;
       
//        const char *C = _strstrip(Buffer);
      
//        // Comment or blank
//        if (C[0] == '#' || C[0] == 0)
// 	 continue;
      	    
//        xexp_append_1 (catalogues, xexp_text_new ("source", C));
//      }
// }

// void
// append_system_source_dir (xexp *catalogues, string Dir)
// {
//    DIR *D = opendir(Dir.c_str());
//    if (D == 0)
//      return;

//    vector<string> List;
   
//    for (struct dirent *Ent = readdir(D); Ent != 0; Ent = readdir(D))
//      {
//        if (Ent->d_name[0] == '.')
// 	 continue;

//        // CNC:2003-12-02 Only accept .list files as valid sourceparts
//        if (flExtension(Ent->d_name) != "list")
// 	 continue;
      
//        // Skip bad file names ala run-parts
//        const char *C = Ent->d_name;
//        for (; *C != 0; C++)
// 	 if (isalpha(*C) == 0 && isdigit(*C) == 0
//              && *C != '_' && *C != '-' && *C != '.')
// 	   break;
//        if (*C != 0)
// 	 continue;
      
//        // Make sure it is a file and not something else
//        string File = flCombine(Dir,Ent->d_name);
//        struct stat St;
//        if (stat(File.c_str(),&St) != 0 || S_ISREG(St.st_mode) == 0)
// 	 continue;

//        // skip our own file
//        if (File == CATALOGUE_APT_SOURCE)
// 	 continue;

//        List.push_back(File);      
//      }
//    closedir(D);
   
//    sort(List.begin(),List.end());

//    // Read the files
//    for (vector<string>::const_iterator I = List.begin(); I != List.end(); I++)
//      append_system_sources (catalogues, *I);
// }

// void
// cmd_get_catalogues ()
// {
//   struct stat buf;
//   int stat_result;
//   xexp *catalogues = NULL;

//   /* Check if there are problems reading the file */
//   stat_result = stat (CATALOGUE_CONF, &buf);
//   if (!stat_result)
//     {
//       /* Map the catalogue report to (maybe) delete error reports from it */
//       xexp *tmp_catalogues = read_catalogues ();
//       catalogues = xexp_list_map (tmp_catalogues, map_catalogue_error_details);
//       xexp_free (tmp_catalogues);

//       /* Add errors to those catalogues inside the failed-catalogues file */
//       merge_catalogues_with_errors (catalogues);
//     }
//   else
//     {
//       /* If there's not a conf file on disk, write a empty one */
//       catalogues = read_catalogues (); /* @check test this scenario */
//       write_user_catalogues (catalogues);
//     }

//   string Main = _config->FindFile("Dir::Etc::sourcelist");
//   if (FileExists(Main) == true)
//     append_system_sources (catalogues, Main);

//   string Parts = _config->FindDir("Dir::Etc::sourceparts");
//   if (FileExists(Parts) == true)
//     append_system_source_dir (catalogues, Parts);

//   response.encode_xexp (catalogues);
//   xexp_free (catalogues);
// }

/* APTCMD_SET_CATALOGUES
 */

// void
// cmd_set_catalogues ()
// {
//   int success = true;

//   xexp *catalogues = request.decode_xexp ();
//   xexp_adel (catalogues, "source");

//   /* Map the catalogue report to delete error reports from it */
//   xexp *mapped_catalogues = xexp_list_map (catalogues,
// 					   map_catalogue_error_details);

//   /* Update failed catalogues file */
//   save_failed_catalogues (catalogues);

//   /* Prepare mapped_catalogues to be used for updating the
//      sources.list file later, freeing catalogues first */
//   xexp_free (catalogues);
//   catalogues = mapped_catalogues;

//   /* Update sources.list file */
//   update_sources_list (catalogues);

//   /* We should update catalogue cache after modify catalogues */
//   need_cache_init ();

//   xexp_free (catalogues);
//   response.encode_int (success);
// }

/* APTCMD_ADD_TEMP_CATALOGUES
 *
 * Stores a new temporal source list in /etc/apt directory
 */
// void
// cmd_add_temp_catalogues ()
// {
//   int success;
//   xexp *tempcat;

//   tempcat = request.decode_xexp ();
//   xexp_adel (tempcat, "source");

//   /* add a temporal sources.list file */
//   success = add_temp_sources_list (tempcat);  

//   xexp_free (tempcat);
//   response.encode_int (success);
// }

// /* APTCMD_RM_TEMP_CATALOGUES
//  *
//  * Remove the temporal source list stored in /etc/apt directory
//  */

// void
// cmd_rm_temp_catalogues ()
// {
//   int success = true;

//   clean_temp_catalogues ();
  
//   response.encode_int (success);
// }

// static bool set_dir_cache_archives (const char *alt_download_root);
// static ResultCode operation (bool check_only,
// 			const char *alt_download_root,
// 			bool download_only,
// 			GError **error,
// 			bool allow_download,
// 			bool with_status,
//             bool *need_reboot);

/* APTCMD_INSTALL_CHECK
 *
 * Check if a package can be installed. It uses the
 * common "operation ()" code, that runs or checks the
 * current operation queue (packages marked for install
 * or uninstall).
 */

// void
// cmd_install_check ()
// {
//   const char *package = request.decode_string_in_place ();
//   bool found = false;
//   int result_code = rescode_failure;
  
//   if (ensure_cache (true))
//     {
//       found = mark_named_package_for_install (package);
//       result_code = operation (true, NULL, false);
//     }

//   response.encode_int (found && result_code == rescode_success);
// }

/* APTCMD_DOWNLOAD_PACKAGE
 *
 * Download a package, using the common "operation ()" code, that
 * installs packages marked for install.
 */

#include <sys/statvfs.h>
#include <mntent.h>

/* global variable to report the download size to the frontend */
static int64_t download_size = 0;

// static bool
// volume_is_readwrite (char* option)
// {
//   g_return_val_if_fail (option, FALSE);

//   enum
//   {
//     RW = 0,
//     END
//   };

//   const char * const suboptions[] =
//     {
//       MNTOPT_RW,
//       NULL
//     };

//   int so;
//   char* argument = NULL;

//   while (*option != 0)
//     {
//       so = getsubopt (&option, (char* const*) suboptions, &argument);
//       if (so == RW)
//         return true;
//     }

//   return false;
// }

// static bool
// volume_path_is_mounted_writable (const gchar *path)
// {
//   bool result = false;

//   g_return_val_if_fail (path, false);

//   FILE *fp = setmntent (MOUNTED, "r");
//   g_return_val_if_fail (fp != NULL, false);

//   struct mntent *mnt;
//   while ((mnt = getmntent (fp)) != NULL)
//     {
//       gboolean usable = (g_strcmp0 (mnt->mnt_dir, path) == 0 &&
//                          volume_is_readwrite (mnt->mnt_opts));

//       if (usable)
//         {
//           /* Try to write a dummy file to be completely sure,
//              since getmntent is not always up-to-date about the
//              read-only status of a partition, which would in fact
//              become read-only only when actually writing to it */
//           gchar *dummyfile = g_strdup_printf ("%s/.ham-dummy-file-XXXXXX",
//                                               path);
//           int fd;

//           /* Try to open a temporary file under the selected path */
//           fd = mkstemp (dummyfile);
//           if (fd != -1)
//             {
//               /* Try to write some data in the file */
//               const char *dummytext = "Dummy text";
//               int n = strlen (dummytext);
//               bool data_written = (write (fd, dummytext, n) == n
//                                    && fsync (fd) != -1);

//               /* Close the descriptor and decide the final result */
//               result = (close (fd) != -1) && data_written;

//               /* Delete the file on success */
//               if (result)
//                 unlink (dummyfile);
//             }
//           g_free (dummyfile);

//           /* Don't keep on iterating if an usable path was found */
//           if (result)
//             break;
//         }
//     }

//   endmntent (fp);

//   return result;
// }

// void
// cmd_download_package ()
// {
//   const char *package = request.decode_string_in_place ();

//   const char *alt_download_root = NULL;
//   int result_code = rescode_out_of_space;

//   const char *internal_mmc_mountpoint = getenv ("INTERNAL_MMC_MOUNTPOINT");
//   if (!internal_mmc_mountpoint)
//     internal_mmc_mountpoint = INTERNAL_MMC_MOUNTPOINT;

//   const char *removable_mmc_mountpoint = getenv ("REMOVABLE_MMC_MOUNTPOINT");
//   if (!removable_mmc_mountpoint)
//     removable_mmc_mountpoint = REMOVABLE_MMC_MOUNTPOINT;

//   if (ensure_cache (true))
//     {
//       if (mark_named_package_for_install (package))
//         {
//           if (flag_download_packages_to_mmc
// 	      && internal_mmc_mountpoint
//               && volume_path_is_mounted_writable (internal_mmc_mountpoint))
//             {
//               alt_download_root = internal_mmc_mountpoint;
//               result_code = operation (false, alt_download_root, true);
//             }

//           if (flag_download_packages_to_mmc
// 	      && result_code == rescode_out_of_space
// 	      && removable_mmc_mountpoint
//               && volume_path_is_mounted_writable (removable_mmc_mountpoint))
//             {
//               alt_download_root = removable_mmc_mountpoint;
//               result_code = operation (false, alt_download_root, true);
//             }

//           /* default or bailout option */
//           if (!flag_download_packages_to_mmc ||
//               result_code == rescode_out_of_space)
//             {
//               alt_download_root = NULL;
//               result_code = operation (false, alt_download_root, true);
//             }
//         }
//       else
//         result_code = rescode_packages_not_found;
//     }

//   response.encode_int (result_code);
//   response.encode_int64 (download_size);
//   response.encode_string (alt_download_root);

//   download_size = 0;
// }

static gint package_version_name_compare (const PackageVersion *v1,
					  const PackageVersion *v2)
{
  return g_strcmp0 (v1->pkg, v2->pkg);
}

static GList *package_version_list_sort_by_name (GList *l)
{
  return g_list_sort (l, (GCompareFunc) package_version_name_compare);
}

static GValue *package_value_list_to_gvalue (GList *l)
{
  const int len = g_list_length (l) + 1;
  gchar **pkgs = (gchar **) g_malloc (len * sizeof (gchar *));
  int i = 0;

  for (GList *p = l; p != NULL; p = g_list_next (p))
    {
      PackageVersion *pv = static_cast<PackageVersion *> (p->data);
      if (pv->error)
        pkgs[i++] = g_strdup_printf ("%s=%s=%s", pv->pkg, pv->version,
                                     pkgmgr_error_to_dbus_error (PkgmgrdError (pv->error->code)));
      else
        pkgs[i++] = g_strdup_printf ("%s=%s", pv->pkg, pv->version);
    }

  pkgs[i] = 0;

  GValue *val = g_new0 (GValue, 1);
  g_value_init (val, G_TYPE_STRV);
  g_value_take_boxed (val, pkgs);

  return val;
}

GHashTable *
cmd_get_current_operation_data (GError **error)
{
  GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);
  cache_operation_lock ();
  OperationData *op = cache_get_current_operation();
  if (!op)
    op = cache_get_compound_operation ();

  if (op)
    {
      g_hash_table_insert_string_value (map, "Operation", operation_type_to_string (op->type));
      if (op->pkg != NULL)
	g_hash_table_insert_string_value (map, "Name", op->pkg);
      if (op->version != NULL)
	g_hash_table_insert_string_value (map, "Version", op->version);
      if (op->filename != NULL)
	g_hash_table_insert_string_value (map, "Filename", op->filename);

      g_hash_table_insert_int64_value (map, "DownloadCurrentBytes", op->current_bytes);
      g_hash_table_insert_int64_value (map, "DownloadTotalBytes", op->total_bytes);
      g_hash_table_insert_int_value (map, "OperationProgressPct", op->operation_progress_pct);

      if (op->pending_pkgs != NULL)
	{
	  GValue *val = package_value_list_to_gvalue (op->pending_pkgs);
	  g_hash_table_insert (map, g_strdup ("PendingPackages"), val);
	}

      if (op->rejected_pkgs != NULL)
	{
	  GValue *val = package_value_list_to_gvalue (op->rejected_pkgs);
	  g_hash_table_insert (map, g_strdup ("RejectedPackages"), val);
	}

    }
  cache_operation_unlock ();

  return map;
}

/* APTCMD_INSTALL_PACKAGE
 *
 * Install a package, using the common "operation ()" code, that
 * installs packages marked for install.
 */

static bool
mark_named_for_remove (const char *package)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);
  pkgCache::PkgIterator pkg = cache.FindPkg (package);

  if (!pkg.end ())
    {
      GError *error = NULL;
      return mark_for_remove (pkg, TRUE, &error) == TRUE;
    }
  else
    return false;
}

static int operation_type_from_str (const char *operation)
{
  static const struct OpType {
    const char *op;
    int type;
  } op2type[] = {
    { "Install", OPERATION_INSTALL },
    { "InstallFile", OPERATION_INSTALL_FILE },
    { "Uninstall", OPERATION_UNINSTALL },
    { "Refresh", OPERATION_REFRESH },
    { "Restore", OPERATION_RESTORE_INSTALL },
    { "Upgrade", OPERATION_UPGRADE },
    { "UpgradeAll", OPERATION_UPGRADE_ALL },
    { "SystemUpgrade", OPERATION_SYSTEM_UPGRADE },
    { "DownloadSystemUpgrade", OPERATION_DOWNLOAD_SYSTEM_UPGRADE },
    { 0, 0 }
  };

  for (const OpType *p = op2type; p->op; ++p)
    if (!strcmp (p->op, operation))
      return p->type;

  return -1;
}

static gboolean
revert_install (OperationData *op)
{
  apt_worker_reset_state ();

  if (!ensure_cache (false))
    {
      return FALSE;
    }

  bool r = mark_named_for_remove (op->pkg);
  int pkgflags;

  for (GList *i = g_list_first (op->installed_user_deps); i != NULL; i = g_list_next (i))
    r |= mark_named_for_remove (static_cast<const PackageVersion *> (i->data)->pkg);

  if (r)
    {
      DBG ("%s", op->pkg);
      cache_operation_start (op->pkg, op->version, op->domain, "", op->type);
      OperationData *op = cache_get_current_operation ();
      op->send_signals = FALSE;
      op->revert_install = TRUE;

      GError *error = NULL;
      ResultCode rc = operation (false, NULL,
				 false, &error, true, true, &pkgflags);

      // Leave the package marked as failed install
      op->update_failed_installs = false;

      need_cache_init ();
      cache_operation_complete (NULL);
      apt_worker_reset_state ();

      DBG ("%s", rc == rescode_success ? "ok" : "fail");

      return rc == rescode_success ? TRUE : FALSE;
    }
  else
    return FALSE;
}

static void
dump_operation_data_installed_user_deps (const OperationData *op)
{
  for (GList *i = g_list_first (op->installed_user_deps); i != NULL; i = g_list_next (i))
    g_debug ("installed dep: %s", static_cast<const PackageVersion *> (i->data)->pkg);
}

void
cmd_ignore_update_package (const char *package, const char *version, GError **error)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

    ensure_ignored_update_pkg_list();
    
    ignored_update_pkgs->addPackage(package, version);
    ignored_update_pkgs->saveToFile( IGNORED_PACKAGE_LIST );

    if (awc->cache->failed_upgrades.clear(package, version))
        awc->cache->failed_upgrades.write();

    send_package_ignored_signal("Update", package, version);

    update_upgrade_notification ();
}

void
create_cache_refreshed_file(const gchar* filename)
{
    DBG("");

    FILE *f = fopen (filename, "w");

    if (f == NULL) {
        DBG("Couldn't open a file '%s' for writing", filename);
        return;
    }

    time_t ts = time(NULL);
    gchar* str = g_strdup_printf("%ld", ts);
    if ( fputs(str, f) == EOF ) {
        DBG("Couldn't write a file '%s'", filename); // Something went wrong :'(
    }

    g_free( str );

    fclose(f);
}

gboolean
cmd_refresh_cache (const char *domain, const char *flags, GError **error)
{
    bool automatic = flag_string_contains (flags, "automatic");
    bool clearignored = flag_string_contains (flags, "clearignored");

    if (!ensure_cache (false)) {
		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
		return FALSE;
	}

    DBG("flags: %s", flags);
	
    if (!connect_network (automatic)) {
		g_set_error (error, PKGMGRD_ERROR, ERROR_DOWNLOAD_FAILED, "No network");
                return FALSE;
    }

    cache_operation_start ("", "", domain, "", OPERATION_REFRESH);
    send_operation_started_signal();
	ResultCode rc = cmd_check_updates (FALSE, domain, clearignored);
	
	OperationData* op = cache_get_current_operation();
	if ( ! op ) {
        cache_operation_start ("", "", domain, "", OPERATION_REFRESH);
    }
	
    if (rc == rescode_partial_success) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PARTIAL_SUCCESS,
                   "Some parts of the cache could not be refreshed");
    } else if (rc == rescode_failure) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_OPERATION_FAILED, "The cache could not be refreshed");
    } else if (rc != rescode_success) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Rescode: %d", rc);
    } else if ( rc == rescode_success ) {
      if ( strcmp(domain, "") == 0 ) {
        create_cache_refreshed_file( CACHE_REFRESHED_FILE );
      }
    }

    cache_operation_complete(*error);

    disconnect_network ();

    return rc == rescode_success;
}

gboolean
cmd_install_package (const char *package, const char *version, const char *domain,
                     const char *alt_download_root, GError **error)
{
	int pkgflags = 0;
	ResultCode result_code = rescode_failure;
	gboolean only_user = TRUE;
//	bool auto_update_package_list = true;
	OperationData *op = NULL;

	if (!ensure_cache (true)) {
		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
		return FALSE;
	}

	cache_operation_start (package, version, domain, "", OPERATION_INSTALL);

        if ( check_battery_level(error, CRITICAL_BATTERY_LEVEL) ) {
            op = cache_get_current_operation();
            if ( op ) {
                op->update_failed_installs = true;
            }
            cache_operation_complete (*error);
            return FALSE;
        }
        free_list(pkgs_to_be_deleted);
        pkgs_to_be_deleted = NULL;
	
	bool installable = mark_named_package_for_install (package, version, domain, only_user, FALSE, FALSE, error);
	DBG ("cmd_install_package: %s, installable: %d", package, installable);

	if (verbosity >= 2)
		dump_operation_data_installed_user_deps (cache_get_current_operation ());

/*	if (auto_update_package_list
	    && ! installable
	    && ((*error)->code == ERROR_PACKAGE_NOT_FOUND
		|| (*error)->code == ERROR_PACKAGE_VERSION_NOT_FOUND
		|| (*error)->code == ERROR_NO_CANDIDATE))
	{
	    // Update package list, try again

	    ResultCode rc = cmd_check_updates (FALSE);
	    if (rc == rescode_success
	    	|| rc == rescode_partial_success)
	    {
		    g_clear_error (error);
    		cache_operation_start (package, version, "", OPERATION_INSTALL);

	    	installable = mark_named_package_for_install (package, version, only_user, FALSE, FALSE, error);
	    }
	}*/

	if (installable) {
		op = operation_data_copy (cache_get_current_operation());
		result_code = operation (false, alt_download_root,
					false, error, true, true, &pkgflags);
	} else {
		result_code = rescode_packages_not_found;
		OperationData *od = cache_get_current_operation();
		if ( od && (*error) && (*error)->code == ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE ) {
		    od->update_failed_installs = true;
		}
	}
	if (result_code != rescode_success) {
		if (!*error) {
			g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Rescode: %d", result_code);
		}
	}

	OperationData *original_op = operation_data_copy(cache_get_current_operation());
	need_cache_init ();
	if (result_code != rescode_success
	    && (*error)->code == ERROR_OPERATION_FAILED) {
		revert_install (op);
	}

	cache_set_current_operation (original_op);
	cache_operation_complete (*error);

	operation_data_free (op);

	update_upgrade_notification ();

	return result_code == rescode_success;
}

/* APTCMD_UPGRADE_PACKAGE
 *
 * Upgrade a package, using the common "operation ()" code, that
 * upgrades packages marked for upgrade.
 */

gboolean
cmd_upgrade_package (const char *package, const char *version, const char *domain, 
                     const char *alt_download_root, GError **error)
{
	ResultCode result_code = rescode_failure;
	gboolean only_user = TRUE;
//	bool auto_update_package_list = true;
	int pkgflags = 0;

	if (!ensure_cache (true)) {
		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
		return FALSE;
	}

	cache_operation_start (package, version, domain, "", OPERATION_UPGRADE);

        if ( check_battery_level(error, CRITICAL_BATTERY_LEVEL) ) {
            OperationData *op = cache_get_current_operation();
            if ( op ) {
                op->update_failed_installs = true;
            }
            cache_operation_complete (*error);
            return FALSE;
        }
        free_list(pkgs_to_be_deleted);
        pkgs_to_be_deleted = NULL;

	bool installable = mark_named_package_for_install (package, version, domain, only_user, FALSE, TRUE, error);
/*	if (auto_update_package_list
	    && ! installable
	    && ((*error)->code == ERROR_PACKAGE_NOT_FOUND
		|| (*error)->code == ERROR_NO_CANDIDATE))
	{
	    // Update package list, try again

	    ResultCode rc = cmd_check_updates (FALSE);
	    if (rc == rescode_success
		|| rc == rescode_partial_success)
	    {
		g_clear_error (error);
		cache_operation_start (package, version, "", OPERATION_UPGRADE);

		installable = mark_named_package_for_install (package, version, only_user, FALSE, TRUE, error);
	    }
	}*/

	DBG ("cmd_upgrade_package: %s, installable: %d", package, installable);
	if (installable) {
		result_code = operation (false, alt_download_root,
					false, error, true, true, &pkgflags);
	} else {
		result_code = rescode_packages_not_found;
		OperationData *od = cache_get_current_operation();
		if ( od && (*error) && (*error)->code == ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE ) {
		    od->update_failed_installs = true;
		}
	}
	if (result_code != rescode_success) {
		if (!*error) {
			g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Rescode: %d", result_code);
		}
	}
	need_cache_init ();
	cache_operation_complete (*error);
	if ( result_code == rescode_success ) {
	    ensure_ignored_update_pkg_list();
        if ( ignored_update_pkgs->removePackage(package, version) ) {
            ignored_update_pkgs->saveToFile( IGNORED_PACKAGE_LIST );
        }
	}

	update_upgrade_notification ();

	return result_code == rescode_success;
}

pkgCache::PkgFileIterator
add_extra_index_from_section (pkgTagSection &section, const char *filename, domain_t domain);

static gboolean
mark_package_for_reinstall (const pkgCache::VerIterator &ver, GError **error)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

  pkgCache::PkgIterator pkg = ver.ParentPkg();
  pkgCache::VerIterator cur = pkg.CurrentVer ();
  
  if (!cur.end ())
    {
      if (!is_user_package (cur))
        {
          g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_USER,
                       "Package '%s' installed version '%s' is not user package",
                       pkg.Name (),
                       cur.VerStr ());
          return false;
        }

      if (is_os_meta_package (cur))
        {
          g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_IS_SYSTEM_PACKAGE,
                       "Package '%s' installed version '%s' is OS metapackage",
                       pkg.Name (),
                       cur.VerStr ());
          return false;
        }
    }

  if (!awc->cache->check_compatible_domains (ver, error))
    return false;

  if (!is_user_package (ver))
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_USER,
		   "Package '%s' version '%s' is not user package",
		   pkg.Name (),
		   ver.VerStr ());
      return false;
    }
  if (is_os_meta_package (ver) || is_os_meta_package (ver))
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_IS_SYSTEM_PACKAGE,
		   "Package '%s' version '%s' is OS metapackage",
		   pkg.Name (),
		   ver.VerStr ());
      return false;
    }

  mark_related (ver);

  pkgCache::VerIterator sys_ver;
  if (!mark_for_apt_install (*awc->cache, ver, true, &sys_ver, error))
    return false;

  if (check_marked_os_meta_package_upgrade (sys_ver, error))
    {
      if (*error)
	g_error_free (*error);

      g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE,
		   "Package requires system upgrade '%s' version '%s'",
		   sys_ver.ParentPkg ().Name (),
		   sys_ver.VerStr ());
      return false;
    }
  else if (check_if_user_packages_marked_for_uninstall (ver, error))
    return false;

  return true;
}

gboolean
cmd_restore_package (const char *package, const char *version, const char *domain, GError **error)
{
    if (!ensure_cache (true)) {
        g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
        return FALSE;
    }

    ensure_restore_pkg_list();
    bool ok = false;
    restore_pkgs->getPackageTimestamp(package, &ok);
    if ( ! ok ) {
//        DBG("%s Restore timestamp can not be found. Package: %s (%s)", __PRETTY_FUNCTION__, package, version);
        g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND, "Cannot find a restore package: %s (%s)", package, version);
        return FALSE;
    }
    bool fromfile = false;
    ok = false;
    char* pkgfn = restore_pkgs->getPackageFilename(package, &ok);
    if ( ok && pkgfn && strcmp(pkgfn, "") != 0 ) {
        fromfile = true;
    } else if ( !pkgfn ) {
        pkgfn = g_strdup("");
    }
    
    OperationType op_type = OPERATION_RESTORE_INSTALL;
//    OperationType op_type = OPERATION_RESTORE;
    
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgCache::PkgIterator pkg;
    pkgCache::VerIterator ver;
    if ( !fromfile && awc->cache->find_version (package, version, domain, pkg, ver) ) {
        pkgCache::VerIterator installed = pkg.CurrentVer();
        if ( !installed.end() ) {
            op_type = OPERATION_RESTORE_UPGRADE;
        }
    }

    const gchar *alt_download_root = NULL;
    int pkgflags = 0;
    ResultCode result_code = rescode_failure;
    gboolean only_user = TRUE;
    OperationData *op = NULL;

    cache_operation_start (package, version, domain, pkgfn, op_type);

    if ( check_battery_level(error, CRITICAL_BATTERY_LEVEL) ) {
        op = cache_get_current_operation();
        if ( op ) {
            op->update_failed_installs = true;
        }
        
        if ( restore_pkgs->removePackage( package ) ) {
            restore_pkgs->saveToFile( RESTORE_PACKAGE_LIST );
        }
        
        cache_operation_complete (*error);
        return FALSE;
    }
    free_list(pkgs_to_be_deleted);
    pkgs_to_be_deleted = NULL;

    bool installable = false;
    
    if ( fromfile ) {
        pkgTagSection section;
        read_tag_section (pkgfn, &section, error);

        {
            string arch = section.FindS("Architecture");
            if (arch != DEB_HOST_ARCH && arch != "all") {
                g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE,
                            "File '%s' has a wrong architecture", pkgfn);
                DBG("File '%s' has a wrong architecture", pkgfn);
            }
        }

        if ( !(*error) ) {
            string p_str = section.FindS("Package");
            string v_str = section.FindS("Version");

            const char *p = p_str.c_str ();
            const char *v = v_str.c_str ();
            domain_t domain = get_file_domain (pkgfn);

            if ( strcmp(p, package) != 0 || strcmp(v, version) != 0 ) {
                g_set_error(error, PKGMGRD_ERROR, ERROR_INVALID_FILE, "Invalid file: '%s'", pkgfn);
            }
            cache_operation_set_domain (domain_get_name (domain));

            if ( !(*error) ) {
                pkgCache::PkgFileIterator pf;
                {
                    scoped_cache_rw_lock locker;
                    pf = add_extra_index_from_section (section, pkgfn, domain);
                    if (pf.end()) {
                        g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Extra index failure");
                    }
                }

                if ( !(*error) ) {
                    pkgDepCache &cache = *(awc->cache);
                    pkgCache::PkgIterator pkg = cache.FindPkg (p);

                    pkgCache::VerIterator ver;
                    if (awc->cache->find_version (p, v, NULL, pf, ver, error)) {
                        installable = mark_package_for_reinstall(ver, error);
                    }
                }
            }
        }
    } else {
        installable = mark_named_package_for_install (package, version, domain, only_user, FALSE, TRUE, error);
    }

    if ( !installable && !fromfile ) {
        g_clear_error (error);
        free_list(pkgs_to_be_deleted);
        pkgs_to_be_deleted = NULL;
        installable = mark_named_package_for_install (package, version, domain, only_user, FALSE, FALSE, error);
    }

    if (installable) {
	op = operation_data_copy (cache_get_current_operation());
	result_code = operation (false, alt_download_root,
				 false, error, true, true, &pkgflags);
    } else {
	result_code = rescode_packages_not_found;
	OperationData *od = cache_get_current_operation();
	if ( od && (*error) && (*error)->code == ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE ) {
	    od->update_failed_installs = true;
	}
    }
    if (result_code != rescode_success) {
        if (!*error) {
            DBG("%s Error. Package: %s (%s)", __PRETTY_FUNCTION__, package, version);
            g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Rescode: %d", result_code);
        }
    }
    need_cache_init ();

    if ( restore_pkgs->removePackage( package ) ) {
//        DBG("%s Package removed: %s (%s)", __PRETTY_FUNCTION__, package, version);
        restore_pkgs->saveToFile( RESTORE_PACKAGE_LIST );
    }

    cache_operation_complete (*error);
    if (result_code != rescode_success && (*error)->code == ERROR_OPERATION_FAILED) {
//        DBG("%s Result code not success and operation failed. Package: %s (%s)", __PRETTY_FUNCTION__, package, version);
	revert_install (op);
    }
    operation_data_free (op);

    update_upgrade_notification ();

    if ( result_code == rescode_success ) {
        ensure_ignored_update_pkg_list();
        if ( ignored_update_pkgs->removePackage(package, version) ) {
//            DBG("%s Package removed from ignore packages list. Package: %s (%s)", __PRETTY_FUNCTION__, package, version);
            ignored_update_pkgs->saveToFile( IGNORED_PACKAGE_LIST );
        }
    }

    return result_code == rescode_success;
}

void
cmd_ignore_restore_package (const char *package, GError **error)
{
    ensure_restore_pkg_list();

    if ( restore_pkgs->removePackage(package) ) {
        restore_pkgs->saveToFile( RESTORE_PACKAGE_LIST );
    } else {
        g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND, "Cannot find a restore package: %s", package);
    }
}


static gboolean
find_os_metapackage (const char *package, const char *version, const char *domain,
		     pkgCache::PkgIterator &pkg,
		     pkgCache::VerIterator &ver,
		     GError **error)
{
	AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
        if (!(awc->cache->find_version (package, version, domain, pkg, ver, error)
              && awc->cache->check_compatible_domains (ver, error)))
          return false;

	if (!is_os_meta_package (ver))
	{
		g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE,
			     "Package '%s' is not system metapackage", pkg.Name ());
		return FALSE;
	}

	return TRUE;
}

gboolean
cmd_download_system_upgrade (const char *package, const char *version, const char *domain,
                             const char *alt_download_root, GError **error)
{
	int pkgflags = 0;
	ResultCode result_code = rescode_failure;

	if (!ensure_cache (true)) {
		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
		return FALSE;
	}

	{
	  pkgCache::PkgIterator pkg;
	  pkgCache::VerIterator ver;
	  if (!find_os_metapackage (package, version, domain,
				    pkg, ver,
				    error))
	    return FALSE;
	}

/*
operation (bool check_only,
	const char *alt_download_root,
	bool download_only,
	GError **error,
	bool allow_download,
	bool with_status, 
    bool *need_reboot)
*/

    cache_operation_start (package, version, domain, "", OPERATION_DOWNLOAD_SYSTEM_UPGRADE);

    if ( check_battery_level(error, LOW_BATTERY_LEVEL) ) {
        cache_operation_complete (*error);
        return FALSE;
    }

	OperationData *op = cache_get_current_operation();
	op->send_dep_signals = FALSE;
	free_list(pkgs_to_be_deleted);
        pkgs_to_be_deleted = NULL;

	if ( mark_named_package_for_install (package, version, domain, TRUE, TRUE, TRUE, error) ) {
		result_code = operation (false, alt_download_root, true, error, true, true, &pkgflags);
	} else {
		cache_operation_complete (*error);
		return FALSE;
	}

//	op->update_failed_installs = TRUE;

	if ( result_code != rescode_success ) {
		switch ( result_code ) {
			case rescode_packages_not_found:
				g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
							 "Package '%s' or its dependencies not found in download cache", package);
			break;
			default:
				g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Unknown error");
		}
	}

    cache_operation_complete (*error);
	return result_code == rescode_success;
}

void reboot_device(bool update_mode, int delay = 0)
{
    char cmd[0x100];

    DBG ("update mode: %d delay: %d)", update_mode, delay);

    const char *args = update_mode ? "-u" : "";
    snprintf(cmd, sizeof(cmd), "%s -d %d -- %s", REBOOT_SCRIPT, delay, args);
    system(cmd);
}

void
cmd_reboot (int pkgflags)
{
  reboot_device(pkgflags & pkgflag_reboot_update, 5);
}

/*!
 * Return Maemo-Flags union of all marked packages.
 */
int
get_flags_union ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  int flags = 0;
  for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++)
    {
      if (cache[p].Keep())
	continue;

      pkgCache::VerIterator ver = cache[p].CandidateVerIter(cache);
      package_record rec (ver);

      flags |= get_flags (rec);
    }

  DBG ("get_flags_union: %s", make_flags_str (flags).c_str());

  return flags;
}

/* APTCMD_UPGRADE_SYSTEM
 *
 * Installs system upgrade package, using the common "operation ()" code, that
 * upgrades packages marked for upgrade.
 */
gboolean
cmd_upgrade_system (const char *package, const char *version, const char *domain,
                    const char *alt_download_root, GError **error)
{
	DBG("Pkg: %s (%s) alt_download_root: %s", package, version, alt_download_root);

	if (strlen(version) == 0)
		version = NULL;

	if (!ensure_cache (true)) {
		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
		return FALSE;
	}

	pkgCache::PkgIterator pkg;
	pkgCache::VerIterator ver;

	if (!find_os_metapackage (package, version, domain,
				  pkg, ver,
				  error))
		return FALSE;

	unlink ("/var/lib/package-manager/rescue-reboot-count");

	cache_operation_start (package, version, domain, "", OPERATION_SYSTEM_UPGRADE);

        if ( check_battery_level(error, LOW_BATTERY_LEVEL) ) {
                cache_operation_complete (*error);
                return FALSE;
        }

	if (!mark_for_install (pkg, ver, TRUE, TRUE, TRUE, 0, error)) {
		cache_operation_complete (*error);
		return FALSE;
	}

	int pkgflags = get_flags_union ();

	/* Try to perform operation with allow_download == false */

        package_record rec (ver);
        if (!in_scratchbox() && !rec.has_flag ("Maemo-Flags", "no-reckless"))
          {
            /* This is how you add an item to the end of a configuration
               item that is a list.
            */
            _config->Set ("DPkg::Options::", "--reckless");
          }

	int tmp_flags;
	ResultCode res = operation (false,
				    alt_download_root,
				    false,
				    error,
				    true, // allow_download
				    true,
				    &tmp_flags);

	bool ok = res == rescode_success;

        _config->Clear ("DPkg::Options", "--reckless");
        
	if (!ok && !*error) {
		switch (res) {
		case rescode_packages_not_found:
			g_set_error (error, PKGMGRD_ERROR, ERROR_NEED_DOWNLOAD,
				     "Package '%s' or its dependencies not found in download cache", pkg.Name ());
			break;
		default:
			g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL,
				     "Rescode: %d", res);
			break;
		}
	}

	bool pm_failed = *error &&
	    ((*error)->code == ERROR_OPERATION_FAILED ||
	     (*error)->code == ERROR_UPGRADE_FAILED);

	// If DPkg failed, keep operation record and reboot to rescue
	// mode.
	if (!pm_failed)
		erase_operation_record ();

	need_cache_init ();

	cache_operation_complete (*error, pkgflags);

	update_upgrade_notification ();

	if (pm_failed || (ok && need_reboot (pkgflags))) {
	    if (fork() == 0)
	    {
		reboot_device(pkgflags & pkgflag_reboot_update, 5);
		exit(0);
	    }
	}

	return ok ? TRUE : FALSE;
}

/*
static bool
is_upgrade_available (const char *package)
{
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *(awc->cache);

    pkgCache::PkgIterator pkg = cache.FindPkg (package);
    if (! pkg.end ()) {
        pkgCache::VerIterator v0 = pkg.CurrentVer ();
        pkgCache::VerIterator v1 = cache[pkg].CandidateVerIter (cache);
        return debVS.CmpVersion(v0.VerStr(), v1.VerStr()) < 0;
    }
    else
        return false;
}*/

/*
static gboolean
upgrade_many (vector<string> const& packages,
	      GError **error)
{
    vector<string>::const_iterator i;

    for (i = packages.begin (); i != packages.end (); ++i) {
	if (is_upgrade_available (i->c_str ()))
	    if (! cmd_upgrade_package (i->c_str (), NULL, NULL, error))
		return FALSE;
    }

    return TRUE;
}*/

bool
is_update_package_ignored(pkgCache::PkgIterator pkg, pkgCache::VerIterator candidate)
{
    if ( candidate.end() ) {
        return false; // ???
    }
    ensure_ignored_update_pkg_list();
    bool ok = false;
    ignored_update_pkgs->getPackageTimestamp(pkg.Name(), candidate.VerStr(), &ok);

    return ok;
}

struct StatusMode
{
  signed char Status;
  unsigned char Mode;
  bool related;

  StatusMode() { }
  StatusMode (pkgDepCache::StateCache const& s, bool related)
    : Status (s.Status), Mode (s.Mode), related (related)
  {
  }
};

typedef std::vector<StatusMode> StatusMode_vector_t;

static void get_package_states (StatusMode_vector_t &S)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &dc = *(awc->cache);
  pkgCache &c = dc;
  S.reserve (c.Head ().PackageCount);
  S.clear ();

  for (pkgCache::PkgIterator p = dc.PkgBegin(); !p.end(); p++)
    S.push_back (StatusMode (dc[p], awc->cache->extra_info[p->ID].related));
}

static void reset_changed_package_states (const StatusMode_vector_t &S)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &dc = *(awc->cache);
  pkgCache &c = dc;
  assert (S.size() == c.Head ().PackageCount);
  StatusMode_vector_t::const_iterator i = S.begin();

  for (pkgCache::PkgIterator p = dc.PkgBegin(); !p.end(); p++, ++i)
    {
      pkgDepCache::StateCache const& s = dc[p];
      if (s.Status != i->Status)
	DBG("Pkg: %s Status=%d (was %d)", p.Name (), i->Status, s.Status);
      if (s.Mode != i->Mode)
	DBG("Pkg: %s Mode=%d (was %d)", p.Name (), i->Mode, s.Mode);

      if (s.Status != i->Status || s.Mode != i->Mode)
	{
	  if (i->Mode == pkgDepCache::ModeDelete)
	    dc.MarkDelete (p);
	  else if (i->Mode == pkgDepCache::ModeInstall)
	    dc.MarkInstall (p);
	  else if (i->Mode == pkgDepCache::ModeKeep)
	    dc.MarkKeep (p);
	}

      awc->cache->extra_info[p->ID].related = i->related;
    }
}

/**
 * Upgrade all upgradeable packages. Done in a loop one-by-one because
 * UI spec says so.
 *
 * @todo Upgrade order, if user package upgrade depends on system meta
 * package upgrade?
 * @todo Implement support for allow_system_upgrade?
 */
gboolean
cmd_upgrade_all (gboolean allow_system_upgrade, GError **error)
{
    DBG ("allow_system_upgrade=%d", allow_system_upgrade);

    //GSList *ssu_pkgs_found = NULL;
    //gboolean only_user = TRUE;
    const char *alt_download_root = NULL;
    bool stop_if_one_fails = false;

    if (!ensure_cache (true))
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");

      return FALSE;
    }

    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *(awc->cache);

    // Mark all available upgrades

    //int n_installable = 0;
    GList *pending = NULL;
    GList *rejected = NULL;
    StatusMode_vector_t sc;
    
    bool battery_low = false;
    free_list(pkgs_to_be_deleted);
    pkgs_to_be_deleted = NULL;

    for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++) {
    	pkgCache::VerIterator v0 = pkg.CurrentVer ();
    	pkgCache::VerIterator v1 = cache[pkg].CandidateVerIter (cache);

#if 0
	// Look for the SSU package if needed
	if (ssu_packages_needs_refresh
	    && (!installed.end () || !candidate.end ()))
        {
	    pkgCache::VerIterator viter = !candidate.end ()
		? candidate
		: installed;
	    if (is_os_meta_package (viter))
            {
		/* Add it to the local GSList */
		ssu_pkgs_found = g_slist_prepend (ssu_pkgs_found,
						  g_strdup (pkg.Name ()));
            }
        }
#endif

	pkgCache::VerIterator sys_ver;

    	if (! __filter_package_list (pkg, v0, v1, PKG_TYPE_UPDATE, NULL, TRUE, &sys_ver)
    	    && ! is_update_package_ignored(pkg, v1)
	    && ! is_os_meta_package (v1))
	{
	    // Version?
// 	    if (mark_named_package_for_install (pkg.Name (),
// 						TRUE,
// 						TRUE,
// 						error))
// 		++n_installable;

	    DBG ("%s", pkg.Name ());
	    GError *err = NULL;
	    get_package_states (sc);

	    bool ok = mark_named_package_for_install (pkg.Name (), v1.VerStr (), NULL, TRUE, FALSE, TRUE, &err);
	    if (err)
	      {
		reset_changed_package_states (sc);

                DBG ("rejecting %s", pkg.Name());

                PackageVersion *pv = package_version_new ();
                pv->pkg = g_strdup (pkg.Name ());
                pv->version = g_strdup (v1.VerStr ());
                pv->error = g_error_copy (err);
                rejected = g_list_prepend (rejected, pv);

		g_error_free (err);
	      }

	    // cache_reset_packages ();
	    if (!ok)
	      continue;

	    PackageVersion *pv = package_version_new ();
	    pv->pkg = g_strdup (pkg.Name ());
	    pv->version = g_strdup (v1.VerStr ());
	    pending = g_list_prepend (pending, pv);
	}
	else
	{
	    if (!v1.end ())
	    {
	    	const char *v = awc->cache->failed_installs.failed_version(pkg.Name());
	    	if (v && debVS.CmpVersion(v1.VerStr(), v) > 0)
	    	{
	    	    PackageVersion *pv = package_version_new ();
	    	    pv->pkg = g_strdup (pkg.Name ());
	    	    pv->version = g_strdup (v1.VerStr ());
	    	    pending = g_list_prepend (pending, pv);
	    	}
	    }
	}
    }

    cache_reset_packages ();

    pending = g_list_reverse (pending);
    rejected = g_list_reverse (rejected);

    OperationData *all_op = operation_data_new ();
    all_op->type = OPERATION_UPGRADE_ALL;
    // all_op->pkg = g_strdup ("");
    // all_op->version = g_strdup ("");
    all_op->pending_pkgs = copy_package_version_list (pending);
    all_op->rejected_pkgs = copy_package_version_list (rejected);
    cache_compound_operation_reset (all_op);

    if (operation_started_func)
      operation_started_func (operation_type_to_string (all_op->type),
    			      all_op->pkg, all_op->version,
    			      signal_data);

    GError *first_error = NULL;
    free_list(pkgs_to_be_deleted);
    pkgs_to_be_deleted = NULL;

    for (GList *p = g_list_first (pending); p != NULL; p = g_list_next (p))
      {
	PackageVersion *pv = static_cast<PackageVersion *> (p->data);
	int pkgflags = 0;

	pending = g_list_next (p);

	all_op = operation_data_new ();
	all_op->type = OPERATION_UPGRADE_ALL;
	// all_op->pkg = g_strdup ("");
	// all_op->version = g_strdup ("");
	all_op->pending_pkgs = copy_package_version_list (pending);
	all_op->rejected_pkgs = copy_package_version_list (rejected);
	cache_compound_operation_reset (all_op);

	cache_operation_start (pv->pkg, pv->version, NULL, "", OPERATION_UPGRADE_ALL, pending, rejected);

    if ( battery_low ) {
        if ( ! (*error) ) {
            *error = g_error_copy(first_error);
        }
    } else if ( check_battery_level(error, CRITICAL_BATTERY_LEVEL) ) {
	    battery_low = true;
        if (!first_error)
            first_error = g_error_copy (*error);
    }

	DBG ("upgrade_all: current package=%s=%s", pv->pkg, pv->version);

	ResultCode result_code;
	bool installable = false;
	if ( !battery_low ) {
	    installable = mark_named_package_for_install (pv->pkg, pv->version, NULL, TRUE, FALSE, TRUE, error);
	}

    if ( installable ) {
        result_code = operation (false, alt_download_root,
                                 false, error, true, true, &pkgflags);

    } else {
        OperationData *od = cache_get_current_operation();
		if ( od && (*error) && ( (*error)->code == ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE || (*error)->code == ERROR_LOW_BATTERY_LEVEL ) ) {
		    od->update_failed_installs = true;
		}
    }

    all_op->pkgflags |= pkgflags;

    need_cache_init ();
    cache_operation_complete (*error, need_reboot (pkgflags) ? TRUE : FALSE);

	if (stop_if_one_fails
	    && result_code != rescode_success
	    && !*error)
	  {
	    g_set_error (error, PKGMGRD_ERROR, ERROR_OPERATION_FAILED,
			 "Rescode: %d", result_code);

	    break;
	  }

	if (*error)
	  {
	    if (!first_error)
	      first_error = g_error_copy (*error);
	    g_clear_error (error);
	  }

	if (!ensure_cache (true))
	  {
	    GError *err;
	    g_set_error (&err, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
	    if (!first_error)
	      first_error = err;
	    else
	      g_error_free (err);

	    break;
	  }
      }

    if (first_error)
      {
	all_op->error = pkgmgr_error_to_dbus_error (PkgmgrdError (first_error->code));
	*error = first_error;
      }

    update_upgrade_notification ();

    send_operation_complete_signal (all_op);
    cache_compound_operation_reset (NULL);

    return *error == NULL;
}

// void
// cmd_remove_check ()
// {
//   const char *package = request.decode_string_in_place ();

//   if (ensure_cache (true))
//     {
//       AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//       pkgDepCache &cache = *(awc->cache);
//       pkgCache::PkgIterator pkg = cache.FindPkg (package);

//       if (!pkg.end ())
// 	{
// 	  mark_for_remove (pkg);
	  
// 	  for (pkgCache::PkgIterator pkg = cache.PkgBegin();
// 	       pkg.end() != true;
// 	       pkg++)
// 	    {
// 	      if (cache[pkg].Delete())
// 		response.encode_string (pkg.Name());
// 	    }
// 	}
//     }

//   response.encode_string (NULL);
// }


gboolean
cmd_pkgmgr_fetch_files_of_installed_package (const char *package, GList **files, GError **error)
{
 
    FILE *f;
    int i;

    char path[PATH_MAX];
    GList *list = NULL;

    i=snprintf(path,PATH_MAX,"%s%s.list",DPKG_INFO_PATH,package);
    //   DBG(path);
    f = fopen (path, "r");
    if (f)
    {
	char *line = NULL;
	size_t len = 0;
	ssize_t n;
	
	while ((n = getline (&line, &len, f)) != -1)
	{
	    if (n > 0 && line[n-1] == '\n') line[n-1] = '\0';
	    list = g_list_append (list, g_strdup(line));
	}
	
	free (line);
	fclose (f);
    }
    else
    {
	g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
		     "Cannot find package '%s'", package);
	return FALSE;
	
    }

    *files=list;
    return TRUE;
}

gboolean
cmd_uninstall_package (const char *package, GError **error)
{
	ResultCode result_code = rescode_failure;
	gboolean only_user = TRUE;
	int pkgflags = 0;

	cache_operation_start (package, "", NULL, "", OPERATION_UNINSTALL);
	if (ensure_cache (true)) {
		AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
		pkgDepCache &cache = *(awc->cache);
		pkgCache::PkgIterator pkg = cache.FindPkg (package);

		if (!pkg.end ()) {
			gboolean res = mark_for_remove (pkg, only_user, error);
			if (!res) {
				if (*error && (*error)->code == ERROR_PACKAGE_NOT_INSTALLED && awc->cache->failed_installs.contains(pkg.Name())) {
					OperationData *op = cache_get_current_operation ();
					op->update_failed_installs = true;
					op->version = g_strdup (awc->cache->failed_installs.failed_version (package));
					g_clear_error (error);
					result_code = rescode_success;
					// Cleared in cache_operation_complete
				}
				else if (!*error) {
					g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_INSTALLED,
						"Package '%s' is not installed", package);
				}
			} else {

				result_code = operation (false, NULL, false,
							error, true, true, &pkgflags);
			}
		} else {
			g_set_error (error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND,
				"Cannot find package '%s'", package);
		}
	}

	need_cache_init ();
	if (result_code != rescode_success && !*error) {
		g_set_error (error, PKGMGRD_ERROR, ERROR_OPERATION_FAILED,
			"Rescode: %d", result_code);
	}
	cache_operation_complete (*error);

	update_upgrade_notification ();

	return result_code == rescode_success;
}


// static void
// reset_new_domains ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

//   pkgDepCache &cache = *(awc->cache);
//   int package_count = cache.Head().PackageCount;

//   for (int i = 0; i < package_count; i++)
//     awc->cache->extra_info[i].new_domain = DOMAIN_UNSIGNED;
// }

// static void
// collect_new_domains ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

//   pkgDepCache &cache = *(awc->cache);
//   int package_count = cache.Head().PackageCount;

//   for (int i = 0; i < package_count; i++)
//     if (awc->cache->extra_info[i].related)
//       awc->cache->extra_info[i].cur_domain
//         = awc->cache->extra_info[i].new_domain;
// }

#ifdef HAVE_APT_TRUST_HOOK
// static int
// index_trust_level_for_package (pkgIndexFile *index,
// 			       const pkgCache::VerIterator &ver)
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &cache = *(awc->cache);
//   pkgCache::PkgIterator pkg = ver.ParentPkg();

//   int cur_level =
// 	  domain_get_trust_level(awc->cache->extra_info[pkg->ID].cur_domain);

//   int index_domain = get_domain (index);
//   int index_level = domain_get_trust_level(index_domain);

//   DBG ("trust_level: cur %d index %d (%s)",
// 	  cur_level, index_level, domain_get_name(index_domain));

//   /* If we have already found a good domain, accept this one only if
//      it is the same domain, or strictly better.
//    */
//   if (awc->cache->extra_info[pkg->ID].new_domain != DOMAIN_UNSIGNED)
//     {
//       int new_level =
// 	      domain_get_trust_level(awc->cache->extra_info[pkg->ID].new_domain);

//       if (index_domain == awc->cache->extra_info[pkg->ID].new_domain
// 	  || index_level > new_level)
// 	{
// 	  awc->cache->extra_info[pkg->ID].new_domain = index_domain;
// 	  return index_level;
// 	}
//       else
// 	return -1;
//     }

//   /* If this is a new install, accept the first domain that comes
//      along.
//    */
//   if (cache[pkg].NewInstall())
//     {
//       DBG ("new: accept index");
//       awc->cache->extra_info[pkg->ID].new_domain = index_domain;
//       return index_level;
//     }

//   /* This is an upgrade, only accept the current domain, or ones that
//      dominate it.
//   */
//   if (index_level >= cur_level)
//     {
//       DBG ("upgrade: accept better");
//       awc->cache->extra_info[pkg->ID].new_domain = index_domain;
//       return index_level;
//     }

//   /* This index is as good as any other.
//    */
//   DBG ("Hmm");
//   return cur_level;
// }
#endif //HAVE_APT_TRUST_HOOK

// static void
// encode_trust_summary ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   pkgDepCache &cache = *(awc->cache);

//   for (pkgCache::PkgIterator pkg = cache.PkgBegin();
//        pkg.end() != true;
//        pkg++)
//     {
//       domain_t cur_domain = awc->cache->extra_info[pkg->ID].cur_domain;
//       domain_t new_domain = awc->cache->extra_info[pkg->ID].new_domain;

//       if (cache[pkg].Upgrade() && !domains[new_domain].is_certified)
// 	{
// 	  DBG ("not certified: %s", pkg.Name());
// 	  response.encode_int (pkgtrust_not_certified);
// 	  response.encode_string (pkg.Name());
// 	}

//       if (cache[pkg].Upgrade() && !cache[pkg].NewInstall()
// 	  && !domain_dominates_or_is_equal (new_domain, cur_domain))
// 	{
// 	  log_stderr ("domain change: %s (%s -> %s)",
// 		      pkg.Name(),
// 		      domains[cur_domain].name,
// 		      domains[new_domain].name);
// 	  response.encode_int (pkgtrust_domains_violated);
// 	  response.encode_string (pkg.Name());
// 	}
//     }

//   response.encode_int (pkgtrust_end);
// }

// static void
// encode_upgrades ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
//   if (ensure_cache (true))
//     {
//       pkgDepCache &cache = *(awc->cache);

//       for (pkgCache::PkgIterator pkg = cache.PkgBegin();
// 	   pkg.end() != true;
// 	   pkg++)
// 	{
// 	  if (cache[pkg].Upgrade() && !cache[pkg].NewInstall())
// 	    {
// 	      response.encode_string (pkg.Name());
// 	      response.encode_string (cache[pkg].CandVersion);
// 	    }
// 	}
//     }

//   response.encode_string (NULL);
// }

/* We modify the pkgDPkgPM package manager so that we can provide our
   own method of constructing the 'order list', the ordered list of
   packages to handle.  We do this to ignore packages that should be
   kept.  Ordinarily, if a package should be kept but needs to be
   reinstalled or configured, libapt-pkg will try to do that with
   every operation.

   XXX - there might be a way to get the same effect by cleverly
         manipulating the pkgDepCache, but I would have to look
         harder...
*/

class myDPkgPM : public pkgDPkgPM
{
public:

  bool CheckDownloadedPkgs (bool clear_corrupted, int *pkgflags);

  bool CreateOrderList ();

  myDPkgPM(pkgDepCache *Cache);
};

bool
myDPkgPM::CreateOrderList ()
{
  if (pkgPackageManager::List != 0)
    return true;
   
  delete pkgPackageManager::List;
  pkgPackageManager::List = new pkgOrderList(&Cache);
   
  // Generate the list of affected packages and sort it
  for (PkgIterator I = Cache.PkgBegin(); I.end() == false; I++)
    {
      // Ignore no-version packages
      if (I->VersionList == 0)
	continue;
      
      // Not interesting
      if ((Cache[I].Keep() == true || 
	   Cache[I].InstVerIter(Cache) == I.CurrentVer()) && 
	  I.State() == pkgCache::PkgIterator::NeedsNothing &&
	  (Cache[I].iFlags & pkgDepCache::ReInstall) != pkgDepCache::ReInstall &&
	  (I.Purge() != false || Cache[I].Mode != pkgDepCache::ModeDelete ||
	   (Cache[I].iFlags & pkgDepCache::Purge) != pkgDepCache::Purge))
	continue;
      
      // Ignore interesting but kept packages, except when they are
      // related to the current operation.

      if (Cache[I].Keep() == true && !is_related (I))
	{
	  log_stderr ("Not handling unrelated package %s.", I.Name());
	  continue;
	}
      
      // DBG ("Handling interesting package %s.", I.Name());

      // Append it to the list
      pkgPackageManager::List->push_back(I);      
    }
   
  return true;
}

bool
myDPkgPM::CheckDownloadedPkgs (bool clean_corrupted, int *pkgflags)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  bool result = true;
  *pkgflags = 0;
  
  for (pkgOrderList::iterator I = pkgPackageManager::List->begin(); 
       I != pkgPackageManager::List->end(); I++)
    {
      bool partial_result = true;
      PkgIterator Pkg(Cache,*I);
      pkgCache::VerIterator cand_ver = Cache[Pkg].CandidateVerIter(Cache);
      package_record rec (cand_ver);
      string File = FileNames[Pkg->ID];
      if (File.empty())
        continue;
      FileFd Fd (File, FileFd::ReadOnly);
      if (_error->PendingError() == true) // return false?
        continue;

      string ExpectedSHA256 = rec.get_string("SHA256");
      int flags = get_flags(rec);
      *pkgflags |= flags;

      if (!ExpectedSHA256.empty())
        {
          SHA256Summation SHA256;
          SHA256.AddFD(Fd.Fd(), Fd.Size());
          string file_sha256 = string(SHA256.Result());

          if (file_sha256 != ExpectedSHA256)
            {
              log_stderr ("File %s is corrupted (SHA256).", File.c_str());
              partial_result = false;
            }
        }
      else
        {
          string ExpectedSHA1 = rec.get_string("SHA1");

          if (!ExpectedSHA1.empty())
            {
              SHA1Summation SHA1;
              SHA1.AddFD(Fd.Fd(), Fd.Size());
              string file_sha1 = string(SHA1.Result());

              if (file_sha1 != ExpectedSHA1)
                {
                  log_stderr ("File %s is corrupted (SHA1).", File.c_str());
                  partial_result = false;
                }
            }
          else
            {
              string ExpectedMD5 = rec.get_string("MD5sum");

              if (!ExpectedMD5.empty())
                {
                  MD5Summation sum;
                  sum.AddFD (Fd.Fd(), Fd.Size());
                  string MD5 = (string)sum.Result();

                  if (MD5 != ExpectedMD5)
                    {
                      log_stderr ("File %s is corrupted (MD5sum).", File.c_str());
                      partial_result = false;
                    }
                }
            }
        }
      Fd.Close();

      OperationData *op = cache_get_current_operation();
      if (op && op->type == OPERATION_SYSTEM_UPGRADE )
      {
        // Check whether Aegis agrees with the expected origin.  If it
        // doesn't, we reject the package as corrupted.  Passing it on
        // for installation might lead to harmfully interrupted
        // installations when Aegis rejects it later, or to confusing
        // situations when Aegis records a different origin for a
        // package than what we expect.

        domain_t real = get_file_domain (File.c_str());
        domain_t expected = awc->cache->find_best_domain (cand_ver);
        if (real != expected)
          {
            log_stderr ("Origin check failed for %s:\n", File.c_str());
            log_stderr (" Expected '%s' but it really is from '%s'\n", 
                        domain_get_name (expected), domain_get_name (real));
            partial_result = false;
          }
      }

      result = result && partial_result;
      if (clean_corrupted && !partial_result)
        unlink (File.c_str());
    }

  return result;
}

myDPkgPM::myDPkgPM (pkgDepCache *Cache)
  : pkgDPkgPM (Cache)
{
}

static ResultCode
combine_rescodes (ResultCode all, ResultCode one)
{
  if (all == rescode_success)
    return one;
  else if (all == one)
    return all;
  else
    return rescode_failure;
}

/* Sets an alternative cache directory to download packages.
   If alt_download_root is NULL, then the default path is set */
static bool
set_dir_cache_archives (const char *alt_download_root)
{
  bool result = false;

  if (alt_download_root == NULL)
    {
      /* Setting default location */

      /* If setting to default set just the relative path */
      _config->Set ("Dir::Cache::Archives", DEFAULT_DIR_CACHE_ARCHIVES);

      result = true;
    }
  else
    {
      char *archives_dir = NULL;
      /* Setting an alternative location */

      /* Prepare data to create a temporary directory */
      archives_dir = g_strdup_printf ("%s/%s",
				       alt_download_root,
				       ALT_DIR_CACHE_ARCHIVES);

      /* Create temporary dir to use it as download location */
      if (mkdir (archives_dir, 0777) < 0 && errno != EEXIST)
	{
	  _error->Error ("Can't create directory at %s: %m", archives_dir);
	}
      else
	{
	  char *partial_dir = g_strdup_printf ("%s/partial", archives_dir);
	  if (mkdir (partial_dir, 0777) < 0  && errno != EEXIST)
	    {
	      _error->Error("Can't create directory at %s: %m", partial_dir);
	    }
	  else
	    {
	      /* If not setting to default set the full path */
	      _config->Set ("Dir::Cache::Archives", archives_dir);

	      result = true;
	    }
	  g_free (partial_dir);
	}
      g_free (archives_dir);
    }

  return result;
}

static int64_t
get_pkg_required_free_space ()
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);

  int64_t retval = 0;

  for (pkgCache::PkgIterator pkg = cache.PkgBegin (); pkg.end () != true; pkg++)
    {
      if ( (!cache[pkg].Keep() || (cache[pkg].iFlags & pkgDepCache::ReInstall) ) && ( (cache[pkg].Upgrade () || pkg.State () != pkgCache::PkgIterator::NeedsNothing) ) )
        {
          pkgCache::VerIterator ver = cache[pkg].CandidateVerIter (cache);
          package_record rec (ver);
	  if (rec.has ("Maemo-Required-Free-Space"))
	    retval += get_required_free_space (rec);
	  else
	    retval += get_installed_size_bytes (ver);
        }
    }

  return retval;
}

static bool
is_there_enough_free_space (const char *archive_dir, int64_t size,
			    GError **error)
{
  struct statvfs buf;

  if (statvfs (archive_dir, &buf) != 0)
    {
      log_stderr ("Couldn't determine free space in %s", archive_dir);
      return false;
    }

  const char *internal_mmc_mountpoint = getenv ("INTERNAL_MMC_MOUNTPOINT");
  if (!internal_mmc_mountpoint)
    internal_mmc_mountpoint = INTERNAL_MMC_MOUNTPOINT;

  const char *removable_mmc_mountpoint = getenv ("REMOVABLE_MMC_MOUNTPOINT");
  if (!removable_mmc_mountpoint)
    removable_mmc_mountpoint = REMOVABLE_MMC_MOUNTPOINT;

  /* what if after downloaded the bytes in DEFAULT_DIR_CACHE_ARCHIVES
   * there's not enough space to install them? */
  if (internal_mmc_mountpoint
      && !strstr (archive_dir, internal_mmc_mountpoint)
      && removable_mmc_mountpoint
      && !strstr (archive_dir, removable_mmc_mountpoint))
    {
      /* Should we add install_user_size_delta value */
      size += get_pkg_required_free_space ();
    }

  if (unsigned (buf.f_bfree) < double (size) / buf.f_bsize)
    {
      log_stderr ("You don't have enough free space in %s", archive_dir);

      uint64_t available = buf.f_bfree;
      available *= buf.f_bsize;

      g_set_error (error, PKGMGRD_ERROR, ERROR_NOT_ENOUGH_SPACE,
		   "Not enough space to install: %Ld MB needed, %Ld MB available",
		   size / (1024*1024),
		   available / (1024*1024));

      return false;
    }

  return true;
}

static void
__purge_check()
{
	AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
	pkgCacheFile &Cache = *(awc->cache);
	if (_config->FindB("APT::Get::Purge",false) == true) {
		pkgCache::PkgIterator I = Cache->PkgBegin();
		for (; I.end() == false; I++)
		{
			if (I.Purge() == false && Cache[I].Mode == pkgDepCache::ModeDelete)
				Cache->MarkDelete(I,true);
		}
	}
}

static string
apt_pkg_error_str ()
{
  ostringstream out;
  string msg;

  while (!_error->empty ())
    {
      if (_error->PopMessage (msg))
	out << "E: ";
      else
	out << "W: ";
      out << msg << '\n';
    }

  return out.str ();
}

/*!
  Perform DPkgPM::DoInstall() and return the dpkg stdout, stderr.
*/
static pkgPackageManager::OrderResult
do_install(pkgDPkgPM &pm, int pmstatus_wfd, gchar **std_out, gchar **std_err)
{
  pkgPackageManager::OrderResult res = pkgPackageManager::Failed;

  char outtmpname[24], errtmpname[24];
  int fd[2];

  strcpy(outtmpname, "/var/tmp/dpkg.XXXXXX");
  fd[0] = mkstemp(outtmpname);
  if (fd[0] < 0)
    {
      // error.set(Error::Internal, "Could not create tmp file");
      return res;
    }

  strcpy(errtmpname, "/var/tmp/dpkg.XXXXXX");
  fd[1] = mkstemp(errtmpname);
  if (fd[1] < 0)
    {
      // error.set(Error::Internal, "Could not create tmp file");
      close(fd[0]);
      unlink(outtmpname);
      return res;
    }

  int outfd, errfd;
  fpos_t outpos, errpos;

  // Temporarily redirect stdout, stderr to file
  fflush(stdout);
  fgetpos(stdout, &outpos);
  outfd = dup(fileno(stdout));
  freopen(outtmpname, "w", stdout);

  fflush(stderr);
  fgetpos(stderr, &errpos);
  errfd = dup(fileno(stderr));
  freopen(errtmpname, "w", stderr);

  /* The following tricks aegis-install (aka dpkg) into believing that
     we are not in Scratchbox.  Otherwise, it would not use the real
     origin of packages.

     XXX - This should be removed once NB#214851 is fixed.
  */
  string old_sbox_dir;
  if (getenv ("_SBOX_DIR"))
    {
      old_sbox_dir = getenv ("_SBOX_DIR");
      unsetenv ("_SBOX_DIR");
    }

  // Do install
  res = pm.DoInstall(pmstatus_wfd);

  if (!old_sbox_dir.empty ())
    setenv ("_SBOX_DIR", old_sbox_dir.c_str(), 1);

  // Restore stdout, stderr
  fflush(stdout);
  dup2(outfd, fileno(stdout));
  close(outfd);
  clearerr(stdout);
  fsetpos(stdout, &outpos);

  fflush(stderr);
  dup2(errfd, fileno(stderr));
  close(errfd);
  clearerr(stderr);
  fsetpos(stderr, &errpos);

  GError *error = NULL;
  if (!g_file_get_contents (outtmpname, std_out, NULL, &error))
    {
      DBG ("g_file_get_contents() failed");
      g_error_free (error);
    }

  error = NULL;
  if (!g_file_get_contents (errtmpname, std_err, NULL, &error))
    {
      DBG ("g_file_get_contents() failed");
      g_error_free (error);
    }

  close(fd[0]);
  close(fd[1]);
  unlink(outtmpname);
  unlink(errtmpname);

  return res;
}

static void
make_acquired_packages_interesting (pkgAcquire &Fetcher)
{
  GList *packages = NULL;
  
  for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin();
       I != Fetcher.ItemsEnd(); I++)
    {
      const char *h = (*I)->HookName();
      const char *a = (*I)->HookArg();
      if (strcmp (h, "Archives") == 0 && a)
        packages = g_list_append (packages, (void *)a);
    }

  add_interesting_packages (packages);
  g_list_free (packages);
}

/* operation () is used to run pending apt operations
 * (removals or installations). If check_only parameter is
 * enabled, it will only check if the operation is doable.
 *
 * operation () is used from cmd_install_package,
 * cmd_install_check, cmd_uninstall_package and
 * cmd_download_system_upgrade
 */

ResultCode
operation (bool check_only,
	const char *alt_download_root,
	bool download_only,
	GError **error,
	bool allow_download,
	bool with_status,
	int *pkgflags)
{
  *pkgflags = 0;
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgCacheFile &Cache = *(awc->cache);
  SPtr<myDPkgPM> Pm;
  bool op_started_sent = false;

  {
    OperationData *op = cache_get_current_operation();
    if (op)
      op->update_failed_installs = true;
  }
  apply_proxy_settings ();

  __purge_check();
  if (Cache->DelCount() == 0 && Cache->InstCount() == 0 &&
      Cache->BadCount() == 0)
    {
      /* Encode trust summary and upgrades information when running
	 APTCMD_INSTALL_CHECK, even when there's nothing to do */
      if (check_only)
	{
// 	  response.encode_int (pkgtrust_end);
// 	  response.encode_string (NULL);
	}
      return rescode_success;
    }

  // Create the text record parser
  pkgRecords Recs (Cache);
  if (_error->PendingError() == true)
    return rescode_failure;

  if (!set_dir_cache_archives (alt_download_root))
    {
      /* Log error, but keep working with default value */
      g_warning ("Failed to use %s to download packages. Using default.",
		 alt_download_root);
    }

  // Lock the archive directory
  FileFd Lock;
  if (_config->FindB("Debug::NoLocking",false) == false)
    {
      Lock.Fd(ForceLock(_config->FindDir("Dir::Cache::Archives") + "lock"));
      if (_error->PendingError() == true)
	{
	  _error->Error("Unable to lock the download directory");
	  return rescode_failure;
	}
    }

  // Create the download object
  DownloadStatus Stat;
  pkgAcquire Fetcher (with_status? &Stat : NULL);

  // Read the source list
  pkgSourceList List;
  if (List.ReadMainList() == false)
    {
      _error->Error("The list of sources could not be read.");
      return rescode_failure;
    }

  pkgIndexFile *extra_index_file = Cache.GetExtraIndexFile ();
  if (extra_index_file)
    List.AddIndexFile (extra_index_file, true);
  
//   set_sources_for_get_domain (&List);

  // Create the package manager
  //
  Pm = new myDPkgPM (Cache);

  // Create the order list explicitely in a way that we like.  We
  // have to do it explicitely since CreateOrderList is not virtual.
  //
  if (!Pm->CreateOrderList ())
    return rescode_failure;

  // Prepare to download
  //
//   reset_new_domains ();
  if (Pm->GetArchives(&Fetcher,&List,&Recs) == false ||
	  _error->PendingError() == true) {
	  DBG("Failed to get archives or pending error");
	  return rescode_failure;
  }

  double FetchBytes = Fetcher.FetchNeeded();
  double FetchPBytes = Fetcher.PartialPresent();

  if (_error->PendingError() == true) {
	  DBG("Failed pending error");
	  return rescode_failure;
  }

  if (check_only)
    {
//       encode_trust_summary ();
//       encode_upgrades ();
      return rescode_success;
    }

//   collect_new_domains ();

  make_acquired_packages_interesting (Fetcher);

  if ((int)(FetchBytes - FetchPBytes) > 0)
    {
      if (!allow_download)
	{
	  log_stderr ("would need to download, but it's not allowed");

#ifdef DEBUG
	  for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin();
	       I != Fetcher.ItemsEnd(); I++)
	    {
	      g_debug ("Would download %s",
		       (*I)->DescURI().c_str());
	    }
#endif

	  return rescode_packages_not_found;
	}

      download_size = FetchBytes - FetchPBytes;
      if (!is_there_enough_free_space
          (_config->FindDir ("Dir::Cache::Archives").c_str (), download_size,
	   error))
	{
	  return rescode_out_of_space;
	}

      if (with_status)
	{
	  send_operation_started_signal();
	  op_started_sent = true;
	}

      if (!connect_network (FALSE)) 
        {
          g_set_error (error, PKGMGRD_ERROR, ERROR_DOWNLOAD_FAILED, "No network");
          return rescode_failure;
        }

    /* Send a status report now if we are going to download
    something.  This makes sure that the progress dialog is
    shown even if the first pulse of the fetcher takes a long
    time to arrive.
    */

    if (with_status)
      send_download_progress_signal(0, (int)(FetchBytes - FetchPBytes), 0);
  }

  if (Fetcher.Run() == pkgAcquire::Failed) {
	  DBG("Fetcher.Run() failed!");
          disconnect_network ();
	  return rescode_failure;
  }

  disconnect_network ();

  /* Check that there is enough free space in the root partition to
     install the packages */

  if (!is_there_enough_free_space ("/", download_size, error))
    {
      return rescode_out_of_space;
    }

  /* Print out errors and distill the failure reasons into a
     apt_proto_rescode.
  */
  std::ostringstream fetch_errors;
  ResultCode result = rescode_success;
  for (pkgAcquire::ItemIterator I = Fetcher.ItemsBegin();
       I != Fetcher.ItemsEnd(); I++)
    {
      if ((*I)->Status == pkgAcquire::Item::StatDone &&
	  (*I)->Complete == true)
	continue;

      if ((*I)->Status == pkgAcquire::Item::StatIdle)
	continue;

      string const& desc_uri = (*I)->DescURI();
      string const& error_text = (*I)->ErrorText;
      DBG ("Failed to fetch %s: %s",
	   desc_uri.c_str(), error_text.c_str());

      fetch_errors << "Failed to fetch " << desc_uri << ": "
		   << error_text << std::endl;

      ResultCode this_result;

      if (g_str_has_prefix (error_text.c_str(), "404"))
	this_result = rescode_packages_not_found;
      else if (g_str_has_prefix (error_text.c_str(), 
				 "Size mismatch"))
	this_result = rescode_package_corrupted;
      else if (g_str_has_prefix (error_text.c_str(), 
				 "MD5Sum mismatch"))
	this_result = rescode_package_corrupted;
      else
	this_result = rescode_failure;

      result = combine_rescodes (result, this_result);
    }

    if (result != rescode_success) {
        OperationData *op = cache_get_current_operation ();
        if (op && op->cancel_flag) {
            if(op->download_failed_flag) {
                g_set_error (error, PKGMGRD_ERROR, ERROR_DOWNLOAD_FAILED, "No network");
            } else {
                g_set_error (error, PKGMGRD_ERROR, ERROR_OPERATION_CANCELED, "Canceled");
            }
        } else {
            g_set_error (error, PKGMGRD_ERROR, ERROR_DOWNLOAD_FAILED,
                         fetch_errors.str ().c_str ());
        }

        return (result == rescode_failure)?rescode_download_failed:result;
    }

  /* Make sure that all the packages are written to disk before
     proceeding.  This helps with retrying the operation in case it is
     interrupted.
  */

  if (with_status)
    {
      int b = (int)(FetchBytes - FetchPBytes);
      send_download_progress_signal(b, b, 1);
    }
  sync ();

  /* Install packages if not just downloading */
  if (!download_only)
    {
      if (with_status && ! op_started_sent ) {
            send_operation_started_signal();
            op_started_sent = true;
        }
      if (with_status)
	send_operation_progress_signal (0);

      if (Pm->CheckDownloadedPkgs (true, pkgflags) == false)
        return rescode_package_corrupted;

      gchar *out = NULL, *err = NULL;

      OperationData *op = cache_get_current_operation();
      if (op && op->type == OPERATION_SYSTEM_UPGRADE)
	save_operation_record (cache_get_current_operation (),
			       alt_download_root);

      /* Do install */
      _system->UnLock();
      pkgPackageManager::OrderResult Res;
      if (flag_collect_dpkg_output)
	Res = do_install (*Pm, status_fd, &out, &err);
      else
	Res = Pm->DoInstall (status_fd);
      _system->Lock();

      awc->cache->save_extra_info ();

      if (Res == pkgPackageManager::Failed ||
	      _error->PendingError() == true)
      {
	      OperationData *op = cache_get_current_operation ();
	      int err_code = op &&
                (op->type == OPERATION_UPGRADE
                 || op->type == OPERATION_UPGRADE_ALL
                 || op->type == OPERATION_SYSTEM_UPGRADE
                 || (op->type == OPERATION_INSTALL && awc->cache->previous_error_was (ERROR_UPGRADE_FAILED)))
		? ERROR_UPGRADE_FAILED
                : ERROR_OPERATION_FAILED;
	      string const& dpkg_error = apt_pkg_error_str ();
	      g_set_error (error, PKGMGRD_ERROR, err_code,
			   "Operation failed: %s\n%s",
			   dpkg_error.c_str (),
			   err);
	      DBG("Pm->DoInstall failed: %s: %s", dpkg_error.c_str (), err);
	      g_free (out);
	      g_free (err);
	      return rescode_failure;
      }

      g_free (out);
      g_free (err);

      if (Res != pkgPackageManager::Completed) {
	      DBG("Pm->DoInstall not completed!");
	      return rescode_failure;
      }
    }

  return rescode_success;
}

/* APTCMD_CLEAN
 */

// void
// cmd_clean ()
// {
//   bool success = true;
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

//   // Try to lock the archive directory.  If that fails because we are
//   // out of space, continue anyway since it is critical to free flash
//   // in that case.
//   //
//   // ForceLock has the same interface as GetLock: it returns -1 in
//   // case of failure with errno set appropriately.  However, errnor is
//   // always EPERM when the open syscall failed.  "Feh.."

//   FileFd Lock;
//   if (_config->FindB("Debug::NoLocking",false) == false)
//     {
//       int fd = ForceLock(_config->FindDir("Dir::Cache::Archives") + "lock");
//       if (fd < 0)
// 	{
// 	  if (errno != EPERM && errno != ENOSPC)
// 	    {
// 	      success = false;
// 	      _error->Error("Unable to lock the download directory");
// 	    }
// 	  else
// 	    _error->Warning("Unable to lock the download directory, but cleaning it anyway.");
// 	}
//       else
// 	Lock.Fd (fd);
//     }
   
//   if (success)
//     {
//       pkgAcquire Fetcher;
//       Fetcher.Clean(_config->FindDir("Dir::Cache::archives"));
//       Fetcher.Clean(_config->FindDir("Dir::Cache::archives") + "partial/");
//     }

//   response.encode_int (success);

//   // As a special case, we try to init the cache again.  Chances are
//   // good that it will now succeed because there might be more space
//   // available now.  We don't use ensure_cache for this since we want
//   // it to happen silently.

//   if (awc->cache == NULL)
//     need_cache_init ();
// }


// XXX - interpret status codes

static bool
__check_dep(pkgCache::PkgIterator& pkg, pkgCache::VerIterator& ver, string &version, unsigned int op)
{
	if (ver.end ()) {
		// might be a virtual package, check the provides list.

		pkgCache::PrvIterator P = pkg.ProvidesList();

		for (; P.end() != true; P++)
		{
			// Check if the provides is a hit
			if (P.OwnerPkg().CurrentVer() != P.OwnerVer())
				continue;

			// Compare the versions.
			if (debVS.CheckDep (P.ProvideVersion(), op, version.c_str ())) {
				return true;
			}
		}
		return false;
	}
	else {
		return debVS.CheckDep (ver.VerStr (), op, version.c_str ());
	}
}
static bool
check_dependency (string &package, string &version, bool repo_pkg, unsigned int op)
{
  if (!ensure_cache (true))
    return false;

  DBG ("check_dependency: %s (%s)", package.c_str(), version.c_str());
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  pkgDepCache &cache = *(awc->cache);
  bool result = false;
  pkgCache::PkgIterator pkg;
  pkgCache::VerIterator ver;

  pkg = cache.FindPkg (package);
  if (pkg.end ())
    return false;

  ver = pkg.CurrentVer ();
  result = __check_dep (pkg, ver, version, op);
  if (!result && repo_pkg) {
	  ver = cache[pkg].CandidateVerIter(cache);
	  result = __check_dep (pkg, ver, version, op);
  }
  return result;
}

static void
add_dep_string (string &str,
		string &package, string &version, unsigned int op)
{
  str += package;
  if (op != pkgCache::Dep::NoOp)
    {
      str += " (";
      str += pkgCache::CompType (op);
      str += " ";
      str += version;
      str += ")";
    }
}

static PkgInstallStatus
check_and_encode_missing_dependencies (const char *deps, const char *end,
				bool only_check,
				bool repo_pkg)
{
  const char *ptr;
  string package, version;
  unsigned int op;
  bool dep_ok = true;

  ptr = deps;
  while (true)
    {
      // check one 'or group'

      bool group_ok = false;
      string group_string = "";

      while (true)
	{
	  ptr = debListParser::ParseDepends (ptr, end,
					     package, version, op,
					     false);
	  if (ptr == NULL)
	    {
	      cerr << "Error parsing depends list\n";
	      return status_unknown;
	    }

	  if (only_check && package == "maemo")
	    return status_incompatible_current;

	  add_dep_string (group_string, package, version, op);

	  if (!group_ok)
		  group_ok = check_dependency (package, version, repo_pkg,
					  op & ~pkgCache::Dep::Or);

	  if ((op & pkgCache::Dep::Or) == 0)
	    break;

	  group_string += " | ";
	}

      if (!group_ok)
	{
	  if (only_check)
	    cerr << "FAILED: " << group_string << "\n";
	  else
	    {
// 	      response.encode_int (sumtype_missing);
// 	      response.encode_string (group_string.c_str ());
	    }
	  dep_ok = false;
	}

      if (ptr == end)
	break;
    }

  DBG ("check_and_encode_missing_dependencies: %d", dep_ok);
  return dep_ok? status_installable : status_missing;
}

static bool
get_field (pkgTagSection *section, const char *field,
	   const char *&start, const char *&end)
{
  if (section->Find (field, start, end))
    {
      // fprintf (stderr, "%s = %.*s\n", field, end-start, start);
      return true;
    }
  else
    {
      // fprintf (stderr, "%s = <undefined>\n", field);
      return false;
    }
}

static gint64
get_field_int (pkgTagSection *section, const char *field, int def)
{
  const char *start, *end;
  if (get_field (section, field, start, end))
	  return g_ascii_strtoll (start, NULL, 10);
  else
    return def;
}

// static void
// encode_field (pkgTagSection *section, const char *field,
// 	      const char *def = "")
// {
//   const char *start, *end;
//   if (get_field (section, field, start, end))
//     response.encode_stringn (start, end-start);
//   else
//     response.encode_string (def);
// }

// static void
// encode_localized_field (pkgTagSection *section, const char *field,
// 			const char *def = "")
// {
//   const char *start, *end;

//   if (lc_messages && *lc_messages)
//     {
//       char *locale_field = g_strdup_printf ("%s-%s", field, lc_messages);
//       if (get_field (section, locale_field, start, end))
// 	{
// 	  response.encode_stringn (start, end-start);
// 	  g_free (locale_field);
// 	  return;
// 	}
//       g_free (locale_field);
//     }
  
//   if (get_field (section, field, start, end))
//     response.encode_stringn (start, end-start);
//   else
//     response.encode_string (def);
// }

static bool
substreq (const char *start, const char *end, const char *str)
{
  return ((size_t)(end-start) == strlen (str)) && !strncmp (start, str, end-start);
}

static PkgInstallStatus
check_installable (pkgTagSection &section, bool only_user, bool only_os_update, bool repo_pkg)
{
  PkgInstallStatus installable_status = status_installable;
  const char *start, *end;

  if (!get_field (&section, "Architecture", start, end)
      || !(substreq (start, end, DEB_HOST_ARCH)
	   || substreq (start, end, "all")))
    installable_status = status_incompatible;
    
  if (only_user && !is_user_package (section))
    {
      /* Put more information for developers into the log.  They will
	 likely be confused by the "incompatible" error message when
	 testing a package that has not been properly flagged as
	 visible.
       */
      g_message ("Package must have \"Maemo-Flags: visible\" "
		 "to be considered compatible.\n");
      return status_incompatible_section;
    }

  // if (only_user && get_field (&section, "Section", start, end))
  //   {
  //     if (only_user_system && !is_user_system_section (start, end))
  // 	return status_incompatible;
  //     else if (!only_user_system && is_user_system_section (start, end))
  // 	return status_incompatible;
  //   }

  if (get_field (&section, "Pre-Depends", start, end)) {
    installable_status =
	    combine_status (check_and_encode_missing_dependencies (start, end, true, repo_pkg),
		      installable_status);
    DBG ("check_installable: Pre-Depends: %d" ,installable_status == status_installable);
  }

  if (get_field (&section, "Depends", start, end)) {
    installable_status = 
      combine_status (check_and_encode_missing_dependencies (start, end, true, repo_pkg),
		      installable_status);
    DBG ("check_installable: Depends: %d" ,installable_status == status_installable);
  }

  return installable_status;
}

Package*
cmd_get_file_details (const char *filename, GError **error)
{
  bool only_user = true;
//   const char *filename = request.decode_string_in_place ();
  pkgTagSection section;
  gchar *record = read_tag_section (filename, &section, error);
  if (!record) {
	  return NULL;
  }

  PkgInstallStatus installable_status = check_installable (section, only_user, false, true);

  const char *installed_version = NULL;
  int64_t installed_size = 0;

  if (ensure_cache (true))
    {
      AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
      pkgDepCache &cache = *(awc->cache);
      pkgCache::PkgIterator pkg = cache.FindPkg (section.FindS ("Package"));
      if (!pkg.end ())
	{
	  pkgCache::VerIterator cur = pkg.CurrentVer ();
	  if (!cur.end ())
	    {
	      installed_version = cur.VerStr ();
	      installed_size = cur->InstalledSize;
	    }
	}
    }
  Package *pkgli = package_new();
  const char *start, *end;
  if (get_field (&section, "Section", start, end) && is_user_section(start, end)) {
	  gchar *sect = g_strndup(start, end-start);
	  pkgli->category = g_strdup(user_section_to_category(sect));
	  g_free (sect);
  } else
          pkgli->category = g_strdup("unknown");

  if (!*error && !is_user_package (section)) {
          g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_USER,
                      "'%s' is not user package!", filename);
  }

  if (get_field (&section, "Package", start, end)) {
	  pkgli->name = g_strndup(start, end-start);
  }
  if (get_field (&section, "Maemo-Display-Name", start, end)) {
	  pkgli->display_name = g_strndup(start, end-start);
  }
//   response.encode_string (installed_version);
//   response.encode_int64 (installed_size);
  if (get_field (&section, "Version", start, end)) {
	  pkgli->version = g_strndup(start, end-start);
  }
//   if (get_field (&section, "Maintainer", start, end)) {
// 	  pkgli->version = g_strndup(start, end-start);
//   }
  if (installed_version) {
	  pkgli->type = PKG_TYPE_INSTALLED;
  } else {
	  pkgli->type = PKG_TYPE_NOT_INSTALLED;
  }
  pkgli->install_status = installable_status;
  pkgli->installed_size = 1024 * get_field_int (&section, "Installed-Size", 0);
//   response.encode_int64 (1024LL * get_field_int (&section, "Installed-Size", 0)
// 		       - installed_size);
  if (get_field (&section, "Description", start, end)) {
	  pkgli->description = format_multiline_field(start, end-start);
  }
  if (get_field (&section, "Maemo-Icon-26", start, end)) {
	  pkgli->icon_data = g_strndup(start, end-start);
  }

  pkgli->repository = g_strdup("local-file");
  pkgli->flag_local_file = TRUE;
  if (installable_status != status_installable) {
//     encode_missing_dependencies (section);
  }

  delete[] record;
  if (!*error &&
	  (!pkgli->name ||
		  !pkgli->version ||
		  !pkgli->category)) {
	  g_set_error(error, PKGMGRD_ERROR, ERROR_INVALID_FILE,
		  "Cannot read package details from '%s'!", filename);
  }
  if (*error) {
	  package_free(pkgli);
	  pkgli = NULL;
  }
  return pkgli;
}

GHashTable *
cmd_get_file_details_map (const char *filename, GError **error)
{
    DBG(": %s", filename);

    cache_reset_packages ();
    current_operation = 0;

    Package *p = cmd_get_file_details (filename, error);
    if (p) {
        GHashTable *map = g_hash_table_new (g_str_hash, g_str_equal);
        pkgTagSection section;
        gchar *record = read_tag_section (filename, &section, error);
        domain_t domain = get_file_domain (filename);
        pkgCache::PkgFileIterator pf;

        g_hash_table_insert_string_value (map, "Name", p->name);
        g_hash_table_insert_string_value (map, "Version", p->version);
        g_hash_table_insert_string_value (map, "Category", p->category);
        g_hash_table_insert_string_value (map, "DisplayName", p->display_name);
        g_hash_table_insert_string_value (map, "Description", p->description);
        g_hash_table_insert_string_value (map, "Repository", p->repository);
        g_hash_table_insert_string_value (map, "IconData", p->icon_data);
        g_hash_table_insert_string_value (map, "Type",
		                                  package_type_to_string (p->type));
        g_hash_table_insert_boolean_value (map, "IsBroken", p->is_broken);
        g_hash_table_insert_boolean_value (map, "FlagReboot", p->flag_reboot);
        g_hash_table_insert_boolean_value (map, "FlagIgnored", p->flag_ignored);
        g_hash_table_insert_boolean_value (map, "FlagSystemUpdate", p->flag_system_update);
        if ( p->flag_local_file ) {
            g_hash_table_insert_boolean_value (map, "FlagLocalFile", p->flag_local_file);
        }
        g_hash_table_insert_int64_value (map, "InstalledSize", p->installed_size);
        g_hash_table_insert_int64_value (map, "Size", p->download_size);
        g_hash_table_insert_int64_value (map, "InstalledTimestamp", p->installed_timestamp);
        g_hash_table_insert_int64_value (map, "AvailableTimestamp", p->available_timestamp);
        g_hash_table_insert_string_value (map, "Filename", filename);
        pkg_map_insert_failed_install_data (map, p->name, p->version);

        if ( !*error ) {
            AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
            pkgDepCache &cache = *awc->cache;
            
            pkgCache::PkgIterator pkg = cache.FindPkg(p->name);
            if ( true /*pkg.end()*/ ) {
                string p_str = section.FindS("Package");
                string v_str = section.FindS("Version");

                pf = add_extra_index_from_section(section, filename, domain);

                if (pf.end()) {
                  g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Extra index error");
                  delete [] record;
                  package_free (p);
                  g_hash_table_unref (map);
                  return NULL;
                }
                pkgDepCache &cache = *awc->cache;
                pkg = cache.FindPkg(p->name);
                need_cache_init ();
            }
            if (!pkg.end ()) {
                pkgCache::VerIterator ver = pkg.CurrentVer ();
                if ( ver.end() ) {
                    //bool ok = 
                  awc->cache->find_version (pkg, NULL, NULL, pf, ver, error);
                }
                if (!ver.end () && strcmp(ver.VerStr(), p->version) == 0) {
                    pkg_map_insert_deps(map, PKG_TYPE_INSTALLED, pkg, ver);
                }
            } else {
                DBG("%s pkg not found", __PRETTY_FUNCTION__);
            }
        }

        // Insert additional fields not listed in skip_fields array.
        static const char *skip_fields[] = {
            "Description",
            "Installed-Size",
            "Maemo-Display-Name",
            "Maemo-Icon-26",
            "Package",
            "Version",
            0
        };

        for (FieldIterator i (section); !i.end (); ++i) {
            string const& key = i.key ();

            if (!_find_string (skip_fields, key.c_str ())) {
                g_hash_table_insert_string_value (map,
                                                key.c_str (),
                                                  format_multiline_field (i.value ().c_str ()));
            }
        }

        delete [] record;
        package_free (p);

	// Don't reset in query.
        // apt_worker_reset_state ();
        
        return map;
    } else {
	if (!*error) {
	    g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL,
			 "cmd_get_file_details(%s) returned NULL but set no error", filename);
	}
        return NULL;
    }
}

pkgCache::PkgFileIterator
add_extra_index_from_section (pkgTagSection &section, const char *filename, domain_t domain)
{
  const char *Start;
  const char *Stop;

  int filename_len = strlen ("Filename: file://") + strlen (filename) + 1;

  section.GetSection (Start, Stop);
  while (Stop > Start
         && Stop[-1] == '\n')
    Stop--;
  if (Stop[0] == '\n')
    Stop++;

  int section_len = Stop - Start;

  char *index = new char[section_len + filename_len + 1];
  memcpy (index, Start, section_len);
  sprintf (index + section_len, "Filename: file://%s\n", filename);

  pkgCache::PkgFileIterator pf = cache_add_extra_index (index, domain, false);

  delete[] index;

  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  if (awc->cache == NULL)
    pf = pkgCache::PkgFileIterator();

  return pf;
}

gboolean
cmd_install_file (const char *filename, GError **error)
{
  AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
  int pkgflags = 0;
  ResultCode result_code = rescode_failure;
  OperationData *op_for_revert = NULL;
  gchar *record = NULL;
  bool installable = false;
  bool is_update = false;

  pkgTagSection section;

  cache_operation_start ("", "", NULL, filename, OPERATION_INSTALL_FILE);

  if ( check_battery_level(error, CRITICAL_BATTERY_LEVEL) ) {
    OperationData *op = cache_get_current_operation();
    if ( op ) {
        op->update_failed_installs = true;
    }
    cache_operation_complete (*error);
    return FALSE;
  }

  if (filename[0] != '/') {
    g_set_error (error, PKGMGRD_ERROR, ERROR_INVALID_OPERATION, "Filename must be absolute");
    goto out;
  }

  record = read_tag_section (filename, &section, error);
  if (record == NULL)
    goto out;

  {
    string arch = section.FindS("Architecture");
    if (arch != DEB_HOST_ARCH && arch != "all")
      {
        g_set_error(error, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE,
                    "Package '%s' has wrong architecture", filename);
        goto out;
      }
  }

  {
    string p = section.FindS("Package");
    string v = section.FindS("Version");

    const char *package = p.c_str ();
    const char *version = v.c_str ();
    domain_t domain = get_file_domain (filename);
    pkgCache::PkgFileIterator pf;

    cache_operation_set_package (package);
    cache_operation_set_version (version);
    cache_operation_set_domain (domain_get_name (domain));

    if (!ensure_cache (true)) {
      g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
      goto out;
    }

    {
      scoped_cache_rw_lock locker;
      pf = add_extra_index_from_section (section, filename, domain);
      if (pf.end()) {
	g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Extra index failure");
	goto out;
      }
    }

    pkgDepCache &cache = *(awc->cache);
    pkgCache::PkgIterator pkg = cache.FindPkg (package);
    if (!pkg.end() && !pkg.CurrentVer().end())
      is_update = true;

    pkgCache::VerIterator ver;
    if (!awc->cache->find_version (package, version, NULL, pf, ver, error))
      goto out;
    
    free_list(pkgs_to_be_deleted);
    pkgs_to_be_deleted = NULL;

    installable = mark_package_for_reinstall (ver, error);
    // pkgCache::VerIterator cand;
    // gboolean installable =
    //   find_version_to_install (package, version, true, cand, error)
    //   && mark_for_apt_install (*awc->cache, cand, FALSE, error);
    DBG ("%s Pkg: %s, installable: %d", __PRETTY_FUNCTION__, package, installable);
    
    if (verbosity >= 2)
      dump_operation_data_installed_user_deps (cache_get_current_operation ());
    
    if (installable) {
      op_for_revert = operation_data_copy (cache_get_current_operation());
      result_code = operation (false, NULL,
                               false, error, true, true, &pkgflags);
    } else {
      result_code = rescode_packages_not_found;
    }
    if (result_code != rescode_success) {
      if (!*error) {
        g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Rescode: %d", result_code);
      }
    }

    need_cache_init ();

    OperationData *original_op = operation_data_copy(cache_get_current_operation());

    if (op_for_revert)
      {
        if (result_code != rescode_success
            && (*error)->code == ERROR_OPERATION_FAILED
            && !is_update)
          revert_install (op_for_revert);
        operation_data_free (op_for_revert);
      }

    if (is_update && (*error) != NULL)
      {
        /* In the case of a failed update, only track errors when the
           package got broken by it.
        */
        
        cache_init (false);
        pkgDepCache &cache = *(awc->cache);
        pkgCache::PkgIterator pkg = cache.FindPkg (package);
        bool broken = !pkg.end() && is_broken (pkg);
        
        original_op->update_failed_installs = broken;
      }
    else
      original_op->update_failed_installs = true;

    cache_set_current_operation (original_op);
  }

 out:
  cache_operation_complete (*error);

  update_upgrade_notification ();

  delete[] record;

  return result_code == rescode_success;
}

gboolean
cmd_cancel (const char *operation, const char *package, const char *version, GError **error)
{
  int op_type = operation_type_from_str (operation);
  if (op_type < 0)
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_INVALID_OPERATION,
		   "Invalid operation '%s'", operation);
      return FALSE;
    }

  cache_operation_lock ();

  OperationData *op = cache_get_current_operation ();
  
  int current_op_type = -1;
  if ( op ) {
      current_op_type = op->type;
      if ( current_op_type == OPERATION_RESTORE_UPGRADE ) {
          current_op_type = OPERATION_RESTORE_INSTALL;
      }
  }

  if ( !op ) {
      if ( op_type == OPERATION_INSTALL || op_type == OPERATION_UPGRADE || op_type == OPERATION_INSTALL_FILE ) {
        DBG("%s", __PRETTY_FUNCTION__);
        if ( ! cache_remove_failed_installation((OperationType)op_type, package, version) )
            g_set_error(error, PKGMGRD_ERROR, ERROR_NO_OPERATION, "No operation");
      } else
          g_set_error(error, PKGMGRD_ERROR, ERROR_NO_OPERATION, "No operation");
  } else if (op->operation_progress_pct >= 0)
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_NOT_CANCELABLE,
		   "Operation is not cancelable");
    }
  else if (!(current_op_type == op_type
             && (!strcmp (op->pkg, package) || (op_type == OPERATION_INSTALL_FILE && !strcmp (op->filename, package)))
	     && !strcmp (op->version, version)))
    {
      g_set_error (error, PKGMGRD_ERROR, ERROR_INVALID_OPERATION,
		   "Operation does not match");
    }
  else
    op->cancel_flag = TRUE;

  cache_operation_unlock();

  return TRUE;
}

GHashTable *
cmd_simulate_operation(const gchar* operation, const gchar* packageidentifier, const gchar* version, const char *domain, const gchar* flags, GError **error)
{
    DBG("Operation: %s, pkg: %s (%s), flags: %s", operation, packageidentifier, version, flags);

    int op_type = operation_type_from_str (operation);
    if (op_type < 0) {
        g_set_error (error, PKGMGRD_ERROR, ERROR_INVALID_OPERATION, "Invalid operation '%s'", operation);
        return NULL;
    }
    
    if (!ensure_cache (false)) {
        g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL, "Cache failure");
        return NULL;
    }
    
    GList *install_deps = NULL;
    GList *install_user_deps = NULL;
    GList *upgrade_deps = NULL;
    GList *upgrade_user_deps = NULL;
    GList *remove_pkgs = NULL;
    GList *remove_user_pkgs = NULL;
    GList *revert_deps = NULL;
    GList *revert_user_deps = NULL;
    
    GList *ignore_list = NULL;
    
    bool trusted = true;
    
    AptWorkerCache *awc = AptWorkerCache::GetCurrent ();
    pkgDepCache &cache = *(awc->cache);
//    pkgCacheFile &cacheFile = *(awc->cache);
    pkgCache::PkgIterator pkg;
    string package_name = packageidentifier;
    
    int n_installable = 0;
    GError *err = NULL;

    StatusMode_vector_t sc;
    int trusted_cnt = 0;
    bool fromfile = false;

    if ( op_type == OPERATION_UPGRADE_ALL ) {
        DBG("upgrade all");
        free_list(pkgs_to_be_deleted);
        pkgs_to_be_deleted = NULL;
        for ( pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++ ) {
            pkgCache::VerIterator v0 = pkg.CurrentVer ();
            pkgCache::VerIterator v1 = cache[pkg].CandidateVerIter (cache);
            pkgCache::VerIterator sys_ver;
            
            if (! __filter_package_list (pkg, v0, v1, PKG_TYPE_UPDATE, NULL, TRUE, &sys_ver)
                    && ! is_update_package_ignored(pkg, v1) 
                    && ! is_os_meta_package (v1) ) {
		DBG("trying %s", pkg.Name ());

                GError *e = NULL;
                get_package_states (sc);
                if ( mark_named_package_for_install(pkg.Name(), v1.VerStr(), NULL, TRUE, FALSE, TRUE, &e) ) {
                    ++n_installable;
                    PackageVersion *pv = package_version_new ();
                    pv->pkg = g_strdup (pkg.Name ());
                    pv->version = g_strdup (v1.VerStr ());
                    ignore_list = g_list_append( ignore_list, pv);
                    DBG("Marked for install. Pkg: %s (%s)", pkg.Name(), v1.VerStr());
                }
                if ( e ) { // TODO: ???
                    DBG("Error.");
                    reset_changed_package_states (sc);
                    g_error_free( e );
                }
            }
        }
    } else {
        pkg = cache.FindPkg (packageidentifier);
        free_list(pkgs_to_be_deleted);
        pkgs_to_be_deleted = NULL;
        
        bool ok = true;
        if ( op_type == OPERATION_RESTORE_INSTALL || op_type == OPERATION_RESTORE_UPGRADE ) {
            DBG("Operation restore");
            
            ensure_restore_pkg_list();
            ok = false;
            restore_pkgs->getPackageTimestamp(packageidentifier, &ok);
            if ( ! ok ) {
                g_set_error(&err, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_FOUND, "Cannot find a restore package: %s (%s)", packageidentifier, version);
            } else {
                fromfile = false;
                ok = false;
                char* pkgfn = restore_pkgs->getPackageFilename(packageidentifier, &ok);
                if ( ok && pkgfn && strcmp(pkgfn, "") != 0 ) {
                    fromfile = true;
                    packageidentifier = g_strdup(pkgfn);
                }
                if ( pkgfn ) {
                    g_free(pkgfn);
                }
            }
        }
        if ( ok ) {
            if ( !fromfile && ( op_type == OPERATION_INSTALL || op_type == OPERATION_UPGRADE || op_type == OPERATION_RESTORE_INSTALL || op_type == OPERATION_RESTORE_UPGRADE ) ) {
                bool upgrade = op_type == OPERATION_UPGRADE || op_type == OPERATION_RESTORE_UPGRADE;
                DBG("%s", upgrade?"Upgrade":"Install");
                if ( mark_named_package_for_install(packageidentifier, version, domain, TRUE, FALSE, upgrade, &err) ) {
                    ++n_installable;
                }
                if ( n_installable == 0 && ( op_type == OPERATION_RESTORE_INSTALL || op_type == OPERATION_RESTORE_UPGRADE ) ) {
                    if ( err ) {
                        g_error_free( err );
                        err = NULL;
                    }
                    free_list(pkgs_to_be_deleted);
                    pkgs_to_be_deleted = NULL;
                    if ( mark_named_package_for_install(packageidentifier, version, domain, TRUE, FALSE, !upgrade, &err) ) {
                        ++n_installable;
                    }
                }
            } else if ( op_type == OPERATION_SYSTEM_UPGRADE || op_type == OPERATION_DOWNLOAD_SYSTEM_UPGRADE ) {
                pkgCache::PkgIterator sys_pkg;
                pkgCache::VerIterator sys_ver;
                if (find_os_metapackage (packageidentifier, version, domain,
                        sys_pkg, sys_ver,
                        &err)
                    && mark_for_install (sys_pkg, sys_ver, TRUE, TRUE, TRUE, &sys_ver, &err))
                {
                    ++n_installable;
                }
            } else if ( op_type == OPERATION_INSTALL_FILE || fromfile ) {
                DBG("Install file (or restore from file)");
                if (packageidentifier[0] != '/') {
                    if ( !err )
                        g_set_error (&err, PKGMGRD_ERROR, ERROR_INVALID_OPERATION, "Filename must be absolute");
                } else {
                    gchar *record = NULL;
                    pkgTagSection section;
                    record = read_tag_section (packageidentifier, &section, &err);
                    if (record != NULL) {
                        string arch = section.FindS("Architecture");
                        if ( arch != DEB_HOST_ARCH && arch != "all" ) {
                            g_set_error(&err, PKGMGRD_ERROR, ERROR_PACKAGE_NOT_COMPATIBLE, "Package '%s' has wrong architecture", packageidentifier);
                        } else {
                            DBG("Correct arch");
                            string p = section.FindS("Package");
                            string v = section.FindS("Version");
                            domain_t domain = get_file_domain (packageidentifier);

                            const char *package = p.c_str ();
                            const char *version = v.c_str ();
                            package_name = p;

                            DBG("adding extra index");
                            // scoped_cache_rw_lock locker;

                            pkgCache::PkgFileIterator pf = 
                              add_extra_index_from_section(section, packageidentifier, domain);
                        
                            if (pf.end()) {
                              g_set_error (&err, PKGMGRD_ERROR, ERROR_INTERNAL, "Extra index failure");
                              return NULL;
                            }

                            /* We need a cache init, since we don't want
                               the changes done by
                               add_extra_index_from_section to leak into
                               the next operation.
                            */
                            need_cache_init ();

                            DBG("extra index added");

                            pkgCache::VerIterator ver;
                            if (!awc->cache->find_version (package, version, NULL, pf, ver, error))
                                return NULL;

                            if ( mark_package_for_reinstall (ver, &err) ) {
                                DBG("package marked");
                                ++n_installable;
                                PackageVersion *pv = package_version_new ();
                                pv->pkg = g_strdup( package );
                                pv->version = g_strdup( version );
                                ignore_list = g_list_append( ignore_list, pv);
                                DBG("package marked out");
                            }
                        }
                    }
                }
            } else if ( op_type == OPERATION_UNINSTALL ) {
                DBG("Uninstall");
                if (pkg.CurrentVer().end()
                    && awc->cache->failed_installs.contains(packageidentifier)) {
                    // Ok
                } else {
                    bool ok = mark_for_remove( pkg, true, &err );
	            DBG("Uninstall ok: %s", ok?"Yes":"No");
	        }
            } else {
                DBG("Unknown operation");
                g_set_error( error, PKGMGRD_ERROR, ERROR_INVALID_OPERATION, "Invalid operation '%s'", operation );
                return NULL;
            }
        }
    }
    DBG("Installable cnt: %d", n_installable);
    
    gint64 download_size = 0;
    gint64 install_size = 0;
    bool add_inst_size = false;
    if ( err ) {
        DBG("Error: %s", err->message);
    } else {
        // DBG("No error");
    }

    {
    pkgDepCache &cache = *(awc->cache);
    pkg = cache.FindPkg (package_name.c_str());

    for (pkgCache::PkgIterator p = cache.PkgBegin (); !p.end (); p++) {
//        DBG("Pkg: %s", p.Name());
        
        if ( cache[p].InstBroken() ) {
            // DBG("Broken package. Pkg: %s ()", p.Name() /*, ver.end()?"":ver.VerStr()*/ );
            //go_through_broken_deps( cacheFile, p.CurrentVer(), &revert_deps );
            if ( pkg != p ) {
                pkgCache::VerIterator ver = cache[p].CandidateVerIter(cache);
//                pkgCache::VerIterator cur = p.CurrentVer();
                PackageVersion *pv = package_version_new ();
                pv->pkg = g_strdup (p.Name ());
                pv->version = g_strdup (ver.VerStr ());
                if ( is_user_package(ver) )
                    revert_user_deps = g_list_append (revert_user_deps, pv);
                else
                    revert_deps = g_list_append (revert_deps, pv);
            }
        }
        
        if ( cache[p].Mode == pkgDepCache::ModeInstall || cache[p].iFlags & pkgDepCache::ReInstall ) {
            pkgCache::VerIterator ver = cache[p].CandidateVerIter(cache);
            pkgCache::VerIterator cur = p.CurrentVer();
            
            if ( !awc->cache->is_trusted(ver) ) {
              trusted = false;
            } else {
              trusted_cnt++;
            }

            // DBG("Mode install. Pkg: %s (%s)", p.Name(), ver.end()?"":ver.VerStr() );
            // DBG("Pkg: %s is %strusted", p.Name(), trusted?"":"NOT" );

            PackageVersion *pv = package_version_new();
            pv->pkg = g_strdup(p.Name() );
            pv->version = g_strdup( ver.VerStr() );
            
            if ( !err && op_type != OPERATION_SYSTEM_UPGRADE && op_type != OPERATION_DOWNLOAD_SYSTEM_UPGRADE && is_os_meta_package(ver) ) {
                g_set_error(&err, PKGMGRD_ERROR, ERROR_PACKAGE_IS_SYSTEM_PACKAGE, "Package '%s' is a system package", p.Name());
            }

            if ( pkg != p && !is_in_list(ignore_list, pv) ) {
                if ( cur.end() ) {
                    if ( is_user_package(ver) )
                        install_user_deps = g_list_append (install_user_deps, pv);
                    else
                        install_deps = g_list_append (install_deps, pv);
                } else {
                    if ( is_user_package(ver) )
                        upgrade_user_deps = g_list_append (upgrade_user_deps, pv);
                    else
                        upgrade_deps = g_list_append (upgrade_deps, pv);
                }
            }

	    if (!cur.end())
		install_size += get_installed_size_delta_bytes (ver, cur);
	    else
		install_size += get_installed_size_bytes (ver);
            add_inst_size = true;
        } else if (cache[p].Mode == pkgDepCache::ModeDelete) {
            pkgCache::VerIterator ver = p.CurrentVer();
            // DBG("Mode delete. Pkg: %s (%s)", p.Name(), ver.end()?"":ver.VerStr() );
            package_record rec (ver);

            if ( pkg != p ) {
                PackageVersion *pv = package_version_new ();
                pv->pkg = g_strdup (p.Name ());
                pv->version = g_strdup (ver.VerStr ());
                if ( is_user_package(ver) )
                    remove_user_pkgs = g_list_append (remove_user_pkgs, pv);
                else
                    remove_pkgs = g_list_append (remove_pkgs, pv);
            }

	    install_size -= get_installed_size_bytes (ver);
            add_inst_size = true;
        } else if ( cache[p].Mode == pkgDepCache::ModeKeep ) {
//            DBG("Mode keep");
        } else {
          // DBG("Mode none");
        }
    }

    if (!err) {
        download_size = (gint64) cache.DebSize (); // FIXME: This is wrong somehow...
    }
    else {
        if(PkgmgrdError(err->code) == ERROR_PACKAGE_REQUIRES_SYSTEM_UPGRADE) {
            download_size = (gint64) cache.DebSize (); // FIXME: This is wrong somehow...
        }
        add_inst_size = false;
    }
    }

    DBG ("Done inspecting");

    cache_reset_packages();

    DBG ("Done resetting");

    GHashTable *map = g_hash_table_new(g_str_hash, g_str_equal);
    
    if ( err ) {
        DBG("Error: %s", err->message);
        g_hash_table_insert_string_value(map, "Error", err->message);
        g_hash_table_insert_string_value(map, "ErrorName", pkgmgr_error_to_dbus_error( PkgmgrdError( err->code )));
    } else {
        // DBG("No error");
    }

    // Add these fields even there's an error.
    if ( install_user_deps ) {
	install_user_deps = package_version_list_sort_by_name (install_user_deps);
        GValue *val = package_value_list_to_gvalue (install_user_deps);
        g_hash_table_insert (map, g_strdup ("InstallUserDeps"), val);
    }
    if ( install_deps ) {
	install_deps = package_version_list_sort_by_name (install_deps);
        GValue *val = package_value_list_to_gvalue (install_deps);
        g_hash_table_insert (map, g_strdup ("InstallDeps"), val);
    }
    if ( upgrade_user_deps ) {
	upgrade_user_deps = package_version_list_sort_by_name (upgrade_user_deps);
        GValue *val = package_value_list_to_gvalue (upgrade_user_deps);
        g_hash_table_insert (map, g_strdup ("UpgradeUserDeps"), val);
    }
    if ( upgrade_deps ) {
	upgrade_deps = package_version_list_sort_by_name (upgrade_deps);
        GValue *val = package_value_list_to_gvalue (upgrade_deps);
        g_hash_table_insert (map, g_strdup ("UpgradeDeps"), val);
    }
    if ( remove_user_pkgs ) {
	remove_user_pkgs = package_version_list_sort_by_name (remove_user_pkgs);
        GValue *val = package_value_list_to_gvalue (remove_user_pkgs);
        g_hash_table_insert (map, g_strdup ("RemoveUserPkgs"), val);
    }
    if ( remove_pkgs ) {
	remove_pkgs = package_version_list_sort_by_name (remove_pkgs);
        GValue *val = package_value_list_to_gvalue (remove_pkgs);
        g_hash_table_insert (map, g_strdup ("RemovePkgs"), val);
    }
    if ( revert_user_deps ) {
	revert_user_deps = package_version_list_sort_by_name (revert_user_deps);
        GValue *val = package_value_list_to_gvalue (revert_user_deps);
        g_hash_table_insert (map, g_strdup ("RevertUserDeps"), val);
    }
    if ( revert_deps ) {
	revert_deps = package_version_list_sort_by_name (revert_deps);
        GValue *val = package_value_list_to_gvalue (revert_deps);
        g_hash_table_insert (map, g_strdup ("RevertDeps"), val);
    }
    
    if ( trusted && trusted_cnt > 0 ) {
        g_hash_table_insert_boolean_value(map, "FlagTrusted", trusted);
    }
    
    if ( download_size > 0 ) {
        g_hash_table_insert_int64_value (map, "DownloadSize", download_size);
    }
    if ( add_inst_size ) {
        g_hash_table_insert_int64_value (map, "TotalInstalledSize", install_size);
    }
    
    DBG("map NULL: %s", map?"No":"Yes");
    return map;
}

// gboolean
// cmd_get_catalogues (GList **catalogues, GError **error)
// {
// 	g_assert (catalogues != NULL);
// 	*catalogues = catalogues_get_all ();
// 	return TRUE;
// }

// static xexp *
// make_catalogue (const gchar *name, const gchar *uri, const gchar *dist, const gchar *comp)
// {
// 	xexp *cat = xexp_list_new ("catalogue");
// 	xexp_append_1 (cat, xexp_text_new ("name", name));
// 	xexp_append_1 (cat, xexp_text_new ("uri", uri));
// 	xexp_append_1 (cat, xexp_text_new ("dist", dist));
// 	xexp_append_1 (cat, xexp_text_new ("components", comp));
// 	return cat;
// }

// static gboolean
// write_catalogues (xexp *cats, GError **error)
// {
// 	if (! write_user_catalogues (cats)) {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL,
// 			     "Writing user catalogues failed");
// 		return FALSE;
// 	}

// 	if (! write_sources_list (CATALOGUE_APT_SOURCE, cats)) {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL,
// 			     "Writing sources.list failed");
// 		return FALSE;
// 	}

// 	apt_worker_reset_state ();

// 	return TRUE;
// }

// static xexp *
// find_catalogue_by_name (xexp *cats, const gchar *name)
// {
// 	for (xexp *cat = xexp_first (cats); cat != 0; cat = xexp_rest (cat)) {
// 		if (! g_strcmp0 (name, xexp_aref_text (cat, "name"))) {
// 			return cat;
// 		}
// 	}

// 	return NULL;
// }

// gboolean
// cmd_add_catalogue (const gchar *name,
// 		   const gchar *uri,
// 		   const gchar *dist,
// 		   const gchar *comp,
// 		   GError **error)
// {
// 	DBG("cmd_add_catalogue(%s, %s, %s, %s)", name, uri, dist, comp);
// 	xexp *cats = read_catalogues ();
// 	xexp *cat;
// 	gchar *catname;
// 	gboolean ret = FALSE;

// 	if (! cats) {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL,
// 			     "read_catalogues () unexpectedly failed");
// 		return FALSE;
// 	}

// 	cat = find_catalogue_by_name (cats, name);
// 	if (cat == NULL) {
// 		catname = find_catalogue_by_info (uri, dist, comp, TRUE);
// 		if (catname == NULL) {
// 			cat = make_catalogue (name, uri, dist, comp);
// 			xexp_append_1 (cats, cat);
// 			ret = write_catalogues (cats, error);
// 		}
// 		else {
// 			g_set_error (error, PKGMGRD_ERROR, ERROR_CATALOGUE_ALREADY_EXISTS,
// 				     "Catalogue '%s' with same info already exists",
// 				     catname);
// 			g_free (catname);
// 		}
// 	}
// 	else {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_CATALOGUE_ALREADY_EXISTS,
// 			     "Catalogue '%s' already exists",
// 			     name);
// 	}
// 	xexp_free (cats);
// 	return ret;
// }

// gboolean
// cmd_remove_catalogue (const gchar *name,
// 		      GError **error)
// {
// 	xexp *cats = read_catalogues ();
// 	xexp *cat;
// 	gboolean ret = FALSE;

// 	if (! cats) {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL,
// 			     "read_catalogues () unexpectedly failed");
// 		return FALSE;
// 	}

// 	cat = find_catalogue_by_name (cats, name);
// 	if (cat) {
// 		if (! xexp_aref_bool (cat, "essential")) {
// 			xexp_del (cats, cat); /* frees cat */
// 			ret = write_catalogues (cats, error);
// 		} else {
// 			g_set_error (error, PKGMGRD_ERROR, ERROR_ACCESS_DENIED,
// 				     "Catalogue '%s' is essential and cannot be removed",
// 				     name);
// 		}
// 	}
// 	else {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_CATALOGUE_NOT_FOUND,
// 			     "Catalogue '%s' not found", name);
// 	}
// 	xexp_free (cats);
// 	return ret;
// }

// gboolean
// cmd_enable_catalogue (const gchar *name, gboolean enable,
// 		      GError **error)
// {
// 	xexp *cats = read_catalogues ();
// 	xexp *cat;
// 	gboolean ret = FALSE;

// 	if (! cats) {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_INTERNAL,
// 			     "read_catalogues () unexpectedly failed");
// 		return FALSE;
// 	}

// 	cat = find_catalogue_by_name (cats, name);
// 	if (cat) {
// 		if (! xexp_aref_bool (cat, "essential")) {
// 			xexp_aset_bool (cat, "disabled", ! enable);
// 			ret = write_catalogues (cats, error);
// 		}
// 		else {
// 			g_set_error (error, PKGMGRD_ERROR, ERROR_ACCESS_DENIED,
// 				     "Catalogue '%s' is essential and cannot be modified",
// 				     name);
// 		}
// 	}
// 	else {
// 		g_set_error (error, PKGMGRD_ERROR, ERROR_CATALOGUE_NOT_FOUND,
// 			     "Catalogue '%s' not found", name);
// 	}
// 	g_free (cats);
// 	return ret;
// }

// */

/* APTCMD_SAVE_BACKUP_DATA

   This method is used to store the list of installed packages. It's
   used in backup machinery to restore the installed packages
   from their repositories.
 */

// static xexp *
// get_backup_packages ()
// {
//   AptWorkerCache *awc = AptWorkerCache::GetCurrent ();

//   if (!ensure_cache (true))
//     return NULL;

//   xexp *packages = xexp_list_new ("backup");

//   pkgDepCache &cache = *(awc->cache);

//   for (pkgCache::PkgIterator pkg = cache.PkgBegin(); !pkg.end (); pkg++)
//     {
//       pkgCache::VerIterator installed = pkg.CurrentVer ();

//       // skip not-installed packages
//       //
//       if (installed.end ())
// 	continue;

//       // skip non user packages
//       //
//       if (!installed.end ()
// 	  && !is_user_package (installed))
// 	continue;

//       xexp_cons (packages, xexp_text_new ("pkg", pkg.Name ()));
//     }

//   return packages;
// }

// void
// cmd_save_backup_data ()
// {
//   backup_catalogues ();

//   xexp *packages = get_backup_packages ();
//   if (packages)
//     {
//       xexp_write_file (BACKUP_PACKAGES, packages);
//       xexp_free (packages);
//     }
// }




// void
// cmd_reboot ()
// {
//   system ("/sbin/reboot");
// }

/* Rescue
 */

// static int
// run_system (bool verbose, const char *fmt, ...)
// {
//   va_list ap;
//   va_start (ap, fmt);
//   char *cmd = g_strdup_vprintf (fmt, ap);
//   va_end (ap);

//   if (verbose)
//     fprintf (stderr, "+ %s\n", cmd);
//   int result = system (cmd);
//   free (cmd);

//   return result;
// }

// static int
// rescue_operation_with_dir (const char *dir)
// {
//   int result;

//   fprintf (stderr, "Rescuing from %s\n", dir);
//   result = operation (false, dir, false, false, false);
//   fprintf (stderr, "Result code %d\n", result);
//   _error->DumpErrors ();
//   return result;
// }

// static int
// rescue_operation_with_dev (const char *dev)
// {
//   const char *dir = "/rescue";

//   fprintf (stderr, "Rescuing from %s\n", dev);

//   if (mkdir (dir, 0777) < 0
//       && errno != EEXIST)
//     log_stderr ("%s: %m", dir);

//   run_system (true, "mount -t vfat '%s' '%s'", dev, dir);
//   int result = rescue_operation_with_dir (dir);
//   run_system (true, "umount '%s'", dir);

//   if (rmdir (dir) < 0)
//     log_stderr ("%s: %m", dir);

//   return result;
// }

// static int
// rescue_operation_with_devnode (int major, int minor)
// {
//   const char *node = "/dev.rescue";

//   if (mknod (node, S_IFBLK | 0600, major << 8 | minor) < 0)
//     log_stderr ("%s: %m", node);

//   fprintf (stderr, "Rescuing from %d:%d\n", major, minor);

//   int result = rescue_operation_with_dev (node);

//   unlink (node);

//   return result;
// }

// static const char* rescue_devs[] = {
//   "/dev/mmcblk0p1", // Internal MMC ~/MyDocs
//   "/dev/mmcblk1p1", // External MMC
//   "/dev/mmcblk0p2", // Internal MMC /home
//   "/dev/mmcblk0p3", // Internal MMC swap
//   "/dev/mmcblk0",
//   "/dev/mmcblk1",
//   NULL
// };

// static int
// rescue_with_all_devs ()
// {
//   int result = rescode_packages_not_found;

//   for (int i = 0;
//        (result == rescode_packages_not_found
// 	&& rescue_devs[i] != NULL);
//        i++)
//     {
//       result = rescue_operation_with_dev (rescue_devs[i]);
//     }

//   return result;
// }

// static struct {
//   int major, minor;
// } rescue_devnodes[] = {
//   { 254, 9 },
//   { 254, 1 },
//   { 254, 8 },
//   { 254, 0 },
//   { -1 }
// };

// static int
// rescue_with_all_devnodes ()
// {
//   int result = rescode_packages_not_found;

//   for (int i = 0;
//        (result == rescode_packages_not_found
// 	&& rescue_devnodes[i].major != -1);
//        i++)
//     {
//       result = rescue_operation_with_devnode (rescue_devnodes[i].major,
// 					      rescue_devnodes[i].minor);
//     }

//   return result;
// }

// static void
// show_fb_text (int line, const char *text)
// {
//   run_system (false,
// 	      "/usr/bin/text2screen -s 2 -x 5 -y %d -B -1 -T 0xF000 -t '%s'",
// 	      400 + 20*line, text);
// }

// static void
// show_fb_status (int percent)
// {
//   run_system (false,
// 	      "/usr/bin/text2screen -s 2 -x 5 -y %d -B -1 -T 0xF000 -t '%3d%%'",
// 	      440, percent);
// }

// static void
// interpret_pmstatus (char *str)
// {
//   float percentage;
//   const char *title;

//   if (!strncmp (str, "pmstatus:", 9))
//     {
//       str += 9;
//       str = strchr (str, ':');
//       if (str == NULL)
// 	return;
//       str += 1;
//       percentage = atof (str);
//       str = strchr (str, ':');
//       if (str == NULL)
// 	title = "Working";
//       else
// 	{
// 	  str += 1;
// 	  title = str;
// 	}

//       show_fb_status ((int)percentage);
//       // show_fb_text (2, title);
//     }
// }

// static void
// fork_progress_process ()
// {
//   int fds[2], child_pid;

//   if (pipe (fds) < 0)
//     {
//       perror ("pipe");
//       return;
//     }

//   if ((child_pid = fork ()) < 0)
//     {
//       perror ("fork");
//       return;
//     }

//   if (child_pid == 0)
//     {
//       close (fds[1]);
//       FILE *f = fdopen (fds[0], "r");
//       if (f)
// 	{
// 	  char *line = NULL;
// 	  size_t len = 0;
// 	  ssize_t n;

// 	  while ((n = getline (&line, &len, f)) != -1)
// 	    {
// 	      if (n > 0 && line[n-1] == '\n')
// 		line[n-1] = '\0';

// 	      interpret_pmstatus (line);
// 	    }
// 	}
//       exit (0);
//     }
//   else
//     {
//       close (fds[0]);
//       status_fd = fds[1];
//     }
// }

// static void
// do_rescue (const char *package, const char *download_root,
// 	   bool erase_record)
// {
//   show_fb_text (0, "Rescuing software update.");
//   show_fb_text (1, "Please do not interrupt.");

//   fork_progress_process ();

//   fprintf (stderr, "Rescuing %s\n", package);

//   /* This is just to clean the dpkg journal.  We let libapt-pkg
//      configure the rest of the packages since we will get better
//      progress reporting that way.
//    */
//   run_system (true, "dpkg --configure dpkg");

//   misc_init ();

//   // @todo Is this really necessary?
//   AptWorkerCache::GetCurrent ()->init_cache_after_request = false;
//   if (ensure_cache (false))
//     {
//       if (mark_named_package_for_install (package))
// 	{
// 	  int result = rescue_operation_with_dir (download_root);

// 	  if (result != rescode_success)
// 	    result = rescue_with_all_devs ();

// 	  if (result != rescode_success)
// 	    result = rescue_with_all_devnodes ();

// 	  /* If we get this far, we have done everything we can.  If
// 	     the rescue operation is interrupted before getting here,
// 	     we try again on the next boot.  This might lead to a
// 	     reboot loop, but the device is borked anyway and the user
// 	     should reflash it.
// 	   */
// 	  if (erase_record)
// 	    erase_operation_record ();

// 	  if (result == rescode_packages_not_found)
// 	    return;

// 	  if (result != rescode_success)
// 	    {
// 	      _system->UnLock();
// 	      run_system (true, "dpkg --configure -a --force-all");
// 	      _system->Lock();
// 	    }

// 	  run_system (true, "/sbin/reboot");
// 	}
//       else
// 	fprintf (stderr, "Package %s not found\n", package);
//     }
//   else
//     fprintf (stderr, "Failed to initialize package cache\n");
// }

// int
// cmdline_rescue (char **argv)
// {
//   if (argv[1] == NULL)
//     {
//       xexp *record = read_operation_record ();

//       if (record == NULL)
// 	{
// 	  fprintf (stderr, "apt-worker: Nothing to rescue.\n");
// 	  return 0;
// 	}

//       const char *package = xexp_aref_text (record, "package");
//       const char *download_root = xexp_aref_text (record, "download-root");

//       do_rescue (package, download_root, true);
//     }
//   else
//     do_rescue (argv[1], argv[2], false);

//   return 0;
// }
